import os
import tempfile

import anydbm
from dogpile import cache

from dogpile.core.readwrite_lock import ReadWriteMutex
from dogpile.cache.backends.file import AbstractFileLock

import logging
logger = logging.getLogger(__name__)


class MutexLock(AbstractFileLock):
    """ Mutex lock for windows, which does not knows about fnctl"""
    def __init__(self, filename):
        self.mutex = ReadWriteMutex()

    def acquire_read_lock(self, wait):
        ret = self.mutex.acquire_read_lock(wait)
        return wait or ret

    def acquire_write_lock(self, wait):
        ret = self.mutex.acquire_write_lock(wait)
        return wait or ret

    def release_read_lock(self):
        return self.mutex.release_read_lock()

    def release_write_lock(self):
        return self.mutex.release_write_lock()

import hashlib
def key_generator(namespace, fn, **kw):
    """ Key generator. Will decode unicode keys to str"""
    fname = fn.__name__
    def generate_key(*arg):
        keys = []
        for a in arg:
            if isinstance(a, unicode):
                a = a.encode('utf-8')
            keys.append(a)
        key = (namespace + "_" if namespace else '') + fname + "_" + hashlib.sha1(''.join( str(s) for s in keys )).hexdigest()
        return key
    return generate_key


# stores caches dbm files in a temporary folder
_cache_data_file = os.path.join(tempfile.gettempdir(), 'gracenote', 'cache', 'data.dbm')
_cache_auth_file = os.path.join(tempfile.gettempdir(), 'gracenote', 'cache', 'auth.dbm')


try:
    backend_data = 'dogpile.cache.dbm'
    backend_auth = 'dogpile.cache.dbm'
    # create cache directories if they does not exists
    if not os.path.exists(os.path.dirname(_cache_data_file)):
        os.makedirs(os.path.dirname(_cache_data_file))
    if not os.path.exists(os.path.dirname(_cache_auth_file)):
        os.makedirs(os.path.dirname(_cache_auth_file))
except Exception:
    logger.exception("Could not create cache directories. Using falback backends.")
    # Something went wrong, disable cache
    backend_data='dogpile.cache.null'
    # use memory backend for authentication
    backend_auth='dogpile.cache.memory'

    _cached_data_args=[]
    _cached_auth_args=[]
else:
    if os.name == "nt":
        # if windows, use MutexLock as lock_factory
        _cached_data_args = { "filename":_cache_data_file, "lock_factory": MutexLock,}
        _cached_auth_args = { "filename":_cache_auth_file, "lock_factory": MutexLock,}
    else:
        # otherwise use the default locking mecanism
        _cached_data_args = { "filename":_cache_data_file}
        _cached_auth_args = { "filename":_cache_auth_file}

cached_data = cache.make_region(function_key_generator=key_generator).configure(
    backend_data,
    expiration_time = 3600 * 24 * 3, # 3 days
    arguments=_cached_data_args
)

cached_auth = cache.make_region(function_key_generator=key_generator).configure(
    backend_auth,
    expiration_time = 3600 * 24 * 10, # 10 days
    arguments=_cached_auth_args
)



def cache_on_arguments(region, function, *args, **kwargs):
    try:
        return region.cache_on_arguments()(function)(*args, **kwargs)
    except Exception as e:
        # Cache file is created by dogpile on initialisation.
        # If we have this message, the cache file had been deleted.
        # It won't be created until the next restart, so without this hack
        # any further call will fail.

        # We call directly the method, bypassing the cache system
        # We could recreate the cache direcotry / files / lock but this seems
        # too aggressive.
        db_corrupted = "db type could not be determined" in getattr(e, 'message', '')
        # file have ben removed
        doesnt_exists = "need 'c' or 'n' flag to open new db" in getattr(e, 'message', '')
        # directory have been removed (errno 2 : no such file or directory)
        doesnt_exists = doesnt_exists or getattr(e, 'errno', -1) == 2

        is_db_error = isinstance(e, anydbm.error)

        if db_corrupted or is_db_error:
            # remove the file. Will be recreated on next startup
            try:
                os.remove(region.backend.filename)
            except Exception:
                # don't care
                pass
        if db_corrupted or doesnt_exists or is_db_error:
            return function(*args, **kwargs)
        else:
            raise
