# coding: utf-8
import os

from gnsdk import _gnsdk
from gnsdk import GnsdkFingerprintError
from gnsdk.http import GracenoteHTTPClient
import threading

from audioread.ffdec import FFmpegAudioFile

import logging
logger = logging.getLogger(__name__)

__all__ = [
    "Gracenote",
    "GnsdkContext",
    "GnsdkAudioStreamFingerPrinter"
]




class Gracenote(object):

    """ Class that encapsulate calls to the underlying GNSDK C class
    and HTTP API wrapper
    """

    def __init__(
            self,
            client_id,
            client_id_tag,
            licence_content,
            client_app_version,
            user_token=None,
            web_user_token=None,
            http_timeout=10
    ):
        self.lock = threading.Lock()
        # config for GNSDK
        self.gnsdk_conf = {}

        if type(client_id) is unicode:
            client_id = client_id.encode("utf-8")
        if type(client_id_tag) is unicode:
            client_id_tag = client_id_tag.encode("utf-8")
        if type(licence_content) is unicode:
            licence_content = licence_content.encode("utf-8")
        if type(client_app_version) is unicode:
            client_app_version = client_app_version.encode("utf-8")
        if type(user_token) is unicode:
            user_token = user_token.encode("utf-8")

        self.gnsdk_conf["client_id"] = client_id
        self.gnsdk_conf["client_id_tag"] = client_id_tag
        self.gnsdk_conf["licence_content"] = licence_content
        self.gnsdk_conf["client_app_version"] = client_app_version
        self.gnsdk_conf["user_token"] = user_token

        # Config for HTTP API client
        self.http_conf = {}
        self.http_conf["client_id"] = client_id
        self.http_conf["client_id_tag"] = client_id_tag
        self.http_conf["user_id"] = web_user_token
        self.http_conf["timeout"] = http_timeout

        self.http_client = GracenoteHTTPClient(**self.http_conf)

    @property
    def user_token(self):
        return self.gnsdk_conf.get("user_token", None)

    @property
    def web_user_token(self):
        return self.http_conf.get("user_id", None)

    def _update_user_token(self, user_token):
        logger.debug("Update user token")
        if type(user_token) is unicode:
            user_token = user_token.encode("utf-8")
        self.gnsdk_conf["user_token"] = user_token

    def _update_api_user_token(self, user_id):
        logger.debug("Update user token")
        self.http_conf["user_id"] = user_id

    def audio_fingerprinter(self, is_stream_mode=True):
        """ Return a fingerprinter context manager """
        if is_stream_mode is True:
            fp = GnsdkAudioStreamFingerPrinter(**self.gnsdk_conf)
        else:
            fp = GnsdkAudioFileFingerPrinter(**self.gnsdk_conf)
        # cannot be passed as constructor argument as we can't bypass __cinit__
        # The context manager will update the user token on exit
        fp.gracenote = self
        return fp


    def submit_file(self, fname, is_stream_mode=True):
        """Compute a fingerprint an submit it on the gracenote server
        """
        if not os.path.exists(fname):
            raise IOError("File '%s' does not exists" % fname)
        fingerprint = None
        with self.audio_fingerprinter(is_stream_mode) as fp:
            fingerprint = fp.fingerprint(fname)
        return self.submit_fingerprint(fingerprint, is_stream_mode)

    def submit_fingerprint(self, fingerprint, is_stream_mode=True):
        """ Search for a fingerprint on the gracenote server
        """
        if fingerprint is not None:
            # use a lock to avoid concurrent registrations
            with self.lock:
                if self.http_client.user_id is None:
                    user_id = self.http_client.register()
                    self._update_api_user_token(user_id)
            return self.http_client.submit_fingerprint(fingerprint, is_stream_mode)

    def search_track(self, artist, album, track):
        """ Search for a fingerprint on the gracenote server
        """
        # use a lock to avoid concurrent registrations
        with self.lock:
            if self.http_client.user_id is None:
                user_id = self.http_client.register()
                self._update_api_user_token(user_id)
        return self.http_client.search_track(artist, album, track)


class GnsdkContext(_gnsdk._Gnsdk):

    """
    GNSDK context. Ensure that the SDK is properly initialized an shutdown
    """

    def __enter__(self):
        self.__init_gnsdk__()
        return self

    def __exit__(self, type, value, traceback):
        self.__shutdown_gnsdk__()

        # update gracenote user token
        new_token = self.get_usertoken()
        _gracenote = getattr(self, "gracenote", None)
        if _gracenote is not None and new_token is not None:
            _gracenote._update_user_token(new_token)


class GnsdkAudioStreamFingerPrinter(GnsdkContext):

    """
    GNSDK audio stream fingerprinter context
    """

    def fingerprint(self, fname):
        """ Return thefingerprint for the given file (stream mode)"""
        with FFmpegAudioFile(fname) as f:
            try:
                # fingerprint expect string, not unicode
                if type(fname) is unicode:
                    fname = fname.encode("utf-8")
                # audioread default sample size : 16
                self.__fingerprint_stream_begin__(
                    fname, f.samplerate, f.channels, 16)
                try:
                    return self._fingerprint(f)
                finally:
                    self.__fingerprint_stream_end__()
            except GnsdkFingerprintError:
                logger.error("Could not fingerprint file '%s'", fname)
                return None

    def _fingerprint(self, audio_datas):
        """ Internal method forfinger printing.

        This method must always be called
         - after self.__fingerprint_stream_begin__
         - before self.__fingerprint_stream_end__
        """
        ret = None
        for buff in audio_datas:
            complete = self.__fingerprint_stream_write__(buff)
            if complete:
                logger.info("Fingerprint complete")
                ret = self.get_fingerprint_stream()
                break
        return ret

class GnsdkAudioFileFingerPrinter(GnsdkContext):

    """
    GNSDK audio stream fingerprinter context
    """

    def fingerprint(self, fname):
        """ Return thefingerprint for the given file (stream mode)"""
        with FFmpegAudioFile(fname) as f:
            try:
                # fingerprint expect string, not unicode
                if type(fname) is unicode:
                    fname = fname.encode("utf-8")
                # audioread default sample size : 16
                self.__fingerprint_file_begin__(
                    fname, f.samplerate, f.channels, 16)
                try:
                    return self._fingerprint(f)
                finally:
                    self.__fingerprint_file_end__()
            except GnsdkFingerprintError:
                logger.error("Could not fingerprint file '%s'", fname)
                return None

    def _fingerprint(self, audio_datas):
        """ Internal method forfinger printing.

        This method must always be called
         - after self.__fingerprint_stream_begin__
         - before self.__fingerprint_stream_end__
        """
        ret = None
        for buff in audio_datas:
            complete = self.__fingerprint_file_write__(buff)
            if complete:
                logger.info("Fingerprint complete")
                ret = self.get_fingerprint_file()
                break
        return ret
