# coding: utf-8

import sys
import requests
# only for text unquote :
import urllib
from xml.etree import ElementTree
from xml.etree.ElementTree import ParseError
from xml.etree.ElementTree import Element


# define a zone for persistent caching
from cache import cached_auth, cached_data, cache_on_arguments

from gnsdk import GnsdkError
from gnsdk import utils

import logging
logger = logging.getLogger(__name__)


class GracenoteHTTPError(GnsdkError):
    pass


class GracenoteResponseError(GracenoteHTTPError):
    pass


class GracenoteAccountError(GracenoteHTTPError):
    pass


class GracenoteNetworkError(GracenoteHTTPError):
    pass


class GracenoteFatalError(GracenoteHTTPError):
    pass


class GracenoteLicenceError(GracenoteFatalError):
    pass


class GnMetadata(dict):

    """
    This class is a dictionary containing metadata fields that are available
    for the queried item.
    """
    # highly inspired from pygn

    def __init__(self):
        # Basic Metadata
        self[u'track_artist_name'] = u''
        self[u'album_artist_name'] = u''
        self[u'album_title'] = u''
        self[u'album_year'] = u''
        self[u'track_title'] = u''
        self[u'track_number'] = u''
        self[u'track_count'] = u''

        # Descriptors
        self[u'genre'] = {}
        self[u'artist_origin'] = {}
        self[u'artist_era'] = {}
        self[u'artist_type'] = {}
        self[u'mood'] = {}
        self[u'tempo'] = {}

        # Related Content
        self[u'album_art_url'] = u''
        self[u'artist_image_url'] = u''
        self[u'artist_bio_url'] = u''
        self[u'review_url'] = u''

        # Gracenote IDs
        self[u'album_gnid'] = u''
        self[u'track_gnid'] = u''

        # Radio ID
        self[u'radio_id'] = u''

        #  External IDs: Special content rights in license required
        self[u'album_xid'] = {}
        self[u'track_xid'] = {}

    def __getattr__(self, name):
        # enable object.attribute accessor
        if name in self.keys():
            return self[name]
        raise AttributeError(
            "Instance of %s has no attribute '%s'" % (self.__class__, name))


class GnQuery(object):
    # highly inspired from pygn

    def __init__(self):
        self.root = Element(u'QUERIES')

    def addQuery(self, cmd):
        query = ElementTree.SubElement(self.root, 'QUERY')
        query.attrib[u'CMD'] = cmd

    def addQueryMode(self, modeStr):
        query = self.root.find(u'QUERY')
        mode = ElementTree.SubElement(query, 'MODE')
        mode.text = modeStr

    def addQueryTextField(self, fieldName, value):
        query = self.root.find(u'QUERY')
        text = ElementTree.SubElement(query, 'TEXT')
        text.attrib[u'TYPE'] = fieldName
        text.text = value

    def addAuth(self, client_id, user_id):
        auth = ElementTree.SubElement(self.root, 'AUTH')
        client = ElementTree.SubElement(auth, 'CLIENT')
        user = ElementTree.SubElement(auth, 'USER')

        client.text = client_id
        user.text = user_id

    def addQueryClient(self, client_id):
        query = self.root.find(u'QUERY')
        client = ElementTree.SubElement(query, 'CLIENT')
        client.text = client_id

    def addQueryOption(self, parameterName, value):
        query = self.root.find(u'QUERY')
        option = ElementTree.SubElement(query, 'OPTION')
        parameter = ElementTree.SubElement(option, 'PARAMETER')
        parameter.text = parameterName
        valueElem = ElementTree.SubElement(option, 'VALUE')
        valueElem.text = value

    def addQueryFingerprint(self, fingerprint,
                            fingerprint_algo="Philips",
                            fingerprint_version="1.1"):

        # TODO : parse the fingerprint to get algo/version

        root = self.root.find(u'QUERY')
        e_fingerprint = ElementTree.SubElement(root, 'FINGERPRINT')
        e_fingerprint.attrib[u'ALGORITHM'] = fingerprint_algo
        e_fingerprint.attrib[u'VERSION'] = fingerprint_version

        e_data = ElementTree.SubElement(e_fingerprint, 'DATA')
        e_data.text = fingerprint

    def addQueryFingerprintFile(self, fingerprint,
                            fingerprint_algo="Cantametrix",
                            fingerprint_version="1.1"):

        # TODO : parse the fingerprint to get algo/version

        root = self.root.find(u'QUERY')
        e_fingerprint = ElementTree.SubElement(root, 'FINGERPRINT')
        e_fingerprint.attrib[u'ALGORITHM'] = fingerprint_algo
        e_fingerprint.attrib[u'VERSION'] = fingerprint_version

        e_data = ElementTree.SubElement(e_fingerprint, 'DATA')
        e_data.text = fingerprint

    def tostring(self):
        return ElementTree.tostring(
            self.root,
            encoding='utf8',
            method="xml"
        )


class GnResponse(object):

    """ A class for parsing the gracenote response"""

    def __init__(self, datasxml):
        self.datasxml = datasxml

        self.responseTree = None
        try:
            self.responseTree = ElementTree.fromstring(self.datasxml)
        except (ParseError, TypeError) as e:
            logger.error("Error parsing the response : %s", e)
        except (Exception) as e:
            logger.error("Unknown error while parsing the response : %s", e)

    @property
    def is_valid(self):
        if self.responseTree is None:
            return False
        responseElem = self.responseTree.find(u'RESPONSE')
        return responseElem is not None

    @property
    def response_status(self):
        if self.is_valid:
            responseElem = self.responseTree.find(u'RESPONSE')
            return responseElem.attrib[u'STATUS']
        return None

    @property
    def is_success(self):
        if self.is_valid:
            return self.response_status == 'OK'
        return False

    @property
    def is_no_match(self):
        if self.is_valid:
            return self.response_status == 'NO_MATCH'
        return False

    @property
    def is_error(self):
        if self.is_valid:
            return self.response_status == 'ERROR'
        return False

    @property
    def is_licence_issue(self):
        if self.is_valid and self.is_error:
            message = self.message or ""
            message = message.lower()
            # handle "[Gracenote Error 130/74]" like messages
            return "[gracenote error" in message

    @property
    def is_fatal(self):
        if self.is_valid and self.is_error:
            message = self.message or ""
            message = message.lower()
            # this is not good
            return "please contact" in message
        return False

    @property
    def message(self):
        if self.responseTree is not None:
            message = self.responseTree.find(u'MESSAGE')
            if message is not None:
                return message.text
        return None

    def parse(self):
        if self.is_success:
            responseElem = self.responseTree.find(u'RESPONSE')
            return self._parse_response_ok(responseElem)
        return None

    def _parse_response_ok(self, responseElem):
        if responseElem is None:
            return
        # Parse response
        metadata = GnMetadata()
        # Find Album element
        albumElem = responseElem.find(u'ALBUM')
        if albumElem is not None:
            self._parse_album(albumElem, metadata)
            self._parse_artist(albumElem, metadata)
            self._parse_track(albumElem, metadata)
            self._parse_tracklist(albumElem, metadata)
        return metadata

    def _parse_album(self, albumElem, metadata):
        if albumElem is None:
            return
        # Parse album metadata
        metadata[u'album_gnid'] = self._getElemText(albumElem, 'GN_ID')
        metadata[u'album_artist_name'] = self._getElemText(
            albumElem, 'ARTIST'
        )

        metadata[u'album_title'] = self._getElemText(albumElem, 'TITLE')
        metadata[u'album_year'] = self._getElemText(albumElem, 'DATE')
        metadata[u'album_art_url'] = self._getElemText(
            albumElem, 'URL', 'TYPE', 'COVERART')
        metadata[u'genre'] = self._getMultiElemText(
            albumElem, 'GENRE', 'ORD', 'ID'
        )
        metadata[u'track_count'] = self._getElemText(
            albumElem, 'TRACK_COUNT'
        )
        self._parse_xid(metadata, u"album_xid", albumElem)

    def _parse_xid(self, metadatas, key, elem, encoding=u'utf-8'):
        xids = elem.findall(u'XID')

        if xids is not None:
            for xid in xids:
                datasource = xid.attrib.get("DATASOURCE", None)
                value = xid.text
                if type(value) is str:
                    value = value.decode(encoding)
                if datasource not in metadatas[key]:
                    metadatas[key][datasource] = []
                metadatas[key][datasource].append(value)

    def _parse_artist(self, albumElem, metadata):
        if albumElem is None:
            return
        metadata[u'artist_image_url'] = self._getElemText(
            albumElem, 'URL', 'TYPE', 'ARTIST_IMAGE'
        )
        metadata[u'artist_bio_url'] = self._getElemText(
            albumElem, 'URL', 'TYPE', 'ARTIST_BIOGRAPHY'
        )
        metadata[u'review_url'] = self._getElemText(
            albumElem, 'URL', 'TYPE', 'REVIEW'
        )

        # Look for OET
        artistOriginElem = albumElem.find(u'ARTIST_ORIGIN')
        if artistOriginElem is not None:
            metadata[u'artist_origin'] = self._getMultiElemText(
                albumElem, 'ARTIST_ORIGIN', 'ORD', 'ID'
            )
            metadata[u'artist_era'] = self._getMultiElemText(
                albumElem, 'ARTIST_ERA', 'ORD', 'ID'
            )
            metadata[u'artist_type'] = self._getMultiElemText(
                albumElem, 'ARTIST_TYPE', 'ORD', 'ID'
            )

    def _parse_track(self, albumElem, metadata):
        if albumElem is None:
            return
        # Parse track metadata
        matchedTrackElem = albumElem.find(u'MATCHED_TRACK_NUM')
        if matchedTrackElem is not None:
            trackElem = albumElem.find(u'TRACK')

            metadata[u'track_number'] = self._getElemText(
                trackElem, 'TRACK_NUM'
            )
            metadata[u'track_gnid'] = self._getElemText(trackElem, 'GN_ID')
            metadata[u'track_title'] = self._getElemText(trackElem, 'TITLE')
            metadata[u'track_artist_name'] = self._getElemText(
                trackElem, 'ARTIST'
            )

            metadata[u'mood'] = self._getMultiElemText(
                trackElem, 'MOOD', 'ORD', 'ID'
            )
            metadata[u'tempo'] = self._getMultiElemText(
                trackElem, 'TEMPO', 'ORD', 'ID'
            )

            # If track-level GOET exists, overwrite metadata from album
            if trackElem.find(u'GENRE') is not None:
                metadata[u'genre'] = self._getMultiElemText(
                    trackElem, 'GENRE', 'ORD', 'ID'
                )
            if trackElem.find(u'ARTIST_ORIGIN') is not None:
                metadata[u'artist_origin'] = self._getMultiElemText(
                    trackElem, 'ARTIST_ORIGIN', 'ORD', 'ID'
                )
            if trackElem.find(u'ARTIST_ERA') is not None:
                metadata[u'artist_era'] = self._getMultiElemText(
                    trackElem, 'ARTIST_ERA', 'ORD', 'ID'
                )
            if trackElem.find(u'ARTIST_TYPE') is not None:
                metadata[u'artist_type'] = self._getMultiElemText(
                    trackElem, 'ARTIST_TYPE', 'ORD', 'ID'
                )
            self._parse_xid(metadata, u"track_xid", trackElem)

    def _parse_tracklist(self, albumElem, metadata):
        if albumElem is None:
            return
        # Parse tracklist
        metadata[u'tracks'] = []
        for trackElem in albumElem.iter(u'TRACK'):
            trackdata = {}

            trackdata[u'track_number'] = self._getElemText(
                trackElem, 'TRACK_NUM'
            )
            trackdata[u'track_gnid'] = self._getElemText(
                trackElem, 'GN_ID'
            )
            trackdata[u'track_title'] = self._getElemText(
                trackElem, 'TITLE'
            )
            trackdata[u'track_artist_name'] = self._getElemText(
                trackElem, 'ARTIST'
            )

            trackdata[u'mood'] = self._getMultiElemText(
                trackElem, 'MOOD', 'ORD', 'ID'
            )
            trackdata[u'tempo'] = self._getMultiElemText(
                trackElem, 'TEMPO', 'ORD', 'ID'
            )

            # If track-level GOET exists, overwrite metadata from album
            if trackElem.find(u'GENRE') is not None:
                trackdata[u'genre'] = self._getMultiElemText(
                    trackElem, 'GENRE', 'ORD', 'ID'
                )
            if trackElem.find(u'ARTIST_ORIGIN') is not None:
                trackdata[u'artist_origin'] = self._getMultiElemText(
                    trackElem, 'ARTIST_ORIGIN', 'ORD', 'ID'
                )
            if trackElem.find(u'ARTIST_ERA') is not None:
                trackdata[u'artist_era'] = self._getMultiElemText(
                    trackElem, 'ARTIST_ERA', 'ORD', 'ID'
                )
            if trackElem.find(u'ARTIST_TYPE') is not None:
                trackdata[u'artist_type'] = self._getMultiElemText(
                    trackElem, 'ARTIST_TYPE', 'ORD', 'ID'
                )
            metadata[u'tracks'].append(trackdata)
        return metadata

    def _getElemText(self,
                     parentElem,
                     elemName,
                     elemAttribName=None,
                     elemAttribValue=None,
                     encoding=u'utf-8'
                     ):
        """
        XML parsing helper function to find child element with a specific name,
        and return the text value
        """
        elems = parentElem.findall(elemName)
        for elem in elems:
            if elemAttribName is not None and elemAttribValue is not None:
                if elem.attrib[elemAttribName] == elemAttribValue:
                    value = urllib.unquote(elem.text)
                    if type(value) is str:
                        value = value.decode(encoding)
                    return value
                else:
                    continue
            else:  # Just return the first one
                value = urllib.unquote(elem.text)
                if type(value) is str:
                    value = value.decode(encoding)
                return value
        return u''

    def _getMultiElemText(self, parentElem, elemName, topKey, bottomKey, encoding=u'utf-8'):
        """
        XML parsing helper function to return a 2-level dict of multiple
        elements by a specified name, using topKey as the first key, and
        bottomKey as the second key
        """
        elems = parentElem.findall(elemName)
        # 2-level dictionary of items, keyed by topKey and then bottomKey
        result = []
        if elems is not None:
            for elem in elems:
                # Encode top key as unicode
                if type(topKey) is str:
                    topKey = topKey.decode(encoding)
                # Encode bottom key as unicode
                if type(bottomKey) is str:
                    bottomKey = bottomKey.decode(encoding)

                # Encode top key value as unicode
                topKeyValue = elem.attrib.get(topKey, None)
                if topKeyValue is not None and type(topKeyValue) is str:
                    topKeyValue = topKeyValue.decode(encoding)

                # Encode bottom key value as unicode
                botKeyValue = elem.attrib.get(bottomKey, None)
                if botKeyValue is not None and type(botKeyValue) is str:
                    botKeyValue = botKeyValue.decode(encoding)

                # Encode text value as unicode
                textValue = elem.text
                if type(textValue) is str:
                    textValue = textValue.decode(encoding)

                result.append(
                    {
                        topKey.lower(): topKeyValue,
                        u'value': textValue,
                        bottomKey.lower(): botKeyValue,
                    }
                )
        return sorted(result, key=lambda x: x[topKey.lower()])


class GracenoteHTTPClient(object):

    def __init__(self, client_id, client_id_tag, user_id=None, timeout=10):
        if client_id is None:
            raise ValueError("You must provide a client_id")
        if client_id_tag is None:
            raise ValueError("You must provide a client_id_tag")

        self.url = "https://c%s.web.cddbp.net/webapi/xml/1.0/" % (
            client_id
        )
        # The web api expect client_id-client_id_tag in the form
        # XXXXXXX-YYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYY
        self.client_id = "-".join([client_id, client_id_tag])
        self.user_id = user_id

        self.timeout = timeout

    def _http_query(self, queryXML):
        response = None
        try:
            response = requests.post(self.url, queryXML, timeout=self.timeout)
        # this catch may be a little too wide
        except(Exception, ):
            raise GracenoteNetworkError, GracenoteNetworkError("Error while talking to gracenote server"), sys.exc_info()[2]
        if response and response.text:
            # response encoding may not be utf-8
            result = response.text.encode(response.encoding)
            return result

    def register(self):
        logger.info("Registering user")

        def cached_register(client_id):
            query = GnQuery()
            query.addQuery(u'REGISTER')
            query.addQueryClient(client_id)

            queryXML = query.tostring()

            datas = self._http_query(queryXML)
            response = GnResponse(datas)

            if response.is_success:
                responseXML = response.responseTree
                responseNode = responseXML.find(u'RESPONSE')
                userNode = responseNode.find(u'USER')
                if userNode is None:
                    raise GracenoteResponseError("Response should have a USER node")
                user_id = userNode.text
            elif not response.is_valid:
                raise GracenoteResponseError("Server response was invalid")
            else:
                # not success and valid : KO
                raise GracenoteAccountError("Server response was KO")
            return user_id
        # pass the client_id as a parameter to avoid caching issues if client_id changes
        self.user_id = cache_on_arguments(cached_auth, cached_register, self.client_id)
        return self.user_id

    def _raise_for_response_error(self, response):
        if response.is_licence_issue:
            logger.error("Gracenote licence error")
            raise GracenoteLicenceError(response.message)
        elif response.is_fatal:
            logger.error("Gracenote fatal error")
            raise GracenoteFatalError(response.message)
        elif response.is_error or response.message is not None:
            logger.error("Server responded with an error")
            message = response.message
            raise GracenoteResponseError("Server responded with an error. Message : '%s'" % (message if message is not None else ""))
        elif not response.is_valid:
            raise GracenoteResponseError("Server response was invalid")
        else:
            # not success and valid : KO
            message = response.response_status
            raise GracenoteResponseError(
                "Server response was KO : %s" %
                ("no message" if message is None else message)
            )

    def submit_fingerprint(self, fingerprint, is_stream_mode):
        def cached_submit_fingerprint(fingerprint, is_stream_mode):
            datas = self._http_submit_fingerprint(fingerprint, is_stream_mode)
            response = GnResponse(datas)

            if response.is_success:
                logger.debug("Found a match")
                return response.parse()
            elif response.is_no_match:
                logger.debug("No match found")
                return None
            else:
                self._raise_for_response_error(response)
        return cache_on_arguments(cached_data, cached_submit_fingerprint, fingerprint, is_stream_mode)

    def _http_submit_fingerprint(self, fingerprint, is_stream_mode):
        if fingerprint is None:
            raise ValueError("Fingerprint can't be None")
        query = GnQuery()

        query.addAuth(self.client_id, self.user_id)
        query.addQuery('ALBUM_FINGERPRINT')
        query.addQueryMode('SINGLE_BEST_COVER')
        if is_stream_mode is True:
            query.addQueryFingerprint(fingerprint, fingerprint_algo="Philips", fingerprint_version="1.1")
        else:
            query.addQueryFingerprint(fingerprint, fingerprint_algo="Cantametrix", fingerprint_version="1.1")
        query.addQueryOption(
            'SELECT_EXTENDED',
            'COVER,REVIEW,ARTIST_BIOGRAPHY,ARTIST_IMAGE,ARTIST_OET,MOOD,TEMPO,LINK'
        )
        query.addQueryOption(
            'SELECT_DETAIL',
            'GENRE:3LEVEL,MOOD:2LEVEL,TEMPO:3LEVEL,ARTIST_ORIGIN:4LEVEL,ARTIST_ERA:2LEVEL,ARTIST_TYPE:2LEVEL'
        )
        queryXML = query.tostring()

        logger.debug("Submitting fingerprint")
        return self._http_query(queryXML)

    def search_track(self, artist, album, track):
        def cached_search_track(artist, album, track):
            datas = self._http_search_artist_album(artist, album, track)
            response = GnResponse(datas)

            if response.is_success:
                logger.debug("Found a match")
                return response.parse()
            elif response.is_no_match:
                logger.debug("No match found")
                return None
            else:
                self._raise_for_response_error(response)
        return cache_on_arguments(cached_data, cached_search_track, artist, album, track)

    def _http_search_artist_album(self, artist, album, track):
        query = GnQuery()

        query.addAuth(self.client_id, self.user_id)
        query.addQuery('ALBUM_SEARCH')
        query.addQueryMode('SINGLE_BEST_COVER')

        if artist:
            query.addQueryTextField('ARTIST', utils.cleanup_string(artist))
        if album:
            query.addQueryTextField('ALBUM_TITLE', utils.cleanup_string(album))
        if track:
            query.addQueryTextField('TRACK_TITLE', utils.cleanup_string(track))

        query.addQueryOption(
            'SELECT_EXTENDED',
            'COVER,REVIEW,ARTIST_BIOGRAPHY,ARTIST_IMAGE,ARTIST_OET,MOOD,TEMPO,LINK'
        )
        query.addQueryOption(
            'SELECT_DETAIL',
            'GENRE:3LEVEL,MOOD:2LEVEL,TEMPO:3LEVEL,ARTIST_ORIGIN:4LEVEL,ARTIST_ERA:2LEVEL,ARTIST_TYPE:2LEVEL'
        )
        queryXML = query.tostring()

        return self._http_query(queryXML)
