# coding: utf-8
import os
import sys
import signal
import multiprocessing
import time
import peewee

import argparse
import warnings
warnings.simplefilter('ignore')

from lima import config
import lima.log
import logging
from logging import NullHandler
import logging.handlers

from lima import system

import migrator
from lima.tools import tmdb
from lima.tools import segment
from lima.meta import storage

from lima import db
from lima.web import webapp, LimaWSGIRefServer

from lima import version

from lima.fs import FileSystem as fs

LOG_FORMAT = u'%(asctime)s - %(levelname)s - %(name)s - %(message)s'
LOG_DATE_FORMAT = u"%Y-%m-%d %H:%M:%S"
logger = logging.getLogger()

# force the python binary path to be added at the PATH
os.environ['PATH'] = os.path.dirname(
    sys.executable) + os.pathsep + os.environ['PATH']

# set process priority to the lower possible
import psutil
p = psutil.Process(os.getpid())
try:
    # windows
    p.nice(psutil.BELOW_NORMAL_PRIORITY_CLASS)
except Exception:
    # *nix
    p.nice(19)


class LimaMetaMain(object):
    def __init__(self):
        self._init_args_()

    def _init_args_(self):
        """ Initialize argparse """
        self.args_parser = argparse.ArgumentParser()
        self.args_parser.add_argument('-d', '--debug',
                                      dest='debug',
                                      help='Process in debug mode',
                                      action='store_true')

        self.args_parser.add_argument('--log-fs',
                                      dest='logfs',
                                      help='Log filesystem operation',
                                      action='store_true')
        self.args_parser.add_argument('--log-sql',
                                      dest='logsql',
                                      help='Log sql operation operation',
                                      action='store_true')

        self.args_parser.add_argument('--log',
                                      dest='logfile',
                                      help='Log to file')
        self.args_parser.add_argument('--log-debug',
                                      dest='logfile_debug',
                                      help='Log to file in debug mode',
                                      action='store_true')
        self.args_parser.add_argument('--no-log-stdout',
                                      dest='nolog_stdout',
                                      help='Do not log to stdout or stderr',
                                      action='store_true')
        self.args_parser.add_argument(
            '-t', '--threads',
            dest='nb_threads',
            help='Number of tasks to process in paralell',
            default=max((multiprocessing.cpu_count()) / 2, 1),
            type=int)

        fs_group = self.args_parser.add_mutually_exclusive_group(required=True)

        fs_group.add_argument('-f', '--fs-port',
                              dest='fs_port',
                              help='Port to use')

        fs_group.add_argument('--fs-unix-socket',
                              dest='fs_unix_socket',
                              help='Unix Domain socket to use')

        self.args_parser.add_argument('-p', '--port',
                                      dest='web_port',
                                      help='Port to use',
                                      required=True)

        # Disable loop file feeder
        self.args_parser.add_argument('-v', '--verbose',
                                      action='store_true',
                                      dest='verbose')

        self.args_parser.add_argument('--disable-fs',
                                      dest='disable_fs_watcher',
                                      help='Disable fs watcher',
                                      action='store_true')
        self.args_parser.add_argument('--disable-fp',
                                      dest='disable_file_processor',
                                      help='Disable file processor',
                                      action='store_true')
        self.args_parser.add_argument('--disable-tasks',
                                      dest='disable_tasks_worker',
                                      help='Disable tasks runner',
                                      action='store_true')

        self.args_parser.add_argument('--disable-index',
                                      dest='disable_indexer',
                                      help='Disable file indexer',
                                      action='store_true')
        self.args_parser.add_argument('--indexer-delay',
                                      dest='indexer_delay',
                                      help='Indexer collector delay in seconds',
                                      type=int,
                                      default=60, )
        self.args_parser.add_argument(
            '--indexer-threshold',
            dest='indexer_files_threshold',
            help='Indexer file threshold to force indexation',
            type=int,
            default=10, )

        self.args_parser.add_argument('--db-root',
                                      dest='db_root',
                                      help='Database root directory',
                                      required=True)

        self.args_parser.add_argument('--disable-gc',
                                      dest='disable_gc',
                                      help='Disable meta garbage collector',
                                      action='store_true')
        self.args_parser.add_argument(
            '--gc-delay',
            dest='gc_delay',
            help='Garbage collector delay in seconds',
            type=int,
            default=3600,  # default : 1 hour
        )
        self.args_parser.add_argument('--gc-fake',
                                      dest='gc_fake',
                                      help='GC won\'t delete metas',
                                      default=False,
                                      action='store_true')

    def _configure_logger(self, args):
        # create formatter and add it to the handlers
        formatter = logging.Formatter(LOG_FORMAT, datefmt=LOG_DATE_FORMAT)
        logger = logging.getLogger()
        logger.handlers = []
        if args.debug:
            logger.setLevel(logging.DEBUG)
        else:
            logger.setLevel(logging.INFO)
        # add the handlers to the logger
        if args.logfile:
            # create file handler which logs even debug messages
            fh = lima.log.GzipRotatingFileHandler(args.logfile,  # 2 Mo
                                                      maxBytes=2 << 20,
                                                      backupCount=5,
                                                      encoding='utf-8')
            if args.logfile_debug:
                fh.setLevel(logging.DEBUG)
            else:
                fh.setLevel(logging.INFO)
            fh.setFormatter(formatter)
            logger.addHandler(fh)
        if not args.nolog_stdout:
            # create console handler (stderr on windows) with a higher log level
            ch = logging.StreamHandler(sys.stderr if system.is_windows() else sys.stdout)
            ch.setLevel(logging.DEBUG if args.debug else logging.INFO)
            ch.setFormatter(formatter)
            logger.addHandler(ch)

        # Disable fsonrpc logging
        logger_fsonrpc = logging.getLogger('fsonrpc')
        logger_fsonrpc.propagate = False
        logger_fsonrpc.addHandler(NullHandler())

        logger_redo = logging.getLogger('redo')
        logger_redo.propagate = False
        logger_redo.addHandler(NullHandler())
        # create console handler and set level to info

        if not args.logfs:
            logger_limafs = logging.getLogger('lima.fs')
            logger_limafs.propagate = False
            logger_limafs.addHandler(NullHandler())

        if not args.logsql:
            logger_sql = logging.getLogger('peewee')
            logger_sql.propagate = False
            logger_sql.addHandler(NullHandler())
            logger_sql = logging.getLogger('lima.db')
            logger_sql.propagate = False
            logger_sql.addHandler(NullHandler())

        logging.getLogger('guessit').propagate = False
        logging.getLogger('guessit').addHandler(NullHandler())

        logging.getLogger('segment').propagate = False
        logging.getLogger('segment').addHandler(NullHandler())

        logging.getLogger('requests').propagate = False
        logging.getLogger('requests').addHandler(NullHandler())

    def main(self):
        signal.signal(signal.SIGINT, self.kill)

        args = self.args_parser.parse_args()
        self._configure_logger(args)
        logger = logging.getLogger(__name__)
        logger.info("Version\t:\t%s", version.full_version)
        logger.info("Commit\t:\t%s\t%s", version.short_commit, version.commit)
        logger.info("Tag\t:\t%s", version.tag)

        config_obj = config.Configuration(args.db_root)
        config.upgrade(config_obj)

        system.configure()
        system.set_process_priority()

        if args.fs_unix_socket:
            fs.configure(unix=args.fs_unix_socket)
        else:
            fs.configure(port=args.fs_port)

        segment.configure(
                "Lima Meta",
                config_obj.get("segmentio", "api_key", True),
                version.full_version,
                args.debug
        )

        db.configure(args.db_root)
        from lima.meta.scheduler import indexer
        indexer.configure()
        from lima.meta.scheduler import killswitch
        killswitch.configure()

        storage.configure()
        tmdb.configure(args.db_root, True)

        segment.track(event="Starting")

        # as the app uses the db, we have to import it after we have configured
        # the database root path
        from lima.meta.scheduler.scheduler import LimaMetaApp
        migrator_kwargs = {
            'print': args.debug,
            'database': db.get_database(),
            'migration_module': "lima.migrations",
            'directory': os.path.join(
                os.path.dirname(os.path.realpath(__file__)), "..",
                "migrations"),
        }

        try:
            migrator.migrate(**migrator_kwargs)
        except peewee.DatabaseError as e:
            if "database disk image is malformed" in e.message:
                # this is very bad, main sqlite db is corrupted.
                # only safe solution : delete the file, let the exception
                # be raised and pray for next run to work
                logger.error(u"Database corrupted")
                segment.track_error(event="Database corrupted")
                try:
                    try:
                        db.get_database().close()
                    except Exception:
                        logger.exception(u"Could not close corrupted db")
                    os.unlink(db.get_db_path())
                except Exception:
                    logger.exception(u"Could not remove corrupted db")
                    segment.track_error(event="Database corrupted : could not remove db file")
                    raise
            raise

        except Exception:
            logger.exception(u"Migration error")
            segment.track_error(event="Migration error")
            raise
        # WSGI server
        self.server = LimaWSGIRefServer(host='localhost', port=args.web_port)
        # Lima meta app

        self.app = LimaMetaApp(
            config_obj,
            pool_size=args.nb_threads,
            disable_fs_watcher=args.disable_fs_watcher,
            disable_file_processor=args.disable_file_processor,
            disable_tasks_worker=args.disable_tasks_worker,
            disable_gc=args.disable_gc,
            gc_delay=args.gc_delay,
            gc_fake=args.gc_fake,
            disable_indexer=args.disable_indexer,
            indexer_delay=args.indexer_delay,
            indexer_files_threshold=args.indexer_files_threshold,
            is_debug=args.debug, )
        try:
            # Adds a reference to the web controller to be able to start/stop/resume/kill it.
            webapp.config['lima_app'] = self.app
            # We need to keep a reference on the webapp to be able to kill it if anythng goes wrong.
            self.app.webapp = webapp
            self.app.start()
            webapp.run(server=self.server, quiet=not args.debug)
        except (KeyboardInterrupt):
            logger.info(u"Stoping Daemon gracefully")
            self.kill()
        logger.info(u"Bye")

    def kill(self, signum=None, frame=None):
        try:
            self.app.kill()
            segment.track(event="Killed")
            segment.flush()
        except:
            logger.exception(u"Error killing lima app")
        # ubber uggly : we should kill the wsgi server
        # but calling shutdown justs hangs
        time.sleep(0.1)
        sys.exit()


def lima_meta():
    LimaMetaMain().main()


if __name__ == "__main__":
    lima_meta()
