# coding: utf-8
import os
import settings

import logging
logger = logging.getLogger(__name__)

from ConfigParser import ConfigParser, NoSectionError, NoOptionError
from StringIO import StringIO

import base64
from itertools import cycle, izip


def xor(ss, key):
    key = cycle(key)
    return ''.join(chr(ord(x) ^ ord(y)) for (x,y) in izip(ss, key))

class ConfigurationError(Exception):
    pass


class Configuration(object):

    def __init__(self, user_config_folder=None):
        self.user_config_folder = user_config_folder
        self.sys_config = None
        self.user_config = None

        self.load_config()

        # TODO : should obfuscate a minimum to avoid a simple 'strings' commands
        # on the pyc.
        self.k='qtIWho_wIcdeHTpkMeCZh83PYgmBsgACNKFY_NlOaa0='

    def reload_config(self):
        logger.debug(u"Reloading configuration")
        self.load_config()

    def load_config(self):
        self.sys_config = ConfigParser()
        self.user_config = ConfigParser()

        # We allow the main config file to be overrided
        system_config_override = os.path.join(
            settings.CONFIG['system_override']['location'],
            settings.CONFIG['system_override']['file_name']
        )
        # If the override file exists, use it
        if os.path.exists(system_config_override):
            self.system_config_folder = settings.CONFIG[ 'system_override']['location']
            self.system_config_file = system_config_override
            logger.info(u"Configuration overrided from %s", self.system_config_file)
        else:
            # if the override file does not exists, use the system config
            self.system_config_folder = settings.CONFIG[ 'system']['location']
            self.system_config_file = os.path.join(
                self.system_config_folder,
                settings.CONFIG['system']['file_name']
            )

        try:
            if self.user_config_folder is None:
                self.user_config_folder = self.system_config_folder
            self._create_user_config_folder()
        except:
            self.user_config_folder = self.system_config_folder

        self.user_config_file = os.path.join(
            self.user_config_folder,
            settings.CONFIG['user']['file_name']
        )
        self._ensure_path(self.system_config_folder)
        self._ensure_path(self.system_config_file)
        self._ensure_path(self.user_config_folder)
        self.sys_config.readfp(open(self.system_config_file))
        logger.debug(
            "System configuration loaded from '%s'", self.system_config_file)

        if os.path.exists(self.user_config_file):
            self.user_config.readfp(open(self.user_config_file))
            logger.debug(
                "User configuration loaded from '%s'", self.user_config_file)
        else:
            self.user_config.readfp(StringIO())

    def _create_user_config_folder(self):
        if not os.path.exists(self.user_config_folder):
            try:
                os.makedirs(self.user_config_folder)
            except (Exception,) as e:
                logger.exception("Could not create user configuration folder")
                raise

    def _ensure_path(self, path):
        if not os.path.exists(path):
            raise ConfigurationError("Path '%s' does not exists" % path)

    def get_system(self, section, option, decrypt=False):
        try:
            value = self.sys_config.get(section, option)
            if value == 'None':
                value = None
            if decrypt is True and value is not None:
                value=xor(base64.b64decode(value), self.k)
            return value
        except(NoSectionError, NoOptionError):
            return None

    def get_user(self, section, option, decrypt=False):
        try:
            value = self.user_config.get(section, option)
            if value == 'None':
                value = None
            if decrypt is True and value is not None:
                value=xor(base64.b64decode(value), self.k)
            return value
        except(NoSectionError, NoOptionError):
            return None

    def set(self, section, option, value, encrypt=False):
        if section not in self.user_config.sections() and section != "DEFAULT":
            logger.debug("Section %s does not exists yet. Create a new one", section)
            self.user_config.add_section(section)
        try:
            if encrypt and value is not None:
                value = base64.b64encode(xor(value, self.k))
            return self.user_config.set(section, option, value)
        # should not happen, we just created the section
        except(NoSectionError):
            logger.exception("Could not find %s : %s", section, option)
            return None

    def get(self, section, option, decrypt=False):
        value = self.get_user(section, option, decrypt)
        # fixes bad None value that may have been saved
        if value is None:
            value = self.get_system(section, option, decrypt)
        if value is None:
            logger.warn("Could not find value for %s.%s", section, option)
        return value

    def getboolean(self, section, option):
        value = self.get(section, option)
        if value is not None:
            if value.lower() in ('1', 'true', 'yes', 'y'):
                return True
            elif value.lower() in ('0', 'false', 'no', 'n'):
                return False
            else:
                raise ValueError("'%s' could not be converted to boolean" % value)

    def save(self):
        with open(self.user_config_file, "w") as fconfig:
            self.user_config.write(fconfig)
        logger.debug("User config saved")


import threading

class SharedConfig(dict):
    def __init__(self, restrict_keys_flag_dirty=None, configuration_dirty_handler=None, *args, **kwargs):
        """
        SharedConfig object, basically a dict with an extra 'dirty' attribute to known when the config had changed.
        restrict_keys_flag_dirty can be set to a list of key which will mark the config as dirty.
        """
        super(SharedConfig, self).__init__(*args, **kwargs)
        if restrict_keys_flag_dirty is None:
            restrict_keys_flag_dirty = []
        self.lock = threading.Lock()
        self.restrict_keys_flag_dirty = restrict_keys_flag_dirty
        self.configuration_dirty_handler = configuration_dirty_handler
        self.is_dirty = False

    def update(self, *args, **kwargs):
        with self.lock:
            super(SharedConfig, self).update(*args, **kwargs)
            self.is_dirty = True
            if self.configuration_dirty_handler:
                self.configuration_dirty_handler(self)


    def __setitem__(self,key, value):
        if key in self.keys():
            # The key is already present : compare with old value
            if self[key] != value:
                if not self.restrict_keys_flag_dirty or key in self.restrict_keys_flag_dirty:
                    self.is_dirty = True
        elif not self.restrict_keys_flag_dirty or key in self.restrict_keys_flag_dirty:
            self.is_dirty = True
        super(SharedConfig, self).__setitem__(key, value)
        if self.is_dirty:
            # Configuraiton is dirty : call the handler
            if self.configuration_dirty_handler:
                self.configuration_dirty_handler(self)





def upgrade(config_obj):
    try:
        sys_cfg_version = int(config_obj.get_system('config', 'version'))
    except TypeError:
        # was None
        sys_cfg_version = -1
    except ValueError:
        # not an integer
        sys_cfg_version = -10
    try:
        user_cfg_version = int(config_obj.get_user('config', 'version'))
    except TypeError:
        # was None
        user_cfg_version = -1
    except ValueError:
        # not an integer
        user_cfg_version = -10

    if user_cfg_version == sys_cfg_version :
        logger.info("Configuration is up to date (v%s)", sys_cfg_version)
        # nothing to do
        return

    # migrate configurations
    if user_cfg_version < sys_cfg_version :
        # user did not had version, this is the first round
        if user_cfg_version == -1:
            # Remove invalidated gracenote web token
            try:
                if config_obj.user_config.remove_option('gracenote', 'web_user_token'):
                    logger.info('Removed old gracenote user token')
            except (NoSectionError, NoOptionError):
                pass

        if sys_cfg_version > 0:
            # set the current version and save the file
            config_obj.set('config', 'version', sys_cfg_version)
        config_obj.save()
    # ensure configuration is reloaded
    config_obj.reload_config()
