# coding: utf-8

import os
import peewee
import sys

from lima.config import Configuration

config = Configuration()

# DB root path
_db_root = None
# database instance
_db = None


class DbNoSpaceError(peewee.OperationalError):
    pass


class LimaSqliteDatabase(peewee.SqliteDatabase):
    def sql_error_handler(self, exception, sql, params, require_commit):
        err_msg = getattr(exception, 'message')
        if 'database or disk is full' in (err_msg or '').lower():
            raise DbNoSpaceError, DbNoSpaceError(err_msg), sys.exc_info()[2]
        # Re-raise the exception.
        raise sys.exc_info()[0], sys.exc_info()[1], sys.exc_info()[2]


def configure(db_root):
    """Set db_root as the root folder for database"""
    global _db_root
    _db_root = db_root


def get_db_path():
    """Returns the database full name """
    global _db_root
    if _db_root is None:
        # allow to pass the value from an environment variable
        # typically for tests
        # this is due to the fact that peewee NEED the db to be
        # created before declaring models
        _db_root = os.environ.get("LIMA_META_DB_ROOT", None)
        if _db_root is None:
            raise ValueError("DB root is not defined.")
    return os.path.expanduser(os.path.join(_db_root,
                                           config.get("database", "filename")))


def ensure_db_folder():
    db_path = os.path.basename(get_db_path())
    if not os.path.exists(db_path) and os.path.isdir(db_path):
        os.makedirs(db_path)


def get_database():
    global _db
    if _db is None:
        ensure_db_folder()
        _db = LimaSqliteDatabase(get_db_path(), threadlocals=True, timeout=10)
    return _db


from contextlib import contextmanager
import time

import logging
logger = logging.getLogger(__name__)


@contextmanager
def database(debug_tag=None):
    logger.debug(u"[%s] Acquiring DB", debug_tag or 'no tag')
    start = time.time()
    try:
        db = get_database()
        db.connect()
        yield db
    finally:
        db.close()
        stop = time.time()
        result = stop - start
        logger.debug(u"[%s] DB unlocked in %s sec", debug_tag or 'no tag',
                     result)


@contextmanager
def atomic(debug_tag=None):
    try:
        with database(debug_tag) as _db:
            yield _db.atomic()
    finally:
        pass


@contextmanager
def transaction(debug_tag=None):
    try:
        with database(debug_tag) as _db:
            yield _db.transaction()
    finally:
        pass


def maintenance(db):
    db.get_conn().execute('VACUUM;')
    db.get_conn().execute('REINDEX;')
