# coding: utf-8

import os
import sys
import errno
import socket


from itertools import tee
from itertools import ifilter

from lowlevel.filesystem import Filesystem
from lowlevel.rpc import RPCClient
from lowlevel.exc import ErrnoException
from limafs import LimaFilesystem

from limafs import (
    LimaFsError,
    LimaFsTimeoutError,
    LimaFsConnectionRefusedError,
    LimaFsNotReadyError,
    LimaFsDoesnotExistError,
    LimaFsNotDirectoryError,
    LimaFsIsDirectoryError,
    LimaFsNoSpaceError,
    LimaFsNotReadableError,
    LimaFsNotMountedError,
    LimaFsMemoryError,
    LimaFsDeviceNotconfigured,
)


import threading
import select

import logging
logger = logging.getLogger(__name__)

import re
# Conflict pattern:
# will match filename in the form
# filename (core N.X).ext
# the extension is optional
PATTERN_CONFLICT_FILENAME = "(?P<filename>.*)(\s+(?P<conflict_flag>\(.*\)))(?P<ext>\.[a-zA-Z]+){0,1}"
RE_CONFLICT_FILENAME = re.compile(PATTERN_CONFLICT_FILENAME)


class FileWrapper(object):

    """ A File wrapper that will handle lowlevel errors and convert them in fs
    errors """

    def __init__(self, fh):
        self.fh = fh

    def __getattribute__(self, key):
        """
        Encapsulate fh methods call to handle ErrnoException and convert them
        to LimaFserror """
        v = object.__getattribute__(object.__getattribute__(self, 'fh'), key)

        def __call_wrapper__(self, *args, **kwargs):
            try:
                return v(*args, **kwargs)
            except Exception:
                exc_type, excep, traceback = sys.exc_info()
                if issubclass(type(excep), ErrnoException):
                    if excep.is_nomemory():
                        raise LimaFsMemoryError, LimaFsMemoryError(excep), traceback
                    elif excep.is_timeout():
                        raise LimaFsTimeoutError, LimaFsTimeoutError(excep), traceback
                    elif excep.is_conn_refused():
                        raise LimaFsConnectionRefusedError, LimaFsConnectionRefusedError(excep), traceback
                    elif excep.is_not_ready():
                        raise LimaFsNotReadyError, LimaFsNotReadyError(excep), traceback
                    elif excep.is_file_does_not_exists():
                        raise LimaFsDoesnotExistError, LimaFsDoesnotExistError(excep), traceback
                    elif excep.is_notdir():
                        raise LimaFsNotDirectoryError, LimaFsNotDirectoryError(excep), traceback
                    elif excep.is_dir():
                        raise LimaFsIsDirectoryError, LimaFsIsDirectoryError(excep), traceback
                    elif excep.is_nospace():
                        raise LimaFsNoSpaceError, LimaFsNoSpaceError(excep), traceback
                    elif excep.is_not_readable():
                        raise LimaFsNotReadableError, LimaFsNotReadableError(excep), traceback
                    elif excep.is_not_configured():
                        raise LimaFsDeviceNotconfigured, LimaFsDeviceNotconfigured(excep), traceback
                    else:
                        raise LimaFsError, LimaFsError(excep), traceback
                else:
                    raise exc_type, excep, traceback
        if callable(v):
            # the attibute is a method, we execute it in the wrapper context
            return __call_wrapper__.__get__(self.fh)
        return v

    def __enter__(self, *args, **kwargs):
        # here we return self instead of self.fh to allo __getattribute__ to be
        # called
        return self

    def __exit__(self, *args, **kwargs):
        return object.__getattribute__(self, 'fh').__exit__(*args, **kwargs)


class FCFS(LimaFilesystem):

    """ Class representing a FCFS file system """

    def __init__(self, yml_config):
        self.yml_config = yml_config
        self.path = u'/'
        self.thread_fs = threading.local()

        # Dict containing path -> meta
        self.metas = {}
        # Dict containing path -> lock for metadatas manipulation
        self.metas_locks = {}

        self._kill_event = threading.Event()

    def __create_connection(self):
        if self.unix:
            sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
            sock.settimeout(10)
            sock.connect(self.unix)
            self.thread_fs.socket = sock
        else:
            self.thread_fs.socket = socket.create_connection((self.host, self.port), timeout=10)

    def __enter__(self):

        if not hasattr(self, 'port') or (self.port is None and self.unix is None):
            raise ValueError('You must configure either `port` or `unix`.')

        # Check socket status / connectivity
        self.__check_socket__()

        if not hasattr(self.thread_fs, 'socket') or self.thread_fs is None:

            try:
                self.__create_connection()
            except socket.timeout as e:
                    traceback = sys.exc_info()[2]
                    raise LimaFsTimeoutError, LimaFsTimeoutError(e), traceback
            except (Exception,) as e:
                if getattr(e, "errno", None) is not None:
                    traceback = sys.exc_info()[2]
                    raise LimaFsConnectionRefusedError, LimaFsConnectionRefusedError(e), traceback
                else:
                    raise

        if not hasattr(self.thread_fs, 'client'):
            self.thread_fs.client = RPCClient(self.thread_fs.socket, self.yml_config, timeout=11)
        if not hasattr(self.thread_fs, 'fs'):
            self.thread_fs.fs = Filesystem(self.thread_fs.client)
        return self

    def __check_socket__(self):
        if hasattr(self.thread_fs, 'socket'):
            try:
                _, readys, _ = select.select(
                    [], [self.thread_fs.socket], [], 1
                )
                if len(readys) == 0:
                    # Can't write on socket.
                    # We close it and delete it
                    self.__clear_socket_client_fs__()
            except (socket.error, ) as e:
                if e.errno == errno.EBADF:
                    # Server likely closed the connection
                    # We close it and delete it
                    self.__clear_socket_client_fs__()
                else:
                    raise

    def __clear_socket_client_fs__(self):
        if hasattr(self.thread_fs, 'socket'):
            # self.thread_fs.socket.close()
            # logger.debug(u"Socket closed")
            del self.thread_fs.socket
        if hasattr(self.thread_fs, 'client'):
            del self.thread_fs.client
        if hasattr(self.thread_fs, 'fs'):
            del self.thread_fs.fs

    def __exit__(self, *exc):
        exc_type, excep, traceback = exc
        if excep is not None:
            if issubclass(type(excep), ErrnoException):
                if excep.is_nomemory():
                    raise LimaFsMemoryError, LimaFsMemoryError(excep), traceback
                elif excep.is_timeout():
                    self.__clear_socket_client_fs__()
                    raise LimaFsTimeoutError, LimaFsTimeoutError(excep), traceback
                elif excep.is_conn_refused():
                    self.__clear_socket_client_fs__()
                    raise LimaFsConnectionRefusedError, LimaFsConnectionRefusedError(excep), traceback
                elif excep.is_not_ready():
                    raise LimaFsNotReadyError, LimaFsNotReadyError(excep), traceback
                elif excep.is_file_does_not_exists():
                    raise LimaFsDoesnotExistError, LimaFsDoesnotExistError(excep), traceback
                elif excep.is_notdir():
                    raise LimaFsNotDirectoryError, LimaFsNotDirectoryError(excep), traceback
                elif excep.is_dir():
                    raise LimaFsIsDirectoryError, LimaFsIsDirectoryError(excep), traceback
                elif excep.is_nospace():
                    raise LimaFsNoSpaceError, LimaFsNoSpaceError(excep), traceback
                elif excep.is_not_readable():
                    raise LimaFsNotReadableError, LimaFsNotReadableError(excep), traceback
                elif excep.is_not_configured():
                    raise LimaFsDeviceNotconfigured, LimaFsDeviceNotconfigured(excep), traceback
                else:
                    raise LimaFsError, LimaFsError(excep), traceback
            else:
                raise exc_type, excep, traceback

    @property
    def fs(self):
        return self.thread_fs.fs

    @property
    def mountpoint(self):
        if not hasattr(self, u'_mountpoint'):
            self._mountpoint = self.status[u'mount_point']
        return self._mountpoint

    @property
    def status(self):
        with self:
            return self.fs.status()

    @property
    def address(self):
        if not hasattr(self, '_address'):
            uri = self.status['uri']
            self._address = uri.split('//')[1].split(':')[0].split('.')[0]
        return self._address

    def get_system_path(self, path):
        if not os.path.exists(self.mountpoint):
            raise LimaFsNotMountedError(u"mountpoint '%s' does not exists" % self.mountpoint)
        return os.path.normpath(
            u'%s/%s' % (self.mountpoint, path)
        )

    def configure(self, port=None, host='127.0.0.1', unix=None):
        self.__clear_socket_client_fs__()
        if not (port or unix):
            raise ValueError('You must configure either `port` or `unix`.')
        self.port = port
        self.host = host
        self.unix = unix

    def hasmeta(self, path, key):
        with self:
            return self.getmeta(path, key) is not None

    def open(self, path, mode=u'r'):
        with self:
            return FileWrapper(self.fs.open(path, mode))

    def open_fork(self, path, fork=u'default', mode=u'r'):
        with self:
            if u'w' in mode and fork != u"default":
                raise ValueError(u"Writing in a non default fork is disabled")
            return FileWrapper(self.fs.open_fork(path, fork, mode))

    def open_fork_meta(self, path, fork=u'meta', mode=u'r'):
        with self:
            fork_meta_path = self.get_meta_fork_path(path, fork)
            dirname = os.path.dirname(fork_meta_path)
            self.fs.makedirs(dirname)
            fork = self.fs.open(fork_meta_path, mode)
            setattr(fork, 'path', fork_meta_path)
            return FileWrapper(fork)

    def hasfork(self, path, fork):
        with self:
            try:
                return self.fs.stat_fork(path, fork) \
                    is not None
            except ErrnoException as e:
                if e.errno == errno.ENOENT:
                    return False
                else:
                    raise

    def ensure_directory(self, path):
        with self:
            if not self.fs.exists(path):
                self.fs.makedirs(path)

    def join(self, *items):
        return self.path.join(items)

    def stat(self, path, fork=None):
        with self:
            return self.fs.stat(path, fork)

    def bulkstat(self, path):
        with self:
            for exc, st in self.fs.bulkstat(path):
                if exc is not None:
                    if exc.is_file_does_not_exists:
                        exc = LimaFsDoesnotExistError(exc)
                    else:
                        exc = LimaFsError(exc)
                yield exc, st

    def unlink(self, path):
        with self:
            return self.fs.unlink(path)

    def unlink_fork(self, path, fork):
        with self:
            return self.fs.unlink_fork(path, fork)

    def isdir(self, path):
        with self:
            return self.fs.isdir(path)

    def pin(self, is_pined, path):
        with self:
            return self.fs.pin(is_pined, path)

    def rename(self, path_from, path_to):
        with self:
            return self.fs.rename(path_from, path_to)

    def isfile(self, st=None, path=None):
        with self:
            if path is not None:
                stat = self.stat(path)
            else:
                stat = st
            if stat is not None:
                return stat.isreg()
            return False

    def exists(self, path):
        with self:
            return self.fs.exists(path)

    def listdir(self, path):
        with self:
            return self.fs.listdir(path)

    def listdirstat(self, path):
        with self:
            return self.fs.listdirstat(path)

    def symlink(self, path_from, path_to):
        with self:
            return self.fs.symlink(path_from, path_to)

    def readlink(self, path):
        with self:
            return self.fs.readlink(path)

    def fork_path(self, fork_ids, flags=0x0000):
        with self:
            return self.fs.fork_path(fork_ids, flags)



    def resolve_symlink(self, path, fallback_last_resolved=False):
        """ Recursively resolve symlinks
        :param fallback_last_valid: if true and the symlink point to file that does not exists,
        will return the last resolved path instead of returning None
        """
        st = self.stat(path)
        try:
            while st.islink():
                path = self.readlink(path)
                st = self.stat(path)
        except LimaFsDoesnotExistError:
            if fallback_last_resolved:
                return path
            return None
        return path

    def dirname(self, path):
        i = path.rfind(u'/')+1
        return path[:i]

    def basename(self, path):
        i = path.rfind(u'/')+1
        return path[i:]

    def is_empty_dir(self, path):
        """ Return true if the directory is empty
        """
        # there should be a better way
        return len( set( ['.', '..'] )  | set( self.listdir(path) ))  == 2

    def walk(self, *args, **kwargs):
        """ Return an iterator on all path for the given root.

        :param root: Root from where to start the walk
        :param ignore_pattern: Regular expression pattern to exclude file from
        the walk.
        """
        for path, st in self.walk_stat(*args, **kwargs):
            yield path

    def walk_stat(self, root, ignore_pattern=None, since_mrtime=-1, exclude_dirs=False, recursive=True):
        """ Return an iterator on all path for the given root.

        :param root: Root from where to start the walk
        :param ignore_pattern: Regular expression pattern to exclude file from
        the walk.
        :param since_mrtime: Return items that have been modified since the
        given mrtime (default : 0)
        """
        ignore_re = None

        if ignore_pattern is not None:
            ignore_re = re.compile(ignore_pattern)

        with self:
            try:
                list_dir_stat = self.fs.listdirstat(root)
                # find the root stat (path == u".")
                try:
                    path, stat = [(path, stat) for path, stat in list_dir_stat if path == u"."][0]
                except IndexError:
                    logger.error(u"Filesystem inconsistency. Could not get root stat")
                    raise StopIteration
                try:
                    list_dir_stat.remove((path, stat))
                except ValueError:
                    # this should not happen
                    logger.error(u"%s already removed", path)

                root_stat = stat
                mrtime = root_stat.st_mrtime
                # we only want nodes with mrtime superior to the last mrtime
                if mrtime <= since_mrtime:
                    logger.debug(
                        u"%s - ignoring mrtime %s - root mrtime : %s", root, mrtime, since_mrtime
                    )
                    raise StopIteration

                # if root match ignore_pattern stop the iteration
                if ignore_re and ignore_re.match(root) is not None:
                    raise StopIteration
                logger.debug(u"%s - changed", root)
                # if exclude_dirs is False:
                #    yield root, root_stat

                # here we start a the second element as the first is the root
                # stat result
                for path, st in list_dir_stat:
                    if self._kill_event.isSet():
                        logger.info(u"killed")
                        raise StopIteration

                    # ignore current directory and parent
                    # should be moved in ignore patterrns
                    if path in [u'.', u'..']:
                        continue

                    # Do not yield files if mrtime is inferior than reference
                    # mrtime
                    if st is not None and st.st_mrtime <= since_mrtime:
                        continue
                    # Concatenate separator
                    if not path.startswith(self.path) and root != self.path:
                        path = self.path + path
                    fullname = root + path

                    # if fullname match ignore_pattern stop the iteration
                    if ignore_re and ignore_re.match(fullname) is not None:
                        continue

                    logger.debug(u"%s - changed", path)

                    if self._kill_event.isSet():
                        logger.info(u"killed")
                        raise StopIteration
                    yield fullname, st

                    # If fullname is a directory, loop recursivly over it
                    # content
                    if st.isdir() and recursive:
                        for subdir, subst in self.walk_stat(fullname, ignore_pattern, since_mrtime, exclude_dirs):
                            if self._kill_event.isSet():
                                logger.info(u"killed")
                                raise StopIteration
                            yield subdir, subst

            except(LimaFsDoesnotExistError, LimaFsNotDirectoryError):
                # LimaFsDoesnotExistError : folder has been removed
                # LimaFsNotDirectoryError : folder has been replaced with
                # something which is not a directory, but has the same name
                raise StopIteration

    def walk_stat_filter(self, *args, **kwargs):
        try:
            path_filter = kwargs.pop('path_filter')
        except KeyError:
            path_filter = None

        for p, st in self.walk_stat(*args, **kwargs):
            if path_filter and path_filter(p, st):
                yield p, st
            elif not path_filter:
                yield p, st

    def isort_walk(self, iwalk):
        for x in self._isort_walk(iwalk):
            yield x

    def _isort_walk(self, iterable):
        iterable = iter(iterable)
        try:
            pivot = iterable.next()
        except:
            return

        a, b = tee(iterable)
        for x in self._isort_walk(ifilter(lambda item: item[1].st_mrtime <pivot[1].st_mrtime, a)):
            yield x
        yield pivot
        for y in self._isort_walk(ifilter(lambda item: item[1].st_mrtime >=pivot[1].st_mrtime, b)):
            yield y


    def rmdirs(self, path):
        with self:
            self.fs.rmdirs(path)

    def rmdir(self, path):
        with self:
            self.fs.rmdir(path)

    def delete(self, path):
        """ Delete a file or a directory """
        try:
            self.unlink(path)
            return
        except LimaFsIsDirectoryError:
            pass
        try:
            self.rmdir(path)
            return
        except LimaFsNotDirectoryError:
            # should not happen because of the first unlink
            pass


    def local(self, path=None, st=None):
        """ Return the result of st_local for a file"""
        if path is not None:
            st = self.stat(path)
        local = getattr(st, 'st_local', 0)
        return local

    def local_fork(self, path, fork):
        """ Return the result of st_local for a fork"""
        try:
            st = self.stat(path, fork)
            return self.local(st=st)
        except ErrnoException as e:
            if e.errno == errno.ENOENT:
                return None
            else:
                raise

    def is_empty(self, path, fork=None):
        local = self.local_fork(path, fork)
        return local is None or local == 0

    def is_complete(self, st=None):
        local = st.st_local # getattr(st, 'st_local', 0)
        size = st.st_size #getattr(st, 'st_size', 0)
        return size == local and 0 not in (size, local)

    def is_content_local(self, st):
        # Check for local available content
        # First thing first, check if it's a regular file
        if st.isdir():
            return True
        # First thing first, check if local data is available
        if self.local(st=st) == 0:
            return False
        # Check if the file is complete on the local fs
        if not self.is_complete(st=st):
            return False
        return True

    def time(self, st):
        return st.st_mrtime

    def statvfs(self):
        with self:
            return self.fs.statvfs()

    def is_conflict(self, path):
        basename = self.basename(path)
        r = RE_CONFLICT_FILENAME.search(basename)
        if r:
            # filename matches conflict pattern
            return True
        return False

    def kill(self):
        self._kill_event.set()
