import errno
import os
from exc import ErrnoException, ErrnoTimeoutException

import logging
logger = logging.getLogger(__name__)

from redo import retriable
RETRY_CONFIG = {
    'attempts': 3,
    'sleeptime': 5,
    'jitter': 2,
    'retry_exceptions': (ErrnoTimeoutException,)
}

class StatResult(object):
    ##
    IFLNK=0120000
    IFREG=0100000
    IFDIR=0040000
    #ISUID 0004000 /* set user id on execution */
    #ISGID 0002000 /* set group id on execution */
    ISVTX=0001000


    def __new__(cls, v):
        if not v:
            return None

        self = super(StatResult, cls).__new__(cls)
        self.v = v
        return self

    def __getattr__(self, name):
        if name.startswith("st_"):
            if name in self.v:
                return self.v[name]

        raise AttributeError(name)

    def __repr__(self):
        return "<StatResult %r>" % (self.v, )

    def isreg(self):
        return self.st_mode & StatResult.IFREG == StatResult.IFREG

    def isdir(self):
        return self.st_mode & StatResult.IFDIR == StatResult.IFDIR

    def islink(self):
        return self.st_mode & StatResult.IFLNK == StatResult.IFLNK


class FileMode:
    READ      = 0x0200
    WRITE     = 0x0400
    APPEND    = 0x0800
    TRUNCATE  = 0x1000
    EXCLUSIVE = 0x2000
    CREATE    = 0x4000

    @classmethod
    def from_fopen(cls, fopen_flags):
        f = 0

        if "w" in fopen_flags or "a" in fopen_flags or "+" in fopen_flags:
            f |= cls.WRITE

        if "r" in fopen_flags or "+" in fopen_flags:
            f |= cls.READ

        if "a" in fopen_flags:
            f |= cls.APPEND

        if "w" in fopen_flags or "a" in fopen_flags:
            f |= cls.CREATE

        if "w" in fopen_flags:
            f |= cls.TRUNCATE

        if "x" in fopen_flags and f & cls.CREATE == cls.CREATE:
            f |= cls.EXCLUSIVE

        assert f & (cls.READ | cls.WRITE)

        return f



class FileMode:
    READ      = 0x0200
    WRITE     = 0x0400
    APPEND    = 0x0800
    TRUNCATE  = 0x1000
    EXCLUSIVE = 0x2000
    CREATE    = 0x4000



    @classmethod
    def from_fopen(cls, fopen_flags):
        f = 0

        if "w" in fopen_flags or "a" in fopen_flags or "+" in fopen_flags:
            f |= cls.WRITE

        if "r" in fopen_flags or "+" in fopen_flags:
            f |= cls.READ

        if "a" in fopen_flags:
            f |= cls.APPEND

        if "w" in fopen_flags or "a" in fopen_flags:
            f |= cls.CREATE

        if "w" in fopen_flags:
            f |= cls.TRUNCATE

        if "x" in fopen_flags and f & cls.CREATE == cls.CREATE:
            f |= cls.EXCLUSIVE

        assert f & (cls.READ | cls.WRITE)

        return f

class FileHandler(object):

    def __init__(self, fs, fh):
        super(FileHandler, self).__init__()

        self.fs = fs
        self.fh = fh
        self.offset = 0

        self._opened = True

    def tell(self):
        """Mimic standard python file tell """
        return self.offset

    def readline(self, num=None):
        """Mimic standard python file readline """
        buf = ''
        char = self.read(1)
        buf += char
        while char not in ('\n', '') and (num is None or len(buf)<num):
                char = self.read(1)
                buf += char
        return buf

    @retriable(**RETRY_CONFIG)
    def pread(self, offset, size=None):
        assert self._opened

        if size is None:
            size = self.stat().st_size - offset

        return self.fs.local_rpc.read([self.fh, size, offset])

    @retriable(**RETRY_CONFIG)
    def pwrite(self, offset, buf):
        assert self._opened
        return self.fs.local_rpc.write([self.fh, offset, buf])

    def read(self, size=None):

        r = self.pread(self.offset, size)

        self.offset += len(r)

        return r

    def write(self, buf):
        r = self.pwrite(self.offset, buf)

        self.offset += r

        return r

    @retriable(**RETRY_CONFIG)
    def stat(self):
        assert self._opened
        st = self.fs.local_rpc.fstat([self.fh])

        return StatResult(st)


    def seek(self, offset, whence=None):
        whence = whence if whence is not None else os.SEEK_SET
        size = self.stat().st_size

        if whence == os.SEEK_SET:
            self.offset = offset

        elif whence == os.SEEK_CUR:
            self.offset += offset
            self.offset = max(0, self.offset)

        elif whence == os.SEEK_END:
            if size + offset < 0:
                raise ErrnoException(errno.EINVAL)

            self.offset = offset + size

        else:
            raise ErrnoException(errno.EINVAL)

        if self.offset > size:
            self.offset = size

    @retriable(**RETRY_CONFIG)
    def close(self):
        if self._opened:
            self._opened = False
            self.fs.local_rpc.close([self.fh])

    def do_not_close(self):
        self._opened = False

    def __enter__(self):
        return self

    def __exit__(self, *exc):
        self.close()

    def __del__(self):
        self.close()


class Filesystem(object):

    def __init__(self, local_rpc):
        super(Filesystem, self).__init__()

        self.local_rpc = local_rpc

    @retriable(**RETRY_CONFIG)
    def status(self):
        return self.local_rpc.status([])

    @retriable(**RETRY_CONFIG)
    def listdir(self, path):
        return self.local_rpc.listdir([path])

    @retriable(**RETRY_CONFIG)
    def listdirstat(self, path, with_stat=None):
        nst = self.local_rpc.listdirstat([path])

        nst = [(name, StatResult(st)) for name, st in nst]

        return nst

    @retriable(**RETRY_CONFIG)
    def stat(self, path, fork=None):
        res = None
        if fork is None:
            st = self.local_rpc.stat([path])
            res= StatResult(st)
        else:
            res = self.stat_fork(path, fork)
        return res

    @retriable(**RETRY_CONFIG)
    def readlink(self, path):
        res = self.local_rpc.readlink([path])
        return res


    @retriable(**RETRY_CONFIG)
    def bulkstat(self, paths):
        stats = self.local_rpc.bulkstat(paths)
        for errno, st in stats:
            if errno == 0:
                exc = None
            else:
                exc = ErrnoException(-errno)
            yield exc, StatResult(st)


    @retriable(**RETRY_CONFIG)
    def open(self, path, flag, mode=None):

        if isinstance(flag, basestring):
            flag = FileMode.from_fopen(flag)

        mode = mode if mode is not None else 0777

        fh = self.local_rpc.open([path, flag, mode])

        return FileHandler(self, fh)

    @retriable(**RETRY_CONFIG)
    def pin(self, path, value=None):
        value = value if value is not None else True

        self.local_rpc.pin([path, value])

    @retriable(**RETRY_CONFIG)
    def mkdir(self, path, mode=None):
        mode = mode if mode is not None else 0755

        self.local_rpc.mkdir([path, mode])

    def makedirs(self, path, mode=None):
        """ Recurive mkdir"""
        full_folder_name=u""
        for folder in path.split(u'/'):
            try:
                full_folder_name = u'/'.join([
                    full_folder_name,
                    folder,
                ]) if full_folder_name != '' else u'/'+folder
                if full_folder_name != u'/' and not self.exists(full_folder_name):
                    self.mkdir(full_folder_name, mode)
            except (ErrnoException,) as e:
                if not e.is_exists():
                    raise

    @retriable(**RETRY_CONFIG)
    def rmdir(self, path):
        """remove a directory"""
        self.local_rpc.rmdir([path, ])

    def rmdirs(self, root):
        """ Recurive rm dir"""
        for path, st in self.listdirstat(root):
            if path in [u".", u".."]:
                continue
            fullpath = u"/".join([root, path])
            try:
                if st.isdir():
                    self.rmdirs(fullpath)
                    self.rmdir(fullpath)
                else:
                    self.unlink(fullpath)
            except (ErrnoException,) as e:
                if not e.is_file_does_not_exists():
                    raise
        self.rmdir(root)


    @retriable(**RETRY_CONFIG)
    def stat_fork(self, path, fork):
        logger.debug(u"stat_fork %s, %s", path, fork)
        if type(fork) == unicode:
            logger.warning(u"stat_fork use CMP_UNPACK_BIN_ALLOC_LEGACY. Encoding to str")
            fork = fork.encode('utf-8')
        st = self.local_rpc.stat_fork([path, fork])
        return StatResult(st)

    @retriable(**RETRY_CONFIG)
    def open_fork(self, path, fork, flag, mode=None):
        logger.debug(u"open_fork %s, %s, %s, %s", path, fork, flag, mode)
        if type(fork) == unicode:
            logger.warning(u"stat_fork use CMP_UNPACK_BIN_ALLOC_LEGACY. Encoding to str")
            fork = fork.encode('utf-8')

        if isinstance(flag, basestring):
            flag = FileMode.from_fopen(flag)
        mode = mode if mode is not None else 0777

        fh = self.local_rpc.open_fork([path, fork, flag, mode])

        return FileHandler(self, fh)

    @retriable(**RETRY_CONFIG)
    def unlink(self, path):
        self.local_rpc.unlink([path,])

    @retriable(**RETRY_CONFIG)
    def unlink_fork(self, path, fork):
        self.local_rpc.unlink_fork([path, fork])

    @retriable(**RETRY_CONFIG)
    def symlink(self, path_from, path_to):
        self.local_rpc.symlink([path_from, path_to])

    @retriable(**RETRY_CONFIG)
    def rename(self, path_from, path_to):
        self.local_rpc.rename([path_from, path_to])

    @retriable(**RETRY_CONFIG)
    def fork_path(self, fork_ids, flags=0x0000):
        return self.local_rpc.fork_path({u"forks": fork_ids, u"flags":flags})

    def exists(self, path):
        try:
            return self.stat(path) is not None
        except ErrnoException as e:
            if e.errno in (errno.ENOENT, errno.ENOTDIR):
                return False
            else:
                raise

    def isdir(self, path):
        try:
            st = self.stat(path)
            return st.isdir()
        except ErrnoException as e:
            if e.errno == errno.ENOENT:
                return False
            else:
                raise

    @retriable(**RETRY_CONFIG)
    def pin(self, is_pined, path):
        self.local_rpc.pin([is_pined, path,])

    @retriable(**RETRY_CONFIG)
    def statvfs(self):
        return self.local_rpc.statvfs([])
