import packet as fson_packet

import threading
import Queue
import logging
import select
import socket
import sys

logger = logging.getLogger("fsonrpc.connection")

RESPONSE_FATAL = -22000
RESPONSE_FORBIDDEN = -22001
RESPONSE_REDIRECT = -22002
RESPONSE_TIMEOUT = -22003

RESPONSE_INV = {
    1: "RESPONSE_FATAL",
    2: "RESPONSE_FORBIDDEN",
    3: "RESPONSE_REDIRECT",
    4: "RESPONSE_TIMEOUT",
}

SOCKET_READ_TIMEOUT = 10


class Exception(Exception):
    pass


class AsyncResultTimeout(Exception):
    pass


class AsyncResult(object):
    def __init__(self):
        super(AsyncResult, self).__init__()

        self.value = None
        self.exception = None
        self.event = threading.Event()

    def _wait(self, timeout=None):
        # loop over self.event.wait to let a CTRL+C
        # halt the program

        _t = 0.0
        while timeout is None or _t < timeout:
            LAPS = 0.2
            res = self.event.wait(LAPS)
            _t += LAPS

            if res:
                break

        return res

    def wait(self, timeout=None):
        if not self._wait(timeout):
            raise AsyncResultTimeout()

    def get(self, block=True, timeout=None):
        if block:
            self.wait(timeout)

        if self.exception:
            raise self.exception

        else:
            return self.value

    def set(self, value=None):
        self.value = value
        self.event.set()

    def set_exception(self, exception):
        self.exception = exception
        self.event.set()


class ResponseException(Exception):
    def __init__(self, status, blob=None):
        self.status = status
        self.blob = blob

        super(ResponseException, self).__init__("response status %d: %r (blob %dB)" % (status, RESPONSE_INV.get(status, "unknown"), self.blob_size, ))

    @property
    def blob_size(self):
        return len(self.blob) if self.blob is not None else 0


class TimeoutException(ResponseException):
    def __init__(self):
        super(TimeoutException, self).__init__(RESPONSE_TIMEOUT)


class LocalRequest(object):
    """ Represents a request performed by the local node. """

    def __init__(self, conn, req, timeout, no_return):
        super(LocalRequest, self).__init__()

        self.conn = conn
        self.req = req
        self.res = None
        self.no_return = no_return
        self.timeout = timeout

        self.result = AsyncResult()

    def _set_response(self, res):
        if self.no_return:
            self.result.set(None)

        else:
            self._remove()
            self.res = res

            if self.res.status == 0:
                self.result.set(self.res.blob)

            else:
                self.result.set_exception(ResponseException(self.res.status, self.res.blob))

    def get_return(self):
        try:
            return self.result.get(block=True, timeout=self.timeout)

        except AsyncResultTimeout:
            self._remove()
            exc_type, excep, traceback = sys.exc_info()
            raise TimeoutException, TimeoutException(), traceback

    def _remove(self):
        return self.conn._remove_call(self)


class RemoteRequest(object):
    """ Represents a request received by a remote node. """
    def __init__(self, conn, req):
        super(RemoteRequest, self).__init__()

        self.conn = conn
        self.req = req
        self.res = None

        self.result = AsyncResult()

    @property
    def cmd(self):
        return self.req.cmd

    @property
    def blob(self):
        return self.req.blob

    def set(self, blob):
        self.res = fson_packet.Response(
            self.req.id_,
            0,
            blob,
        )

        # XXX to set priority
        self.conn.outgoing_queue.put((0, self))

    def set_exception(self, e):
        if not isinstance(e, ResponseException):
            logger.exception(e)
            e = ResponseException(RESPONSE_FATAL)

        self.res = fson_packet.Response(
            self.req.id_,
            e.status,
            e.blob,
        )

        # XXX to set priority
        self.conn.outgoing_queue.put((0, self))


class Connection(object):
    """ A fsonrpc.Connection instace is initializes over a file-like object
        (commonly a socket, can be a pipe, ...) where FSONRPC protocol is
        going to be transmitted or received.

        Send a command is done through the 'call' method. Asynchronous call
        (i.e. to do several calls at the same time) are performed by
        the method 'async_call'.

        Every received request calls a callback passed as the 'cb' parameter.
        It takes two parameters: the first is the Connection instance receiving
        the remote request, the second is the RemoteRequest instance which
        contains information about the request, and how to answer it, or call
        back the Connection.

        The first parameter is never None. The second can be, which indicates
        the Connection had been closed and will not receive anymore remote
        requests.

        Note that, if both parameters are not None,
        connection_inst == remotereq_inst.conn.

        If you'd like to implement a full FSONRPC server, have a look
        at fsonrpc.server.Server. If you'd like to implement a nice
        dispatcher, have a look at fsonrpc.route.RouteDispatch. If you'd
        like to create a higher-level callback function, see
        the fsonrpc.route.handler decorator.
    """

    DEFAULT_PRIORITY = 0

    def __init__(self, sock, cb=None, auto_start=None):
        auto_start = auto_start if auto_start is not None else True

        super(Connection, self).__init__()

        self.logger = logger

        self.sock = sock

        self.id_counter = 0
        self.id_counter_lock = threading.RLock()

        self.outgoing_queue_kill = False
        self.reader_kill = False
        self.outgoing_queue = Queue.PriorityQueue()

        self.writer_thread = None
        self.reader_thread = None

        self.calls_table = {}
        self.calls_table_lock = threading.RLock()

        self.cb = cb
        self.closed = False

        if auto_start:
            self.start()

    def _generate_id(self):
        with self.id_counter_lock:
            self.id_counter = self.id_counter + 1

            return self.id_counter

    # higher priority, first out
    def async_call(self, cmd, blob=None, priority=None, timeout=None, no_return=None):
        priority = priority if priority is not None else self.DEFAULT_PRIORITY
        no_return = no_return if no_return is not None else False

        req = fson_packet.Request(
            self._generate_id(),
            cmd,
            blob,
        )

        call = LocalRequest(self, req, timeout, no_return)

        self.logger.debug("call %r", req)

        self.outgoing_queue.put((-priority, call))

        return call

    def call(self, *kargs, **kwargs):
        call = self.async_call(*kargs, **kwargs)

        return call.get_return()

    def start(self):
        self.writer_thread = threading.Thread(target=self._writer_run)
        self.reader_thread = threading.Thread(target=self._reader_run)

        self.writer_thread.daemon = True
        self.reader_thread.daemon = True

        self.writer_thread.start()
        self.reader_thread.start()

    def _kill_writer_thread(self):
        self.outgoing_queue_kill = True
        self.outgoing_queue.put((0, None))

        if threading.current_thread() is not self.writer_thread and self.writer_thread is not None:
            self.writer_thread.join()

    def _kill_reader_thread(self):
        self.reader_kill = True
        self.sock.close()

        if threading.current_thread() is not self.reader_thread and self.reader_thread is not None:
            self.reader_thread.join()

    def stop(self):
        self.closed = True

        if self.reader_thread:
            self._kill_reader_thread()
            self.reader_thread = None

        if self.writer_thread:
            self._kill_writer_thread()
            self.writer_thread = None

        if self.cb:
            try:
                self.cb(self, None)

            except Exception, e:
                self.logger.exception(e)

    def _writer_run(self):
        logger = self.logger.getChild("writer")

        while not self.closed:
            try:
                priority, call = self.outgoing_queue.get()

                if self.outgoing_queue_kill:
                    return

                pck = call.req if isinstance(call, LocalRequest) else call.res

                logger.debug("send %r (priority: %d)", pck, priority)
                # we check isinstance here to be faster
                is_local_request = isinstance(call, LocalRequest)
                try:
                    # no need to lock socket, this is the only place
                    # the socket is written)
                    _, writables, _ = select.select([], [self.sock], [], SOCKET_READ_TIMEOUT)
                    if len(writables) == 0:
                        # can't write anything to ocket, stop
                        self.stop()
                        return
                    try:
                        if is_local_request and not call.no_return:
                            with self.calls_table_lock:
                                self.calls_table[call.req.id_] = call
                        pck.dump(self.sock.sendall)
                        if is_local_request and call.no_return:
                            call._set_response(None)
                    except IOError as e:
                        # socket may have ben closed in the other thread
                        self.stop()
                        return
                    except socket.error as e:
                        self.stop()
                        return
                except Exception, e:
                    self.stop()
                    if is_local_request:
                        call.result.set_exception(e)
                        continue
                    else:
                        raise

                logger.debug("%r sent", pck)

            except (Exception, ) as e:
                logger.exception(e)
                continue

    def _reader_run(self):
        logger = self.logger.getChild("reader")

        while not self.closed:
            try:
                # check if socket is readable
                readables, _, _ = select.select([self.sock], [], [], SOCKET_READ_TIMEOUT)
                if len(readables) == 0:
                    # can't read anything from socket, stop
                    self.stop()
                    return
                try:
                    packet = fson_packet.Packet.load(self.sock.recv)
                except IOError:
                    # socket may have ben closed in the other thread
                    self.stop()
                    return
                except socket.error:
                    self.stop()
                    return

                if self.reader_kill:
                    return

                if packet is None:
                    self.stop()
                    return

                logger.debug("received %r", packet)

                if packet is None:
                    continue

                elif isinstance(packet, fson_packet.Request):
                    self._handle_request(packet)

                elif isinstance(packet, fson_packet.Response):
                    self._handle_response(packet)

                else:
                    assert(0)

            except Exception:
                logger.exception("Could not handle fson packet")
                continue

    def _handle_request(self, req):
        call = RemoteRequest(self, req)

        if self.cb:
            try:
                self.cb(self, call)

            except Exception, e:
                call.set_exception(e)

    def _handle_response(self, res):
        with self.calls_table_lock:
            call = self.calls_table.get(res.id_)

        if call is None:
            logger.warning("unknown response: %r", res)
        else:
            call._set_response(res)

    def _remove_call(self, call):
        # on success, register call
        try:
            with self.calls_table_lock:
                self.calls_table.pop(call.req.id_)

        except KeyError:
            pass
