import fson
import cStringIO
import logging
logger = logging.getLogger("fsonrpc.packet")

class Exception(Exception):
    pass

class InvalidPacketException(Exception):
    pass

class Packet(object):
    """ FSON generic packet. """
    FORGETBOX_TAG = 0xFB
    VERSION = 1

    def __init__(self, type_, id_, version = None):
        if version is None: version = self.VERSION

        super(Packet, self).__init__()

        assert 0 <= version < 2 ** 16
        assert 0 <= type_ < 2 ** 16
        assert 0 <= id_ < 2 ** 32

        self.version = version
        self.type_ = type_
        self.id_ = id_
        self.blob = ""

        if self.version > self.VERSION:
            raise InvalidPacketException("unknown version '%d'." % (self.version, ))

    @classmethod
    def load(cls, reader):
        """ Load from a raw source a packet, deserializes the data and
            instances the 'Request' or 'Response' class with the
            deserialized values.

            The 'reader' parameter should be a file-like object.
        """

        tag, size, version, type_ = fson.load(reader, "1422")

        if None in (tag, size, version, type_):
            if any((tag, size, version, type_)):
                logger.warning(
                    "received invalid packet (%d %d %d %d)",
                    tag, size, version, type_
                )
            return

        if tag != Packet.FORGETBOX_TAG:
            raise InvalidPacketException("bad tag '%X'." % tag)

        size -= 4 # deduce header and tail, already read

        if type_ == Request.TYPE:
            return Request.load(reader, size, version)
        elif type_ == Response.TYPE:
            return Response.load(reader, size, version)
        else:
            raise InvalidPacketException("unknown message type '%s'." % type_)

    @property
    def blob_size(self):
        return len(self.blob)

    # Helpers
    @classmethod
    def loads(cls, st):
        """ Same than the load() method, but first parameter is a string """
        return cls.load(cStringIO.StringIO(st).read)

    def dumps(self):
        """ Serialize the packet in raw. Returns a string """
        w = cStringIO.StringIO()
        self.dump(w.write)
        return w.getvalue()

class Request(Packet):
    TYPE = 0x01

    def __init__(self, id_, cmd, blob = None, version = None):
        super(Request, self).__init__(self.TYPE, id_, version = version)

        assert 0 <= cmd <= 2 ** 16

        self.cmd = cmd
        self.blob = blob or ""

    @property
    def size(self):
        # Size contains the number of bytes *following* it
        # (i.e. minus the initial 5 bytes of (tag,size))
        return 2 + 2 + 4 + 2 + self.blob_size

    @classmethod
    def load(cls, reader, size, version):
        if size < 6:
            logger.warning("invalid request size: %d", size)
            return

        id_, cmd = fson.load(reader, "42")
        size -= 6

        if None in (id_, cmd):
            logger.warning("invalid request id:%s cmd:%s", id_, cmd)
            return

        logger.debug("got request id:%d cmd:%d", id_, cmd)

        blob = None
        if size:
            blob, = fson.load(reader, "B", size)
            assert isinstance(blob, str)

        return cls(id_, cmd, blob, version = version)

    def dump(self, writer):
        assert 0 <= self.size < 2 ** 24

        if self.blob:
            return fson.dump(writer, "142242B",
                self.FORGETBOX_TAG, self.size, self.version, self.TYPE,
                                        self.id_, self.cmd, self.blob
            )
        else:
            return fson.dump(writer, "142242",
                self.FORGETBOX_TAG, self.size, self.version, self.TYPE,
                                        self.id_, self.cmd
            )

    def __repr__(self):
        return "<Request #%.8X, cmd=%.4X, blob=(%dB) (ver%d)>" % (
            self.id_, self.cmd, self.blob_size, self.version,
        )

    def __eq__(self, other):
        return (True
            and self.version == other.version
            and self.type_ == other.type_
            and self.id_ == other.id_
            and self.blob == other.blob
            and self.cmd == other.cmd
        )

class Response(Packet):
    TYPE = 0x02

    def __init__(self, id_, status, blob = None, version = None):
        super(Response, self).__init__(self.TYPE, id_, version = version)

        assert -2 ** 15 <= status < 2 ** 15

        self.status = status
        self.blob = blob or ""

    @property
    def size(self):
        # Size contains the number of bytes *following* it
        # (i.e. minus the initial 5 bytes of <tag,size>)
        return 2 + 2 + 4 + 2 + self.blob_size

    @classmethod
    def load(cls, reader, size, version):
        if size < 6:
            logger.warning("invalid response size: %d", size)
            return

        id_, status = fson.load(reader, "42")
        size -= 6

        if status is not None:
            status = fson.u16_cast_s16(status)

        if None in (id_, status):
            logger.warning("invalid response id:%s status:%s", id_, status)
            return

        logger.debug("got response id:%d status:%d", id_, status)

        blob = None
        if size:
            blob, = fson.load(reader, "B", size)
            assert isinstance(blob, str)

        return cls(id_, status, blob, version = version)

    def dump(self, writer):
        assert 0 <= self.size < 2 ** 24

        status = fson.s16_cast_u16(self.status)

        if self.blob:
            return fson.dump(writer, "142242B",
                self.FORGETBOX_TAG, self.size, self.version, self.TYPE,
                                        self.id_, status, self.blob
            )
        else:
            return fson.dump(writer, "142242",
                self.FORGETBOX_TAG, self.size, self.version, self.TYPE,
                                        self.id_, status
            )

    def __repr__(self):
        return "<Response #%.8X, status=%d, blob=(%dB) (ver%d)>" % (
            self.id_, self.status, self.blob_size, self.version,
        )

    def __eq__(self, other):
        return (True
            and self.version == other.version
            and self.type_ == other.type_
            and self.id_ == other.id_
            and self.blob == other.blob
            and self.status == other.status
        )

def main():
    req = Request(1, 0xDEAD, "Hello, world! :-)")
    print "req", req
    req_raw = req.dumps()
    print "req raw", repr(req_raw)
    req2 = Packet.loads(req_raw)
    print "req'", req2

    res = Response(1, -1, "How are you today?")
    print "res", res
    res_raw = res.dumps()
    print "res raw", repr(res_raw)
    res2 = Packet.loads(res_raw)
    print "res'", res2

    assert req == req2
    assert res == res2

if __name__ == "__main__":
    main()
