# coding: utf-8


import sys
from lima.meta.base.readers import MetadataReader, MetadataFileReader, MetadataExtractor
from lima.meta.base.exceptions import ExtractorError
import threading

from gnsdk.gracenote import Gracenote
from gnsdk import GnsdkError
from gnsdk.http import GracenoteNetworkError, GracenoteFatalError

import logging
logger = logging.getLogger(__name__)


# TODO : split fingerprint extraction / submission to
# make possible multiples agents
class GracenoteExtractorError(ExtractorError):
    pass


class GracenoteExtractorNetworkError(ExtractorError):
    pass


# gracenote client is a singleton
_gracenote_client_ = None
lock = threading.Lock()

def init_gracenote(client_id, client_id_tag, licence_content, client_app_version, user_token=None, web_user_token=None):
    """ Initialize Gracenote client, singleton style"""
    global _gracenote_client_
    with lock:
        if _gracenote_client_ is None:
            _gracenote_client_ = Gracenote(client_id, client_id_tag, licence_content, client_app_version, user_token, web_user_token)
    return _gracenote_client_


class HasGracenoteMixin(object):

    def __init__(self, gracenote):
        self.gracenote = gracenote
        super(HasGracenoteMixin, self).__init__()

    @property
    def user_token(self):
        return self.gracenote.user_token

    @property
    def web_user_token(self):
        return self.gracenote.web_user_token


class GracenoteFingerprintExtractor(HasGracenoteMixin, MetadataExtractor):

    def extract(self, fname, is_stream_mode):
        try:
            if isinstance(fname, unicode):
                fname = fname.encode(sys.getfilesystemencoding() or 'utf-8')
            datas = None
            with self.gracenote.audio_fingerprinter(is_stream_mode) as fp:
                datas = fp.fingerprint(fname)
        except GracenoteFatalError:
            logger.error("Fatal error while fingerprinting file")
            raise
        except (GracenoteNetworkError, ) as e:
            logger.error("Network error while fingerprinting file")
            raise
        except (GnsdkError) as e:
            logger.error("Error while fingerprinting file")
            raise GracenoteExtractorError, GracenoteExtractorError(e), sys.exc_info()[2]
        else:
            return {'audio.track.fingerprint': datas}


class GracenoteTagsExtractor(HasGracenoteMixin, MetadataExtractor):
    # HSA tags : hsaudiotags tags attributes names
    OUTPUT_TAGS = {
        u"audio.album.picture.url": u"album_art_url",
        u"audio.album.artist": u"album_artist_name",
        u"audio.album.gnid": u"album_gnid",
        u"audio.album.title": u"album_title",
        u"audio.album.year": u"album_year",
        u"audio.album.track_count": u"track_count",

        u"audio.track.artist": u"track_artist_name",
        u"audio.track.gnid": u"track_gnid",
        u"audio.track.number": u"track_number",
        u"audio.track.title": u"track_title",
        u"audio.track.genres": u"genre",
        u"audio.track.moods": u"mood",
        u"audio.track.tempos": u"tempo",

        u"audio.artist.picture.url": u"artist_image_url",
        u"audio.artist.bio.url": u"artist_bio_url",
        u"audio.album.review.url": u"review_url",

        u"audio.album.xids": u"album_xid",
        u"audio.track.xids": u"track_xid",

        # First interface with core
        # u"audio.title": u"track_title",
        # u"audio.album": u"album_title",
        # u"audio.artist": u"track_artist_name",
        # u"audio.track": u"track_number",
    }

    def normalize(self, tags):
        """ Convert extracted tags as a dict

        :param tags: object representing the tags
        :returns: a dict containging tags
        :rtype: dict
        """
        datas = {}

        if tags is not None:
            for limatag, gntag in GracenoteTagsExtractor.OUTPUT_TAGS.items():
                datas[limatag] = tags.get(gntag, None)
        return {u'tags': datas}


class GracenoteFingerprintSearch(GracenoteTagsExtractor):

    def extract(self, fingerprint, is_stream_mode):
        try:
            datas = self.gracenote.submit_fingerprint(fingerprint, is_stream_mode)
        except (GracenoteFatalError,):
            logger.error("Fatal error while fingerprinting file")
            raise
        except (GracenoteNetworkError, ) as e:
            logger.error("Network error while fingerprinting file")
            raise
        except (GnsdkError, Exception) as e:
            logger.error("Error while searching with gracenote : %s", getattr(e, 'message', 'Unknown error'))
            raise GracenoteExtractorError, GracenoteExtractorError(e), sys.exc_info()[2]
        else:
            return self.normalize(datas)


class GracenoteAudioSearch(GracenoteTagsExtractor):

    def extract(self, artist, album, track):
        try:
            datas = self.gracenote.search_track(artist, album, track)
        except (GracenoteFatalError,):
            logger.error("Fatal error while fingerprinting file")
            raise
        except (GnsdkError, Exception) as e:
            logger.error("Error while searching with gracenote : %s", getattr(e, 'message', 'Unknown error'))
            raise GracenoteExtractorError, GracenoteExtractorError(e), sys.exc_info()[2]
        else:
            return self.normalize(datas)


class GracenoteExtractor(HasGracenoteMixin, MetadataExtractor):

    """ Class for gracenote tags extraction.
    """

    def __init__(self, gracenote):
        super(GracenoteExtractor, self).__init__(gracenote)
        self.fp_extractor = GracenoteFingerprintExtractor(self.gracenote)
        self.tags_extractor = GracenoteTagsExtractor(self.gracenote)

    def extract(self, fname, is_stream_mode):
        tags = None
        if isinstance(fname, unicode):
            fname = fname.encode(sys.getfilesystemencoding() or "utf-8")
        fp_datas = self.fp_extractor.extract(fname, is_stream_mode)
        if fp_datas:
            fp = fp_datas.get('audio.track.fingerprint', None)
            tags = self.tags_extractor.extract(fp, is_stream_mode)
        return tags


class GracenoteReaderBase(object):
    @property
    def gn_config(self):
        gn_config = {}
        gn_config["client_id"] = self.config.get("gracenote.client_id", None)
        gn_config["client_id_tag"] = self.config.get("gracenote.client_id_tag", None)
        gn_config["licence_content"] = self.config.get("gracenote.licence", None)
        gn_config["client_app_version"] = self.config.get("gracenote.app_version", None)

        user_token = self.config.get("gracenote.user_token", None) or ''
        if len(user_token) == 0:
            user_token = None
        gn_config["user_token"] = user_token

        web_user_token = self.config.get("gracenote.web_user_token", None) or ''
        if len(web_user_token) == 0:
            web_user_token = None
        gn_config["web_user_token"] = web_user_token
        return gn_config

    @property
    def user_token(self):
        return self.extractor.user_token

    @property
    def web_user_token(self):
        return self.extractor.web_user_token


class GracenoteReader(GracenoteReaderBase, MetadataFileReader):

    """" A reader specialized with an AudioTagsExtractor """

    def init_extractor(self):
        """ Initialize the metadata extractor """
        gracenote = init_gracenote(**self.gn_config)
        return GracenoteExtractor(gracenote)


class GracenoteFingerprintReader(GracenoteReaderBase, MetadataFileReader):

    """" A reader specialized with a GracenoteFingerprintExtractor """

    def init_extractor(self):
        """ Initialize the metadata extractor """
        gracenote = init_gracenote(**self.gn_config)
        return GracenoteFingerprintExtractor(gracenote)


class GracenoteTagsReader(GracenoteReaderBase, MetadataReader):

    """" A reader specialized with a GracenoteFingerprintExtractor """

    def init_extractor(self):
        """ Initialize the metadata extractor """
        gracenote = init_gracenote(**self.gn_config)
        return GracenoteFingerprintSearch(gracenote)


class GracenoteAudioSearchReader(GracenoteReaderBase, MetadataReader):
    def init_extractor(self):
        """ Initialize the metadata extractor """
        gracenote = init_gracenote(**self.gn_config)
        return GracenoteAudioSearch(gracenote)
