# coding: utf-8
import os
import sys
import json
from hsaudiotag import auto

from lima.meta.base.readers import MetadataFileReader, MetadataExtractor
from lima.meta.base.exceptions import MetadataFileError

__all__ = [
    "AudioTagsReader",
    "AudioTagsExtractor",
    "AudioTagsThumbExtractor",
    "AudioTagThumbReader",

]


class AudioTagsExtractor(MetadataExtractor):

    """ Class for audio tags metadata """

    # HSA tags : hsaudiotags tags attributes names
    OUTPUT_TAGS = {
        # general infos
        u"audio.album.title": u"album",
        u"audio.track.artist": u"artist",
        u"audio.track.genre": u"genre",
        # track infos
        u"audio.track.duration": u"duration",
        u"audio.track.year": u"year",
        u"audio.track.title": u"title",
        u"audio.track.number": u"track",
        u"audio.track.sample_rate": u"sample_rate",
        u"audio.track.size":u"size",
        u"audio.track.audio_offset":u"audio_offset",
        u"audio.track.audio_size":u"audio_size",
        u"audio.track.bitrate":u"bitrate",
        u"audio.track.comment":u"comment",
        u"audio.track.description":u"description",
        u"audio.track.composer": u"composer",
        # not supported yet
        # album infos
        #u"track_position":u"position",
        #u"album_release_year":u"year",
        #u"album_cover_path": u"album_cover_path",
        #u"album_is_compilation": u"is_compilation",
        #u"disk_number":u"disk_number",

        # First interface with core
        # u"audio.title": u"title",
        # u"audio.album": u"album",
        # u"audio.artist": u"artist",
        # u"audio.duration": u"duration",
        # u"audio.track": u"track",
    }

    def extract(self, fname, ext=None):
        tags = None
        try:
            # hsaudioread is able to open files with unicode names
            if ext is None:
                tags = auto.File(fname)
            else:
                # if we have an extension given, access to the corresponding
                # class directly to avoid muliples reads to determine which class to use
                _cls = auto.EXT2CLASS.get(ext, None)
                if _cls:
                    tags = _cls(fname)
        except (Exception) as e:
            # raising custom exception with traceback, python2 style
            raise MetadataFileError, MetadataFileError(e), sys.exc_traceback
        else:
            # so far so good
            return self.normalize(tags)

    def normalize(self, tags):
        """ Convert extracted tags as a dict

        :param tags: object representing the tags
        :returns: a dict containging tags
        :rtype: dict
        """
        datas = {}
        for limatag, hstag in AudioTagsExtractor.OUTPUT_TAGS.items():
            value = getattr(tags, hstag)
            if type(value) is str:
                value = value.decode(u'utf-8')
            datas[limatag] = value
        return {u'tags': datas}

class AudioTagsReader(MetadataFileReader):

    """" A reader specialized with an AudioTagsExtractor """

    def init_extractor(self):
        """ Initialize the metadata extractor """
        return AudioTagsExtractor()

class AudioTagsThumbExtractor(MetadataExtractor):
    def extract(self, fname):
        try:
            if isinstance(fname, unicode):
                fname = fname.encode(sys.getfilesystemencoding() or 'utf-8')
            tags = auto.File(fname)
        except (Exception) as e:
            # raising custom exception with traceback, python2 style
            raise MetadataFileError, MetadataFileError(e), sys.exc_traceback
        else:
            # so far so good
            return self.normalize(tags)

    def normalize(self, tags):
        """ Convert extracted tags as a dict

        :param tags: object representing the tags
        :returns: a dict containging tags
        :rtype: dict
        """
        picture = None
        if tags.picture is not None and len(tags.picture) > 0:
            picture = tags.picture
        return {u'audio.track.picture': picture}



class AudioTagThumbReader(MetadataFileReader):

    """" A reader specialized with an AudioTagsExtractor """

    def init_extractor(self):
        """ Initialize the metadata extractor """
        return AudioTagsThumbExtractor()
