# coding: utf-8
import abc
import os
import json

from lima.meta.base.exceptions import (
    MetadataFileError,
    ReaderError
)

__all__ = [
    "MetadataFileError",
    "MetadataFileReader",
    "MetadataExtractor"
]


class MetadataReaderError(ReaderError):

    """ Base exception for MetadataReader """
    pass


class MetadataReader(object):

    __metaclass__ = abc.ABCMeta

    """ Base class for meta data worker  """

    def __init__(self, config = None):
        """ Metadata file worker constructor.

        """
        self.config = config if config is not None else {}
        self.extractor = self.init_extractor()

    @abc.abstractmethod
    def init_extractor(self):
        """ Initialize the metadata extractor.

        This method is abstract and must be implemented by sublasses.
        """
        return  # pragma: no cover

    def execute(self, fname, raw_datas, *args, **kwargs):
        datas = {}
        datas[u"file"] = fname
        datas.update(self.extractor.extract(raw_datas, *args, **kwargs))
        return datas


class MetadataFileReader(MetadataReader):


    def execute(self, fname, input_enc='utf-8', *args, **kwargs):
        """ Method called to execute a worker against a file.

        :param fname: the path to the file. If fname is of type str, we will try to convert it to unicode.


        As writing to output is an effective output of this program, the result will be encoded with output_encoding.
        As returning the value is not considered as an output, the unicode string will be returned.
        """

        if type(fname) == str:
            fname = fname.decode(input_enc)
        self._sanity_checks(fname)
        datas = {}
        datas[u"file"] = fname
        datas.update(self.extractor.extract(fname, *args, **kwargs))

        return datas

    # privates methods
    def _sanity_checks(self, fname):
        self._worker_sanity_checks()
        self._file_sanity_checks(fname)

    def _worker_sanity_checks(self):
        """ Process sanity check on the worker state"""
        if self.extractor is None:
            raise MetadataReaderError(u"Metadata extractor is not initialized")

    def _file_sanity_checks(self, fname):
        """ Process some sanity checks against the file"""
        if type(fname) not in (str, unicode):
            return
        if any([fname is None, fname is not None and fname.replace(" ", '') == '']):
            raise MetadataFileError(u"You must provide a file name")
        if not os.path.exists(fname):
            raise MetadataFileError(u"File '%s' does not exists" % fname)
        if not os.path.isfile(fname):
            raise MetadataFileError(u"'%s' is not a file" % fname)


class MetadataExtractor(object):

    """ Base class for meta data extractor  """
    __metaclass__ = abc.ABCMeta

    @abc.abstractmethod
    def extract(self, *args, **kwargs):
        """ Method called to extract metadata against a file.
        Assumes that the file exists and all sanity checks are done against
        the file name.

        :param fname: path as a string
        :returns: dictionary containing metadatass
        :rtype: dict
        """
        return  # pragma: no cover

    def _sanity_checks(self):
        pass # pragma : no cover
