# coding: utf-8

import os
import string

from lima.meta.audio.tags import AudioTagsReader, AudioTagThumbReader

from lima.fs import FileSystem as fs

from lima.meta.blackboard.agents.base import BlackBoardAgent, AgentAbortedError
from lima.meta.blackboard.agents.common import FileTypeMixin
from lima.meta.blackboard.agents.classes import NetworkClass, LowCPUClass, VeryLowCPUClass

from lima.meta.storage import AlbumStorage, ArtistStorage
from lima.tools import segment
from lima.tools import strings
from lima.fs.limafs import (
    LimaFsTimeoutError, LimaFsConnectionRefusedError,
    LimaFsNotReadyError,
    LimaFsNoSpaceError, LimaFsMemoryError, LimaFsNotMountedError,
)
from lima.db import DbNoSpaceError

__all__ = [
    'AudioLocalMusicTagAgent',
    'AudioAlbumLocalCoverAgent',
    'AudioTagMergerAgent',
    'AudioGracenoteTagAgent',
    'AudioGracenoteFingerprintAgent',
    'AudioGracenoteSearchAgent',
    'AudioGracenoteSearchArtistAgent',
    'AudioCoverDownloadAgent',
    'AudioAlbumCoverLink',
    'AudioNoAlbumLocalCoverAgent',
    'AudioArtistCoverDownloadAgent',
    'AudioTrackArtistCoverDownloadAgent',
]

DEFAULT_THUMBS_CONFIG = [
        ('pvw', 320, 320),  # pvw : preview for client

        # ('pvw32', 32, 32), # thumb32 : windows thumbnail
        # ('pvw96', 96, 96), # thumb96 : windows thumbnail
        # ('pvw256', 256, 256), # thumb256 : windows thumbnail
        # ('pvw1024', 1024, 1024), # thumb1024 : windows thumbnail
]


class AudioAgent(FileTypeMixin, BlackBoardAgent):

    """ Base class for audio agent """

    def _precondition_(self):
        """ Check if the file is an audio file"""
        activate = self._check_mime(self.meta_storage, 'audio')
        # ignore .m3u playlists which are reported as audio/x-mpegurl
        is_playlist = self._check_mime(
            self.meta_storage, 'audio/x-mpegurl'
        )
        if is_playlist:
            self.set_content_type('playlist')

        activate = activate and not is_playlist
        if activate:
            self.logger.debug(u"File is audio type. Enable.")
        return activate


class AudioTagsMixin(object):

    WANTED = [
        # Common
        u'audio.track.title', u'audio.track.artist',
        u'audio.track.duration', u'audio.track.number',
        u'audio.album.title', u'audio.album.artist',
        u'audio.track.genre', u'audio.album.year',
        u'audio.track.composer'
    ]
    WANTED_GRACENOTE = [
        # Common
        u'audio.track.title', u'audio.track.artist',
        u'audio.track.duration', u'audio.track.number',
        u'audio.album.title', u'audio.album.artist',
        u'audio.track.genres', u'audio.album.year',
        # Gracenote
        u'audio.track.gnid',
        u'audio.album.picture.url',
        u'audio.artist.picture.url',
    ]

    TAGS_FIXES = {
            u'audio.track.title': lambda x: string.capwords(x),
            u'audio.album.title': lambda x: string.capwords(x),
            u'audio.album.artist': lambda x: string.capwords(x),
            u'audio.track.artist': lambda x: string.capwords(x),
            u'audio.track.genre': lambda x: string.capwords(x),
    }

    def _filter_tags(self, tags):
        """ Return a dictionary filtered with keys presents in
        self.WANTED"""
        return self._filter_tags_with(tags, self.WANTED)

    def _filter_tags_gracenote(self, tags):
        """ Return a dictionary filtered with keys presents in
        self.WANTED_GRACENOTE"""
        return self._filter_tags_with(tags, self.WANTED_GRACENOTE)

    def _filter_tags_with(self, tags, _filter):
        return {
            key: value
            for key, value in tags.items() if key in _filter
        }

    def _fixes_tags(self, tags):
        # apply fixes on track/album title and track/album artist
        return {
                key: self.TAGS_FIXES.get(key, lambda x: x)(value)
            for key, value in tags.items()
        }


class AudioLocalMusicTagAgent(AudioTagsMixin, AudioAgent):

    """ Agent that will extract tags from an audio file"""

    KEY = u'audio.tags.local'
    priority = 1
    _id = u'audio.tags'

    CLASSES = [LowCPUClass, ]

    def _check(self):
        activate = not self.meta_storage.hasmeta_internal(AudioLocalMusicTagAgent.KEY)\
                or self.is_content_modified

        if not activate:
            self.logger.debug(u"File already has audio tags")
            return False

        # check if tags already had been merged
        activate = not self.meta_storage.hasmeta(u'audio.track.title')
        if not activate and not self.is_content_modified:
            self.logger.debug(u"File has merged tags")
            return False

        self.logger.info(u"Activate tags extraction")
        return activate

    def _process(self):
        self.logger.debug(u"Processing tags")
        self._check_abort()

        reader = AudioTagsReader()

        try:
            datas = reader.execute(fs.get_system_path(self.path))
        except IOError as e:
            # errno 22 : invalid mode or filename
            # we may have problem to access the file throught filesystem
            # in case of encoding inconsistencies in filename
            if getattr(e, 'errno', -1) == 22:
                # as we give a readable, we specify the extension
                # to avoid too many reads to determine which
                ext = os.path.splitext(self.path)[1][1:]
                # fallback : read through RPC
                with fs.open(self.path) as audiofile:
                    datas = reader.execute(audiofile, ext)
            else:
                raise

        if datas is None:
            return False

        self._check_abort()
        tags = datas.get("tags")
        if tags is None or len(tags) == 0:
            return False
        self.logger.debug(u"Got tags")

        self._check_abort()
        filterdtags = self._filter_tags(tags)
        filterdtags = self._fix_missings(filterdtags)

        self.meta_storage.setmeta_internal(
            AudioLocalMusicTagAgent.KEY, filterdtags
        )

        self.context[AudioTagMergerAgent.KEY_DIRTY] = True
        self.logger.info(u"Processed audio tags")
        return True

    def _fix_missings(self, tags):
        ret = {}
        ret.update(tags)

        # if u"audio.album.artist" not in tags and u"audio.track.artist" in tags:
        #    ret[u"audio.album.artist"] = tags[u"audio.track.artist"]
        return ret

from lima.meta.audio.gracenote import (
    GracenoteFingerprintReader,
    GracenoteTagsReader,
    GracenoteAudioSearchReader,
    GracenoteExtractorNetworkError,
    GracenoteFatalError,
)


class GracenoteDisabledError(AgentAbortedError):
    pass


class GracenoteAgent(AudioAgent):
    # As gracenote can be enabled/disabled/re enabled, we mark the
    # GracenoteDisabledError as not being takied in account for classes errors
    EXCEPTIONS_NO_CLASS_ERROR = [GracenoteDisabledError, ]

    KEY_TAGS = 'audio.fingerprint.gn.tags'

    def is_gracenote_enabled(self):
        enabled = False
        with self.config.lock:
            enabled = self.config.get('gracenote.enabled', False)
            if enabled is not True:
                self.logger.debug(u"Gracenote is not activated")
        return enabled

    def disable_gracenote(self):
        with self.config.lock:
            self.config['gracenote.enabled'] = False
            self.logger.warn(u"Gracenote disabled")
            segment.track(event="Gracenote disabled")

    def _precondition_(self):
        activate = super(GracenoteAgent, self)._precondition_()
        if not activate:
            return False
        # Check if the filename encoding is consistent.
        # If its not the case, we won't be able to encode it to a bytestring to use it with ffmpeg
        fname_enc_consistent = strings.is_filename_encoding_consistent(fs.get_system_path(self.path))
        if not fname_enc_consistent:
            self.logger.warning(u"Filename encoding is inconsistent. Not using gracenote")
            return False

        # avoid processing identified podcasts
        genre = self.meta_storage.getmeta('audio.track.genre', '')
        if genre.lower() == 'podcast':
            self.logger.debug(u"File as genre '%s'. Not activating gracenote.", genre)
            return False

        return self.is_gracenote_enabled()

    def _check_abort(self):
        super(GracenoteAgent, self)._check_abort()
        if self.is_gracenote_enabled() is not True:
            raise GracenoteDisabledError(u"Gracenote disabled")


class AudioGracenoteSearchAgent(AudioTagsMixin, GracenoteAgent):
    CLASSES = [LowCPUClass, ]
    priority = 10
    _id = u'audio.search.gn'
    # KEY = u'audio.tags.gn'
    KEY_NO_MATCH = u"audio.search.gn.no_match"

    threshold = 80

    def _check(self):
        # first, see if we already have a result
        activate = not self.meta_storage.hasmeta_internal(self.KEY_TAGS)
        if not activate and not self.is_content_modified:
            self.logger.debug(u"File has gracenote results")
            return False

        # check if tags already had been merged
        activate = not self.meta_storage.hasmeta(u'audio.track.title')
        if not activate and not self.is_content_modified:
            self.logger.debug(u"File has tags")
            return False

        no_match = self.meta_storage.getmeta_internal(AudioGracenoteSearchAgent.KEY_NO_MATCH, None)
        if no_match is True:
            self.logger.debug(u"Did not found any match. Not trying again")
            return False

        # then check if we have enought infos for a gracenote research
        local_tags = self.meta_storage.getmeta_internal(AudioLocalMusicTagAgent.KEY) or {'tags': {}}
        activate = 'audio.track.title' in local_tags and ('audio.track.artist' in local_tags or 'audio.album.title' in local_tags)
        if not activate and not self.is_content_modified:
            self.logger.debug(u"File has not enough info for search")
            return False

        self.logger.info(u"Activate audio search")
        return activate

    def _validate_result(self, track_artist, album_artist, track_composer, album_title, track_title, result):
        if not result:
            return False
        if 'tags' not in result:
            return False
        if len(result['tags']):
            return False

        ratio_track_title = 0
        if track_title and result['tags'].get('audio.track.title'):
            ratio_track_title = strings.score(track_title, result['tags'].get('audio.track.title'))

        ratio_album_title = 0
        if album_title and result['tags'].get('audio.album.title'):
            ratio_album_title = strings.score(album_title, result['tags'].get('audio.album.title'))

        ratio_artist = 0
        gn_album_artist = result['tags'].get('audio.album.artist')
        gn_track_artist = result['tags'].get('audio.track.artist')

        local_artists = (album_artist, track_artist, track_composer)
        gn_artists = (gn_album_artist, gn_track_artist)

        ratio_artist = 0
        for local_art in local_artists:
            for gn_art in gn_artists:
                ratio_artist = max(ratio_artist, strings.score(local_art, gn_art))
                if ratio_artist == 100:
                    break  # no need to continue

        match = False
        if ratio_track_title == ratio_album_title == ratio_artist == 100:
            self.logger.info(u"Perfect match")
            match = True
        else:
            # split condition in 2 for readability
            match = (ratio_track_title > self.threshold and ratio_album_title > self.threshold)
            match = match or (ratio_track_title > self.threshold and ratio_artist > self.threshold)
        return match

    def _search_full(self, reader, track_artist, album_artist, track_composer, album_title, track_title):
        """Search gracenote with all parameter"""
        track_artist = (track_artist or u'').strip()
        album_artist = (album_artist or u'').strip()
        album_title = (album_title or u'').strip()
        track_composer = (track_composer or u'').strip()
        track_title = (track_title or u'').strip()

        has_artist = track_artist or album_artist or track_composer
        if not has_artist and not album_title and not track_title:
            return {}
        return reader.execute(self.path, album_artist or track_artist or track_composer, album_title, track_title)

    def _search_album_track(self, reader, album_title, track_title):
        """Search gracenote with album / track"""
        album_title = (album_title or u'').strip()
        track_title = (track_title or u'').strip()

        if not album_title and not track_title:
            return {}
        return reader.execute(self.path, None, album_title, track_title)

    def _search_artist_track(self, reader, track_artist, album_artist, track_composer, track_title):
        """Search gracenote with all artist / track"""

        track_artist = (track_artist or u'').strip()
        album_artist = (album_artist or u'').strip()
        track_composer = (track_composer or u'').strip()
        track_title = (track_title or u'').strip()

        has_artist = track_artist or album_artist or track_composer
        if not has_artist and not track_title:
            return {}
        return reader.execute(self.path, album_artist or track_artist or track_composer, None, track_title)

    def search_track_and_validate(self, reader, track_artist, album_artist, track_composer, album_title, track_title):
        try:
            result = self._search_full(reader, track_artist, album_artist, track_composer, album_title, track_title)
            if self._validate_result(track_artist, album_artist, track_composer, album_title, track_title, result):
                return result
        except (LimaNetworkError):
            self.logger.error(u"Could not search artist / album / track on gracenote (no network)")
            raise
        try:
            result = self._search_album_track(reader, album_title, track_title)
            if self._validate_result(track_artist, album_artist, track_composer, album_title, track_title, result):
                self.logger.info("Found with fallback album / track")
                return result
        except (LimaNetworkError):
            self.logger.error(u"Could not search artist / album / track on gracenote (no network)")
            raise
        try:
            result = self._search_artist_track(reader, track_artist, album_artist, track_composer, track_title)
            if self._validate_result(track_artist, album_artist, track_composer, album_title, track_title, result):
                self.logger.info("Found with fallback artist / track")
                return result
        except (LimaNetworkError):
            self.logger.error(u"Could not search artist / album / track on gracenote (no network)")
            raise

    def _process(self):
        self.logger.debug(u"Processing audio search")
        self._check_abort()

        local_tags = self.meta_storage.getmeta_internal(AudioLocalMusicTagAgent.KEY) or {'tags': {}}

        track_artist = local_tags.get('audio.track.artist')
        album_artist = local_tags.get('audio.album.artist')
        track_composer = local_tags.get('audio.track.composer')

        track_title = local_tags.get('audio.track.title')
        album_title = local_tags.get('audio.album.title')

        reader = GracenoteAudioSearchReader(self.config)
        try:
            result = self.search_track_and_validate(reader, track_artist, album_artist, track_composer, album_title, track_title)
        except GracenoteFatalError:
            self.logger.error("Gracenote search error (fatal)")
            segment.track_error(event="Gracenote search error (fatal)")
            self.disable_gracenote()
            return False
        except (LimaNetworkError):
            self.logger.error("Gracenote search error (no network)")
            raise
        except (LimaFsTimeoutError):
            self.logger.error("Gracenote search error (timeout)")
            raise
        except (LimaFsNotMountedError, LimaFsConnectionRefusedError):
            self.logger.error("Gracenote search error (core unreachable)")
            raise
        except (LimaFsNoSpaceError, DbNoSpaceError):
            self.logger.exception("Gracenote search error (no space)")
            raise
        except (LimaFsMemoryError,):
            self.logger.exception("Gracenote search error (no memory)")
            raise
        except (GracenoteExtractorNetworkError,):
            self.logger.error("Gracenote search error (network)")
            raise
        except Exception as e:
            # segment.track_error(event="Gracenote search error")
            self.logger.error("Could not get gracenote tags : %s", getattr(e, 'message', 'Unknown error'))
            raise
        finally:
            if reader.web_user_token:
                with self.config.lock:
                    self.config['gracenote.web_user_token'] = reader.web_user_token

        self._check_abort()
        if result:
            tags = result.get(u"tags", None)
            if tags:
                self.logger.info(u"Gracenote result matched.")
                filterdtags = self._filter_tags_gracenote(tags)

                self.meta_storage.setmeta_internal(self.KEY_TAGS, filterdtags)
                self.context[AudioTagMergerAgent.KEY_DIRTY] = True
                self.logger.info(u"Fetched audio tags")
                self.meta_storage.setmeta_internal(AudioGracenoteSearchAgent.KEY_NO_MATCH, False)
                return True
            else:
                self.meta_storage.setmeta_internal(AudioGracenoteSearchAgent.KEY_NO_MATCH, True)
        else:
            self.meta_storage.setmeta_internal(AudioGracenoteSearchAgent.KEY_NO_MATCH, True)
        self.logger.info(u"Gracenote search : no match")
        return False


class AudioGracenoteFingerprintAgent(GracenoteAgent):
    KEY = u'audio.fingerprint.gn.fileid'
    priority = 11
    _id = u'audio.fingerprint.gn'

    CLASSES = [LowCPUClass, ]

    def _check(self):
        activate = self._check_mime_not(self.meta_storage, 'audio/midi', 'audio/vnd.dts')
        if not activate:
            self.logger.warn("Explicitly ignoring file")
            return False
        activate = not self.meta_storage.hasmeta_internal(AudioGracenoteFingerprintAgent.KEY)
        if not activate and not self.is_content_modified:
            self.logger.debug(u"File has fingerprint")
            return False

        activate = self.meta_storage.getmeta_internal(AudioGracenoteSearchAgent.KEY_NO_MATCH, False)
        if not activate and not self.is_content_modified:
            return False

        self.logger.info(u"Activate file fingerprinting")
        return activate

    def _process(self):
        self.logger.debug(u"Processing file fingerprinting")
        self._check_abort()

        reader = GracenoteFingerprintReader(self.config)
        try:
            datas = reader.execute(fs.get_system_path(self.path), is_stream_mode=False)
        finally:
            if reader.web_user_token:
                with self.config.lock:
                    self.config['gracenote.web_user_token'] = reader.web_user_token

        if datas is None:
            return False

        self._check_abort()
        try:
            fingerprint = datas.get("audio.track.fingerprint", None)
        # except Exception as u:
        except Exception:
            # segment.track_error(event="Gracenote fingerprint error")
            raise

        if fingerprint is None or len(fingerprint) == 0:
            self.logger.error(u"Could not get fingerprint")
            return False

        self.logger.debug(u"Got fingerprint")
        self._check_abort()
        self.meta_storage.setmeta_internal(AudioGracenoteFingerprintAgent.KEY, fingerprint)

        self.logger.info(u"File fingerprinted")
        return True


class AudioGracenoteTagAgent(AudioTagsMixin, GracenoteAgent):

    # KEY = u'audio.tags.gn'
    priority = AudioGracenoteFingerprintAgent.priority + 1
    KEY_NO_MATCH = 'audio.tags.gn.no_match'

    CLASSES = [NetworkClass, LowCPUClass]

    _id = u'audio.tags.gn'

    def _check(self):
        activate = not self.meta_storage.getmeta_internal(AudioGracenoteTagAgent.KEY_NO_MATCH, False)\
                or self.is_content_modified
        if not activate:
            self.logger.debug(u"No match for fingerprint. Not retrying")
            return False

        activate = not self.meta_storage.hasmeta_internal(self.KEY_TAGS) \
            or self.is_content_modified
        if not activate:
            self.logger.debug(u"File has tags")
            return False
        activate = self.meta_storage.hasmeta_internal(AudioGracenoteFingerprintAgent.KEY)
        if not activate:
            self.logger.debug(u"No gracenote tags present")
            return False

        self.logger.info(u"Activate tags fetching")
        return True

    def _process(self):
        self.logger.debug(u"Processing tags")
        self._check_abort()
        fp_data = self.meta_storage.getmeta_internal(
            AudioGracenoteFingerprintAgent.KEY
        )
        if not fp_data:
            return False

        self._check_abort()
        reader = GracenoteTagsReader(self.config)

        try:
            datas = reader.execute(self.path, fp_data, is_stream_mode=False)
        except GracenoteFatalError:
            self.logger.error("Could not get gracenote tags (fatal)")
            segment.track_error(event="Could not get gracenote tags (fatal)")
            self.disable_gracenote()
            return False
        except (LimaFsTimeoutError):
            self.logger.error("Could not get gracenote tags (timeout)")
            raise
        except (LimaFsNotMountedError, LimaFsNotReadyError, LimaFsConnectionRefusedError):
            self.logger.exception("Could not get gracenote tags (core unreachable)")
            raise
        except (LimaFsNoSpaceError, DbNoSpaceError):
            self.logger.exception("Could not get gracenote tags (no space)")
            raise
        except (LimaFsMemoryError,):
            self.logger.exception("Could not get gracenote tags (no memory)")
            raise
        except (GracenoteExtractorNetworkError,):
            self.logger.exception("Could not get gracenote tags (network)")
            raise
        except Exception as e:
            # segment.track_error(event="Gracenote submission error")
            self.logger.error("Could not get gracenote tags : %s", getattr(e, 'message', 'Unknown error'))
            raise
        finally:
            if reader.web_user_token:
                with self.config.lock:
                    self.config['gracenote.web_user_token'] = reader.web_user_token

        if datas is None:
            self.logger.debug('Gracenote extractor returned None')
            self.meta_storage.setmeta_internal(AudioGracenoteTagAgent.KEY_NO_MATCH, True)
            return False

        self._check_abort()
        tags = datas.get(u"tags")
        if tags is None or len(tags) == 0:
            self.logger.debug('Gracenote extractor returned 0 tags')
            self.meta_storage.setmeta_internal(AudioGracenoteTagAgent.KEY_NO_MATCH, True)
            return False

        self.logger.debug(u"Got tags")
        self._check_abort()

        filterdtags = self._filter_tags_gracenote(tags)

        self.meta_storage.setmeta_internal(AudioGracenoteTagAgent.KEY_TAGS, filterdtags)
        self.context[AudioTagMergerAgent.KEY_DIRTY] = True
        self.logger.info(u"Fetched audio tags")
        return True


class AudioTagMergerAgent(AudioTagsMixin, AudioAgent):

    KEY = u'audio.tag'
    KEY_DIRTY = u'audio.tags.dirty'

    KEY_NOTAG = u'NOTAG'

    priority = AudioGracenoteTagAgent.priority + \
        AudioLocalMusicTagAgent.priority

    CLASSES = [VeryLowCPUClass, ]

    _id = u'audio.tags.merge'

    def _check(self):
        dirty = self.context.get(AudioTagMergerAgent.KEY_DIRTY, False)
        # assume dirty if the value is None
        dirty = dirty if dirty is not None else False

        activate = dirty or self.is_content_modified
        if not activate:
            self.logger.debug(u"Tags seems up to date")
            return False

        local_tags = self.meta_storage.getmeta_internal(AudioLocalMusicTagAgent.KEY)
        gn_tags = self.meta_storage.getmeta_internal(AudioGracenoteTagAgent.KEY_TAGS)
        activate = ((local_tags or gn_tags) is not None)

        if not activate:
            self.logger.debug(u"No tags to merge")
            return False

        self.logger.info(u"Enabling tag merge")
        return True

    def _process(self):
        self.logger.info(u"Processing tag merge")

        self._check_abort()
        tags = self.meta_storage.getmeta(AudioTagMergerAgent.KEY, {})

        self._check_abort()

        local_tags = self.meta_storage.getmeta_internal(
            AudioLocalMusicTagAgent.KEY
        ) or {}
        self._check_abort()
        gn_tags = self.meta_storage.getmeta_internal(AudioGracenoteTagAgent.KEY_TAGS, {})

        should_merge = True
        local_genre = local_tags.get('audio.track.genre', '')
        # if the file is identified as a podcast, do not merge
        if local_genre.lower() == 'podcast':
            self.logger.info(u"Track is a podcast. Using local tags only.")
            should_merge = False
        else:
            # if the local track name and fetched track name are too differents,
            # do not merge
            local_track = local_tags.get('audio.track.title')
            gn_track = gn_tags.get('audio.track.title')

            track_score = 100
            if local_track and gn_track:
                track_score = strings.score(local_track, gn_track)

            should_merge = track_score > 40
            if not should_merge:
                self.logger.warn(u"Local tags and fetched tags are too differents (score : %s, local : %s, result : %s). Not merging.", track_score, local_track, gn_track)

        if should_merge:
            tags = self._merge_tags(local_tags, gn_tags)
        else:
            tags = self._merge_tags(local_tags, {})

        tags = self._filter_tags(tags)
        tags = self._fixes_tags(tags)

        merged_tags = {}

        for meta_key in tags.keys():
            self._check_abort()
            value = tags[meta_key]
            # Do not store empty values
            # if we can check the length (str, unicode, dict, list,.. iterables)
            # we not store items with lenght = 0
            if value is not None\
                    and (not hasattr(value, '__len__')
                         or (hasattr(value, '__len__') and len(value) > 0)
                         ):
                merged_tags[meta_key] = value
        try:
            self.meta_storage.update(merged_tags)
        except Exception:
            self.logger.exception(u"Could not set meta %s", meta_key)
            raise

        self.context[AudioTagMergerAgent.KEY_DIRTY] = False
        self.context[self.KEY_NEED_REINDEX] = True
        self.set_content_type('music')
        # cleanup local tags + gracenote tags
        title = merged_tags.get("audio.track.title", None)
        album = merged_tags.get("audio.album.title", None)
        artist = merged_tags.get("audio.album.artist", None)
        track_artist = merged_tags.get("audio.track.artist", None)

        self.logger.debug(u"Audio tags has been merged")

        if title or album or artist or track_artist:
            self.logger.info(5 * u"-" + "Identified as '%s' by '%s' on '%s'",
                    title or 'Unknow title',
                    artist or track_artist or 'Unknow artist',
                    album or 'Unknow album'
            )

        artist_picture_url = tags.get('audio.artist.picture.url', None)
        if artist and artist_picture_url:
            artist_storage = ArtistStorage.get_from_pool(artist)
            if not artist_storage.hasmeta('picture.url'):
                artist_storage.set_meta('picture.url', artist_picture_url)
                artist_storage.unload(persist=True)
        return True

    def _merge_tags(self, local_tags, gn_tags):
        tags = self.meta_storage.getmeta(AudioTagMergerAgent.KEY) or {}
        # first, add gracenote tags
        for key, value in gn_tags.items():
                tags[key] = value

        # Second, override gracenotes fields with local tags non null fields
        for key, value in local_tags.items():
            if value is not None and \
                    (not hasattr(value, '__len__') or (hasattr(value, '__len__') and len(value) > 0)):
                        tags[key] = value
        return tags


class AudioGracenoteSearchArtistAgent(AudioTagsMixin, GracenoteAgent):
    """ Extract track artist url"""
    CLASSES = [LowCPUClass, NetworkClass]
    priority = AudioTagMergerAgent.priority + 1

    _id = u'audio.search_artist.gn'
    KEY_NO_MATCH = u"audio.search_artist.gn.no_match"
    KEY_TRACK_ARTIST_COVER_URL = 'audio.track_artist.picture.url'

    threshold = 80

    def _check(self):
        # check if we already failed with no match
        activate = not self.meta_storage.getmeta_internal(AudioGracenoteSearchArtistAgent.KEY_NO_MATCH, False)\
                or self.is_content_modified
        if not activate:
            self.logger.debug(u"No match found")
            return False

        # check if we have track artist and not trak artist url
        track_artist = self.meta_storage.getmeta('audio.track.artist', None)

        if track_artist and len(track_artist.strip()):
            artist_storage = ArtistStorage.get_from_pool(track_artist)
            activate = not artist_storage.hasmeta('picture.url')
            if not activate:
                self.logger.debug(u"%s already have url", track_artist)
            if activate:
                activate = not artist_storage.exists_non_empty('pvw')
                if not activate:
                    self.logger.debug(u"%s already have pvw", track_artist)
        else:
            activate = False

        activate = activate or self.is_content_modified

        return activate

    def _search_track_artist(self, reader, track_artist):
        """Search gracenote with all parameter"""
        track_artist = (track_artist or u'').strip()
        if not track_artist:
            return {}
        return reader.execute(self.path, track_artist, None, None)

    def _validate_result(self, track_artist, result):
        ratio_track_artist = 0
        if not result:
            return False
        other_artist = result['tags'].get('audio.album.artist')
        if track_artist and result['tags'].get('audio.album.artist'):
            ratio_track_artist = strings.score(track_artist, other_artist)

            if ratio_track_artist > self.threshold:
                self.logger.info("%s found with score %s", other_artist, ratio_track_artist)
                return True
            elif ratio_track_artist > 50:
                ratio_track_artist = strings.partial_score(track_artist, other_artist)
                if ratio_track_artist > self.threshold:
                    self.logger.info("%s found with score %s", other_artist, ratio_track_artist)
                    return True

        return False

    def search_track_artist_and_validate(self, reader, track_artist):
        result = self._search_track_artist(reader, track_artist)
        if self._validate_result(track_artist, result):
            return result

    def _process(self):
        self.logger.debug(u"Processing artist search")
        self._check_abort()

        track_artist = self.meta_storage.getmeta('audio.track.artist', None)
        if not track_artist:
            return False

        reader = GracenoteAudioSearchReader(self.config)
        try:
            result = self.search_track_artist_and_validate(reader, track_artist)
        except GracenoteFatalError:
            self.logger.error("Gracenote artist search error (fatal)")
            segment.track_error(event="Gracenote artist search error (fatal)")
            self.disable_gracenote()
            return False
        except (LimaFsTimeoutError):
            self.logger.error("Gracenote album cover download error (timeout)")
            raise
        except (LimaFsNotMountedError, LimaFsNotReadyError, LimaFsConnectionRefusedError):
            self.logger.exception("Gracenote artist search error (core unreachable)")
            raise
        except (LimaFsNoSpaceError, DbNoSpaceError):
            self.logger.exception("Gracenote artist search error (no space)")
            raise
        except (LimaFsMemoryError,):
            self.logger.exception("Gracenote artist search error (no memory)")
            raise
        except (GracenoteExtractorNetworkError,):
            self.logger.exception("Gracenote artist search error (network)")
            raise
        except Exception as e:
            # segment.track_error(event="Gracenote artist search error")
            self.logger.error("Gracenote artist search error : %s", getattr(e, 'message', 'Unknown error'))
            raise
        finally:
            if reader.web_user_token:
                with self.config.lock:
                    self.config['gracenote.web_user_token'] = reader.web_user_token

        if result:
            self.logger.info(u"Found preview url for artist %s", track_artist)
            artist_storage = ArtistStorage.get_from_pool(track_artist)
            artist_storage.setmeta('picture.url', result['tags'].get('audio.artist.picture.url'))
            artist_storage.unload(persist=True)
            return True
        else:
            self.logger.info(u"No results for preview url for artist %s", track_artist)
            self.meta_storage.setmeta_internal(AudioGracenoteSearchArtistAgent.KEY_NO_MATCH, True)


from image import WithThumbnailsMixin
from StringIO import StringIO


class AudioAlbumMixin(object):
    thumbs_config = DEFAULT_THUMBS_CONFIG

    def has_album_datas(self):
        return self.artist is not None and self.album is not None

    def has_artist_datas(self):
        return self.artist is not None

    def has_track_artist_datas(self):
        return self.track_artist is not None

    @property
    def album_storage(self):
        if not hasattr(self, '_album_storage'):
            self._album_storage = AlbumStorage.get_from_pool(self.artist, self.album)
        return self._album_storage

    @property
    def artist_storage(self):
        if not hasattr(self, '_artist_storage'):
            self._artist_storage = ArtistStorage.get_from_pool(self.artist)
        return self._artist_storage

    @property
    def track_artist_storage(self):
        if not hasattr(self, '_track_artist_storage'):
            self._track_artist_storage = ArtistStorage.get_from_pool(self.track_artist)
        return self._track_artist_storage

    @property
    def artist(self):
        return self.meta_storage.getmeta(u"audio.album.artist")

    @property
    def track_artist(self):
        return self.meta_storage.getmeta(u"audio.track.artist")

    @property
    def album(self):
        return self.meta_storage.getmeta(u"audio.album.title")


class LocalCoverAgent(AudioAgent):

    """Base class for agent extracting cover"""
    KEY_NO_PICTURE = 'no_picture'

    thumbs_config = DEFAULT_THUMBS_CONFIG
    CLASSES = [LowCPUClass, ]

    def _check(self):
        thumb_names = [thumbname for thumbname, _, _ in self.thumbs_config]
        # cover is not present in the storage
        has_cover = self.has_all_thumbs_non_empty(self.cover_storage, *thumb_names)

        activate = not has_cover\
            or self.is_content_modified
        if not activate:
            self.logger.debug(u"File has cover")
            return False

        self.logger.info(u"Activate cover extraction")
        return activate

    def _process(self):
        self.logger.debug(u"Processing cover")
        self._check_abort()
        cover = self.get_picture()

        if cover is None:
            # set a flag indicating that the file does not have any picture
            self.meta_storage.setmeta_internal(self.KEY_NO_PICTURE, True)
            return False
        self.logger.debug(u"Got cover")

        self._check_abort()
        for thumbname, width, height in self.thumbs_config:
            coverimg = StringIO(cover)
            with self.cover_storage.get_storage(thumbname, mode='w') as fork:
                self.resize(coverimg, fork, width, height)
                self.logger.debug(u"Generated %sx%s thumbnail", width, height)
            self.meta_storage.setmeta_internal(self.KEY_NO_PICTURE, False)

        self.context[self.KEY_NEED_REINDEX] = True
        self.logger.info(u"Cover extracted")
        return True

    def get_picture(self):
        reader = AudioTagThumbReader()

        self._check_abort()

        try:
            datas = reader.execute(fs.get_system_path(self.path))
        except IOError as e:
            # errno 22 : invalid mode or filename
            # we may have problem to access the file throught filesystem
            # in case of encoding inconsistencies in filename
            if getattr(e, 'errno', -1) == 22:
                # as we give a readable, we specify the extension
                # to avoid too many reads to determine which
                ext = os.path.splitext(self.path)[1][1:]
                # fallback : read through RPC
                with fs.open(self.path) as audiofile:
                    datas = reader.execute(audiofile, ext)
            else:
                raise

        if datas is None:
            return None

        self._check_abort()
        # Get raw datas from file
        return datas.get('audio.track.picture', None)


class AudioAlbumLocalCoverAgent(AudioAlbumMixin, WithThumbnailsMixin, LocalCoverAgent):

    """ Extract the file cover and store it at artist/album level."""

    priority = AudioTagMergerAgent.priority + 1

    _id = u'audio.cover.extract'

    def _check(self):
        if not self.has_album_datas():
            return False
        return super(AudioAlbumLocalCoverAgent, self)._check()

    @property
    def cover_storage(self):
        return self.album_storage


from lima.meta.blackboard.agents.network import ImageDownloadAgent, LimaNetworkError


class AudioCoverDownloadAgent(AudioAlbumMixin, AudioAgent, ImageDownloadAgent):

    """Agent that will download and resize the posted for an audio file.

    The file will be downladed only if the thumbnail does not already exists.
    The file will be stored at albumartist/album level
    """
    priority = AudioAlbumLocalCoverAgent.priority + 1

    COVER_URL_KEY = u'audio.album.picture.url'

    _id = u'music.cover.download'

    CLASSES = [NetworkClass, ]

    def __init__(self, *args, **kwargs):
        super(AudioCoverDownloadAgent, self).__init__(
            self.thumbs_config, timeout=5,
            *args, **kwargs
        )

    @property
    def download_storage(self):
        return self.album_storage

    def _check(self):
        if not self.has_album_datas():
            return False

        return super(AudioCoverDownloadAgent, self)._check()

    def _process(self):
        try:
            result = super(AudioCoverDownloadAgent, self)._process()
            if result:
                self.context[self.KEY_NEED_REINDEX] = True
            return result
        except GracenoteFatalError:
            self.logger.error("Gracenote album cover download error (fatal)")
            segment.track_error(event="Gracenote album cover download error (fatal)")
            self.disable_gracenote()
            return False
        except (LimaNetworkError):
            self.logger.error("Gracenote album cover download error (no network)")
            raise
        except (LimaFsTimeoutError):
            self.logger.error("Gracenote album cover download error (timeout)")
            raise
        except (LimaFsNotMountedError, LimaFsConnectionRefusedError, LimaFsTimeoutError):
            self.logger.error("Gracenote album cover download error (core unreachable)")
            raise
        except (LimaFsMemoryError,):
            self.logger.exception("Gracenote album cover download error (no memory)")
            raise
        except (GracenoteExtractorNetworkError,):
            self.logger.exception("Gracenote artist search error (network)")
            raise
        except Exception:
            segment.track_error(event="Gracenote album cover download error")
            raise

    @property
    def url(self):
        gn_tags = self.meta_storage.getmeta_internal(AudioGracenoteTagAgent.KEY_TAGS)
        if not gn_tags:
            return None
        url = gn_tags.get(self.COVER_URL_KEY, None)
        if url is not None and len(url.replace(' ', '')) == 0:
            return None
        return url


class AudioArtistCoverDownloadAgent(AudioAlbumMixin, AudioAgent, ImageDownloadAgent):

    """Agent that will download and resize the posted for an audio file.

    The file will be downladed only if the thumbnail does not already exists.
    The file will be stored at albumartist/album level
    """
    priority = AudioAlbumLocalCoverAgent.priority

    COVER_URL_KEY = u'audio.artist.picture.url'

    _id = u'music.artist.cover.download'

    CLASSES = [NetworkClass, ]

    def __init__(self, *args, **kwargs):
        super(AudioArtistCoverDownloadAgent, self).__init__(
            self.thumbs_config, *args, **kwargs
        )

    @property
    def download_storage(self):
        return self.artist_storage

    def _check(self):
        if not self.has_artist_datas():
            return False

        return super(AudioArtistCoverDownloadAgent, self)._check()

    def _process(self):
        try:
            result = super(AudioArtistCoverDownloadAgent, self)._process()
            if result:
                self.context[self.KEY_NEED_REINDEX] = True
            return result
        except GracenoteFatalError:
            self.logger.exception("Gracenote artist cover download error (fatal)")
            segment.track_error(event="Gracenote artist cover download error (fatal)")
            self.disable_gracenote()
            return False
        except (LimaNetworkError):
            self.logger.error("Gracenote artist cover download error (no network)")
            raise
        except (LimaFsTimeoutError):
            self.logger.error("Gracenote artist cover download error (timeout)")
            raise
        except (LimaFsNotMountedError, LimaFsConnectionRefusedError):
            self.logger.error("Gracenote artist cover download error (core unreachable)")
            raise
        except (LimaFsNoSpaceError, DbNoSpaceError):
            self.logger.exception("Gracenote artist cover download error (no space)")
            raise
        except (LimaFsMemoryError,):
            self.logger.exception("Gracenote artist cover download error (no memory)")
            raise
        except (GracenoteExtractorNetworkError,):
            self.logger.exception("Gracenote artist cover download error (network)")
            raise
        except Exception:
            segment.track_error(event="Gracenote artist cover download error")
            self.logger.exception("Gracenote artist cover download error")
            raise

    @property
    def url(self):
        if not self.has_artist_datas():
            return None
        url = self.download_storage.getmeta('picture.url')
        if url is not None and len(url.replace(' ', '')) == 0:
            return None
        return url


class AudioTrackArtistCoverDownloadAgent(AudioArtistCoverDownloadAgent):
    priority = AudioGracenoteSearchArtistAgent.priority + 1

    """Agent that will download and resize the artist poster for an audio file.

    The file will be downladed only if the thumbnail does not already exists.
    The file will be stored at artist level
    """
    priority = AudioAlbumLocalCoverAgent.priority

    COVER_URL_KEY = AudioGracenoteSearchArtistAgent.KEY_TRACK_ARTIST_COVER_URL

    _id = u'music.track_artist.cover.download'

    def has_artist_datas(self):
        return self.has_track_artist_datas()

    @property
    def download_storage(self):
        return self.track_artist_storage


class AudioNoAlbumLocalCoverAgent(AudioAlbumMixin, WithThumbnailsMixin, LocalCoverAgent):

    """ Extract the file cover and store it at the file level when no artist / album is available"""

    priority = AudioGracenoteTagAgent.priority + AudioGracenoteSearchAgent.priority + AudioGracenoteSearchArtistAgent.priority + AudioTagMergerAgent.priority \
            + AudioArtistCoverDownloadAgent.priority + AudioTrackArtistCoverDownloadAgent.priority \
            + AudioAlbumLocalCoverAgent.priority

    _id = u'audio.cover.extract.noalbum'

    def _check(self):
        dirty = self.context.get(AudioTagMergerAgent.KEY_DIRTY, None)
        activate = dirty is False
        if not activate:
            return
        # do not activate if we already knows that there is no picture present
        activate = not self.meta_storage.getmeta_internal(self.KEY_NO_PICTURE, False)
        if not activate:
            self.logger.info(u"%s no preview embeded.", self.path)
            return

        # only activate if we do not have any info about album / artist / track artist
        activate = activate and (not self.has_album_datas() and self.track_artist is None)
        if not activate:
            return
        activate = activate and super(AudioNoAlbumLocalCoverAgent, self)._check()

        if activate:
            self.logger.info(u"%s have a preview but no artist / album / track artist info.", self.path)
        return activate

    @property
    def cover_storage(self):
        # will store at the file level
        return self.meta_storage


class AudioAlbumCoverLink(AudioAlbumMixin, WithThumbnailsMixin, AudioAgent):

    CLASSES = [VeryLowCPUClass, ]

    """Agent that will link artist/album previews to the file"""

    priority = AudioTagMergerAgent.priority + \
        AudioAlbumLocalCoverAgent.priority + AudioCoverDownloadAgent.priority + AudioNoAlbumLocalCoverAgent.priority

    _id = u'audio.album.cover.link'

    def _check(self):
        thumb_names = [thumbname for thumbname, _, _ in self.thumbs_config]
        if not self.has_album_datas():
            return False
        album_has_all_thumbs = self.has_all_thumbs_non_empty(self.album_storage, *thumb_names)
        file_has_all_thumbs = self.has_all_thumbs_non_empty(self.meta_storage, *thumb_names)

        # Check if linked pvw point to the album pvw
        already_same_link = True
        for thumbname in thumb_names:
            already_same_link = already_same_link and self.meta_storage.is_same_link(self.album_storage, thumbname, thumbname)

        return album_has_all_thumbs and (not file_has_all_thumbs or (self.is_content_modified and not already_same_link))

    def _process(self):
        self.logger.debug(u"Processing cover link")

        for thumbname, _, _ in self.thumbs_config:
            # symlink artist/album preview to file
            self.meta_storage.link(self.album_storage, thumbname, thumbname)
            self.logger.info(u"%s/%s/%s linked", self.artist, self.album, thumbname)
        self.context[self.KEY_NEED_REINDEX] = True
        return True
