# coding: utf-8
import os
import sys
import abc
import threading

from lima.meta.base.exceptions import LimaMetaError
from lima.fs import FileSystem as fs
from lima.fs.limafs import LimaFsDoesnotExistError
from lima.meta.scheduler import killswitch


from lima.tools import logs
logger = logs.getLogger(__name__)

__all__ = [
    'BlackBoardAgent',
]


from lima.meta.scheduler.tasks import (
    TaskAbortedError, TaskKilledError, TaskStopedError, TaskRetryError,
    TooManyFailureError, TooManyRetryError
)

from lima.meta.blackboard.agents.storage import AgentStorage


class AgentAbortedError(TaskAbortedError):
    pass


class AgentKilledError(TaskKilledError):
    pass


class AgentStopedError(TaskStopedError):
    pass


class KeepTask(object):

    def __init__(self, processor, task):
        self.logger = logger
        self.processor = processor
        self.task = task
        self.processor.tasks.append(task)

    def __enter__(self):
        return self.task

    def __exit__(self, *exc):
        if self.task in self.processor.tasks:
            self.processor.tasks.remove(self.task)
        del self.task
        del self.processor

        exc_type, exc_inst, traceback = exc
        if exc_type is not None:
            # raise original exception
            raise exc_type, exc_inst, traceback


from classes import (AgentLock,)


class BlackBoardAgent(object):

    """Base class for Blackboard agents"""

    __metaclass__ = abc.ABCMeta

    # Agent classses
    CLASSES = []
    # Exceptionshat will raise a TaskRetryError
    EXCEPTIONS_RETRY = []
    # Exceptionshat will not be counted in the class error count
    EXCEPTIONS_NO_CLASS_ERROR = []
    # Key to flag file as to be reindexed
    KEY_NEED_REINDEX = u'file.need_reindex'

    def __init__(self, path, parent=None, is_update=False, is_content_modified=False, is_path_modified=False, is_basename_modified=False, config=None, context=None, is_debug=False):
        log_name = u"lima.meta.agent.%s" % (self._id)
        self.path = path
        self.logger = logs.getLogger(log_name, self)
        self.logger.name = log_name
        self.is_debug = is_debug
        # DIRTY little hack : getLogger will convert the logger name in base
        # string '' even if the name is given as unicode u''. This conversion
        # in logging/__init__.py:1032 will just screw our logs when trying to
        # log unicode string containing accents (typicallly a filename or a path)
        #
        # Here, we force it to be a unicode string.
        self.parent = parent
        self.lock = threading.Lock()

        if config is None:
            raise ValueError("Config should be initialized")

        self.meta_storage = AgentStorage(self.path, is_debug=self.is_debug)
        # here, the folder wiil be pinned only if needed
        try:
            self.meta_storage.pin()
        except Exception:
            self.logger.error(u"Could not pin meta folder")

        self.config = config
        if context is None:
            context = {}
        self.context = context

        self.is_update = is_update
        self.is_content_modified = is_content_modified
        self.is_path_modified = is_path_modified
        self.is_basename_modified = is_basename_modified

        # How may failed attemp are tollerated for this agent
        # default : 3, may be overrided by by children
        self.max_errors = 3
        self.max_unsuccessfull_runs = 1
        self.failed_runs = 0
        self.unsuccessfull_runs = 0

        self.tasks = []
        self._stop = False
        self._kill = False

    @property
    def agent_lock(self):
        if getattr(self, '_agent_lock', None) is None:
            self._agent_lock = AgentLock(
                self, self.CLASSES, self.EXCEPTIONS_RETRY, self.EXCEPTIONS_NO_CLASS_ERROR
            )
        return self._agent_lock

    @abc.abstractproperty
    def _id(self):
        """ Define agent priority over others"""
        return None  # pragma: no cover

    def __str__(self):
        fname = os.path.basename(self.path)
        return "%s - %s" % (fname.encode('utf-8'), self._id.encode('utf-8'))

    def __unicode__(self):
        return str(self).decode('utf-8')

    @abc.abstractproperty
    def priority(self):
        """ Define agent priority over others"""
        return sys.maxint  # pragma: no cover

    def run(self):
        return self.process()

    def process(self):
        """ Starts the agent
        returns False if too many errors occured
        """
        self.resume(False)
        res = False
        try:
            self._check_exists()
            with self.agent_lock:
                res = self._process()
                self.is_update = False
                # handle multiples levels of file update
                self.is_content_modified = False
                self.is_path_modified = False
                self.is_basename_modified = False

            self.on_stop()
        except(TaskRetryError, ) as e:
            self.logger.error("Agent delayed : %s", e.time_to_wait)
            self.on_stop(e)
            raise
        except(AgentAbortedError,) as e:
            self.logger.error("Agent aborted : %s", e.cause)
            self.on_stop(e)
            raise
        except(LimaFsDoesnotExistError,) as e:
            # In the case where the file has been deleted during the processing,
            # we raise an AgentAbortedError.
            self.logger.error(
                "File was moved / deleted during processing. Aborting."
            )
            self.on_stop(e)
            raise AgentAbortedError, AgentAbortedError(u"File has been deleted"), sys.exc_info()[2]
        except(AgentKilledError,) as e:
            self.logger.error("Agent killed : %s", e.cause)
            self.on_stop(e)
            raise
        except(AgentStopedError,) as e:
            self.logger.error("Agent stoped : %s", e.cause)
            self.on_stop(e)
            raise
        except(TooManyRetryError, TooManyFailureError) as e:
            self.on_stop(e)
            raise
        except(LimaMetaError,) as e:
            self.logger.error("Agent failed : %s", getattr(e, "message", "Unknown error"))

            self.failed_runs += 1
            self.on_stop(e)
            raise
        except(Exception,) as e:
            self.logger.exception("Agent failed hard")
            self.failed_runs += 1
            self.on_stop(e)
            raise
        except:
            self.logger.exception("OMG this should not happen")
            self.on_stop(e)
            raise

        else:
            if res is not True:
                self.unsuccessfull_runs += 1
        return res is True

    def check(self):
        """ Check if this agent can handle the file.
        returns False if too many errors occured
        """
        if not self._check_exists(raises=False):
            self.logger.debug(u"The file does not exists")
            return False
        if self.failed_runs >= self.max_errors:
            self.logger.warn("Failed (%s trials). Aborting", self.max_errors)
            return False
        if self.unsuccessfull_runs >= self.max_unsuccessfull_runs:
            self.logger.debug("Unsuccesfull (%s trials). Aborting", self.unsuccessfull_runs)
            return False
        activate = self._precondition_()
        if not activate:
            return False
        res = self._check()
        try:
            self._check_abort()
        except(AgentKilledError, AgentAbortedError, AgentStopedError):
            return False
        # we expect an explicit True
        return res is True

    def stop(self):
        self.logger.info(u"Stopping agent %s", self._id)
        with self.lock:
            self._stop = True
            for t in self.tasks:
                try:
                    t.stop()
                except:
                    self.logger.exception("Error while stopping current task")

    def resume(self, notice=True):
        if notice:
            self.logger.info(u"Resuming agent %s", self._id)
        with self.lock:
            self._stop = False

    def kill(self):
        self.logger.info(u"Killing agent %s", self._id)
        self._kill = True
        for t in self.tasks[:]:
            try:
                t.kill()
            except:
                self.logger.exception(u"Error while stopping current task")
            finally:
                # removes references to tasks
                self.tasks.remove(t)
                del t

    @property
    def _should_continue(self):
        return not self._stop

    def _check_abort(self):
        if self._kill is True:
            self.logger.error(u"Agent have been killed")
            raise AgentKilledError(u"Agent is dying")
        if self._stop is True:
            self.logger.error(u"Agent is stoping")
            raise AgentStopedError(u"Agent stoped")
        if killswitch.switch.check():
            logger.warn(u"Kill switch detected. Aborting")
            raise AgentAbortedError(u"Kill switch detected")

    def _check_exists(self, raises=False):
        if not fs.exists(self.path):
            if raises is True:
                raise AgentAbortedError(u"The file does not exists")
            return False
        return True

    def run_task(self, task):
        ''' Return a context manager wich keep a trace over executed task by the
        agent.
        The task will be added to the list self.tasks on enter and removed when
        the task finish and exits from the context manager
        This allow to stop / kill underlying tasks
        '''
        return KeepTask(self, task)

    @abc.abstractmethod
    def _check(self):
        return False  # pragma: no cover

    @abc.abstractmethod
    def _process(self):
        return False  # pragma: no cover

    def _precondition_(self):
        return True

    def on_stop(self, exc=None):
        if self.parent is not None:
            self.parent.on_agent_stop(self, exc)

    def set_content_type(self, content_type):
        self.context["content_type"] = content_type
        self.meta_storage.setmeta("content_type", content_type)
