# coding: utf-8

import abc
import sys
import threading

import logging
logger = logging.getLogger(__name__)
import warnings


from lima.fs.fcfs import LimaFsTimeoutError
from lima.meta.scheduler.tasks import (
    TaskRetryError, TooManyFailureError, TooManyRetryError, TaskDisabledError,
    TaskStopedError, TaskKilledError, TaskAbortedError
)


class AgentClassDisabledError(TaskDisabledError):

    """
    Agents form this class should be disabled when this exception is raised
    """

    def __init__(self, tryied_times, classes):
        super(AgentClassDisabledError, self).__init__(tryied_times)
        self.classes = classes


class AgentClassLock(object):
    PARALLEL_COUNT = 1

    # MAX Agent failures
    # this will only be used for knowns exceptions
    CLASS_MAX_FAILURE = None

    CLASS_FAILURE_COUNT = 0

    RETRY_WAIT_TIME = 2

    @classmethod
    def instance(cls, *args, **kwargs):
        if not hasattr(cls, '_instance'):
            cls._instance = cls(*args, **kwargs)
        return cls._instance

    @abc.abstractproperty
    def _id(self):
        pass

    def __init__(self):
        """
        :param known_exceptions_retry: list of exceptions types to catch and raise a retry
        error.
        If self.failure_count reach MAX_ERROR_RETRY, the agent class wil be
        disabled.
        """
        self._lock = threading.Lock()
        self.semaphore = self.PARALLEL_COUNT

        # How many times the item is retryied after a fail
        # for knowm exceptions

    def check_and_lock(self):
        if self.semaphore is None:
            return
        with self._lock:
            """
            if self.CLASS_MAX_FAILURE is not None\
                    and self.CLASS_FAILURE_COUNT >= self.CLASS_MAX_FAILURE:
                raise AgentClassDisabledError(
                    self.CLASS_FAILURE_COUNT, [self, ])
            """
            # Semaphore locked ?
            if self.semaphore == 0:
                logger.debug(
                    u"[%s] Already %s tasks running in parralell", self._id, self.PARALLEL_COUNT)
                raise TaskRetryError(min(self.RETRY_WAIT_TIME, 300))
            self.semaphore -= 1

    def unlock(self):
        if self.semaphore is None:
            return
        with self._lock:
            self.semaphore += 1

    def __enter__(self):
        self.check_and_lock()
        return self

    def __exit__(self, *exc):
        self.unlock()

        if exc[0] is None:
            self.class_failure_count = 0
        else:
            """
            # Trace how many time agents from this class have failed
            AgentClassLock.CLASS_FAILURE_COUNT += 1
            if AgentClassLock.CLASS_MAX_FAILURE is not None\
                    and AgentClassLock.CLASS_FAILURE_COUNT >= AgentClassLock.CLASS_MAX_FAILURE:
                raise AgentClassLockDisabledError(
                    AgentClassLock.CLASS_FAILURE_COUNT
                )
            else:
            """
            raise exc[0], exc[1], exc[2]


class DefaultClass(AgentClassLock):
    _id = "default"
    PARALLEL_COUNT = 1


class NetworkClass(AgentClassLock):
    _id = "network"
    PARALLEL_COUNT = 4
    CLASS_MAX_FAILURE = 3
    RETRY_WAIT_TIME = 5


class HighCPUClass(AgentClassLock):
    _id = "cpu.high"
    PARALLEL_COUNT = 1


class MediumCPUClass(AgentClassLock):
    _id = "cpu.medium"
    PARALLEL_COUNT = 2


class VideoPictureExtractor(MediumCPUClass):
    _id = "video.preview.extract"


class LowCPUClass(AgentClassLock):
    _id = "cpu.low"
    PARALLEL_COUNT = 15


class VeryLowCPUClass(AgentClassLock):
    _id = "cpu.low"
    PARALLEL_COUNT = 20


class GroupClassLock(AgentClassLock):

    _id = 'classes.group'

    def __init__(self, classes):
        super(GroupClassLock, self).__init__()
        if len(classes) == 0:
            self.classes = [DefaultClass.instance(), ]
        else:
            self.classes = [_class.instance()
                            for _class in classes]

    def check_and_lock(self):
        locked_classes = []
        try:
            for _class in self.classes:
                _class.check_and_lock()
                locked_classes.append(_class)
        except:
            [_class.unlock() for _class in locked_classes]
            raise

    def unlock(self):
        # will raise the first seen exception
        exc_info = None
        for _class in self.classes:
            try:
                _class.unlock()
            except(TaskKilledError, TaskStopedError, TaskAbortedError):
                if exc_info is None:
                    exc_info = sys.exc_info()
            except Exception as e:
                logger.error(
                    "Error unlocking agent class unlock, continuing. (%s)",
                    getattr(e, 'message', 'Unknown error')
                )
                if exc_info is None:
                    exc_info = sys.exc_info()
        if exc_info is not None:
            raise exc_info[0], exc_info[1], exc_info[2]

    def __exit__(self, *exc):
        exc_info = None
        for _class in self.classes:
            try:
                _class.__exit__(*exc)
            except(TaskKilledError, TaskStopedError, TaskAbortedError):
                if exc_info is None:
                    exc_info = sys.exc_info()
            except Exception as e:
                logger.error(
                    "Error exiting agent class lock, continuing. (%s)",
                    getattr(e, "message", "Unknown error")
                    )
                if exc_info is None:
                    exc_info = sys.exc_info()
        if exc_info is not None:
            raise exc_info[0], exc_info[1], exc_info[2]


class AgentLock(object):
    MAX_ERROR_RETRY = 3
    # Time to wait in second
    RETRY_WAIT_TIME = 2
    # How many time we should wait and retry
    MAX_RETRY = 10

    def __init__(self, agent, classes, known_exceptions_retry=[], known_exceptions_ignore_class_errors=[]):
        self.agent = agent
        self.known_exceptions_retry = known_exceptions_retry

        if LimaFsTimeoutError not in self.known_exceptions_retry:
            self.known_exceptions_retry.append(LimaFsTimeoutError)

        self.known_exceptions_ignore_class_errors = known_exceptions_ignore_class_errors

        if len(classes) == 0:
            warnings.warn(u"%s has no class" % agent._id)
        self.class_lock = GroupClassLock(classes)

    def __enter__(self):
        failure_count = getattr(self.agent, 'failure_count', 0)

        if failure_count >= self.MAX_ERROR_RETRY:
            failure_count = self.agent.failure_count
            self.agent = None
            raise TooManyFailureError(failure_count)

        self.class_lock.__enter__()

        return self

    def __exit__(self, *exc):
        noneExc = (None, None, None)
        if exc[0] is not None:
            if exc[0] in self.known_exceptions_retry:
                failure_count = getattr(self.agent, 'failure_count', 0)
                failure_count += 1
                self.agent.failure_count = failure_count
                self.class_lock.__exit__(*noneExc)
                # Raise a retry eror if the exception is in known exceptions
                raise TaskRetryError(
                    min(self.RETRY_WAIT_TIME ** failure_count, 300)
                )
            else:
                if not exc[0] in self.known_exceptions_ignore_class_errors:
                    # Removes reference to agent
                    self.agent = None
                    self.class_lock.__exit__(*exc)
                else:
                    self.class_lock.__exit__(*noneExc)
                raise exc[0], exc[1], exc[2]
        else:
            # Removes reference to agent
            self.agent = None
            self.class_lock.__exit__(*noneExc)
