# coding: utf-8

import sys
from PIL import Image
from PIL import ImageFile
# HACK to avoid PIL “IOError: image file truncated”
ImageFile.LOAD_TRUNCATED_IMAGES = True


from lima.tools import logs
logger = logs.getLogger(__name__)

from lima.meta.blackboard.agents.base import BlackBoardAgent
from lima.meta.blackboard.agents.common import (
    FileTypeMixin, WithForkMixin
)

from lima.fs import FileSystem as fs

from lima.meta.base.exceptions import LimaMetaError

__all__ = [
    'ImageThumbnailsAgent',
]


DEFAULT_THUMBS_CONFIG = [
    ('pvw', 320, 320),  # pvw : preview for client
]


class LimaImageProcessingError(LimaMetaError):

    """ Base exception class for image processing errors """
    pass


class LimaImageNotSupporteError(LimaImageProcessingError, ValueError):
    pass


def flip_horizontal(im):
    return im.transpose(Image.FLIP_LEFT_RIGHT)


def flip_vertical(im):
    return im.transpose(Image.FLIP_TOP_BOTTOM)


def rotate_180(im):
    return im.transpose(Image.ROTATE_180)


def rotate_90(im):
    return im.transpose(Image.ROTATE_90)


def rotate_270(im):
    return im.transpose(Image.ROTATE_270)


def transpose(im):
    return rotate_90(flip_horizontal(im))


def transverse(im):
    return rotate_90(flip_vertical(im))

orientation_funcs = [lambda x: x,
                     lambda x: x,
                     flip_horizontal,
                     rotate_180,
                     flip_vertical,
                     transpose,
                     rotate_270,
                     transverse,
                     rotate_90
                     ]
KEY_ORIENTATION_EXIF = 0x0112


class ImageResizeMixin(object):

    ''' Mixin for resizing images'''

    def __init__(self, *args, **kwargs):
        super(ImageResizeMixin, self).__init__(*args, **kwargs)
        if not hasattr(self, 'logger'):
            self.logger = logger

    def resize(self, filein, fileout, width, height, format='JPEG'):
        """
        Resize the giver filein to fileout at eh given width/height.

        :param filein: path or file like object for the image to resize
        :param fileout: path or file like object for the output image

        :param width: target width
        :param height: target height
        :param format: target format. Must be a Pillow supported format.
        Default to 'JPEG'

        :returns: None
        :raises: Anything that Pillow can raise
        """
        size = (width, height)
        image = None
        try:
            try:
                image = Image.open(filein)
            except Exception as e:
                message = getattr(e, 'message', '')
                if isinstance(message, Exception):
                    # message is the original exception, lets raise
                    raise

                message = str(message).lower()
                if "cannot identify image" in message \
                        or 'unsupported' in message:
                    raise LimaImageNotSupporteError("File may not be an image or format is unsupported.")
                raise

            original_mode = image.mode
            try:
                image = self.apply_rotation(image)
            except Exception:
                self.logger.error(u"Could not apply rotation")

            if original_mode == '1':
                image = image.convert('L')
            elif original_mode == 'I':
                # 8-bit pixels, black and white
                # convert to 8-bit pixels, black and white
                table = [i / 256 for i in range(65536)]
                image = image.point(table, 'L')
            elif original_mode != 'RGBA':
                # convert everything else to RGBA
                image = image.convert('RGBA')

            image.thumbnail(size, Image.ANTIALIAS)
            if original_mode == 'P' and image.format == 'GIF':
                image = image.convert('P')

            try:
                # convert alpha to a white background
                im = Image.new("RGB", image.size, (255, 255, 255))
                im.paste(image, image)
                image = im
            except Exception:
                self.logger.error("Could not convert alpha (%s)", original_mode)

            try:
                image.save(fileout, format, optimize=True, progressive=True)
            except IOError:
                ImageFile.MAXBLOCK = width * height
                image.save(fileout, format, optimize=True, progressive=True)
            # so far so good
            return True
        except LimaImageNotSupporteError:
            raise
        except(ValueError, TypeError, ) as e:
            self.logger.exception("Could not resize image")
            raise LimaImageProcessingError, LimaImageProcessingError(e), sys.exc_info()[2]
        finally:
            if image is not None:
                image.close()

    def apply_rotation(self, image):
        if not hasattr(image, '_getexif'):
            return image
        tags = image._getexif()
        if tags is None:
            return image
        orientation = tags.get(KEY_ORIENTATION_EXIF, 0)
        f = orientation_funcs[orientation]
        return f(image)


class ImageAgent(FileTypeMixin, BlackBoardAgent):

    """ Base class for image agent."""

    def _precondition_(self):
        """ Precondition for image agent.
        Will check for supported image.

        :returns: True if the file is supported, False otherwise
        """
        activate = self._check_mime(self.meta_storage, 'image')
        activate = activate and self._check_mime_not(self.meta_storage,
                'image/svg', #  ignore SVG
                'image/sgi', 'image/x-rgb', #  ignore SGI
                'image/vnd.djvu', #  ignore djvu
                )
        if activate:
            self.logger.debug("File is a supported image. Activate.")
        return activate


class WithThumbnailsMixin(WithForkMixin, ImageResizeMixin):


    def has_all_thumbs_non_empty(self, storage, *thumbnails):
        """
        :returns: True if all forks are presents with a size > 0 for thumbnails
        """
        return self.has_all_forks_non_empty(storage, *thumbnails)

    def can_thumbnail(self, storage, images, thumbnails, is_update=False):
        """ Return True if the given images can be generated.
        An image can be generated if :

        - A fork for at least one image name exists
        - A fork for at least one thumbnail name does not exists

        :param images: List of images fork names
        :param thumbnails: List of thumbnails fork names

        """
        has_all = self.has_all_thumbs_non_empty(storage, *thumbnails)
        if has_all:
            self.logger.debug('All thumbs already presents')
        return (not has_all) or is_update

    def thumbnail_image(self, path, storage, meta_storage, image, thumbs_config):
        """ Generate thumbnails for the given image fork name

        :param image: the image fork name
        :param thumbs_config: list containing, for each item, (name, widht,
        height)
        """
        result = True
        with fs.open_fork(path, image, 'r') as fimage:
            self.logger.debug("Got original image")
            for forkname, width, height in thumbs_config:
                fimage.seek(0)
                with storage.get_storage(forkname, mode='w') as forkThumb:
                    self.resize(fimage, forkThumb, width, height)
                    self.logger.info('Generated %sx%s thumbnail', width, height)
                result = result and storage.exists(forkname)
        return result

    def thumbnail_from_image(self, storage, meta_storage, image_path, thumbs_config):
        """ Generate thumbnails for the given path with the given image_path

        :param image_path: the image fork name
        :param thumbs_config: list containing, for each item, (name, widht,
        height)
        """
        result = True
        with fs.open(image_path, 'r') as fimage:
            self.logger.debug("Got original image")
            for forkname, width, height in thumbs_config:
                fimage.seek(0)
                with storage.get_storage(forkname, mode='w') as forkThumb:
                    self.resize(fimage, forkThumb, width, height)
                    self.logger.debug('Generated %sx%s thumbnail', width, height)
                result = result and storage.exists(forkname)
        return result

from lima.meta.storage import MetaForkStorage


class ImageThumbnailsAgentBase(WithThumbnailsMixin, BlackBoardAgent):

    """ """

    priority = 10

    def __init__(self, thumbs_config, forks_sources=None, *args, **kwargs):
        """

        :param forks_sources: List of fork names and fallback fork to generate thumbnails from.
        Only the first fork that we can thumbnail will be processed.

        :param thumbs_config: a list containting (name, width, height) for each
        thumbnail to generate
        """
        super(ImageThumbnailsAgentBase, self).__init__(*args, **kwargs)
        if forks_sources is None:
            forks_sources = (u'default',)

        self.forks_sources = forks_sources
        self.thumbs_config = thumbs_config
        self.storage = MetaForkStorage(self.path)

    def _check(self):
        thumb_names = [name for name, _, _ in self.thumbs_config]
        return self.can_thumbnail(self.storage, self.forks_sources, thumb_names, self.is_content_modified)

    def _process(self):

        thumb_names = [name for name, _, _ in self.thumbs_config]
        result = False
        for image in self.forks_sources:
            # It's the job of _check to decide if the processing happens
            # in process method, we always considerate that we process
            # no mattter what is the value of self.is_update
            # Hence the 'True' argument for is_update in the call to
            # can_thumbnail
            if self.can_thumbnail(self.storage, (image,), thumb_names, True):
                try:
                    result = self.thumbnail_image(self.path, self.storage, self.meta_storage, image, self.thumbs_config)
                    if result:
                        # We stop on the first we can generate thumbnail
                        break
                # only catch known errors here
                except(LimaImageProcessingError, ):
                    self.logger.error("Could not create thumbnail %s for %s", image, self.path)
        return result


from classes import VeryLowCPUClass


class ImageThumbnailsAgent(ImageThumbnailsAgentBase, ImageAgent):

    priority = 1

    _id = u'picture.thumbnails'
    CLASSES = [VeryLowCPUClass, ]

    def __init__(self, *args, **kwargs):
        super(ImageThumbnailsAgent, self).__init__(
            DEFAULT_THUMBS_CONFIG, *args, **kwargs
        )
        self.max_errors = 1
