# coding: utf-8

import os
import datetime
import re
import unicodedata

from lima.tools.tmdb import tmdb3
from lima.tools import segment

from redo import retrying

from lima.fs import FileSystem as fs


from lima.meta.blackboard.agents.video import (
    VideoAgent,
    VideoInfosAgent,
    VideoNameExtractInfosAgent,
    VideoOsdbInfosAgent,
    TmdbDisabledError,
)
from lima.meta.blackboard.agents.classes import (
    VeryLowCPUClass, LowCPUClass, NetworkClass
)


__all__ = [
    'MovieCoverDownloadAgent',
    'MovieLocalCoverAgent',
    'MovieNfoExtract',
    'MovieSubmitTmdb',
]


class MovieAgent(VideoAgent):

    """ Base class for a movie agent """

    KEY_TITLE = u'movie.title'
    KEY_YEAR = u'movie.year'
    KEY_PICTURE_URL = u'movie.picture.url'
    KEY_ACTORS = u'movie.actors'
    KEY_DIRECTORS = u'movie.directors'
    KEY_PLOT = u'movie.synopsis'

    def _precondition_(self):
        """ Return True if the file is not marked as TV show"""
        activate = super(MovieAgent, self)._precondition_()
        if not activate:
            return False
        is_tv = self.meta_storage.getmeta_internal(
            VideoNameExtractInfosAgent.KEY_ISTV
        )

        activate = is_tv is None or is_tv is False
        if not activate:
            self.logger.debug(u"File is TV show, ignoring")

        fifty_mb = 50 << 20  # 50 * 1024 * 1024
        filesize = fs.stat(self.path).st_size
        if filesize <=fifty_mb:
            self.logger.debug(u"File is too small, ignoring")
            self.context[self.KEY_NEED_REINDEX]=False
            activate = False


        return activate

import re
import xmltodict


class MovieNfoExtract(MovieAgent):

    """ Agent that will try to extract infos from nfo files.

    If nfo file is in XBMC format, this agent will try to extract :
    - imdb id
    - original title
    - actors / directors

    If nfo file is not in XBMC format, it will apply a regular expression to
    find imdbid
    """
    priority = 1
    _id = u"movie.nfo.extract"
    CLASSES = [VeryLowCPUClass, ]

    IMDBID_PATTERN = "(tt\d+)"

    @property
    def nfo_fname(self):
        """Returns the .nfo file name """
        if not hasattr(self, '_nfo_fname'):
            # Get the name without extension
            fname = os.path.splitext(self.path)[0]
            self._nfo_fname = fname + u".nfo"
        return self._nfo_fname

    def _check(self):
        """ This agent activate itself if :
            - a .nfo file with the same name exists
            - there is no imdb id present
        """
        activate = not self.meta_storage.hasmeta(self.KEY_IMDBID)
        if not activate:
            self.logger.debug(u"File already has imdbid")
            return False
        activate = fs.exists(self.nfo_fname)
        if activate:
            self.logger.debug(u"NFO file found")
        return activate

    def _process(self):
        with fs.open(self.nfo_fname, 'r') as f:
            datas = f.read()

        if datas is None or len(datas) == 0:
            self.logger.error(u"NFO content was empty")
            return False
        try:
            nfo = xmltodict.parse(datas)
            if u"movie" in nfo:
                self.logger.debug(u"NFO is XBMC format")
                return self._parse_xbmc_nfo(nfo)
            else:
                # format unknown, will try to extract an imdb id
                return self._parse_regexp(datas)
        except:
            # could not parse NFO file, will try to extract
            # at least an imdb id
            return self._parse_regexp(datas)

    def _parse_xbmc_nfo(self, nfo):
        """ This method extract datas from a nfo file in XBMC format
        see https://code.google.com/p/moviejukebox/wiki/NFO_Files for details.
        """
        imdbid = nfo.get('movie', {}).get('id', None)
        if imdbid is not None and imdbid.startswith(u"tt"):
            self.logger.debug(u"Found imdb id %s", imdbid)
            self._check_abort()
            self.meta_storage.setmeta(self.KEY_IMDBID, imdbid)
            self.set_content_type('movie')
            self.context[self.KEY_NEED_REINDEX] = True

        title = nfo.get("movie", {}).get(u"originaltitle", None)
        if title is not None:
            self.logger.debug(u"Found title %s", title)
            self._check_abort()
            self.meta_storage.setmeta(self.KEY_TITLE, title)
            self.set_content_type('movie')
            self.context[self.KEY_NEED_REINDEX] = True

        year = nfo.get('movie', {}).get('year', None)
        if year is not None:
            self.logger.debug(u"Found year %s", year)
            self._check_abort()
            self.meta_storage.setmeta(self.KEY_YEAR, year)
            self.context[self.KEY_NEED_REINDEX] = True

        if title or year :
            self.logger.info(u"Found movie '%s' (%s)", title or '', year or '')

        actors = [
            {u"name": actor[u"name"], u"character": actor[u"role"]}
            for actor in nfo.get('movie', {}).get('actor')
        ]
        if len(actors) > 0:
            self.logger.debug(u"Found %s actors", len(actors))
            self._check_abort()
            self.meta_storage.setmeta(self.KEY_ACTORS, actors)
            self.context[self.KEY_NEED_REINDEX] = True

        director = nfo.get('movie', {}).get('director', None)
        if director is not None:
            directors = [{u"name": director}]
            self.logger.debug(u"Found %s directors", len(directors))
            self._check_abort()
            self.meta_storage.setmeta(self.KEY_DIRECTORS, directors)
            self.context[self.KEY_NEED_REINDEX] = True

        plot = nfo.get('movie', {}).get('plot', None)
        if plot is not None:
            self._check_abort()
            self.meta_storage.setmeta(self.KEY_PLOT, plot)
            self.context[self.KEY_NEED_REINDEX] = True

        return True

    def _parse_regexp(self, nfo):
        """This methods extract an imdb id from any file with a regexp."""

        regex = re.compile("(tt\d+)")
        imdb_ids = list(set(regex.findall(nfo)))

        if len(imdb_ids) == 1:
            imdbid = imdb_ids[0]
            if imdbid is not None:
                self.logger.debug(u"Found imdb id %s", imdbid)
                self._check_abort()
                self.meta_storage.setmeta(self.KEY_IMDBID, imdbid)
                self.context[self.KEY_NEED_REINDEX] = True
                return True
        return False


from image import WithThumbnailsMixin

from lima.meta.storage import MetaForkStorage


class MovieLocalCoverAgent(MovieAgent, WithThumbnailsMixin):

    """Clas for agent in charge to extract local cover.
    A local cover is a file with the same name as the movie file, with a .jpg
    extension.
    """

    priority = 1
    _id = u'movie.cover.local'

    CLASSES = [VeryLowCPUClass, ]

    thumb_config = [('cover', 320, 320), ]

    def __init__(self, *args, **kwargs):
        super(MovieLocalCoverAgent, self).__init__(*args, **kwargs)
        self.storage = MetaForkStorage(self.path)

    @property
    def local_cover_name(self):
        if not hasattr(self, '_local_cover'):
            # Get the name without extension
            fname = os.path.splitext(self.path)[0]
            self._local_cover = fname + u".jpg"
        return self._local_cover

    def _check(self):
        thumbs_name = [name for name, _, _ in self.thumb_config]

        activate = not self.has_all_thumbs_non_empty(self.storage, *thumbs_name)

        if not activate:
            self.logger.debug(u"All thumbnails already presents")
            return False

        activate = fs.exists(self.local_cover_name)
        if activate:
            self.logger.debug(u"Local cover found : %s", self.local_cover_name)
        return activate

    def _process(self):
        # cover may have been deleted
        if fs.exists(self.local_cover_name):
            self.thumbnail_from_image(
                self.meta_storage,
                self.meta_storage,
                self.local_cover_name, self.thumb_config
            )
            return True
        return False

from tmdb3.tmdb_exceptions import TMDBError, TMDBRequestError




class TmdbMovieAgent(MovieAgent):

    """ Base class for TMDB agents"""

    def is_tmdb_enabled(self):
        enabled = False
        with self.config.lock:
            enabled = self.config.get('tmdb.enabled', False)
            if enabled is not True:
                self.logger.debug(u"TMDB is not activated")
        return enabled

    def _precondition_(self):
        activate = super(TmdbMovieAgent, self)._precondition_()
        if not activate:
            return False
        api_key = self.config.get(u"tmdb.api_key", None)
        if api_key:
            tmdb3.set_key(api_key)
        return self.is_tmdb_enabled()

    def _check_abort(self):
        super(TmdbMovieAgent, self)._check_abort()
        if self.is_tmdb_enabled() is not True:
            raise TmdbDisabledError(u"TMDB disabled")

    def _process_movie(self, movie):
        """ Process a movie found on TMDB """
        try:
            self._process_imdbid(movie)
            self._process_year(movie)
            self._process_title(movie)
            self._process_poster_url(movie)
            self._process_directors(movie)
            self._process_actors(movie)
            self._process_plot(movie)

            title = self.meta_storage.getmeta(self.KEY_TITLE)
            year = self.meta_storage.getmeta(self.KEY_YEAR)
            if title:
                self.logger.info(
                        5*u"-" + u" Identified movie as %s (%s)", title, year or 'unknown'
                )
            return True
        except (TMDBRequestError, ) as e:
            tmdberrno = getattr(e, "tmdberrno", 0)
            # 3 : Authentication failed
            # 7 : Invalid API key
            # 10 : Suspended API key
            # 14 : Authentication failed
            # 16 : Device denied
            # 17 : Session denied
            # 31 : Accound disabled
            if tmdberrno in (3, 7, 10, 14, 16, 17, 31):
                segment.track_error(event="TMDB panic")

                with self.config.lock:
                    self.config['tmdb.enabled']=False
                self.logger.exception("TMDB panic")

            elif tmdberrno not in (9, 21, 34):
                # 9 : service offline
                # 21 : Entry not found
                # 34 : Resource could not be found
                self.logger.exception("Not found")
                # segment.track_error(event="TMDB error")
            return None
        except (TMDBError, ) as e:
            self.logger.exception("Could not process movie")
            # segment.track_error(event="TMDB error")
            return False

    def _process_imdbid(self, movie):
        self._check_abort()
        if not self.meta_storage.hasmeta(self.KEY_IMDBID):
            self.meta_storage.setmeta(self.KEY_IMDBID, movie.imdb)
            self.set_content_type('movie')
            self.context[self.KEY_NEED_REINDEX] = True

    def _process_title(self, movie):
        self._check_abort()
        title = movie.title
        if type(title) is str:
            title = title.decode(u'utf-8')
        self.logger.debug(u"Found movie title %s", title)

        self._check_abort()
        self.meta_storage.setmeta(self.KEY_TITLE, title)
        self.set_content_type('movie')
        self.context[self.KEY_NEED_REINDEX] = True

    def _process_year(self, movie):
        self._check_abort()
        year = None
        try:
            if isinstance(movie.releasedate, datetime.date):
                year = movie.releasedate.year
            elif type(movie.releasedate) in (str, unicode):
                year_pattern = '.*(?P<year>\d{4}).*'
                result = re.match(year_pattern, movie.releasedate)
                if result:
                    year = result.groupdict().get('year')
        except TMDBRequestError:
            # could not obtain movie year
            pass
        if not year:
            return
        if type(year) is int:
            year = unicode(year)
        if type(year) is str:
            year = year.decode(u'utf-8')
        self.logger.debug(u"Found movie year %s", year)

        self._check_abort()
        self.meta_storage.setmeta(self.KEY_YEAR, year)
        self.context[self.KEY_NEED_REINDEX] = True

    def _process_poster_url(self, movie):
        # Accessing poster make a network query
        self._check_abort()
        poster_url = movie.poster.geturl()
        if type(poster_url) is str:
            poster_url = poster_url.decode(u'utf-8')
        self.logger.debug(u"Found poster url %s", poster_url)
        self._check_abort()
        self.meta_storage.setmeta_internal(self.KEY_PICTURE_URL, poster_url)

    def _process_directors(self, movie):
        self._check_abort()
        # Directors
        directors = []
        for director in movie.crew:
            if director.job == "Director":
                directors.append({u'name': director.name})

        self.logger.debug(u"Found %s directors", len(directors))
        self._check_abort()
        self.meta_storage.setmeta(self.KEY_DIRECTORS, directors)
        self.context[self.KEY_NEED_REINDEX] = True

    def _process_actors(self, movie):
        self._check_abort()
        # Actors
        actors = []
        for actor in movie.cast:
            actors.append({u"name": actor.name, u"character": actor.character})
        self.logger.debug(u"Found %s actors", len(actors))

        self._check_abort()
        self.meta_storage.setmeta(self.KEY_ACTORS, actors)
        self.context[self.KEY_NEED_REINDEX] = True

    def _process_plot(self, movie):
        plot = movie.overview
        if plot is not None:
            self._check_abort()
            self.meta_storage.setmeta(self.KEY_PLOT, plot)
            self.context[self.KEY_NEED_REINDEX] = True


class MovieSubmitTmdb(TmdbMovieAgent):

    """ Agent that will try to find a movie on TMDB with a IMDB id"""
    _id = u'movie.tmdb.submit'
    priority = MovieNfoExtract.priority + \
        VideoNameExtractInfosAgent.priority + 1

    CLASSES = [NetworkClass, ]

    # duration threshold for matching with duration in the case tmdb
    # return more than one result
    # 1 to 5 minutes
    DURATION_MATCHING_THRESHOLDS=xrange(1,6)
    MINIMUM_MOVIE_DURATION_MIN = 5

    @property
    def datas(self):
        return self.meta_storage.getmeta_internal(
            VideoNameExtractInfosAgent.KEY_VIDEO_INFO
        ) or {}

    @property
    def _video_title(self):
        return self.datas.get('title', None)

    @property
    def _video_year(self):
        return self.datas.get('year', None)

    @property
    def _movie_title(self):
        return self.meta_storage.getmeta(self.KEY_TITLE)

    @property
    def _movie_imdbid(self):
        return self.meta_storage.getmeta(self.KEY_IMDBID)

    @property
    def _video_duration(self):
        return self.meta_storage.getmeta(VideoInfosAgent.KEY_DURATION)

    @property
    def osdb_datas(self):
        return self.meta_storage.getmeta_internal(
            VideoOsdbInfosAgent.KEY
        ) or {}

    @property
    def _osdb_title(self):
        return self.osdb_datas.get('movie_name', None)

    @property
    def _osdb_year(self):
        return self.osdb_datas.get('movie_year', None)

    @property
    def _title(self):
        return self._osdb_title or self._video_title

    @property
    def _year(self):
        return self._osdb_year or self._video_year

    def _check(self):
        # If movie title is already present, abort
        activate = self._movie_title is None or (self.is_basename_modified or self.is_content_modified)
        if not activate:
            self.logger.debug(u"Movie already have title.")
            return False

        has_title = (self._title is not None and self._title.strip() != '')

        has_imdbid = self._movie_imdbid is not None
        # If we don't have vide title and video year or imdb id, abort
        activate = (has_title or has_imdbid)
        if not activate:
            self.logger.debug(u"Video does not has sufficient infos")
            return False
        if not self._movie_imdbid and len(self._title or '') < 3:
            self.logger.debug(u"No imdb id and title too short")
            return False

        # if we have the duration and the duration is < to a 5 min threshold,
        # do not process
        # note : we may not have a duration at this point.
        # also, we may never have a duration if problems occurs with ffprobe
        video_duration = self._video_duration

        if video_duration is not None:
            if video_duration < self.MINIMUM_MOVIE_DURATION_MIN * 60:
                self.logger.debug(u"Video file too short (%s sec)", video_duration)
                return False

        return activate

    def _process(self):
        self._check_abort()

        movie = None

        imdb_id = self._movie_imdbid
        if imdb_id is not None:
            movie = self._get_by_imdb_id(imdb_id)

        if movie is None:
            self._check_abort()
            title = self._title
            year = self._year
            if title is not None:
                self._check_abort()
                if year is not None:
                    movie = self._get_by_title_year(title, year)
                else:
                    movie = self._get_by_title(title)


        if movie is not None:
            self._check_abort()
            return self._process_movie(movie)
        return False

    def _get_by_title(self, title):
        return self._get_by_title_year(title, None)

    def _get_by_title_year(self, title, year):
        try:
            with retrying(tmdb3.searchMovie, attempts=3, sleeptime=5, jitter=2) as tmdb3_searchMovie:
                res = tmdb3_searchMovie(title, year=year)

            if len(res) == 0:
                # if we can't find a result with the extracted result, we try
                # with an unicode free parameter if different from original
                # title.
                # ex : if "2001 l'Odysée de l'espace" gives no result, we try
                # with "2001 l'Odysee de l'espace" (accent is removed)
                if type(title) is str:
                    title=title.decode("utf-8")
                str_cleaned_title = unicodedata.normalize('NFKD', title).encode('ascii', 'ignore')
                if str_cleaned_title != title.encode('utf-8'):
                    with retrying(tmdb3.searchMovie, attempts=3, sleeptime=5, jitter=2) as tmdb3_searchMovie:
                        res = tmdb3_searchMovie(str_cleaned_title, year=year)

            if len(res) == 1:
                return res[0]
            elif len(res) >= 1:
                if len(res) > 10:
                    # this will case too many queries and potentially too many
                    # processing
                    self.logger.info("Too many TMDB results (%s results)", len(res))
                    return

                movie = self._match_by_year(res)
                if movie is None:
                    movie  = self._match_by_duration(res)

                if movie is not None:
                    return movie
                else:
                    self.logger.info(u"No match for movie")
                    return None

            return None
        except (TMDBRequestError, ) as e:
            tmdberrno = getattr(e, "tmdberrno", 0)
            # 3 : Authentication failed
            # 7 : Invalid API key
            # 10 : Suspended API key
            # 14 : Authentication failed
            # 16 : Device denied
            # 17 : Session denied
            # 31 : Accound disabled
            if tmdberrno in (3, 7, 10, 14, 16, 17, 31):
                segment.track_error(event="TMDB panic")

                self.logger.error("TMDB panic")
                with self.config.lock:
                    self.config['tmdb.enabled']=False

            elif tmdberrno not in (9, 21, 34):
                # 9 : service offline
                # 21 : Entry not found
                # 34 : Resource could not be found
                # segment.track_error(event="TMDB error")
                pass
            return None
        except (TMDBError,) as e:
            self.logger.exception("TMDB error")
            # segment.track_error(event="TMDB error")
            return None

    def _match_by_year(self, res):
        year = self._year
        if not year:
            self.logger.debug(u"year not available - not comparing it to results")
            return None

        for movie in res:
            if not movie.releasedate or not movie.releasedate.year:
                continue
            diff = abs(int(movie.releasedate.year) - int(year))
            # allow a +1/-1 year error
            if diff <= 1 :
                title, originaltitle = movie.title, movie.originaltitle
                year = movie.releasedate
                self.logger.info(u"Movie '%s (%s)' ('%s') found with %s diff on year",
                        title, year, movie.originaltitle, diff
                )
                return movie
        return None

    def _match_by_duration(self, res):
        video_duration = self._video_duration
        if video_duration is None or video_duration==u'':
            return None

        # convert duration in minutes
        video_duration = video_duration / 60

        for threshold in self.DURATION_MATCHING_THRESHOLDS:
            for movie in res:
                if not movie.runtime:
                    continue
                diff = abs(int(movie.runtime) - int(video_duration))
                title, originaltitle = movie.title, movie.originaltitle
                year = movie.releasedate
                if diff <= threshold:
                    self.logger.info(u"Movie '%s (%s)' ('%s') found with %s sec diff on duration",
                            title, year, movie.originaltitle, diff
                    )
                    return movie
                else:
                    self.logger.debug(u"Movie '%s (%s)' ('%s') does not match (%s sec) ",
                            title, year, originaltitle, diff
                    )

        return None

    def _get_by_imdb_id(self, imdb_id):
        try:
            with retrying(tmdb3.Movie, attempts=3, sleeptime=5, jitter=2) as tmdb3_Movie:
                movie = tmdb3_Movie(imdb_id)
                return movie
        except (TMDBRequestError, ) as e:
            tmdberrno = getattr(e, "tmdberrno", 0)
            # 3 : Authentication failed
            # 7 : Invalid API key
            # 10 : Suspended API key
            # 14 : Authentication failed
            # 16 : Device denied
            # 17 : Session denied
            # 31 : Accound disabled
            if tmdberrno in (3, 7, 10, 14, 16, 17, 31):
                segment.track_error(event="TMDB panic")

                self.logger.error("TMDB panic")
                with self.config.lock:
                    self.config['tmdb.enabled']=False

            elif tmdberrno not in (9, 21, 34):
                # 9 : service offline
                # 21 : Entry not found
                # 34 : Resource could not be found
                # segment.track_error(event="TMDB error")
                pass
            return None
        except (TMDBError,) as e:
            self.logger.error("TMDB error")
            # segment.track_error(event="TMDB error")
            return None



from lima.meta.blackboard.agents.network import ImageDownloadAgent


class MovieCoverDownloadAgent(MovieAgent, ImageDownloadAgent):

    """Agent that will download and resize the posted for a movie. """

    COVER_URL_KEY = u'movie.picture.url'

    priority = MovieLocalCoverAgent.priority +1

    _id = u'movie.cover.download'

    CLASSES = [NetworkClass, ]

    thumb_config = [('cover', 320, 320), ]

    def __init__(self, *args, **kwargs):
        super(MovieCoverDownloadAgent, self).__init__(
            self.thumb_config, timeout=5,
            *args, **kwargs
        )

    def _process(self):
        success = super(MovieCoverDownloadAgent, self)._process()
        if success:
            self.context[self.KEY_NEED_REINDEX] = True
        return success

    @property
    def download_storage(self):
        return self.meta_storage

    @property
    def url(self):
        url = self.meta_storage.getmeta_internal(self.COVER_URL_KEY)
        if url is not None and len(url.replace(' ', '')) == 0:
            return None
        return url
