# coding: utf-8

import sys
import requests
from redo import retrying

from lima.meta.blackboard.agents.base import BlackBoardAgent
from lima.fs import FileSystem as fs
from lima.fs.utils import DeleteForkOnError

import abc

import logging
logger = logging.getLogger(__name__)


class LimaNetworkError(Exception):
    pass


class DownladMixin(object):
    def __init__(self, *args, **kwargs):
        super(DownladMixin, self).__init__(*args, **kwargs)
        if not hasattr(self, 'logger'):
            self.logger = logger

    def download(self, url, expected_content_type, timeout=1):
        response = None
        try:
            with retrying(requests.get, attempts=3, sleeptime=5, jitter=2) as requests_get:
                response = requests_get(url, timeout=timeout)
        except(Exception, ) as e:
            self.logger.exception('Could not get the url %s', url)
            raise LimaNetworkError, LimaNetworkError(e), sys.exc_info()[2]

        content_type = response.headers.get('content-type', '')
        if content_type is not None \
                and not content_type.startswith(expected_content_type):
            self.logger.debug("Content was not as expected : '%s'", content_type)
            return None
        # so far so good
        return response.content


from classes import NetworkClass


class ContentDownloadAgent(DownladMixin, BlackBoardAgent):

    __metaclass__ = abc.ABCMeta

    CLASSES = [NetworkClass, ]
    EXCEPTIONS_RETRY = [LimaNetworkError, ]

    @abc.abstractproperty
    def url(self):
        return None  # pragma : no cover

    @abc.abstractproperty
    def content_type(self):
        return None  # pragma : no cover

    def _check(self):
        if self.url is None:
            self.logger.debug("No url available")
            return False

        return True

    def _get_content(self, timeout=1):
        self.logger.debug("Processing  download")
        content = None

        self._check_abort()
        if self.url is not None:
            content = self.download(self.url, self.content_type, timeout)
        return content



from StringIO import StringIO
from image import ImageResizeMixin, WithThumbnailsMixin


class ImageDownloadAgent(WithThumbnailsMixin, ImageResizeMixin, ContentDownloadAgent):

    __metaclass__ = abc.ABCMeta

    def __init__(self, thumbs_config, timeout=1, *args, **kwargs):
        self.thumbs_config = thumbs_config
        self.timeout = timeout
        super(ImageDownloadAgent, self).__init__(*args, **kwargs)

    @abc.abstractproperty
    def download_storage(self):
        pass

    @property
    def content_type(self):
        return 'image/'

    def _check(self):
        activate =  super(ImageDownloadAgent, self)._check()
        if activate:
            thumb_names = [thumbname for thumbname, _, _ in self.thumbs_config]
            activate = not self.has_all_thumbs_non_empty(self.download_storage, *thumb_names)
        return activate

    def _process(self):
        content = self._get_content(timeout=self.timeout)
        if content is None:
            return False

        fimage = StringIO(content)
        for forkname, width, height in self.thumbs_config:
            fimage.seek(0)
            with self.download_storage.get_storage(forkname, mode='w') as forkThumb:
                self.resize(fimage, forkThumb, width, height)
                logger.debug('Generated %sx%s thumbnail from downloaded file', width, height)
        return True
