# coding: utf-8
from lima.meta.storage import MetaKeyValueStorage


class AgentStorage(object):
    """Proxy class which mimics MetaKeyValueStorage but allow to store 'internal' metadata
    in a second fork.

    All original methods can be called with suffix '_internal' to store in the
    internal key value storage.
    """
    debug = False

    def __init__(self, path, is_debug=False, *args, **kwargs):
        self.storage = MetaKeyValueStorage.get_from_pool(path, 'meta', *args,
                                                         **kwargs)

        AgentStorage.debug = is_debug
        self.internal_storage = MetaKeyValueStorage.get_from_pool(
            path, 'internal', *args, **kwargs)
        # force the path to help debug
        if is_debug:
            self.internal_storage.setmeta(u'path', path)

    def __getattr__(self, attr):
        if '_internal' in attr:
            return getattr(self.internal_storage, attr.replace('_internal', ''))
        else:
            return getattr(self.storage, attr)

    def persist(self, *args, **kwargs):
        self.storage.persist(*args, **kwargs)
        if self.debug:
            self.internal_storage.persist(*args, **kwargs)

    @classmethod
    def persist_metas(cls, path, *args, **kwargs):
        MetaKeyValueStorage.persist_metas(path, u'meta', *args, **kwargs)
        if cls.debug:
            # As an optim, we could avoid to persists the internal storage
            MetaKeyValueStorage.persist_metas(path, u'internal', *args, **
                                              kwargs)

        if kwargs.get('forget', False):
            # as we do not persists the internal storage in non debug mode,
            # we have to unload it manually
            MetaKeyValueStorage.forget(path, u'internal', persist=cls.debug)

    @classmethod
    def forget(cls, path, *args, **kwargs):
        try:
            MetaKeyValueStorage.forget(path, u'meta', *args, **kwargs)
        finally:
            MetaKeyValueStorage.forget(path, u'internal', *args, **kwargs)

    @classmethod
    def remove_key(cls, *args, **kwargs):
        MetaKeyValueStorage.remove_key(*args, **kwargs)

    @classmethod
    def remove_empty(cls, *args, **kwargs):
        MetaKeyValueStorage.remove_empty(*args, **kwargs)

    @classmethod
    def iter_files(cls):
        for f in MetaKeyValueStorage.iter_files():
            yield f

    @classmethod
    def iter_files_stat(cls, *args, **kwargs):
        for f, stat in MetaKeyValueStorage.iter_files_stat(*args, **kwargs):
            yield f, stat

    @classmethod
    def unload_all(cls):
        MetaKeyValueStorage.unload_all()
