# coding: utf-8
from lima.tools.tmdb import tmdb3
from lima.fs import FileSystem as fs

from lima.meta.storage import SerieStorage, SeasonStorage
from lima.meta.blackboard.agents.video import (
    VideoAgent,
    VideoNameExtractInfosAgent,
    VideoOsdbInfosAgent,
    TmdbDisabledError,
)
from lima.meta.blackboard.agents.classes import (
    VeryLowCPUClass, NetworkClass
)
from lima.tools import segment

__all__ = [
    "TvInfoExtractor",
    "TvSeasonCoverDownloadAgent",
    "TvSerieCoverDownloadAgent",
    "TvEpisodeCoverLink",
]


class TvAgent(VideoAgent):

    thumbs_config = [('cover', 320, 320), ]

    """ Base class for a serie agent """

    KEY_SERIE_NAME = u'serie.name'
    KEY_SERIE_ORIGINAL_NAME = u'serie.original_name'
    KEY_SERIE_OVERVIEW = u'serie.overview'

    KEY_SEASON_NAME = u'serie.season.name'
    KEY_SEASON_NUMBER = u'serie.season.number'
    KEY_SEASON_OVERVIEW = u'serie.season.overview'

    KEY_EPISODE_NAME = u'serie.episode.name'
    KEY_EPISODE_NUMBER = u'serie.episode.number'
    KEY_EPISODE_OVERVIEW = u'serie.episode.overview'

    KEY_SERIE_PICTURE_URL = u'serie.picture.url'
    KEY_SEASON_PICTURE_URL = u'season.picture.url'

    KEY_ACTORS = u'actors'
    KEY_DIRECTORS = u'directors'
    KEY_WRITER = u'writers'
    KEY_PRODUCERS = u'producers'
    KEY_NETWORKS = u'networks'

    def _precondition_(self):
        """ Return True if the file is not marked as TV show"""
        activate = super(TvAgent, self)._precondition_()
        if not activate:
            return False
        is_tv = self.meta_storage.getmeta_internal(
            VideoNameExtractInfosAgent.KEY_ISTV
        )

        activate = is_tv is True
        if not activate:
            self.logger.debug(u"File is not TV show, ignoring")

        fifty_mb = 50 << 20  # 50 * 1024 * 1024
        filesize = fs.stat(self.path).st_size
        if filesize <= fifty_mb:
            self.logger.debug(u"File is too small, ignoring")
            self.context[self.KEY_NEED_REINDEX] = False
            activate = False
        return activate

    @property
    def has_serie_name(self):
        return self.meta_storage.hasmeta(self.KEY_SERIE_NAME)

    @property
    def has_season_number(self):
        return self.meta_storage.hasmeta(self.KEY_SEASON_NUMBER)

    @property
    def serie_name(self):
        return self.meta_storage.getmeta(self.KEY_SERIE_NAME)

    @property
    def season_number(self):
        return self.meta_storage.getmeta(self.KEY_SEASON_NUMBER)

    @property
    def serie_storage(self):
        if not hasattr(self, '_serie_storage'):
            self._serie_storage = SerieStorage.get_from_pool(self.serie_name)
        return self._serie_storage

    @property
    def season_storage(self):
        if not hasattr(self, '_season_storage'):
            self._season_storage = SeasonStorage.get_from_pool(self.serie_name, self.season_number)
        return self._season_storage


class TmdbTvAgent(TvAgent):

    """ Base class for TMDB agents"""

    def is_tmdb_enabled(self):
        enabled = False
        with self.config.lock:
            enabled = self.config.get('tmdb.enabled', False)
            if enabled is not True:
                self.logger.debug(u"TMDB is not activated")
        return enabled

    def _precondition_(self):
        activate = super(TmdbTvAgent, self)._precondition_()
        if not activate:
            return False
        api_key = self.config.get(u"tmdb.api_key", None)
        if api_key:
            tmdb3.set_key(api_key)
        return self.is_tmdb_enabled()

    def _check_abort(self):
        super(TmdbTvAgent, self)._check_abort()
        if self.is_tmdb_enabled() is not True:
            raise TmdbDisabledError(u"TMDB disabled")

    def process_serie(self, serie, season, episode):
        """ Process a serie found on TMDB """
        try:
            self._process_serie(serie)
            self._process_season(season)
            self._process_episode(episode)
            return True
        except (tmdb3.TMDBError, ):
            # segment.track_error(event="TMDB error")
            return False

    def _process_serie(self, serie):
        try:
            self.meta_storage.setmeta(self.KEY_SERIE_NAME, serie.name)
            self.set_content_type('episode')

            self.serie_storage.setmeta(self.KEY_SERIE_NAME, serie.name)
            self.serie_storage.setmeta(self.KEY_SERIE_ORIGINAL_NAME, serie.original_name)
            self._process_networks(serie)
            self._process_actors(serie, self.serie_storage)
            self._process_crew(serie, self.serie_storage, u'Director', self.KEY_DIRECTORS)
            self._process_crew(serie, self.serie_storage, u'Writer', self.KEY_WRITER)
            self._process_crew(serie, self.serie_storage, u'Producer', self.KEY_PRODUCERS)

            should_persist = False
            try:
                if not self.serie_storage.hasmeta(self.KEY_SERIE_PICTURE_URL)\
                        and serie.poster is not None:
                    self.serie_storage.setmeta(self.KEY_SERIE_PICTURE_URL, serie.poster.geturl())
                    should_persist = True
                if not self.serie_storage.hasmeta(self.KEY_SERIE_OVERVIEW):
                    self.serie_storage.setmeta(self.KEY_SERIE_OVERVIEW, serie.overview)
                    should_persist = True

                if should_persist:
                    self.serie_storage.persist()
            finally:
                self.serie_storage.unload()
        except (tmdb3.TMDBError, ):
            # segment.track_error(event="TMDB error")
            self.logger.exception(u"Error while processing serie ")

    def _process_season(self, season):
        try:
            # also store a file level
            self.meta_storage.setmeta(self.KEY_SEASON_NUMBER, season.season_number)
            self.meta_storage.setmeta(self.KEY_SEASON_NAME, season.name)
            #
            self.season_storage.setmeta(self.KEY_SEASON_NAME, season.name)
            self.season_storage.setmeta(self.KEY_SEASON_NUMBER, season.season_number)

            should_persist = False
            try:
                if not self.season_storage.hasmeta(self.KEY_SEASON_PICTURE_URL)\
                        and season.poster is not None:
                    self.season_storage.setmeta(self.KEY_SEASON_PICTURE_URL, season.poster.geturl())
                    should_persist = True
                if not self.season_storage.hasmeta(self.KEY_SEASON_OVERVIEW):
                    self.season_storage.setmeta(self.KEY_SEASON_OVERVIEW, season.overview)
                    should_persist = True

                if should_persist:
                    self.season_storage.persist()
            finally:
                self.season_storage.unload()

        except (tmdb3.TMDBError, ):
            # segment.track_error(event="TMDB error")
            self.logger.exception(u"Error while processing season")

    def _process_episode(self, episode):
        try:
            self.meta_storage.setmeta(self.KEY_EPISODE_NAME, episode.name)
            self.meta_storage.setmeta(self.KEY_EPISODE_NUMBER, episode.episode_number)
            self.meta_storage.setmeta(self.KEY_EPISODE_OVERVIEW, episode.overview)
            self._process_actors(episode, self.meta_storage)
            self._process_crew(episode, self.meta_storage, u'Director', self.KEY_DIRECTORS)
            self._process_crew(episode, self.meta_storage, u'Writer', self.KEY_WRITER)
        except tmdb3.TMDBError:
            # segment.track_error(event="TMDB error")
            self.logger.exception(u"Error while processing season")

    def _process_actors(self, item, storage):
        self._check_abort()
        # Actors
        actors = []
        for actor in item.cast:
            actors.append({u"name": actor.name, u"character": actor.character})
        self.logger.debug(u"Found %s actors", len(actors))

        self._check_abort()
        if len(actors) > 0:
            storage.setmeta(self.KEY_ACTORS, actors)
            self.context[self.KEY_NEED_REINDEX] = True

    def _process_crew(self, item, storage, job, key):
        self._check_abort()
        # persons
        persons = []
        for person in item.crew:
            if person.job == job:
                persons.append({u'name': person.name})

        self.logger.debug(u"Found %s %s", len(persons), job)
        self._check_abort()
        storage.setmeta(key, persons)
        if len(persons) > 0:
            self.context[self.KEY_NEED_REINDEX] = True

    def _process_networks(self, serie):
        self._check_abort()
        # persons
        networks = [n.name for n in serie.networks]
        if len(networks) > 0:
            self.serie_storage.setmeta(self.KEY_NETWORKS, networks)
            self.context[self.KEY_NEED_REINDEX] = True


class TvInfoExtractor(TmdbTvAgent):

    _id = u"tv.info.extract"
    priority = VideoNameExtractInfosAgent.priority + 1

    ####################################
    # Video infos extracted infos
    ####################################
    @property
    def datas(self):
        return self.meta_storage.getmeta_internal(
            VideoNameExtractInfosAgent.KEY_VIDEO_INFO
        ) or {}

    @property
    def video_infos_series_name(self):
        return self.datas.get('series', None)

    @property
    def video_infos_episode_number(self):
        ep = self.datas.get('episodeNumber', None)
        if ep is not None:
            return int(ep)
        return None

    @property
    def video_infos_season_number(self):
        ep = self.datas.get('season', None)
        if ep is not None:
            return int(ep)
        return None

    ##################
    # OSDB fetched infos
    ##################
    @property
    def osdb_datas(self):
        return self.meta_storage.getmeta_internal(
            VideoOsdbInfosAgent.KEY
        ) or {}

    @property
    def osdb_series_name(self):
        return self.osdb_datas.get('serie_title', None)

    @property
    def osdb_episode_number(self):
        ep = self.osdb_datas.get('episode_number', None)
        if ep is not None:
            return int(ep)
        return None

    @property
    def osdb_season_number(self):
        ep = self.osdb_datas.get('season_number', None)
        if ep is not None:
            return int(ep)
        return None

    def _check(self):
        serie = self.meta_storage.getmeta(self.KEY_SERIE_NAME, None)
        # if the serie name is not found, no chance that the rest was found;
        return serie is None or\
            self.is_basename_modified

    def _process(self):
        try:
            # If we have a result from OMDB, use it against TMDB
            serie_name = self.osdb_series_name
            episode_number = self.osdb_episode_number
            season_number = self.osdb_season_number

            if not serie_name:
                serie_name = self.video_infos_series_name
                episode_number = self.video_infos_episode_number
                season_number = self.video_infos_season_number

            result = tmdb3.searchSeries(serie_name)
            if len(result) == 0:
                self.logger.warn(u"No match found")
                return self._process_fallback()
            serie = result[0]

            season = None
            if len(serie.seasons) > season_number:
                season = serie.seasons[season_number]
            if season is not None:
                if len(season.episodes) > episode_number:
                    episode = season.episodes[episode_number]
                    self.process_serie(serie, season, episode)
                    return True
            # if we get there, we ha not found it on TMDB
            # set what we have found as fallback
            return self._process_fallback()

        except (tmdb3.TMDBError, ):
            # segment.track_error(event="TMDB error")
            return False
        return False

    def _process_fallback(self):
        """ Fallback for infos """
        # we ;ay have a result on osdb
        res = self._process_fallback_osdb_infos()
        if not res:
            # last resort : we may have a result with guessit
            res = self._process_fallback_videos_infos()
        return res

    def _process_fallback_osdb_infos(self):
        """ If no info found on TMDB at all but results found with OSDB, use
        OSDB infos as fallback
        fall back.
        """
        res = True
        osdb_infos = self.meta_storage.getmeta_internal(VideoOsdbInfosAgent.KEY)

        if osdb_infos.get('serie_title', None):
            self.meta_storage.setmeta(self.KEY_SERIE_NAME, osdb_infos.get('serie_title', None))
        else:
            res = False
        if osdb_infos.get('season_number', None):
            self.meta_storage.setmeta(self.KEY_SEASON_NUMBER, osdb_infos.get('season_number', None))
        else:
            res = False
        if osdb_infos.get('episode_number', None):
            self.meta_storage.setmeta(self.KEY_EPISODE_NUMBER, osdb_infos.get('episode_number', None))
            res = res and True
        else:
            res = False

        if osdb_infos.get('episode_title', None):
            # bonus : we may have an episode title
            self.meta_storage.setmeta(self.KEY_EPISODE_NAME, osdb_infos.get('episode_title', None))
            res = res and True

        if res is True:
            self.context[self.KEY_NEED_REINDEX] = True
            self.set_content_type('episode')

        return res

    def _process_fallback_videos_infos(self):
        """ If no info found on TMDB at all, use infos extracted by guessit as a
        fall back.
        We should be able to find serie name, season number, episode number
        """
        res = True
        video_infos = self.meta_storage.getmeta_internal('video.infos')

        if video_infos.get('series', None):
            self.meta_storage.setmeta(self.KEY_SERIE_NAME, video_infos.get('series', None))
        else:
            res = False
        if video_infos.get('season', None):
            self.meta_storage.setmeta(self.KEY_SEASON_NUMBER, video_infos.get('season', None))
        else:
            res = False
        if video_infos.get('episodeNumber', None):
            self.meta_storage.setmeta(self.KEY_EPISODE_NUMBER, video_infos.get('episodeNumber', None))
            res = res and True
        else:
            res = False

        if res is True:
            self.context[self.KEY_NEED_REINDEX] = True
            self.set_content_type('episode')

        return res


from lima.meta.blackboard.agents.network import ImageDownloadAgent


class TvCoverDownloadAgent(TvAgent, ImageDownloadAgent):
    CLASSES = [NetworkClass, ]

    thumb_config = [('cover', 320, 320), ]

    def __init__(self, *args, **kwargs):
        super(TvCoverDownloadAgent, self).__init__(
            self.thumb_config, timeout=5,
            *args, **kwargs
        )

    @property
    def url(self):
        url = self.download_storage.getmeta(self.COVER_URL_KEY)
        if url is not None and len(url.replace(' ', '')) == 0:
            return None
        return url

    def _process(self):
        success = super(TvCoverDownloadAgent, self)._process()
        if success:
            self.context[self.KEY_NEED_REINDEX] = True
            self.download_storage.persist()
            self.download_storage.unload()
        return success


class TvSerieCoverDownloadAgent(TvCoverDownloadAgent):

    """Agent that will download and resize the poster for a serie. """

    COVER_URL_KEY = TvAgent.KEY_SERIE_PICTURE_URL

    priority = TvInfoExtractor.priority

    _id = u'tv.serie.picture.url'

    def _check(self):
        if not self.has_serie_name:
            return False
        return super(TvSerieCoverDownloadAgent, self)._check()

    @property
    def download_storage(self):
        return self.serie_storage


class TvSeasonCoverDownloadAgent(TvCoverDownloadAgent):

    """Agent that will download and resize the poster for a season."""

    COVER_URL_KEY = TvAgent.KEY_SEASON_PICTURE_URL

    priority = TvInfoExtractor.priority

    _id = u'tv.season.picture.url'

    def _check(self):
        if not self.has_serie_name:
            return False
        if not self.has_season_number:
            return False
        return super(TvSeasonCoverDownloadAgent, self)._check()

    @property
    def download_storage(self):
        return self.season_storage


from image import WithThumbnailsMixin


class TvEpisodeCoverLink(WithThumbnailsMixin, TvAgent):

    CLASSES = [VeryLowCPUClass, ]

    """Agent that will link artist/album previews to the file"""

    priority = TvSeasonCoverDownloadAgent.priority + 1

    _id = u'tv.season.cover.link'

    def _check(self):
        thumb_names = [thumbname for thumbname, _, _ in self.thumbs_config]
        if not self.has_serie_name:
            return False
        if not self.has_season_number:
            return False
        season_has_all_thumbs = self.has_all_thumbs_non_empty(
            self.season_storage, *thumb_names
        )
        file_has_all_thumbs = self.has_all_thumbs_non_empty(
            self.meta_storage, *thumb_names
        )
        already_same_link = True
        for thumbname in thumb_names:
            already_same_link = already_same_link and self.meta_storage.is_same_link(self.season_storage, thumbname, thumbname)
        return season_has_all_thumbs and (not file_has_all_thumbs or (self.is_content_modified and not already_same_link))

    def _process(self):
        self.logger.debug(u"Processing cover link")
        for thumbname, _, _ in self.thumbs_config:
            # symlink artist/album preview to file
            self.meta_storage.link(self.season_storage, thumbname, thumbname)
            self.logger.info(
                u"%s/%s/%s linked", self.serie_name, self.season_number, thumbname
            )
        self.context[self.KEY_NEED_REINDEX] = True
