# coding: utf-8

import os
import guessit
from redo import retrying
from lima.fs import FileSystem as fs

from lima.meta.blackboard.agents.base import BlackBoardAgent
from lima.meta.blackboard.agents.common import FileTypeMixin

from lima.meta.blackboard.agents.base import AgentAbortedError
from lima.tools import segment
from lima.tools import strings

__all__ = [
    'VideoOsdbImdbIdAgent',
    'VideoPreviewExtractorAgent',
    'VideoInfosAgent',
    'VideoNameExtractInfosAgent',
    'VideoOsdbInfosAgent'
]

# Default size for extracted image.
# -1 means keep ratio
DEFAULT_PREVIEW_IMAGE_SIZE = (320, -1)
# Time in second where to find the thumbnail
DEFAULT_THUMB_OFFSET = 100
# If the video length is < DEFAULT_THUMB_OFFSET,
# we will take the total duration divided by
# this value
DEFAULT_THUMB_DIV_OFFSET_FALLBACK = 2
# default thumbail format
DEFAULT_THUMB_FORMAT = 'jpg'

# DEFAULT_THUMBS_CONFIG format :
# preview name, width, height
# not that -1 is for ffmpeg only  and must not be used with PIL as it will
# cause errors
DEFAULT_THUMBS_CONFIG = [
    ('pvw', 320, -1),  # pvw : preview for client

    # ('pvw32', 32, 32),  # thumb32 : windows thumbnail
    # ('pvw96', 96, 96),  # thumb96 : windows thumbnail
    # ('pvw256', 256, 256),  # thumb256 : windows thumbnail
    # ('pvw1024', 1024, -1),  # thumb1024 : windows thumbnail.
]


class VideoAgent(FileTypeMixin, BlackBoardAgent):
    KEY_IMDBID = u'movie.imdb.id'
    KEY_ISTV = u'video.is_tv'
    KEY_TITLE = u'title'
    KEY_YEAR = u'year'

    def _precondition_(self):
        activate = self._check_mime(self.meta_storage, 'video')
        if activate:
            self.logger.debug(u"File is video. Activate.")
        return activate

DEFAULT_THUMBS_SIZES = (
    (320, -1),
)
from lima.tools.ffmpeg import (
    VideoFile, FFmpegError
)


from lima.meta.blackboard.agents.classes import (
    VideoPictureExtractor, LowCPUClass, NetworkClass, VeryLowCPUClass
)


from image import WithThumbnailsMixin


class VideoPreviewExtractorAgent(WithThumbnailsMixin, VideoAgent):
    CLASSES = [VideoPictureExtractor, ]

    """ Class that extract a thumbnail from a video file
    with a call to ffmpeg.

    By spawning ffmpeg, this agent do a heavy work.
    If previews in multiples sizes are needed, there should only be one call to
    this agent and a seond one who will just resize the extracted picture.
    """

    priority = 10

    _id = u'video.preview.extract'

    def __init__(self, *args, **kwargs):
        super(VideoPreviewExtractorAgent, self).__init__(*args, **kwargs)
        # sorts thumbs config by size DESC
        self.thumbs_config = sorted(DEFAULT_THUMBS_CONFIG, key=lambda x: -x[1])
        self.max_errors = 1

    def _check(self):
        thumb_names = [thumbname for thumbname, _, _ in self.thumbs_config]
        # cover is not present in the storage
        # or content had been modified
        activate = not self.has_all_thumbs_non_empty(self.meta_storage, *thumb_names)\
                or self.is_content_modified

        if not activate:
            self.logger.debug(u"File already have preview")
            return False

        # Check if the filename encoding is consistent.
        # If its not the case, we won't be able to encode it to a bytestring to use it with ffmpeg
        fname_enc_consistent = strings.is_filename_encoding_consistent(fs.get_system_path(self.path))
        if not fname_enc_consistent:
            self.logger.warning(u"Filename encoding is inconsistent. Not generating thumbnail")
            return False

        return activate

    def _process(self):
        try:
            self._process_with_path()
            self.logger.info(u'Success.')
        except(FFmpegError, ):
            self.logger.error(u'Could not extract video preview for %s', self.path)
            return False
        return True

    def _process_with_path(self):
        self._check_abort()
        self._process_file(fs.get_system_path(self.path))

    def _process_file(self, fp):
        self.logger.debug(u"Starting preview extraction")
        config = {}
        config['format'] = DEFAULT_THUMB_FORMAT

        offset = DEFAULT_THUMB_OFFSET
        video_file = VideoFile(fp)
        # thumb config is sorted from bigger to smaller
        # in this way, we only process the biggest image with ffmpeg
        # and process others with PIL
        biggest_config = self.thumbs_config[0]
        with self.run_task(video_file):
            self._check_abort()
            if video_file.duration < offset:
                # video is too short, we arbitrary go to 1/3
                offset = int(
                    video_file.duration / DEFAULT_THUMB_DIV_OFFSET_FALLBACK
                )
            config['offset'] = offset
            config['size'] = (biggest_config[1], biggest_config[2])
            biggest_image_name = biggest_config[0]

            self._check_abort()

            with self.meta_storage.get_storage(biggest_image_name, mode=u'w') as biggest_image:
                video_file.thumbnail(fp_out=biggest_image, **config)

            with self.meta_storage.get_storage(biggest_image_name, mode=u'r') as biggest_image:
                for thumbname, width, height in self.thumbs_config[1:]:
                    biggest_image.seek(0)
                    with self.meta_storage.get_storage(thumbname, mode='w') as thumbnail:
                        self.resize(biggest_image, thumbnail, width, height)
                        self.logger.debug(
                            u"Generated %sx%s thumbnail", width, height)


class VideoInfosAgent(VideoAgent):

    """ Class that extract infos from a video file with a call to ffmpeg.
    For now only the duration is extracted.
    """

    CLASSES = [LowCPUClass, ]

    KEY_DURATION = u'video.duration'
    priority = 5

    _id = u'video.duration.extract'

    def _check(self):
        activate = not self.meta_storage.hasmeta(VideoInfosAgent.KEY_DURATION)\
                or self.is_content_modified

        if not activate:
            self.logger.debug(u"Video already have duration",)
            return False

        return activate

    def _process(self):
        self.logger.debug(u"Starting video infos extraction")
        self._check_abort()
        video_file = VideoFile(fs.get_system_path(self.path))
        duration = None
        try:
            duration = video_file.duration
        except(FFmpegError, ):
            self.logger.error(u'Could not extract duration')
            return False

        if duration is None:
            self.logger.error(u'Duration was None')
            return False

        self._check_abort()
        old_duration = self.meta_storage.getmeta(VideoInfosAgent.KEY_DURATION, duration)
        # Duration is a Decimal, convert to float.
        self.meta_storage.setmeta(VideoInfosAgent.KEY_DURATION, float(duration))
        # flag as to be reindexed if the file is a movie
        content_type = self.meta_storage.getmeta('content_type', None)
        if old_duration != duration and content_type:
            self.context[self.KEY_NEED_REINDEX]=True
        self.logger.info(u'Duration extracted')
        return True


class VideoNameExtractInfosAgent(VideoAgent):

    """ Class that extract infos from a movie file name.
    For now only the title is extracted.
    """

    _id = u'video.infos.extract'
    CLASSES = [LowCPUClass, ]

    KEY_VIDEO_INFO = u'video.infos'

    WANTED = (u"season", u"series", u"title", "episodeNumber", u"year")

    # Priority : arbitrary to 5
    priority = 5

    def _check(self):
        # year may not be present, we only check for title
        # also, if the filename has changed we activate the agent
        activate = not self.meta_storage.hasmeta_internal(self.KEY_VIDEO_INFO) or\
                (self.is_basename_modified)

        if not activate:
            self.logger.debug(u"File already has video info")
            return False

        return activate

    def _filter(self, datas):
        return {
            key: value
            for key, value in datas.items() if key in self.WANTED
        }

    def _process(self):
        self.logger.info(u"Extract video infos from filename")
        self._check_abort()
        fname = os.path.basename(self.path)

        video_info = guessit.guess_video_info(fname)
        if u'date' in video_info:
            del video_info['date']
        if u'language' in video_info:
            del video_info['language']
        is_tv = False

        # We should implement a real tv show detection
        if u'series' in video_info and u"season" in video_info and u"episodeNumber" in video_info:
            is_tv = True
        elif video_info.get('type', None) == "movie":
            is_tv = False
        self._check_abort()
        self.meta_storage.setmeta_internal(self.KEY_ISTV, is_tv)

        filtered = self._filter(video_info)

        self.meta_storage.setmeta_internal(self.KEY_VIDEO_INFO, filtered)

        self.logger.info(u"Extracted video infos from filename.")

        return True


class TmdbDisabledError(AgentAbortedError):

    """ Exception raised when tmdb is disabled """
    pass


class OsdbDisabledError(AgentAbortedError):

    """ Exception raised when osdb is disabled """
    pass

from osdbinfos import (
    OpenSutitles,
    OpenSutitlesError,
    OpenSutitlesTimeoutError, OpenSutitlesServiceUnavailable, OpenSutitlesNetworkError,
    DisabledUserAgentOpenSubtitlesError, InvalidUserAgentOpenSubtitlesError
)


class OsdbAggent(VideoAgent):

    """ Base class for OSDB agent """

    KEY_DISABLED = u'osdb.disabled'

    def __init__(self, *args, **kwargs):
        super(OsdbAggent, self).__init__(*args, **kwargs)
        self.osdb = OpenSutitles()

    def is_osdb_enabled(self):
        enabled = False
        if self.meta_storage.getmeta_internal(self.KEY_DISABLED):
            return False
        with self.config.lock:
            enabled = self.config.get('osdb.enabled', False)
            if enabled is not True:
                self.logger.debug(u"OSDB is not activated")
        return enabled

    def osdb_disable(self):
        self.meta_storage.setmeta_internal(self.KEY_DISABLED, True)

    def _precondition_(self):
        activate = super(OsdbAggent, self)._precondition_()
        if not activate:
            return False
        fifty_mb = 50 << 20  # 50 * 1024 * 1024
        filesize = fs.stat(self.path).st_size
        if filesize <= fifty_mb:
            self.logger.debug(u"File is too small, ignoring")
            self.context[self.KEY_NEED_REINDEX] = False
            activate = False
        if not activate:
            return False
        return self.is_osdb_enabled()

    def _check_abort(self):
        super(OsdbAggent, self)._check_abort()
        if self.is_osdb_enabled() is not True:
            raise OsdbDisabledError(u"OSDB disabled")


class VideoOsdbInfosAgent(OsdbAggent):

    """ Class that submit computed hash against OSDB to find imdbid."""
    _id = u'movie.imdb.id'

    KEY = 'osdb.infos'

    CLASSES = [LowCPUClass, NetworkClass]
    # If OSDB is in maintenancem we may get this exception
    EXCEPTIONS_RETRY = [OpenSutitlesTimeoutError, OpenSutitlesServiceUnavailable]

    priority = VideoInfosAgent.priority + 1

    def _check(self):
        # We activate this agent only if we have a conputed hash
        # and not an imdb already
        activate = (not self.meta_storage.hasmeta_internal(VideoOsdbInfosAgent.KEY) and not self.meta_storage.hasmeta(self.KEY_IMDBID))\
                or self.is_content_modified

        if not activate:
            self.logger.debug(u"OSDB infos already presents")
        else:
            self.logger.info(u"Activate fetching osdb infos")
        return activate

    def _process(self):

        self._check_abort()

        full_path = fs.get_system_path(self.path)

        # Submit to OSDB
        try:
            with retrying(self.osdb.get_files_infos, attempts=3, sleeptime=5, jitter=2, retry_exceptions=OpenSutitlesTimeoutError) as osdb_get_files_infos:
                results = osdb_get_files_infos([full_path])
        except (OpenSutitlesServiceUnavailable, OpenSutitlesTimeoutError, OpenSutitlesNetworkError):
            self.logger.info(u"OSDB unavailable")
            raise
        except (DisabledUserAgentOpenSubtitlesError, InvalidUserAgentOpenSubtitlesError):
            segment.track_error(event="OSDB panic : useragent disabled or invalid")
            with self.config.lock:
                self.config['osdb.enabled'] = False
            raise
        except OpenSutitlesError as e:
            # parse error as string
            message = str(getattr(e, "message", ''))
            try:
                if '401' in message or 'unauthorized' in message.lower():
                    with self.config.lock:
                        self.config['osdb.enabled'] = False
                    segment.track_error(event="OSDB error : unauthorized")
                if '403' in message or 'forbidden' in message.lower():
                    with self.config.lock:
                        self.config['osdb.enabled'] = False
                    segment.track_error(event="OSDB error : forbidden")
                if '407' in message or 'limit reached' in message.lower():
                    with self.config.lock:
                        self.config['osdb.enabled'] = False
                    segment.track_error(event="OSDB error : unauthorized")
            except Exception:
                pass
            raise
        except Exception:
            # segment.track_error(event="OSDB error")
            raise

        if len(results) > 0 and full_path in results:
            osdb_infos = results[full_path]
            self.meta_storage.setmeta_internal(VideoOsdbInfosAgent.KEY, osdb_infos)

            kind = osdb_infos.get('kind', None)
            if kind == u'movie':
                self.meta_storage.setmeta_internal(self.KEY_ISTV, False)
            elif kind == u'episode':
                self.meta_storage.setmeta_internal(self.KEY_ISTV, True)
            return True
        else:
            # not found, store an empty dict
            self.meta_storage.setmeta_internal(VideoOsdbInfosAgent.KEY, {})
        return False


class VideoOsdbImdbIdAgent(OsdbAggent):

    """ Class that submit computed hash against OSDB to find imdbid."""
    _id = u'movie.imdb.id'
    CLASSES = [VeryLowCPUClass, NetworkClass]
    # If OSDB is in maintenancem we may get this exception

    priority = VideoOsdbInfosAgent.priority + 1

    def _check(self):
        # We activate this agent only if we have a conputed hash
        # and not an imdb already
        activate = self.meta_storage.hasmeta_internal(VideoOsdbInfosAgent.KEY)
        if not activate:
            self.logger.debug(u"No osdb infos present")

        activate = activate and not self.meta_storage.hasmeta(self.KEY_IMDBID)
        if not activate:
            self.logger.debug(u"IMDB Id already present")

        if activate:
            self.logger.info(u"Activate imdb id extraction")
        return activate

    def _process(self):
        self._check_abort()
        osdb_infos = self.meta_storage.getmeta_internal(VideoOsdbInfosAgent.KEY)

        # Only one result : Success
        if osdb_infos is not None and len(osdb_infos) > 0:
            imdb_id = osdb_infos.get("imdb_id", None)
            if imdb_id is not None:
                self._check_abort()
                self.meta_storage.setmeta(
                    self.KEY_IMDBID, imdb_id.decode(u'utf-8')
                )
                self.context[self.KEY_NEED_REINDEX] = True
                self.logger.info(u"imdb id found : %s", imdb_id)
                return True
        return False
