# coding: utf-8

import datetime

from playhouse.shortcuts import case
import peewee
from peewee import fn

from lima import db
from lima.db import DbNoSpaceError
from lima.fs import FileSystem as fs
from lima.settings import CONFIG
from lima.fs.fcfs import (
    LimaFsError,
    LimaFsTimeoutError,
    LimaFsConnectionRefusedError,
    LimaFsNotReadyError,
    LimaFsDoesnotExistError,
    LimaFsNoSpaceError,
    LimaFsNotReadableError,
    LimaFsNotMountedError,
    LimaFsMemoryError,
    LimaFsDeviceNotconfigured,
)

from lima.meta.blackboard.agents import (
    AudioLocalMusicTagAgent,
    AudioAlbumLocalCoverAgent,
    AudioTagMergerAgent,
    AudioGracenoteTagAgent,
    AudioGracenoteFingerprintAgent,
    AudioGracenoteSearchAgent,
    AudioGracenoteSearchArtistAgent,
    AudioTrackArtistCoverDownloadAgent,
    AudioCoverDownloadAgent,
    AudioAlbumCoverLink,
    AudioNoAlbumLocalCoverAgent,
    AudioArtistCoverDownloadAgent,

    # Image
    ImageThumbnailsAgent,
    # Video
    VideoInfosAgent,
    VideoPreviewExtractorAgent,
    VideoNameExtractInfosAgent,
    VideoOsdbInfosAgent,
    # Movies
    MovieNfoExtract,
    MovieSubmitTmdb,
    VideoOsdbImdbIdAgent,
    MovieCoverDownloadAgent,
    MovieLocalCoverAgent,
    # TV
    TvInfoExtractor,
    TvSeasonCoverDownloadAgent,
    TvSerieCoverDownloadAgent,
    TvEpisodeCoverLink,
)

from lima.tools import logs
logger = logs.getLogger(__name__)

from lima.tools.threads import threading

from lima.meta.blackboard.agents.classes import (
    TooManyRetryError, TooManyFailureError,
    AgentClassDisabledError,
)

from lima.meta.blackboard.agents.base import (
    BlackBoardAgent,
    TaskRetryError
)

from lima.models import LimaNode

from lima.meta import storage
from lima.meta.storage import MetaKeyValueStorage
from lima.meta.blackboard.agents.storage import AgentStorage
from lima.meta.scheduler import indexer
from lima.meta.scheduler import killswitch
from lima.meta import locks

###############################################
COMMON_AGENTS = (
    # General puprose
)
AUDIO_AGENTS = (
    # Audio
    AudioLocalMusicTagAgent,
    AudioAlbumLocalCoverAgent,
    AudioTagMergerAgent,
    AudioGracenoteFingerprintAgent,
    AudioGracenoteTagAgent,
    AudioCoverDownloadAgent,
    AudioGracenoteSearchAgent,
    AudioGracenoteSearchArtistAgent,
    AudioTrackArtistCoverDownloadAgent,
    AudioAlbumCoverLink,
    AudioNoAlbumLocalCoverAgent,
    AudioArtistCoverDownloadAgent,
)
IMAGE_AGENTS = (
    # Image
    ImageThumbnailsAgent,
)
VIDEO_AGENTS = (
    # Video
    VideoInfosAgent,
    VideoPreviewExtractorAgent,
    VideoNameExtractInfosAgent,
    VideoOsdbInfosAgent,
)
MOVIE_AGENTS = (
    # Movies
    MovieNfoExtract,
    MovieSubmitTmdb,
    VideoOsdbImdbIdAgent,
    MovieCoverDownloadAgent,
    MovieLocalCoverAgent,
)
TV_AGENTS = (
    # TV
    TvInfoExtractor,
    TvSerieCoverDownloadAgent,
    TvSeasonCoverDownloadAgent,
    TvEpisodeCoverLink,
)

ALL_AGENTS = COMMON_AGENTS +\
    AUDIO_AGENTS +\
    IMAGE_AGENTS +\
    VIDEO_AGENTS +\
    MOVIE_AGENTS +\
    TV_AGENTS


# Agents presets for each mimetypes
AGENTS_PRESETS = {
        'audio': list(set(COMMON_AGENTS + AUDIO_AGENTS)),
        'video': list(set(COMMON_AGENTS + VIDEO_AGENTS + MOVIE_AGENTS + TV_AGENTS)),
        'image': list(set(COMMON_AGENTS + IMAGE_AGENTS)),
}


import os


MAX_WAIT = 60


class BlackboardConfig(object):

    def __init__(self, path, is_update, is_content_modified, is_path_modified, is_basename_modified, agents_cls):

        if type(path) is str:
            path = path.decode(u"utf-8")
        self.path = path
        self.is_update = is_update
        self.is_content_modified = is_content_modified
        self.is_path_modified = is_path_modified
        self.is_basename_modified = is_basename_modified
        self.agents_cls = agents_cls

    def __str__(self):
        return repr(self)

    def __repr__(self):
        fname = os.path.basename(self.path)
        fname = fname.encode(u"utf-8")
        return "%s - update=%s" % (
            fname, self.is_update
        )


class Blackboard(object):

    """Blackboard for a file.
    This will execute a serie of specialized agents on the file to extract its
    meta datas.
    """

    EXCEPTIONS_DISABLING_AGENT = (
        TooManyRetryError,
        TooManyFailureError,
        TaskRetryError,
        LimaFsError,
    )

    DISABLED_CLASSES = []

    def __init__(self, config, blbk_config, is_debug=False):
        self.lock = threading.Lock()
        logger_name = u"lima.meta.tasks.provider"
        self.logger = logs.getLogger(logger_name, fs.basename(blbk_config.path))
        self.logger.name = logger_name

        self.blbk_config = blbk_config
        self.config = config
        self.is_debug = is_debug

        self.agents = []
        self.task_queue = []

        self.terminated_queue = []

        # context : will be persisted
        self.context = {}
        self.context_storage = AgentStorage(self.blbk_config.path, is_debug=self.is_debug)

        self.pendings_agents = []
        self.disabled_agents = []

        self._kill_event = threading.Event()

    def __enter__(self):
        return self

    def __exit__(self, *exc):
        if exc[0] is not None:
            self.logger.warn(u"Error occured for %s", self.blbk_config.path)
            # an exception occured
            # we need to reset the node status if in an intermediate state
            try:
                limaNode = None
                with db.transaction():
                    limaNode = LimaNode.get(path=self.blbk_config.path)
                if not limaNode:
                    return
                if limaNode.status in (
                        LimaNode.STATUS_WAIT_PROCESSING,
                        LimaNode.STATUS_PROCESSING,
                        LimaNode.STATUS_ERROR,
                        LimaNode.STATUS_NO_ERROR,
                        ):
                    self.logger.warn(u"%s is in an intermediate state", self.blbk_config.path)
                    limaNode.status = LimaNode.STATUS_NEED_RETRY
                    limaNode.try_count = (limaNode.try_count or 0) + 1
                    limaNode.next_try = datetime.datetime.now() + datetime.timedelta(seconds=60 * limaNode.try_count)

                    with db.transaction():
                        limaNode.save()
                        self.logger.info(u"Marked %s as to be reprocessed", self.blbk_config.path)
            except Exception:
                self.logger.error("Could not reset status")

    def load_context(self):
        self.context = self.context_storage.get_context_internal()

    def save_context(self):
        self.context_storage.save_context_internal(self.context)

    def stop(self):
        with self.lock:
            for agent in self.pendings_agents:
                try:
                    agent.stop()
                except:
                    self.logger.exception(u"Error while stopping agent")
            self.save_context()
        try:
            AgentStorage.persist_metas(
                self.blbk_config.path, forget=True
            )
        except LimaFsDoesnotExistError:
            self.logger.info("%s : removed", self.blbk_config.path)

    def kill(self, persists=True):
        self._kill_event.set()
        with self.lock:
            for agent in self.pendings_agents[:]:
                try:
                    agent.kill()
                except Exception:
                    self.logger.exception(u"Error while kiling agent")
            for agent in self.agents[:]:
                if agent in self.pendings_agents:
                    self.pendings_agents.remove(agent)
                if agent in self.disabled_agents:
                    self.disabled_agents.remove(agent)
                if agent in self.agents:
                    self.agents.remove(agent)

            if persists:
                self.save_context()
                try:
                    AgentStorage.persist_metas(self.blbk_config.path, forget=True)
                except LimaFsDoesnotExistError:
                    self.logger.info("%s : removed", self.blbk_config.path)
            else:
                # we unload the in memory metas
                self.logger.warn("Won't persist metas for %s", self.blbk_config.path)
                AgentStorage.forget(self.blbk_config.path, persist=False)

    def _init_agents_(self):
        """Initialize agent instances from given classes.
        """
        with self.lock:
            self.load_context()
            for agent_cls in self.blbk_config.agents_cls:
                agent = agent_cls(
                    path=self.blbk_config.path,
                    parent=self,
                    is_update=self.blbk_config.is_update,
                    is_content_modified=self.blbk_config.is_content_modified,
                    is_path_modified=self.blbk_config.is_path_modified,
                    is_basename_modified=self.blbk_config.is_basename_modified,
                    config=self.config,
                    context=self.context,
                    is_debug=self.is_debug,
                )
                self.agents.append(agent)

    def availables_tasks(self):
        if not fs.exists(self.blbk_config.path):
            self.logger.error(u"%s does not exists", self.blbk_config.path)
            return []
        if len(self.agents) == 0:
            self._init_agents_()
        tasks = []
        self.logger.debug(u"Init task queue")
        with self.lock:
            agents = list(
                set(self.agents) -
                set(self.pendings_agents) -
                set(self.disabled_agents)
            )
            for agent in agents[:]:
                for _cls in agent.CLASSES:
                    if _cls in Blackboard.DISABLED_CLASSES:
                        agents.remove(agent)

        for agent in agents:
            with self.lock:
                if agent not in self.pendings_agents:
                    if not self.should_continue:
                        break
                    if agent.check() is True:
                        if not self.should_continue:
                            break
                        self.pendings_agents.append(agent)
                        tasks.append((agent.priority, agent))
        if not self.should_continue:
            return []
        tasks = sorted(tasks, key=lambda prio_t: prio_t[0])
        # if tasks availables, update the status to 'processing'
        if len(tasks) > 0:
            with db.transaction():
                LimaNode.update(
                    status=case(LimaNode.status, (
                        (LimaNode.STATUS_WAIT_PROCESSING,
                         LimaNode.STATUS_PROCESSING),
                        (LimaNode.STATUS_WAIT_PROCESSING_UPDATE,
                         LimaNode.STATUS_PROCESSING_UPDATE),
                    ), LimaNode.status)
                ).where(
                    (LimaNode.path == self.blbk_config.path)
                    & (LimaNode.status << [LimaNode.STATUS_WAIT_PROCESSING, LimaNode.STATUS_WAIT_PROCESSING_UPDATE])
                ).execute()
        with self.lock:
            if len(tasks) == 0 and len(self.pendings_agents) == 0:
                self.logger.info("-- " + u"%s : nothing to do.", self.blbk_config.path)

                # All work is done, we ask for metas to be persisted in the
                # core
                persisted = False

                content_type = None
                try:
                    storage = MetaKeyValueStorage.get_from_pool(self.blbk_config.path, 'meta')
                    content_type = storage.getmeta('content_type', None)
                except Exception:
                    self.logger.exception("Could not obtain content_type")

                try:
                    AgentStorage.persist_metas(
                        self.blbk_config.path, forget=True
                    )
                    persisted = True
                except Exception:
                    self.logger.exception("Could not persist metas")

                with db.transaction():
                    try:
                        limaNode = LimaNode.get(path=self.blbk_config.path)
                    except peewee.DoesNotExist:
                        self.logger.error("file had been deleted before the end of the process")
                        limaNode = None

                if not limaNode:
                    return

                # get identified content type if any
                if limaNode.content_type is None:
                    if content_type:
                        logger.info("set content type %s to %s", content_type, fs.basename(self.blbk_config.path))
                        limaNode.content_type = content_type

                need_reindex = self.context.get(BlackBoardAgent.KEY_NEED_REINDEX, False)
                # agent may not have flagged the item to be reindexed
                # but if the path changed, we need to force it.
                need_reindex = need_reindex or (limaNode.is_path_modified is True and limaNode.content_type is not None)

                if need_reindex:
                    # flag as to be reindexed
                    limaNode.status_index = LimaNode.STATUS_INDEX_NEED_REINDEX
                else:
                    # flag index status as NONE if not already flagged as indexed, to be ignored or to be indexed
                    if limaNode.status_index not in (LimaNode.STATUS_INDEX_INDEXED, LimaNode.STATUS_INDEX_IGNORE, LimaNode.STATUS_INDEX_NEED_REINDEX):
                        limaNode.status_index = LimaNode.STATUS_INDEX_NONE

                ####
                if limaNode.status in (
                        LimaNode.STATUS_NO_ERROR,
                        LimaNode.STATUS_ERROR,
                        LimaNode.STATUS_WAIT_PROCESSING,
                        LimaNode.STATUS_WAIT_PROCESSING_UPDATE,
                ):
                    if persisted and limaNode.status in (
                            LimaNode.STATUS_NO_ERROR,
                            LimaNode.STATUS_WAIT_PROCESSING,
                            LimaNode.STATUS_WAIT_PROCESSING_UPDATE):
                        limaNode.status = LimaNode.STATUS_PROCESSED
                        limaNode.try_count = 0
                        limaNode.next_try = None

                        limaNode.is_content_modified = False
                        limaNode.is_path_modified = False
                        limaNode.is_basename_modified = False

                    elif not persisted or limaNode.status == LimaNode.STATUS_ERROR:
                        limaNode.status = LimaNode.STATUS_NEED_RETRY
                        limaNode.try_count = limaNode.try_count + 1
                        limaNode.next_try = limaNode.last_try + \
                            datetime.timedelta(
                                seconds=60 * limaNode.try_count
                            )

                    with db.transaction():
                        limaNode.save()
        return tasks

    def __str__(self):
        return "[%s]" % str(self.blbk_config)

    @property
    def has_pendings(self):
        return len(self.pendings_agents) > 0

    @property
    def has_terminated(self):
        return len(self.terminated_queue) > 0

    def is_success(self):
        return len(self.disabled_agents) == 0

    def on_agent_stop(self, agent, exc=None):
        if not issubclass(type(exc), TaskRetryError):
            with self.lock:
                if agent in self.pendings_agents:
                    self.pendings_agents.remove(agent)
                    # kep a trace of agent / exception if any
                    self.terminated_queue.append((agent, exc))

    def _process_terminated_agent(self, agent, exc):
        if exc is None:
            # so far so good
            with db.transaction():
                LimaNode.update(
                    status=LimaNode.STATUS_NO_ERROR,
                ).where(
                    (LimaNode.path == agent.path) &
                    (LimaNode.status != LimaNode.STATUS_ERROR)
                ).execute()
        else:
            # agent ended with an error, update its status accordinaly
            if issubclass(type(exc), self.EXCEPTIONS_DISABLING_AGENT):
                self.logger.error(
                    u"Agent %s has been disabled and won't be spawned any more (%s)",
                    str(agent).decode('utf-8'),
                    str(type(exc))
                )

                with self.lock:
                    if agent not in self.disabled_agents:
                        self.disabled_agents.append(agent)

            if issubclass(type(exc), AgentClassDisabledError):
                with self.lock:
                    if agent not in self.disabled_agents:
                        self.disabled_agents.append(agent)
                    Blackboard.DISABLED_CLASSES += exc.classes

            if not issubclass(type(exc), TaskRetryError):
                with db.transaction():
                    LimaNode.update(
                        status=LimaNode.STATUS_ERROR,
                        last_try=datetime.datetime.now(),
                    ).where(
                        (LimaNode.path == agent.path)
                    ).execute()

    def process_terminated(self):
        """ Process terminated agent and update file status accordinally"""
        with self.lock:
            terminated = self.terminated_queue[:]

        for agent, exc in terminated:
            # pop from the terminated queue
            with self.lock:
                self.terminated_queue.remove((agent, exc))
            try:
                self._process_terminated_agent(agent, exc)
            except Exception:
                self.logger.error(u"Could not process terminated agent")

    @property
    def should_continue(self):
        return not self._kill_event.isSet()

import itertools


class BlackboardController(threading.Thread):

    WAIT = 1
    WAIT_STEP = 5
    WAIT_TIMEOUT = 20
    WAIT_REFUSED = 30
    WAIT_NOT_READY = 60 * 3

    WAIT_PIN = 60 * 5  # we will pin meta folder every 5 minutes.

    def __init__(self, config, task_runner, pool_size, is_debug=False):
        threading.Thread.__init__(self)
        self.lock = threading.Lock()
        self.daemon = True
        self.logger = logs.getLogger(u"lima.meta.tasks.controller")
        self.logger.name = u"lima.meta.tasks.controller"

        self.config = config
        self.is_debug = is_debug

        self.task_runner = task_runner

        # Tasks index by path
        self.backlog_files = {}

        self.pool_size = pool_size

        self.wait_delay = self.WAIT

        self.last_pin = None

    def __repr__(self):
        return u"lima.meta.tasks.controller"

    def populate_blackboards(self):
        if not self.should_continue:
            return
        now = datetime.datetime.now()
        item_sent = 0
        wait_threshold = 10
        wait_betwen_files = 0.1

        nodes = []
        two_sec_ago = datetime.datetime.now() - datetime.timedelta(seconds=2)
        ten_sec_ago = datetime.datetime.now() - datetime.timedelta(seconds=10)
        twenty_sec_ago = datetime.datetime.now() - datetime.timedelta(seconds=20)
        thirty_sec_ago = datetime.datetime.now() - datetime.timedelta(seconds=30)
        fourty_sec_ago = datetime.datetime.now() - datetime.timedelta(seconds=40)
        one_min_ago = datetime.datetime.now() - datetime.timedelta(seconds=60)

        hundred_kb = 100 << 10  # 100 * 1024
        five_mb = 5 << 20  # 5 * 1024 * 1024
        fifty_mb = 50 << 20  # 50 * 1024 * 1024
        hundred_mb = 100 << 20  # 50 * 1024 * 1024
        one_gb = 1 << 30  # 50 * 1024 * 1024

        with db.transaction():
            """
            Pseudo sql :
            select *
            from limanode
            where (
                    status in (NEED_UPDATE, CREATED)
                or (status = NEED_RETRY and next_try <= now)
                )
            and ( last_update <= minutes_ago )
            order by last_updated desc
            limit N
            """
            nodes = LimaNode.select().where(
                    (
                        (
                        LimaNode.status << [
                            LimaNode.STATUS_NEED_UPDATE,
                            LimaNode.STATUS_CREATED,
                        ]) | (
                            (LimaNode.status == LimaNode.STATUS_NEED_RETRY)
                            & (LimaNode.next_try <= now)
                        )
                ) & (
                    ((LimaNode.size <= hundred_kb) & (LimaNode.last_updated <= two_sec_ago))
                    | ((LimaNode.size > hundred_kb) & (LimaNode.size <= five_mb) & (LimaNode.last_updated <= ten_sec_ago))
                    | ((LimaNode.size > five_mb) & (LimaNode.size <= fifty_mb) & (LimaNode.last_updated <= twenty_sec_ago))
                    | ((LimaNode.size > fifty_mb) & (LimaNode.size <= hundred_mb) & (LimaNode.last_updated <= thirty_sec_ago))
                    | ((LimaNode.size > hundred_mb) & (LimaNode.size <= one_gb) & (LimaNode.last_updated <= fourty_sec_ago))
                    | ((LimaNode.size > one_gb) & (LimaNode.last_updated <= one_min_ago))
                ) & (
                        # filter on mimetype : only use
                        # 'video'/'image'/'audio'
                        fn.substr(LimaNode.mimetype, 0, 6) << CONFIG['supported_mimetypes']
                )
            ).order_by(
                    # small files first
                    LimaNode.size.asc(),
                    # is_local_source is boolean : we first fetch item with value True (1), then False (0)
                    LimaNode.is_local_source.desc(),
                    LimaNode.last_updated.desc()
            ).limit(self.pool_size - len(self.backlog_files))
            if not self.should_continue:
                return
            # force sql query to be fully loaded instead of using a DB blocking
            # iterator
            nodes = nodes[:]

        paths = [node.path for node in nodes]
        stats = fs.bulkstat(paths)
        nodes = [(path, st_result[0], st_result[1]) for path, st_result in itertools.izip(nodes, stats)]

        for limaNode, exc, stat in nodes:
            if not self.should_continue:
                break
            if isinstance(exc, LimaFsDoesnotExistError):
                with db.transaction():
                    limaNode.status = LimaNode.STATUS_DELETED
                    limaNode.save()
                continue

            if not self.should_continue:
                break
            if not fs.isfile(st=stat):
                self.logger.warn(u"%s : not a file", limaNode.path)
                with db.transaction():
                    limaNode.status = LimaNode.STATUS_NOT_A_FILE
                    limaNode.save()
                continue

            fname = limaNode.path
            if item_sent >= self.pool_size:
                self.wait(wait_threshold)
                item_sent = 0

            if not self.should_continue:
                break

            is_update = limaNode.status == LimaNode.STATUS_NEED_UPDATE

            status = LimaNode.STATUS_WAIT_PROCESSING if not is_update else LimaNode.STATUS_WAIT_PROCESSING_UPDATE

            if not self.should_continue:
                break
            with db.transaction():
                limaNode.status = status
                limaNode.save()
            #
            kind = limaNode.mimetype.split('/')[0] if limaNode.mimetype else None
            agents_cls = AGENTS_PRESETS.get(kind, [])
            if len(agents_cls) == 0:
                self.logger.warn(u"No agents for mime %s", kind)

            blbk_config = BlackboardConfig(
                path=fname,

                is_update=is_update,
                is_content_modified=limaNode.is_content_modified,
                is_path_modified=limaNode.is_path_modified,
                is_basename_modified=limaNode.is_basename_modified,

                agents_cls=agents_cls
            )
            with self.lock:
                existing_bboard = self.backlog_files.get(limaNode.path, None)
                if existing_bboard is not None:
                    self.logger.debug(
                        u"path %s already in process", blbk_config.path)
                    continue

            if killswitch.switch.check():
                self.logger.error(u"Kill switch detected. not continuing.")
                self.stop()
                break
            try:
                bboard = Blackboard(self.config, blbk_config, is_debug=self.is_debug)
                #
                if not self.should_continue:
                    break
                with self.lock:
                    self.backlog_files[blbk_config.path] = bboard
                #
                self.logger.info(u"Working on %s (mode update : %s)", fname, is_update)
                item_sent += 1
            except LimaFsNotReadableError:
                # TODO : we may want to check this before blackboard initialization and force the file reprocessing
                # by erasing old unreadable metadatas files. Using the simpliest solution for now.
                self.logger.warning(u"Meta exists but is not readable. Won't work on %s", fname)
                limaNode.status = LimaNode.STATUS_NEED_RETRY
                # increase retry count by 5 minutes
                limaNode.try_count = (limaNode.try_count or 0) + 5
                limaNode.last_try = datetime.datetime.now()
                limaNode.next_try = datetime.datetime.now() + datetime.timedelta(seconds=60 * limaNode.try_count)

                with db.transaction():
                    limaNode.save()
                    self.logger.info(u"Marked %s as to be reprocessed", limaNode.path)
                #
            self.wait(wait_betwen_files)

    def register_tasks_or_remove(self, blackboard):
        tasks = blackboard.availables_tasks()

        if len(tasks) == 0 and not blackboard.has_pendings:
            self.logger.debug(
                u"No task for provider %s (%s task in queue - next in %s)",
                str(blackboard).decode('utf-8'),
                len(self.task_runner),
                self.task_runner.task_queue.next_task_in
            )
            persists = True
            try:
                blackboard.process_terminated()
            except (LimaFsNotMountedError, LimaFsConnectionRefusedError):
                self.logger.warn("Can't talk with the core. Metas won't be persisted")
                persists = False
            except Exception:
                self.logger.error(u"Could not process terminated agents")

            with self.lock:
                self.kill_blackboard_and_remove(blackboard, persists=persists)
            self.logger.debug(u"%s has been removed", str(blackboard).decode('utf-8'))
        else:
            for prio, task in tasks:
                if not self.should_continue:
                    return
                self.logger.debug(
                    u"Add task %s with priority %s (%s tasks in queue)",
                    str(task).decode('utf-8'),
                    prio, len(self.task_runner)
                )
                # Push task to the task runner
                self.task_runner.push(prio, task)

    def process_blackboards(self):
        for path in self.backlog_files.keys():
            if not self.should_continue:
                return
            # Blackboard may have been removed
            blackboard = self.backlog_files.get(path, None)
            if blackboard is None:
                continue
            try:
                if blackboard.has_terminated:
                    blackboard.process_terminated()
                if not blackboard.has_pendings:
                    if not self.should_continue:
                        return
                    with blackboard:
                        self.register_tasks_or_remove(blackboard)
            except(LimaFsDoesnotExistError,):
                # The file have been moved/deleted : we kill the blackboard
                # and remove it.
                self.logger.error(u"File has been moved or deleted during processing -  %s", path)
                self.kill_blackboard_and_remove(blackboard)
            except (LimaFsTimeoutError):
                self.logger.error(u"FS timeout -  %s", path)
                self.kill_blackboard_and_remove(blackboard)
            except (LimaFsConnectionRefusedError):
                self.logger.error(u"Connection refused")
                self.kill_blackboard_and_remove(blackboard, persists=False)
                raise
            except (LimaFsNotMountedError):
                self.logger.error(u"Lima not mounted")
                self.kill_blackboard_and_remove(blackboard, persists=False)
                raise
            except(LimaFsError, ):
                # Unknown filesystem error.
                # We remove the blackboard and the file
                # We add the file to the backlog
                self.logger.exception(u"FS error %s", str(blackboard).decode('utf-8'))
                self.logger.error(u"Info : removing %s", str(blackboard).decode('utf-8'))
                self.kill_blackboard_and_remove(blackboard)
            except:
                # Unknown exception.
                # We remove the blackboard and the file
                self.logger.exception(u"Unknown error, removing blackboard")
                self.kill_blackboard_and_remove(blackboard)
                raise

    def execute(self):
        nb_in_process = 0
        last_in_process = 0

        has_computed_files = False

        while not self._kill_event.isSet():
            # Wait for resume event
            if not self._resume_event.wait(0.1):
                continue
            if not self._resume_event.isSet():
                continue

            # query db for file to manage
            try:
                with locks.BLACKBOARD_LOCK:
                    self.check_continue()
                    if len(self.backlog_files) < self.pool_size:
                        self.populate_blackboards()

                    self.check_continue()
                    if len(self.backlog_files) > 0:
                        has_computed_files = True

                    self.check_continue()
                    if killswitch.switch.check():
                        self.logger.error(u"Kill switch detected. not continuing.")
                        self.stop()
                        break
                    self.check_continue()
                    self.process_blackboards()

                    if has_computed_files and len(self.backlog_files) == 0:
                        has_computed_files = False
                    self.check_continue()
                self._pin_if_needed()

            except threading.ThreadKilledError:
                self.logger.info("task provider killed")
                break
            except threading.ThreadStoppedError:
                self.logger.info("task provider stopped")
            except(LimaFsTimeoutError):
                # RPC timeout or filesystem not ready, we just wait
                self.logger.error(u"Timeout")
                self.wait(self.WAIT_TIMEOUT)
            except(LimaFsConnectionRefusedError):
                # RPC timeout or filesystem not ready, we just wait
                self.logger.error(u"Connection refused")
                self.wait(self.WAIT_REFUSED)
            except(LimaFsNotReadyError):
                # RPC timeout or filesystem not ready, we just wait
                self.logger.error(u"Core not ready")
                self.wait(self.WAIT_NOT_READY)
            except(LimaFsNoSpaceError,):
                # No space on device : we kill ourself
                self.logger.error(u"No space left on device. Killing myself")
                try:
                    self.kill(fatal=True)
                except:
                    self.logger.exception(u"Error when killing task feeder on 'no space' error")
            except (LimaFsMemoryError, ):
                # No memory: we kill ourself
                self.logger.error(u"No memory. Killing myself")
                try:
                    self.kill(fatal=True)
                except:
                    self.logger.exception(u"Error when killing task feeder on 'no memory' error")

            except (DbNoSpaceError,):
                # No space on device : we kill ourself
                # original exception is raised
                self.logger.error(u"DB full or no left space on device.")
                try:
                    self.kill(fatal=True)
                except:
                    self.logger.exception(u"Error when killing task feeder on 'db no space' error")
            except(LimaFsNotMountedError,):
                # No space on device : we kill ourself
                # original exception is raised
                self.logger.error(u"Lima is not mounted. Killing myself.")
                try:
                    # kill with backtrace will inform the scheduler
                    self.kill(fatal=True)
                except:
                    self.logger.exception(u"Error when killing blackboard task feeded on 'lima not mounted' error")
            except(LimaFsDeviceNotconfigured,):
                # Device not configured. we kill ourself
                self.logger.error(u"Device not configured. Killing myself.")
                try:
                    # kill with backtrace will inform the scheduler
                    self.kill(fatal=True)
                except:
                    self.logger.exception(u"Error when killing blackboard task feeded on 'lima device not configured' error")
            except(LimaFsError, ):
                # Unknown filesystem error.
                # We remove the blackboard and the file
                # We add the file to the backlog
                self.logger.exception(u"Unknown FS error")
            except:
                self.logger.error(u"Unknown error. Killing myself")
                try:
                    self.kill()
                except:
                    self.logger.error(u"Error when killing task feeder")
                raise
            finally:
                if len(self.backlog_files) == 0:
                    AgentStorage.unload_all()
                    # clear all storages
                indexer.ENABLED.set()

            nb_in_process = len(self.backlog_files)
            # TODO : log task number
            nb_task_process = len(self.task_runner)

            if nb_in_process != last_in_process:
                last_in_process = nb_in_process
                self.logger.info(u"%s files in process", nb_in_process)
                self.logger.info(u"%s tasks waiting to be processed. next in %s sec",
                                 nb_task_process, self.task_runner.task_queue.next_task_in
                                 )

            self.wait(self.wait_delay)

        self.logger.info("Bye.")

    def _pin_if_needed(self):
        """ Pin folders if needed."""

        last_pin_delta = datetime.timedelta(seconds=self.WAIT_PIN + 1)
        # se when we last tried to pin folders
        if self.last_pin is not None:
            last_pin_delta = datetime.datetime.now() - self.last_pin

        if last_pin_delta.seconds > self.WAIT_PIN:
            if self.should_continue:
                storage.pin()
                self.last_pin = datetime.datetime.now()
            if self.should_continue:
                killswitch.pin()
                self.last_pin = datetime.datetime.now()
            if self.should_continue:
                indexer.pin()
                self.last_pin = datetime.datetime.now()

    def kill_blackboard_and_remove(self, blackboard, persists=True):
        self.kill_blackboard(blackboard, persists)
        self.remove_blackboard(blackboard)

    def kill_blackboard(self, blackboard, persists=True):
        try:
            blackboard.kill(persists=persists)
        except LimaFsDoesnotExistError:
            self.logger.exception(u"File did not exists anymore")
        except LimaFsError:
            raise
        except Exception:
            self.logger.exception(u"Error killing blackboard")

    def remove_blackboard(self, blackboard):
        try:
            del self.backlog_files[blackboard.blbk_config.path]
        except(KeyError,):
            self.logger.error(u"Already removed")

    def kill(self, *args, **kwargs):
        try:
            is_core_alive = True
            self.task_runner.kill()
            for path in self.backlog_files.keys():
                blackboard = self.backlog_files.get(path, None)
                if blackboard is None:
                    # blackboard already removed
                    continue
                try:
                    # core is not alive : do not persist
                    self.kill_blackboard(blackboard, persists=is_core_alive)
                except (LimaFsConnectionRefusedError, LimaFsNotMountedError):
                    # We can't talk to fcfs, we do not want to try to persists others files
                    is_core_alive = False
                finally:
                    self.remove_blackboard(blackboard)
        except Exception:
            self.logger.error("Error occured while killing subtasks")
        finally:
            super(BlackboardController, self).kill(*args, **kwargs)

    def stop(self):
        self.task_runner.stop()
        super(BlackboardController, self).stop()

    def resume(self):
        self.task_runner.resume()
        super(BlackboardController, self).resume()

    def is_processing(self):
        return len(self.backlog_files) > 0

    def all_files_count(self):
        with db.transaction():
            return LimaNode.select().where(
                        # filter on mimetype : only use
                        # 'video'/'image'/'audio'
                        fn.substr(LimaNode.mimetype, 0, 6) << CONFIG['supported_mimetypes']
                ).count()

    def processed_files_count(self):
        with db.transaction():
            return LimaNode.processed_nodes().where(
                        # filter on mimetype : only use
                        # 'video'/'image'/'audio'
                        fn.substr(LimaNode.mimetype, 0, 6) << CONFIG['supported_mimetypes']
                    ).count()
