# coding: utf-8
import os
import datetime

import logging
logger = logging.getLogger(__name__)

import peewee

from lima.fs import FileSystem as fs
from lima.fs.limafs import (LimaFsDoesnotExistError,
                            LimaFsConnectionRefusedError, LimaFsNotReadyError,
                            LimaFsError)
from lima.models import LimaNode

from lima.meta.blackboard.agents.storage import AgentStorage
from lima.meta.storage import (AlbumStorage, ArtistStorage, SeasonStorage,
                               SerieStorage)
from lima.meta.scheduler import killswitch
from lima.meta.storage import MetaKeyValueStorage

from lima.mimes import mimetypes


def purge_all(database, root="/%", fake=False):
    if killswitch.switch.check():
        logger.warn(u"Kill switch detected. Aborting")
        return {}
    result = purge_deleted_files(database, root, fake)
    orphans = remove_orphan_metas(database, fake)
    result['orphans'] = orphans

    return result


def remove_status_deleted(database):
    """ Remove nodes with status deleted"""
    with database.transaction():
        LimaNode.delete().where(LimaNode.status == LimaNode.STATUS_DELETED).execute()


def purge_deleted_files(database, root="/", recursive=False, fake=False):
    """" Removes files from local index that are not on filesystem anymore"""

    logger.debug(u"Removing deleted files from db for path %s", root)
    try:
        return _purge_deleted_files(database, root, recursive, fake)
    except LimaFsNotReadyError:
        logger.error(u"Core not ready")
    except LimaFsConnectionRefusedError:
        logger.error(u"Connection refused")
    except LimaFsError:
        logger.error("Filesystem error")


def _purge_deleted_files(database, root="/", recursive=False, fake=False):
    """" Removes files from local index that are not on filesystem anymore"""

    result = {}
    purged_files = 0
    # how many files to remove in a row
    chunk_size = 100

    to_delete = []

    with database.transaction():
        files = LimaNode.list(root, recursive)[:]

    sub_purge = {}
    for limaNode in files:
        if killswitch.switch.check():
            return
        if not fs.exists(limaNode.path):
            _remove_metas(limaNode, fake)
            to_delete.append(limaNode.path)
            logger.info(u"Removing from local index : %s ", limaNode.path)
            # TODO : limaNode.path may be a removd directory.
            # wee nedd to delete his children
            sub_purge = _purge_deleted_files(database, limaNode.path,
                                             recursive=True)
            content_type = limaNode.content_type
            if content_type is None:
                # if content type is null, try to guess from mimetype
                mimetype, _ = mimetypes.guess_type(limaNode.path)
                content_type = mimetype.split('/')[0] if mimetype else None
            try:
                result[content_type] += 1
            except KeyError:
                result[content_type] = 1

            if len(to_delete) >= chunk_size:
                _remove_files(database, to_delete, fake)
                purged_files += len(to_delete)
                to_delete = []
        else:
            try:
                _purge_obsolete_files(limaNode)
            except Exception:
                logger.exception(u"Could not purge obsolete files")

    if len(to_delete) > 0:
        _remove_files(database, to_delete, fake)
        to_delete = []

    purged_files = sum([result[_content_type] for _content_type in result
                        ]) + sum([sub_purge[_content_type]
                                  for _content_type in sub_purge])
    if purged_files > 0:
        logger.info(
            u"Removed %s files from local index for path %s",
            # if content type is null, try to guess from mimetype
            purged_files,
            root,
        )

    # merge subpurge + purge
    for k, v in sub_purge.items():
        try:
            result[k] += v
        except KeyError:
            result[k] = v

    for content_type in result:
        if content_type != 'total':
            logger.info(u"Purged %s of content type %s", result[content_type],
                        content_type or 'unknown')

    result['total'] = purged_files
    return result


def _purge_obsolete_files(limaNode):
    """ Remove obsolete files"""
    if limaNode.mimetype and limaNode.mimetype.startswith('image/'):
        try:
            with MetaKeyValueStorage.get_from_pool(limaNode.path, 'meta') as storage:
                # crc32 : generated by mobile : do not remove
                if storage.exists('meta') and not storage.hasmeta('crc32'):
                    meta_file = storage.get_storage_fname('meta')
                    st = fs.stat(meta_file)
                    if st.st_file_local_origin:
                        fs.unlink(meta_file)
                        logger.info(u"Removed obsolete file %s for %s", meta_file, limaNode.path)
        except Exception:
            logger.error(u"Could not remove obsolete file for %s", limaNode.path)


def _remove_files(database, to_delete, fake=False):
    """ Removes the given files from the local index"""
    if fake is True:
        logger.debug("Fake enable, not deleting")
        return
    if len(to_delete) == 0:
        return
    if killswitch.switch.check():
        return
    with database.transaction():
        logger.debug(u"Removing %s files", len(to_delete))
        LimaNode.delete().where(LimaNode.path << to_delete).execute()


def _remove_metas(limaNode, fake=False):
    """ Remove metas for the given node """
    logger.info(u"Removing metas for %s", limaNode.path)
    if fake is True:
        logger.debug(u"Fake enable, not deleting")
        return

    try:
        if limaNode.default_forkid is not None:
            if killswitch.switch.check():
                logger.warn(u"Kill switch detected. Aborting")
                return
            AgentStorage.remove_key(limaNode.default_forkid, delete_empty_parent=True)
            logger.debug(u"Removing metas for %s : DONE", limaNode.path)
    except LimaFsDoesnotExistError:
        pass
    except LimaFsNotReadyError:
        logger.error(u"Core not ready")
    except LimaFsConnectionRefusedError:
        logger.error(u"Connection refused")
    except LimaFsError:
        logger.error("Filesystem error")


def remove_orphan_metas(database, fake=False):
    """ Removes metasdatas where the file does not exists anymore in the db"""
    count = 0
    try:
        for f, stat in AgentStorage.iter_files_stat():
            meta_directory = os.path.dirname(f)
            forkid = os.path.basename(meta_directory)
            if len(forkid) != 40:
                try:
                    dirname = os.path.dirname(meta_directory)
                    if fs.is_empty_dir(dirname):
                        fs.rmdirs(dirname)
                        logger.info(u"Removed empty dir %s", dirname)
                except Exception:
                    logger.error(u"Could not remove empty dir %s", dirname)
                continue
            try:
                # if this fail, jimp to the except
                with database.transaction():
                    LimaNode.get(default_forkid=forkid.upper())
                if stat.isreg() and fs.is_conflict(f):
                    if killswitch.switch.check():
                        logger.warn(u"Kill switch detected. Aborting")
                        return
                    fs.unlink(f)
                    logger.info(u"Removed conflict %s", f)
            except peewee.DoesNotExist:
                try:
                    if fake is True:
                        logger.debug("Fake enable, not deleting")
                    else:
                        # Check for ctime / mtime : must be > 5 minutes
                        now = datetime.datetime.now()
                        ctime = stat.st_ctime / 1000000  # microseconds
                        ctime = datetime.datetime.fromtimestamp(ctime)

                        if (now - ctime).seconds < 30 * 60:
                            logger.debug(u"Metas created less than 30 min ago. Won't be deleted now")
                        else:
                            if killswitch.switch.check():
                                logger.warn(u"Kill switch detected. Aborting")
                                return
                            try:
                                res = fs.fork_path([forkid])
                                if not len(res) :
                                    fs.rmdirs(meta_directory)
                                    logger.info(u"Orphan : %s - %s. Has been deleted.", forkid, f)
                                else:
                                    filepath = res[0][2]
                                    logger.warning(u"%s is seen as and orphan metadata, but %s still exists. Not deleting.", f, filepath)
                            except LimaFsDoesnotExistError as e:
                                fs.rmdirs(meta_directory)
                                logger.info(u"Orphan : %s - %s. Has been deleted.", forkid, f)

                    count += 1
                except LimaFsDoesnotExistError:
                    pass
        AgentStorage.remove_empty()
    except LimaFsDoesnotExistError:
        logger.error(u"meta folder does not exists")
    except LimaFsNotReadyError:
        logger.error(u"Core not ready")
    except LimaFsConnectionRefusedError:
        logger.error(u"Connection refused")
    except LimaFsError:
        logger.error("Filesystem error")
    return count


def remove_storages_conflicts():
    storages_cls = [AlbumStorage, ArtistStorage, SerieStorage, SeasonStorage]
    for storage_cls in storages_cls:
        _remove_storages_conflicts(storage_cls)


def _remove_storages_conflicts(storage_cls):
    try:
        for f in storage_cls.iter_files(files_only=True):
            if fs.is_conflict(f):
                try:
                    fs.unlink(f)
                    logger.info(u"Removed conflict %s", f)
                except Exception:
                    logger.warn(u"Could not remove conflict %s", f)
    except LimaFsDoesnotExistError:
        pass


def clean_up_album(artist_name, album_name):
    storage = None
    try:
        storage = AlbumStorage.get_from_pool(artist_name, album_name)
        if killswitch.switch.check():
            logger.warn(u"Kill switch detected. Aborting")
            return
        storage.remove()
        logger.info("Removed metas for album %s - %s", artist_name, album_name)
    except Exception:
        logger.debug("Could not remove metas for album %s - %s", artist_name,
                     album_name)
    finally:
        if storage is not None:
            storage.unload()


def clean_up_artist(artist_name):
    storage = None
    try:
        storage = ArtistStorage.get_from_pool(artist_name)
        if killswitch.switch.check():
            logger.warn(u"Kill switch detected. Aborting")
            return
        storage.remove()
        logger.info("Removed metas for artist %s", artist_name)
    except Exception:
        logger.debug("Could not remove metas for album %s", artist_name)
    finally:
        if storage is not None:
            storage.unload()


def clean_up_serie(serie_name):
    storage = None
    try:
        storage = SerieStorage.get_from_pool(serie_name)
        if killswitch.switch.check():
            logger.warn(u"Kill switch detected. Aborting")
            return
        storage.remove()
        logger.info("Removed metas for serie %s", serie_name)
    except Exception as e:
        logger.debug("Could not remove metas for serie  %s : %s", serie_name,
                     e.message)
    finally:
        if storage is not None:
            storage.unload()


def clean_up_season(serie_name, season_number):
    storage = None
    try:
        storage = SeasonStorage.get_from_pool(serie_name, str(season_number))
        if killswitch.switch.check():
            logger.warn(u"Kill switch detected. Aborting")
            return
        storage.remove()
        logger.info("Removed metas for season %s - %s", serie_name,
                     season_number)
    except Exception as e:
        logger.debug("Could not remove metas for season %s - %s : %s",
                     serie_name, season_number, e.message)
    finally:
        if storage is not None:
            storage.unload()
