# coding: utf-8

import os
import tempfile
import gzip
import sqlite3
from StringIO import StringIO

import abc

from lima.fs import FileSystem as fs
from lima.fs.limafs import (LimaFsError, LimaFsDoesnotExistError, )
from lima.models import LimaNode
from lima import db
from lima.tools import segment
from lima.settings import CONFIG

import logging
logger = logging.getLogger(__name__)


class BaseIndexer(object):
    __metaclass__ = abc.ABCMeta

    # main indexes directory
    INDEXES_FOLDER = fs.join(CONFIG['meta_root'], u'indexes')

    # per host index directory

    def __init__(self, filesdb):
        self.filesdb = filesdb
        self.filesdb_name = 'dbfiles'

    @classmethod
    def HOST_INDEXES_FOLDER(cls):
        return fs.join(cls.INDEXES_FOLDER, 'hosts', fs.address)


    @abc.abstractproperty
    def CONTENT_TYPE(self):
        """ Abstract property to force childern to have a CONTENT_TYPE defined
        """
        pass

    @abc.abstractproperty
    def DBNAME(self):
        pass

    @classmethod
    def DB_ZIP(cls):
        """ Abstract property to force childern to have a DB_ZIP defined
        """
        return fs.join(cls.INDEXES_FOLDER, u"%s.db.gz" % cls.DBNAME)

    @classmethod
    def DB(cls):
        return fs.join(cls.HOST_INDEXES_FOLDER(), u"%s.db" % cls.DBNAME)

    @classmethod
    def DB_TEMP(cls):
        return fs.join(cls.HOST_INDEXES_FOLDER(), u"%s.db.tmp" % cls.DBNAME)

    @classmethod
    def DB_ZIP_TEMP(cls):
        return fs.join(cls.HOST_INDEXES_FOLDER(), u"%s.db.gz.tmp" % cls.DBNAME)

    @abc.abstractmethod
    def _do_index(self, *args, **kwargs):
        """ Abstract property to force children to have a '_do_index()' method
        defined"""
        pass

    @abc.abstractmethod
    def _mark_missing_from_existing(self, *args, **kwargs):
        pass

    # database specific operations
    def execute(self, sql, vals=None, cursor=None):
        try:
            if vals:
                return cursor.execute(sql, vals)
            else:
                return cursor.execute(sql)
        except Exception:
            logger.error("%s - %s", sql, vals)
            raise

    def execute_many(self, idx_db, *sqls):
        cursor = idx_db.cursor()
        for sql in sqls:
            self.execute(sql, None, cursor)

    def create_or_load(self, idx_db, gzip_file_name):
        """ Create the idx_db or load it from gz file if present """
        if os.path.exists(gzip_file_name):
            try:
                self._load_from_gzip(idx_db, gzip_file_name)
            except Exception:
                logger.warn(u"Could not read existing index. Rebuilding")
                try:
                    # segment.track_error(event="%s index read error" % self.CONTENT_TYPE)
                    # try to recreate from
                    # every indexes from the most recent
                    # in INDEXES_FOLDER until success
                    # if no success, rereate the idx_db
                    try:
                        os.unlink(gzip_file_name)
                    except Exception:
                        # segment.track_error(event="Could not remove broken gzip index (%s)" % self.CONTENT_TYPE)
                        logger.error(u"Could not remove broken gzip index")
                    try:
                        self.force_reindexation()
                    except Exception:
                        # segment.track_error(event="Could not force reindexation (%s)" % self.CONTENT_TYPE)
                        logger.error(u"Could not remove broken gzip index")
                    try:
                        self._add_reindeaxation_flags()
                    except Exception:
                        # segment.track_error(event="Could not flag reindexation (%s)" % self.CONTENT_TYPE)
                        logger.error(u"Could not flag reindexation")
                    # try to create the db anyway
                except LimaFsError:
                    raise
                except Exception:
                    # segment.track_error(event="%s index force rebuild error" % self.CONTENT_TYPE)
                    logger.error(u"Could not rebuild index")
            finally:
                self._create(idx_db)
        else:
            self._create(idx_db)
        return idx_db

    def _create(self, idx_db):
        self.execute_many(idx_db, *self.create_statements)

    def _load_from_gzip(self, idx_db, fname):
        with gzip.open(fname, 'rb') as fgz:
            tmp_path = os.path.join(tempfile.gettempdir(), "%s.tmp" % fs.basename(self.DB()))
            try:
                if os.path.exists(tmp_path):
                    os.remove(tmp_path)
            except Exception:
                logger.error(u"Could not remove old temporary file %s", tmp_path)

            with open(tmp_path, 'wb') as tmpfile:
                tmpfile.write(fgz.read())

            tmpdb = None
            try:
                tmpdb = sqlite3.connect(tmp_path)
                tempsql = StringIO()
                for line in tmpdb.iterdump():
                    tempsql.write('%s\n' % line)
                tempsql.seek(0)

                idx_db.cursor().executescript(tempsql.read())
            finally:
                if tmpdb is not None:
                    tmpdb.close()


    def _load_db_gzip(self):
        try:
            if fs.exists(self.DB_ZIP()):
                db_index = sqlite3.connect(":memory:")
                db_zip_path = fs.get_system_path(self.DB_ZIP())
                self._load_from_gzip(db_index, db_zip_path)
                return db_index
        except Exception:
                return None

    def attach(self, original_db, otherdb, otherdb_alias):
        class wrapper(object):
            def __enter__(_self):
                self._attach_filesdb(original_db, otherdb, otherdb_alias)
                return _self

            def __exit__(_self, *exc):
                self._detach_filesdb(original_db, otherdb_alias)
                if exc[0] is not None:
                    raise exc[0], exc[1], exc[2]

        return wrapper()

    def _attach_filesdb(self, original_db, otherdb, otherdb_alias):
        original_db.cursor().execute("attach database ? as ?", [otherdb, otherdb_alias])

    def _detach_filesdb(self, original_db, otherdb_alias):
        original_db.cursor().execute("detach database ?", [otherdb_alias, ])

    # / database specific operations

    @classmethod
    def pin(cls):
        # TODO : pin all indexes
        try:
            stat = fs.stat(cls.DB_ZIP())
            if stat.st_pin is not True:
                fs.pin(True, cls.DB_ZIP())
                logger.debug(u"%s pined", cls.DB_ZIP())
        except LimaFsDoesnotExistError:
            logger.debug(u"%s does not exists", cls.DB_ZIP())

    def force_reindexation(self):
        """ Force reindexation for child CONTENT_TYPE"""
        with db.transaction():
            nb_forced = LimaNode.force_reindexation(self.CONTENT_TYPE)
        logger.info(u"%s %s marked for reindexation", nb_forced, self.CONTENT_TYPE)

    def _check_force_reindexation_(self):
        """Method executed before do_index """
        flag_files = ['.'.join([self.DB(), 'force']),
                      '.'.join([self.DB_ZIP(), 'force'])]
        for flag in flag_files:
            if fs.exists(flag):
                logger.info(u"Force %s reindexation", self.CONTENT_TYPE)
                self.force_reindexation()
                break

    def __enter__(self):
        self._check_force_reindexation_()

    def __exit__(self, *exc):
        """Method executed before do_index """
        flag_files = ['.'.join([self.DB(), 'force']),
                      '.'.join([self.DB_ZIP(), 'force'])]
        for flag in flag_files:
            if fs.exists(flag):
                logger.info(u"%s reindexation flag removed", self.CONTENT_TYPE)
                fs.unlink(flag)

    def do_index(self, *args, **kwargs):
        with self:
            return self._do_index(*args, **kwargs)

    def _add_reindeaxation_flags(self):
        """ Drop a flag in /.lima/meta/indexes/{address} for other machine to
        set them as to be reindexed"""
        root = self.INDEXES_FOLDER
        db_basename = os.path.basename(self.DB())

        for dirname in fs.listdir(root):
            if dirname in [u".", u".."]:
                continue

            if dirname.endswith(u"/"):
                dirname = dirname[:-1]

            other_address = dirname

            if other_address != fs.address:
                other_db = fs.join(root, other_address, db_basename)
                if fs.exists(other_db):
                    flag = '.'.join([other_db, 'force'])
                    try:
                        with fs.open(flag, 'w'):
                            logger.info(u"#" * 10 + u"Reindexation flag set for %s", dirname)
                    except Exception as e:
                        logger.error(u"Could not flag %s as to be reindexed", other_address)

    def mark_missings(self, lima_db):
        try:
            self._mark_missing_from_existing(lima_db)
        except (sqlite3.OperationalError) as e:
            # db is in a weird state.
            # we remove it and flag files a to be reindexed.
            if 'no such table' in e.message:
                try:
                    fs.unlink(self.DB())
                except LimaFsDoesnotExistError:
                    pass
                with db.transaction():
                    LimaNode.force_reindexation(self.CONTENT_TYPE)
            else:
                raise
        except (sqlite3.DatabaseError) as e:
            # db is dead.
            # we remove it and flag files a to be reindexed.
            try:
                fs.unlink(self.DB())
            except LimaFsDoesnotExistError:
                pass
            with db.transaction():
                LimaNode.force_reindexation(self.CONTENT_TYPE)
