# coding: utf-8

import sqlite3
from lima import db
from lima.tools import bin_to_hex
from lima.meta.indexers.base import BaseIndexer
from lima.fs import FileSystem as fs
from lima.models import LimaNode

from lima.fs.limafs import (LimaFsError, LimaFsConnectionRefusedError,
                            LimaFsNotReadyError, LimaFsDoesnotExistError)
from lima.meta.blackboard.agents.storage import (AgentStorage, )
from lima.meta.storage import (SerieStorage, SeasonStorage, )
from lima.meta import cleanup_utils
from lima.settings import CONFIG

import logging
logger = logging.getLogger(__name__)


class TvIndexer(BaseIndexer):

    CONTENT_TYPE = u'episode'
    DBNAME = u'series'


    def __repr__(self):
        return u"lima.meta.files.indexers.episodes"

    @property
    def create_statements(self):
        create_table_network = """CREATE TABLE IF NOT EXISTS Network(
            id INTEGER PRIMARY KEY AUTOINCREMENT,
            name TEXT
        )
        """

        create_indexes_network_name = "CREATE INDEX IF NOT EXISTS network_name_ix on Network(name)"

        create_table_serie = """CREATE TABLE IF NOT EXISTS Serie(
            id INTEGER PRIMARY KEY AUTOINCREMENT,
            name TEXT NOT NULL,
            original_name TEXT,

            synopsis TEXT,
            cover TEXT
        )
        """

        create_indexes_serie_name = "CREATE INDEX IF NOT EXISTS serie_name_ix on Serie(name)"

        create_networks_series = """ CREATE TABLE IF NOT EXISTS NetworksSeries(
        network_id INTEGER REFERENCES Network(id) NOT NULL,
        serie_id INTEGER REFERENCES Serie(id) NOT NULL
        )
        """

        create_table_season = """CREATE TABLE IF NOT EXISTS Season(
            id INTEGER PRIMARY KEY AUTOINCREMENT,

            number INTEGER,
            name TEXT,

            synopsis TEXT,
            cover TEXT,
            serie_id INT REFERENCES Serie(id)
        )
        """

        create_indexes_season_name = "CREATE INDEX IF NOT EXISTS season_name_ix on Season(name)"
        create_indexes_seaon_serie_id = "CREATE INDEX IF NOT EXISTS season_serie_id_ix on Season(serie_id)"

        create_table_episode = """
        CREATE TABLE IF NOT EXISTS Episode(
            id INTEGER PRIMARY KEY AUTOINCREMENT,
            number INTEGER NULL,

            name TEXT,
            synopsis TEXT,
            cover TEXT,
            path TEXT NOT NULL,
            pvw TEXT NULL,
            season_id INT REFERENCES Season(id)
        )
        """

        create_indexes_episode_number = "CREATE INDEX IF NOT EXISTS episode_name_ix on Episode(number)"
        create_indexes_serie_season_id = "CREATE INDEX IF NOT EXISTS episode_serie_id_ix on Episode(season_id)"

        create_table_person = """
        CREATE TABLE IF NOT EXISTS Person(
            id INTEGER PRIMARY KEY AUTOINCREMENT,
            name TEXT
        )
        """

        create_indexes_person_name = "CREATE INDEX IF NOT EXISTS artist_name_ix on Person(name)"

        create_table_casting = """
        CREATE TABLE IF NOT EXISTS SerieCasting(
            person_id REFERENCES Person(id) NOT NULL,
            serie_id REFERENCES Serie(id) NOT NULL,
            type TEXT NOT NULL,

            role TEXT,
            PRIMARY KEY (person_id, serie_id, type, role)
        )
        """

        create_indexes_scasting_person_id = "CREATE INDEX IF NOT EXISTS serie_casting_person_id on SerieCasting(person_id)"
        create_indexes_scasting_serie_id = "CREATE INDEX IF NOT EXISTS serie_casting_serie_id on SerieCasting(serie_id)"
        create_indexes_scasting_type = "CREATE INDEX IF NOT EXISTS serie_casting_type on SerieCasting(type)"

        create_table_episode_casting = """
        CREATE TABLE IF NOT EXISTS EpisodeCasting(
            person_id REFERENCES Person(id) NOT NULL,
            episode_id REFERENCES Episode(id) NOT NULL,
            type TEXT NOT NULL,

            role TEXT,
            PRIMARY KEY (person_id, episode_id, type, role)
        )
        """

        create_indexes_ecasting_person_id = "CREATE INDEX IF NOT EXISTS episode_casting_person_id on EpisodeCasting(person_id)"
        create_indexes_ecasting_episode_id = "CREATE INDEX IF NOT EXISTS episode_casting_episode_id on EpisodeCasting(episode_id)"
        create_indexes_ecasting_type = "CREATE INDEX IF NOT EXISTS episode_casting_type on EpisodeCasting(type)"

        tables = [create_table_network,
                  create_table_serie,
                  create_networks_series,
                  create_table_season,
                  create_table_episode,
                  create_table_person,
                  create_table_casting,
                  create_table_episode_casting, ]
        indexes = [
            create_indexes_network_name, create_indexes_serie_name,
            create_indexes_season_name, create_indexes_seaon_serie_id,
            create_indexes_episode_number, create_indexes_serie_season_id,
            create_indexes_person_name, create_indexes_scasting_type,
            create_indexes_scasting_serie_id, create_indexes_scasting_person_id,
            create_indexes_ecasting_type, create_indexes_ecasting_episode_id,
            create_indexes_ecasting_type, create_indexes_ecasting_person_id
        ]

        return tables + indexes

    def insert_networks(self, cursor, networks):
        insert = """ INSERT OR REPLACE INTO Network(id, name)
        VALUES (
                (select id from Network where UPPER(name) = ? limit 1),
                ?
        )"""

        for network in networks:
            values = (network.upper(), network)
            self.execute(insert, values, cursor)

    def insert_serie(self, cursor, name, original_name, synopsis, cover):

        if name is None:
            name = original_name

        if name is None:
            logger.debug(u"Series does not have a name. Ignored from index")

        insert = """ INSERT OR REPLACE INTO Serie(id, name, original_name, synopsis, cover)
        VALUES (
                (select id from Serie where name = ? and (original_name = ? or (original_name is NULL and ? IS NULL)) limit 1),
                ?,
                ?,
                ?,
                ?
        )"""

        values = (name, original_name, original_name, name, original_name,
                  synopsis, cover)
        self.execute(insert, values, cursor)

    def insert_serie_network(self, cursor, serie_id, networks):
        insert = """ INSERT OR REPLACE INTO NetworksSeries(serie_id, network_id)
        VALUES (
                ?,
                (select id from network where upper(name) = ? limit 1)
        )"""

        for network in networks:
            values = (serie_id, network.upper())
            self.execute(insert, values, cursor)

    def insert_season(self, cursor, serie_id, name, number, synopsis, cover):
        insert = """ INSERT OR REPLACE INTO Season(
                id,
                name,
                number,
                synopsis,
                cover,
                serie_id
        )
        VALUES (
                (select sea.id from season sea where sea.serie_id = ? and sea.number = ?),
                ?,
                ?,
                ?,
                ?,
                ?
        )"""

        values = (serie_id, number, name, number, synopsis, cover, serie_id)
        self.execute(insert, values, cursor)

    def insert_serie_casting(self, cursor, serie_id, person_id, _type,
                             role=None):
        insert = """
        insert or replace into SerieCasting(serie_id, person_id, type, role)
        VALUES (
        ?,
        ?,
        ?,
        ?
        )
        """

        values = (serie_id, person_id, _type, role)
        self.execute(insert, values, cursor)

    def insert_episode_casting(self, cursor, episode_id, person_id, _type,
                               role=None):
        insert = """
        insert or replace into EpisodeCasting(episode_id, person_id, type, role)
        VALUES (
        ?,
        ?,
        ?,
        ?
        )
        """

        if type(role) in (str, unicode) or role is None:
            values = (episode_id, person_id, _type, role)
            self.execute(insert, values, cursor)
        elif type(role) is list:
            for _role in role:
                values = (episode_id, person_id, _type, _role)
                self.execute(insert, values, cursor)
                self.execute(insert, values, cursor)

    def insert_person(self, cursor, serie_id, episode_id, name, _type,
                      role=None):
        if name is None:
            logger.warn(
                u"Person does not have a name - not including in the index")
            return
        else:
            insert = """ INSERT OR REPLACE INTO Person(
                id, name
            )
            VALUES(
                (SELECT id from Person where name = ?),
                ?
            )
            """

            values = [name, name]
            self.execute(insert, values, cursor)
            if serie_id is not None:
                self.insert_serie_casting(cursor, serie_id, cursor.lastrowid,
                                          _type, role)
            if episode_id is not None:
                self.insert_episode_casting(cursor, episode_id,
                                            cursor.lastrowid, _type, role)

    def _insert_crew_serie(self, cursor, serie_id, actors, producers, writers):

        for actor in actors:
            try:
                self.insert_person(cursor, serie_id, None, actor['name'],
                                   'actor', actor['character'])
            except Exception:
                logger.exception(u'Could not insert actor')
        for producer in producers:
            try:
                self.insert_person(cursor, serie_id, None, producer['name'],
                                   'producer')
            except Exception:
                logger.exception(u'Could not insert producer')
        for writer in writers:
            try:
                self.insert_person(cursor, serie_id, None, writer['name'],
                                   'writer')
            except Exception:
                logger.exception(u'Could not insert writer')

    def _insert_crew_episode(self, cursor, episode_id, actors, producers,
                             writers):
        for actor in actors:
            try:
                self.insert_person(cursor, None, episode_id, actor['name'],
                                   'actor', actor['character'])
            except Exception:
                logger.exception(u'Could not insert actor')
        for producer in producers:
            try:
                self.insert_person(cursor, None, episode_id, producer['name'],
                                   'producer')
            except Exception:
                logger.exception(u'Could not insert producer')
        for writer in writers:
            try:
                self.insert_person(cursor, None, episode_id, writer['name'],
                                   'writer')
            except Exception:
                logger.exception(u'Could not insert writer')

    def insert_episode(self, cursor, *args, **kwargs):
        try:
            self.insert_networks(cursor, kwargs['serie_networks'])
        except Exception:
            logger.exception(u"Could not insert networks")

        self.insert_serie(cursor, kwargs['serie_name'],
                          kwargs['serie_original_name'],
                          kwargs['serie_synopsis'], kwargs['serie_cover'], )
        serie_id = cursor.lastrowid
        try:
            self.insert_serie_network(cursor, serie_id,
                                      kwargs['serie_networks'], )
        except Exception:
            logger.exception("Could not insert network")

        try:
            self._insert_crew_serie(cursor, serie_id, kwargs['serie_actors'],
                                    kwargs['serie_producers'],
                                    kwargs['serie_writers'], )
        except Exception:
            logger.exception("Could not insert serie crew")

        self.insert_season(cursor, serie_id, kwargs['season_name'],
                           kwargs['season_number'], kwargs['season_synopsis'],
                           kwargs['season_cover'], )
        season_id = cursor.lastrowid

        insert = """insert or replace into Episode(
            id,
            name,
            number,
            synopsis,
            cover,
            pvw,
            path,
            season_id
        ) VALUES (
            (select id from episode where path = ? limit 1),
            ?,
            ?,
            ?,
            ?,
            ?,
            ?,
            ?
        )
        """

        values = (kwargs['path'], kwargs['episode_name'],
                  kwargs['episode_number'], kwargs['episode_synopsis'],
                  kwargs['episode_cover'], kwargs['episode_pvw'],
                  kwargs['path'], season_id, )
        self.execute(insert, values, cursor)
        episode_id = cursor.lastrowid
        try:
            self._insert_crew_episode(cursor, episode_id,
                                      kwargs['episode_actors'],
                                      kwargs['episode_producers'],
                                      kwargs['episode_writers'], )
        except Exception:
            logger.exception(u"Could not insert episode crew")

    def _mark_missing_from_existing(self, lima_db):
        if fs.exists(self.DB()):
            db_index = self._load_db_gzip()
            if not db_index:
                return
            try:
                with self.attach(db_index, self.filesdb, self.filesdb_name):
                        mark_missings = """
                            update %(files_db)s.limanode
                            set status_index = %(new_status_index)s
                            where %(files_db)s.limanode.node_id in (
                                    select node.node_id
                                    from %(files_db)s.limanode as node
                                    left join Episode e on node.path = e.path
                                    where node.content_type = '%(content_type)s'
                                        and node.status_index != %(old_status_index)s
                                        and node.status = %(status)s
                                        and e.id is null
                            )
                        """ % {
                                'files_db': self.filesdb_name,
                                'content_type': self.CONTENT_TYPE,
                                'status': LimaNode.STATUS_PROCESSED,
                                'new_status_index': LimaNode.STATUS_INDEX_NEED_REINDEX,
                                'old_status_index': LimaNode.STATUS_INDEX_IGNORE
                                }
                        with lima_db.transaction():
                            cursor = db_index.cursor()
                            self.execute(mark_missings, cursor=cursor)
                        if cursor.rowcount:
                            logger.warn(u"Marked %s %s missing from index as to be reindexed", cursor.rowcount, self.CONTENT_TYPE)
            finally:
                db_index.close()

    def do_index_episodes(self, idx_db, lima_db):
        cursor = idx_db.cursor()
        indexed = 0
        # Here, we take : identified musics (content_type == 'movie' in the db)
        # and non identified files (content_type is null)
        # as the non identified files may have been processed and identified by
        # another host.
        # The file will be added if and only if its content type is set to
        # 'movie' in metas files
        with lima_db.transaction():
            episodes = LimaNode.reindexables("video", self.CONTENT_TYPE)[:]
            # episodes += LimaNode.processed_video(None)[:]
        for node in episodes:
            try:
                storage = None
                serie_storage = None
                season_storage = None
                try:
                    st = fs.stat(node.path)
                except LimaFsDoesnotExistError:
                    try:
                        # flag file as deleted
                        with lima_db.transaction():
                            node.status_index = LimaNode.STATUS_INDEX_IGNORE
                            node.status = LimaNode.STATUS_DELETED
                            node.save()
                    except Exception as e:
                        logger.exception(u"Could not flag not existing file as not be reindexed %s : %s", node.path, getattr(e, 'message', 'Unknown error'))
                    continue
                except (LimaFsNotReadyError, LimaFsConnectionRefusedError):
                    logger.error("Filesystem not ready. Can't index")
                    break
                except LimaFsError:
                    logger.error("Filesystem error. Can't index")
                    break

                node_id = bin_to_hex(st.v['id'])
                # GC may not have run
                if node.node_id != node_id:
                    continue
                datas = {}
                storage = AgentStorage(node.path)
                if storage.getmeta('content_type') != self.CONTENT_TYPE:
                    logger.debug("%s : not identified as a %s", node.path,
                                 self.CONTENT_TYPE)
                    continue

                datas['path'] = node.path

                datas['serie_name'] = storage.getmeta("serie.name")
                serie_storage = SerieStorage.get_from_pool(datas['serie_name'])
                datas['serie_original_name'] = serie_storage.getmeta(
                    "serie.original_name")
                datas['serie_synopsis'] = serie_storage.getmeta("serie.overview")
                datas['serie_actors'] = serie_storage.getmeta("actors", [])
                datas['serie_producers'] = serie_storage.getmeta("producers", [])
                datas['serie_writers'] = serie_storage.getmeta("writers", [])
                datas['serie_networks'] = serie_storage.getmeta("networks", [])

                datas['serie_cover'] = None
                if serie_storage.exists('cover'):
                    datas['serie_cover'] = serie_storage.get_storage_fname('cover')

                datas['season_number'] = storage.getmeta("serie.season.number")
                season_storage = SeasonStorage.get_from_pool(datas['serie_name'], datas['season_number'])
                datas['season_name'] = season_storage.getmeta("serie.season.name")
                datas['season_synopsis'] = season_storage.getmeta(
                    "serie.season.overview")
                datas['season_cover'] = None
                if season_storage.exists('cover'):
                    datas['season_cover'] = season_storage.get_storage_fname(
                        'cover')

                datas['episode_name'] = storage.getmeta("serie.episode.name")
                datas['episode_number'] = storage.getmeta("serie.episode.number")
                datas['episode_actors'] = storage.getmeta("actors", [])
                datas['episode_writers'] = storage.getmeta("writers", [])
                datas['episode_producers'] = storage.getmeta("producers", [])
                datas['episode_synopsis'] = storage.getmeta("serie.episode.overview")

                datas['episode_cover'] = None
                if storage.exists('cover'):
                    datas['episode_cover'] = storage.get_storage_fname('cover')
                datas['episode_pvw'] = None
                if storage.exists('pvw'):
                    datas['episode_pvw'] = storage.get_storage_fname('pvw')

                logger.debug(u"%s - %s : S%sE%s", node.path, datas['serie_name'],
                             datas['season_number'], datas['episode_number'], )

                self.insert_episode(cursor, **datas)
                indexed += 1
                try:
                    with lima_db.transaction():
                        node.status_index = LimaNode.STATUS_INDEX_INDEXED
                        node.save()
                except Exception:
                    logger.exception(u"Could not flag as indexed %s", node.path)
            except Exception:
                logger.exception(u"Could not index %s", node.path)
            finally:
                if storage is not None:
                    storage.unload()
                if serie_storage is not None:
                    serie_storage.unload()
                if season_storage is not None:
                    season_storage.unload()
        else:
            logger.info(u"episode indexation : done (%s new(s))", indexed)
        return indexed

    def _find_deleteable_items(self, idx_db):
        with self.attach(idx_db, self.filesdb, self.filesdb_name):
            cursor = idx_db.cursor()
            create_temptable = "CREATE TEMP TABLE episodes_to_delete(id int);"
            self.execute(create_temptable, cursor=cursor)

            insert_episode_delete = """ insert into episodes_to_delete
            select episode.id
            from episode
            left join %s.limanode on limanode.path = episode.path
            where limanode.node_id is null
            """ % self.filesdb_name  # assume this attribute is safe
            with idx_db:
                self.execute(insert_episode_delete, cursor=cursor)
            logger.debug(u"Will remove %s episodes from %s index", cursor.rowcount, self.CONTENT_TYPE)

    def do_cleanup(self, idx_db):
        total_cleaned = 0
        #
        self._find_deleteable_items(idx_db)

        cursor = idx_db.cursor()

        delete_episodes_casting = """
        delete from EpisodeCasting
        where EpisodeCasting.episode_id in (
            select episodes_to_delete.id
            from episodes_to_delete
        )
        """

        self.execute(delete_episodes_casting, cursor=cursor)
        total_cleaned += cursor.rowcount
        logger.debug(u"Will remove %s episodes castings from %s index", cursor.rowcount, self.CONTENT_TYPE)

        delete_episodes = """
        delete from episode
        where episode.id in (
            select episodes_to_delete.id
            from episodes_to_delete
        )
        """

        self.execute(delete_episodes, cursor=cursor)
        total_cleaned += cursor.rowcount
        logger.debug(u"Will remove %s episodes from %s index", cursor.rowcount, self.CONTENT_TYPE)

        create_temptable = "CREATE TEMP TABLE seasons_to_delete(id int, serie_name TEXT, season_number int);"
        self.execute(create_temptable, cursor=cursor)

        insert_delete_seasons = """
        insert into seasons_to_delete(id, serie_name, season_number)
          select season.id, serie.name, season.number
          from season
          inner join serie on serie.id = season.serie_id
          left join episode on episode.season_id = season.id
          group by season.id
          having count(distinct episode.id) == 0
        """

        self.execute(insert_delete_seasons, cursor=cursor)

        select_seasons_to_clean = "select serie_name, season_number from seasons_to_delete"
        for line in cursor.execute(select_seasons_to_clean):
            serie_name, season_number = line[0], line[1]
            try:
                cleanup_utils.clean_up_season(serie_name, season_number)
            except Exception:
                logger.error("Could not remove metas for %s - %s", serie_name, season_number)

        delete_seasons = """
        delete from season where season.id in (
            select s.id
            from seasons_to_delete s
        )
        """

        self.execute(delete_seasons, cursor=cursor)
        total_cleaned += cursor.rowcount
        logger.debug(u"Removed %s seasons from %s index", cursor.rowcount, self.CONTENT_TYPE)

        create_temptable = "CREATE TEMP TABLE series_to_delete(id int, serie_name TEXT);"
        self.execute(create_temptable, cursor=cursor)

        insert_series_delete = """insert into series_to_delete
        select serie.id, serie.name
        from serie
        left join season on season.serie_id = serie.id
        group by serie.id
        having count(distinct season.id) == 0
        """

        self.execute(insert_series_delete, cursor=cursor)
        logger.debug(u"Will remove %s serie(s) from %s index", cursor.rowcount, self.CONTENT_TYPE)

        delete_serie_casting = """
        delete from SerieCasting
        where SerieCasting.serie_id in (
            select series_to_delete.id
            from series_to_delete
        )
        """

        self.execute(delete_serie_casting, cursor=cursor)
        total_cleaned += cursor.rowcount
        logger.debug(u"Will remove %s serie castings from %s index", cursor.rowcount, self.CONTENT_TYPE)

        delete_serie_network = """
        delete from NetworksSeries
        where NetworksSeries.serie_id in (
            select series_to_delete.id
            from series_to_delete
        )
        """

        self.execute(delete_serie_network, cursor=cursor)
        total_cleaned += cursor.rowcount
        logger.debug(u"Removed %s serie network from %s index", cursor.rowcount, self.CONTENT_TYPE)

        delete_networks = """
        delete from Network
        where Network.id in (
            select nt.id
            from Network nt
            left join NetworksSeries on NetworksSeries.network_id = nt.id
            group by nt.id
            having count(distinct NetworksSeries.serie_id) == 0
        )
        """

        self.execute(delete_networks, cursor=cursor)
        total_cleaned += cursor.rowcount
        logger.debug(u"Removed %s network(s) from %s index", cursor.rowcount, self.CONTENT_TYPE)

        select_series_to_clean = "select serie_name from series_to_delete"
        for line in cursor.execute(select_series_to_clean):
            serie_name = line[0]
            try:
                cleanup_utils.clean_up_serie(serie_name)
            except Exception:
                logger.error("Could not remove metas for %s - %s", serie_name, season_number)

        delete_series = """
        delete from serie where serie.id in (
            select id from series_to_delete
        )
        """

        self.execute(delete_series, cursor=cursor)
        total_cleaned += cursor.rowcount
        logger.debug(u"Removed %s serie(s) from %s index", cursor.rowcount, self.CONTENT_TYPE)
        return total_cleaned

    def indexables(self):
        with db.transaction():
            return LimaNode.reindexables(u"video", "episode")

    def _do_index(self, idx_db, lima_db):
        self.create_or_load(idx_db, fs.get_system_path(self.DB_ZIP()))
        try:
            indexed = self.do_index_episodes(idx_db, lima_db)
            cleaned = 0
            with idx_db:
                cleaned = self.do_cleanup(idx_db)
            return ((indexed or 0) != 0) or ((cleaned or 0) != 0)
        except Exception:
            logger.error(u"Indexation error")
            raise
        finally:
            SerieStorage.unload_all(persist=False)
            SeasonStorage.unload_all(persist=False)
