# coding: utf-8
import sqlite3
from lima import db
from lima.tools import bin_to_hex
from lima.meta.indexers.base import BaseIndexer
from lima.fs import FileSystem as fs
from lima.models import LimaNode

from lima.fs.limafs import (LimaFsError, LimaFsConnectionRefusedError,
                            LimaFsNotReadyError, LimaFsDoesnotExistError)
from lima.meta.blackboard.agents.storage import (AgentStorage, )
from lima.settings import CONFIG

import logging
logger = logging.getLogger(__name__)


class MovieIndexer(BaseIndexer):

    CONTENT_TYPE = u'movie'
    DBNAME=u'movies'


    def __repr__(self):
        return u"lima.meta.files.indexers.movies"

    @property
    def create_statements(self):
        create_table_person = """
        CREATE TABLE IF NOT EXISTS Person(
            id INTEGER PRIMARY KEY AUTOINCREMENT,
            name TEXT
        )
        """

        create_indexes_person_name = "CREATE INDEX IF NOT EXISTS artist_name_ix on Person(name)"

        create_table_movies = """
        CREATE TABLE IF NOT EXISTS Movie(
            id INTEGER PRIMARY KEY AUTOINCREMENT,
            path TEXT NOT NULL,
            title TEXT NOT NULL,
            year INTEGER NOT NULL,
            duration FLOAT,
            synopsis TEXT,
            pvw TEXT,
            cover TEXT
        )
        """

        create_indexes_movie_name = "CREATE INDEX IF NOT EXISTS artist_name_ix on Movie(name)"

        create_table_casting = """
        CREATE TABLE IF NOT EXISTS Casting(
            person_id REFERENCES Person(id) NOT NULL,
            movie_id REFERENCES Movie(id) NOT NULL,
            type TEXT NOT NULL,

            role TEXT,
            PRIMARY KEY (person_id, movie_id, type, role)
        )
        """

        create_indexes_casting_person_id = "CREATE INDEX IF NOT EXISTS casting_person_id on Casting(person_id)"
        create_indexes_casting_movie_id = "CREATE INDEX IF NOT EXISTS casting_movie_id on Casting(movie_id)"
        create_indexes_casting_persone_type = "CREATE INDEX IF NOT EXISTS casting_person_type on Casting(type)"

        tables = [create_table_person, create_table_movies,
                  create_table_casting]
        indexes = [create_indexes_person_name,
                   create_indexes_movie_name,
                   create_indexes_casting_person_id,
                   create_indexes_casting_movie_id,
                   create_indexes_casting_persone_type, ]

        return tables + indexes

    def insert_person(self, cursor, name):
        if name is None:
            logger.warn(
                u"Person does not have a name - not including in the index")
            return
        else:
            insert = """ INSERT OR IGNORE INTO Person(
                id, name
            )
            VALUES(
                (SELECT id from Person where name = ?),
                ?
            )
            """

            values = [name, name]
            self.execute(insert, values, cursor)

    def insert_casting(self, cursor, movie_id, person_id, role, kind):
        if movie_id is None:
            logger.warn(
                u"Casting does not have a movie id- not including in the index")
            return
        if person_id is None:
            logger.warn(
                u"Casting does not have a person id - not including in the index")
            return
        insert = """ INSERT OR REPLACE INTO Casting(
            person_id,
            movie_id,
            role,
            type
        )
        VALUES(
            ?,
            ?,
            ?,
            ?
        )
        """

        if type(role) in (str, unicode) or role is None:
            values = [person_id, movie_id, role, kind]
            self.execute(insert, values, cursor)
        elif type(role) is list:
            for _role in role:
                values = [person_id, movie_id, _role, kind]
                self.execute(insert, values, cursor)

    def insert_movie(self, cursor, path, title, year, duration, directors,
                     actors, synopsis, pvw, cover):
        if title is None or year is None:
            logger.warn(u"%s as no title/year - not including in the index",
                        path)
            return
        else:
            insert = """
            INSERT OR REPLACE INTO Movie(
                id,
                path,
                title,
                year,
                duration,
                pvw,
                cover,
                synopsis
            )
            VALUES(
                (SELECT id FROM Movie where path = ?),
                ?,
                ?,
                ?,
                ?,
                ?,
                ?,
                ?
            )
            """

            values = (path, path, title, year, duration, pvw, cover, synopsis)
            self.execute(insert, values, cursor)

            movie_id = cursor.lastrowid

            for director in directors:
                name = director.get('name', None)
                try:
                    self.insert_person(cursor, name)
                    person_id = cursor.lastrowid
                    self.insert_casting(cursor, movie_id, person_id, None,
                                        'director')
                except Exception:
                    logger.exception(u"Could not insert director %s", name)
                    pass

            for actor in actors:
                name = actor.get('name', None)
                role = actor.get('character', None)
                try:
                    self.insert_person(cursor, name)
                    person_id = cursor.lastrowid
                    self.insert_casting(cursor, movie_id, person_id, role,
                                        'actor')
                except Exception:
                    logger.exception(u"Could not insert %s - %s", name, role)

    def _mark_missing_from_existing(self, lima_db):
        if fs.exists(self.DB()):
            db_index = self._load_db_gzip()
            if not db_index:
                return
            try:
                with self.attach(db_index, self.filesdb, self.filesdb_name):
                        mark_missings = """
                            update %(files_db)s.limanode
                            set status_index = %(new_status_index)s
                            where %(files_db)s.limanode.node_id in (
                                    select node.node_id
                                    from %(files_db)s.limanode as node
                                    left join movie m on node.path = m.path
                                    where node.content_type = '%(content_type)s'
                                        and node.status_index != %(old_status_index)s
                                        and node.status = %(status)s
                                        and m.id is null
                            )
                        """ % {
                                'files_db':self.filesdb_name,
                                'content_type':self.CONTENT_TYPE,
                                'status':LimaNode.STATUS_PROCESSED,
                                'new_status_index': LimaNode.STATUS_INDEX_NEED_REINDEX,
                                'old_status_index': LimaNode.STATUS_INDEX_IGNORE
                                }
                        with lima_db.transaction():
                            cursor = db_index.cursor()
                            self.execute(mark_missings, cursor=cursor)
                        if cursor.rowcount:
                            logger.warn(u"Marked %s %s missing from index as to be reindexed", cursor.rowcount, self.CONTENT_TYPE)
            finally:
                db_index.close()

    def do_index_movies(self, idx_db, lima_db):
        cursor = idx_db.cursor()
        indexed = 0
        # Here, we take : identified musics (content_type == 'movie' in the db)
        # and non identified files (content_type is null)
        # as the non identified files may have been processed and identified by
        # another host.
        # The file will be added if and only if its content type is set to
        # 'movie' in metas files
        with lima_db.transaction():
            movies = LimaNode.reindexables("video", self.CONTENT_TYPE)[:]
            # movies += LimaNode.processed_video(None)[:]

        for node in movies:
            try:
                storage = None
                try:
                    st = fs.stat(node.path)
                except LimaFsDoesnotExistError:
                    try:
                        # flag file as deleted
                        with lima_db.transaction():
                            node.status_index = LimaNode.STATUS_INDEX_IGNORE
                            node.status = LimaNode.STATUS_DELETED
                            node.save()
                    except Exception as e:
                        logger.exception(u"Could not flag not existing file as not be reindexed %s : %s", node.path, getattr(e, 'message', 'Unknown error'))
                    continue
                except (LimaFsNotReadyError, LimaFsConnectionRefusedError):
                    logger.error("Filesystem not ready. Can't index")
                    break
                except LimaFsError:
                    logger.error("Filesystem error. Can't index")
                    break
                node_id = bin_to_hex(st.v['id'])
                # GC may not have run
                if node.node_id != node_id:
                    continue
                datas = {}
                storage = AgentStorage(node.path)
                if storage.getmeta('content_type') != self.CONTENT_TYPE:
                    logger.debug("%s : not identified as a %s", node.path, self.CONTENT_TYPE)
                    with lima_db.transaction():
                        node.status_index = LimaNode.STATUS_INDEX_IGNORE
                        node.save()
                    continue
                datas['path'] = node.path
                datas['title'] = storage.getmeta("movie.title")
                datas['year'] = storage.getmeta("movie.year")
                datas['directors'] = storage.getmeta("movie.directors", [])
                datas['actors'] = storage.getmeta("movie.actors", [])
                datas['synopsis'] = storage.getmeta("movie.synopsis")
                datas['duration'] = storage.getmeta("video.duration")

                datas['pvw'] = None
                if storage.exists(u"pvw"):
                    datas['pvw'] = storage.get_storage_fname(u"pvw")
                datas['cover'] = None
                if storage.exists("cover"):
                    datas['cover'] = storage.get_storage_fname(u"cover")
                logger.debug(u"%s - %s (%s)", node.path, datas['title'], datas['year'])

                with idx_db:
                    self.insert_movie(cursor, **datas)
                    indexed += 1
                try:
                    with lima_db.transaction():
                        node.status_index = LimaNode.STATUS_INDEX_INDEXED
                        node.save()
                except Exception:
                    logger.exception(u"Could not flag as indexed %s", datas['path'])
            except Exception:
                logger.exception(u"Could not index %s", node.path)
            finally:
                if storage is not None:
                    storage.unload()
        else:
            logger.info(u"Movie indexation : done (%s new(s))", indexed)
        return indexed

    def indexables(self):
        with db.transaction():
            return LimaNode.reindexables(u"video", "movie")

    def _find_deleteable_items(self, idx_db):
        """ Insert deleteables movies in a temporary table"""
        with self.attach(idx_db, self.filesdb, self.filesdb_name):
            cursor = idx_db.cursor()
            create_temptable = "CREATE TEMP TABLE movies_to_delete(id int);"
            self.execute(create_temptable, cursor=cursor)
            insert_movies_delete = """
            insert into movies_to_delete
                select m.id
                from movie m
                left join %s.limanode node on node.path = m.path
                where node.node_id is null
            """ % self.filesdb_name  # assume this attribute is safe
            with idx_db:
                self.execute(insert_movies_delete, cursor=cursor)
        logger.debug(u"Will remove %s movies(s) from %s index", cursor.rowcount, self.CONTENT_TYPE)

    def do_cleanup(self, idx_db):
        total_cleaned = 0
        self._find_deleteable_items(idx_db)

        delete_cast_query = """
        delete from casting
        where movie_id in (
            select m.id
            from movies_to_delete m
        )
        """

        delete_person_query = """
        delete from person
        where id in (
            select p.id
            from person p
            left join casting _cast on _cast.person_id = p.id
            where _cast.person_id is NULL
        )
        """

        delete_movie_query = """
        delete from movie
        where movie.id in (
            select m.id
            from movies_to_delete m
        )
        """

        cursor = idx_db.cursor()
        self.execute(delete_cast_query, cursor=cursor)
        total_cleaned += cursor.rowcount
        self.execute(delete_person_query, cursor=cursor)
        total_cleaned += cursor.rowcount
        logger.debug(u"%s person(s) removed from %s index", cursor.rowcount, self.CONTENT_TYPE)
        self.execute(delete_movie_query, cursor=cursor)
        total_cleaned += cursor.rowcount
        logger.info(u"%s movie(s) removed from %s index", cursor.rowcount, self.CONTENT_TYPE)

        return total_cleaned

    def _do_index(self, idx_db, lima_db):
        self.create_or_load(idx_db, fs.get_system_path(self.DB_ZIP()))
        try:
            indexed = self.do_index_movies(idx_db, lima_db)
            cleaned = 0
            with idx_db:
                cleaned = self.do_cleanup(idx_db)
            # return True if items as been added or if anythong as been removed
            return ((indexed or 0) != 0) or ((cleaned or 0) != 0)
        except Exception:
            logger.error(u"Indexation error")
            raise
        else:
            return False
