# coding: utf-8

import sqlite3

from lima import db
from lima.fs import FileSystem as fs
from lima.models import LimaNode
from lima.tools import bin_to_hex
from lima.meta.indexers.base import BaseIndexer

from lima.meta.blackboard.agents.storage import (AgentStorage, )
from lima.meta.storage import (MetaKeyValueStorage, AlbumStorage,
                               ArtistStorage, )
from lima.fs.limafs import (LimaFsError, LimaFsConnectionRefusedError,
                            LimaFsNotReadyError, LimaFsDoesnotExistError)
from lima.meta import cleanup_utils

import logging
logger = logging.getLogger(__name__)


class MusicIndexer(BaseIndexer):

    CONTENT_TYPE = u'music'
    DBNAME = u'musics'

    def __repr__(self):
        return u"lima.meta.files.indexers.music"

    @property
    def create_statements(self):
        create_table_artist = """
        CREATE TABLE IF NOT EXISTS Artist(
            id INTEGER PRIMARY KEY AUTOINCREMENT,
            name TEXT NOT NULL,
            pvw TEXT
        )
        """

        create_indexes_artist_name = "CREATE INDEX IF NOT EXISTS artist_name_ix on Artist(name)"
        create_table_album = """
        CREATE TABLE IF NOT EXISTS Album(
            id INTEGER PRIMARY KEY AUTOINCREMENT,
            name TEXT NOT NULL,
            artist_id INT,

            year INTEGER,
            pvw TEXT,
            FOREIGN KEY (artist_id) REFERENCES Artist(id)
        )
        """

        create_indexes_album_artist = "CREATE INDEX IF NOT EXISTS album_artist_id_ix on Album(artist_id)"
        create_indexes_album_name = "CREATE INDEX IF NOT EXISTS album_name_ix on Album(name)"

        create_table_music = """
        CREATE TABLE IF NOT EXISTS Music(
            id INTEGER PRIMARY KEY AUTOINCREMENT,

            track_artist_id INT,
            album_id INT,

            path TEXT NOT NULL,

            title TEXT NULL,
            number INT,
            duration FLOAT,

            pvw TEXT,
            mimetype TEXT,
            FOREIGN KEY (track_artist_id) REFERENCES Artist(id),
            FOREIGN KEY (album_id) REFERENCES Artist(id)
        )
        """

        create_indexes_music_album = "CREATE INDEX IF NOT EXISTS music_album_id_ix on Music(album_id)"
        create_indexes_music_track_artist = "CREATE INDEX IF NOT EXISTS music_track_artist_id_ix on Music(track_artist_id)"
        create_indexes_music_mimetype = "CREATE INDEX IF NOT EXISTS music_mimetype_ix on Music(mimetype)"

        tables = (create_table_artist, create_table_album,
                  create_table_music, )
        indexes = (create_indexes_artist_name, create_indexes_album_artist,
                   create_indexes_album_name, create_indexes_music_album,
                   create_indexes_music_track_artist,
                   create_indexes_music_mimetype, )
        return tables + indexes

    def insert_artist(self, cursor, artist, pvw_path):
        if artist is not None:
            insert = """
            INSERT OR REPLACE INTO Artist(id, name, pvw) VALUES (
                (SELECT art.id FROM Artist art where name = ? LIMIT 1),
                ?,
                ?
            )
            """

            vals = (artist, artist, pvw_path)
            self.execute(insert, vals, cursor)

    def insert_album(self, cursor, album_artist, album, year, pvw_path):
        # TODO :we should have a N-N relationship between artists and album.
        # this modification have to be synchronized with mobiles clients

        # if album is not None and album_artist is not None:
        if album is not None:
            insert = """
            INSERT OR REPLACE INTO Album(
                id,
                name,
                year,
                artist_id,
                pvw
            ) VALUES (
                (SELECT alb.id FROM Album alb where name = ? LIMIT 1),
                ?,
                COALESCE( (select alb.year from album alb where name = ? limit 1), ?),
                COALESCE( (select alb.artist_id from album alb where name = ? limit 1), (SELECT art.id FROM Artist art where art.name = ? LIMIT 1)),
                COALESCE( (select alb.pvw from album alb where name = ? limit 1), ? )
            )
            """

            vals = (
                    album,
                    album,
                    album, year,
                    album, album_artist,
                    album, pvw_path
                    )
            self.execute(insert, vals, cursor)

    def insert_music(self, cursor, path, pvw, album_artist_pvw, album_artist, track_artist_pvw, track_artist,
                     album, album_year, album_pvw, title, number, duration,
                     mimetype):
        self.insert_artist(cursor, album_artist, album_artist_pvw)
        self.insert_artist(cursor, track_artist, track_artist_pvw)

        self.insert_album(cursor, album_artist, album, album_year, album_pvw)

        if path is not None:
            storage = MetaKeyValueStorage.get_from_pool(path, u'meta')
            pvw_path = None
            if storage.exists("pvw"):
                pvw_path = storage.get_storage_fname(u"pvw")
            insert = """
            INSERT OR REPLACE INTO Music(
                id,
                path,
                album_id,
                track_artist_id,
                title,
                number,
                duration,
                pvw,
                mimetype
            )
            VALUES(
                (SELECT mus.id from Music mus where path = ? LIMIT 1),
                ?,
                (SELECT alb.id FROM Album alb where alb.name = ? LIMIT 1),
                (SELECT art.id FROM Artist art where art.name = ? LIMIT 1),
                ?,
                ?,
                ?,
                ?,
                ?
            )
            """

            vals = (path, path, album, track_artist, title, number, duration, pvw_path, mimetype)
            self.execute(insert, vals, cursor)

    def insert_raw(self, cursor, *args, **kwargs):
        """ insert the line in the table RawMusic. For debug purpose only """

        create = "CREATE TABLE IF NOT EXISTS RawMusic("
        for k in kwargs.keys():
            create += "%s TEXT," % (k)
        create = create[:-1]
        create += ")"
        self.execute(create, None, cursor)

        insert = "insert into RawMusic("
        for k, v in kwargs.items():
            insert += "%s," % (k)
        insert = insert[:-1]
        insert += ")"
        insert += " VALUES ("
        for k, v in kwargs.items():
            insert += "?,"
        insert = insert[:-1]
        insert += ")"
        self.execute(insert, kwargs.values(), cursor)

    def do_index_audio(self, index_db, lima_db):
        cursor = index_db.cursor()
        indexed = 0

        # Here, we take : identified musics (content_type == 'music' in the db)
        # and non identified files (content_type is null)
        # as the non identified files may have been processed and identified by
        # another host.
        # The file will be added if and only if its content type is set to
        # 'music' in metas files
        with lima_db.transaction():
            # musics = LimaNode.reindexables("audio", self.CONTENT_TYPE)[:]
            musics = LimaNode.reindexables("audio")[:]

        for node in musics:
            try:
                storage = None
                album_artist_storage = None
                track_artist_storage = None
                album_storage = None
                try:
                    st = fs.stat(node.path)
                except LimaFsDoesnotExistError:
                    try:
                        # flag file as deleted
                        with lima_db.transaction():
                            node.status_index = LimaNode.STATUS_INDEX_IGNORE
                            node.status = LimaNode.STATUS_DELETED
                            node.save()
                    except Exception as e:
                        logger.exception(u"Could not flag not existing file as not be reindexed %s : %s", node.path, getattr(e, 'message', 'Unknown error'))
                    #
                    continue
                except (LimaFsNotReadyError, LimaFsConnectionRefusedError):
                    logger.error("Filesystem not ready. Can't index")
                    break
                except LimaFsError:
                    logger.error("Filesystem error. Can't index")
                    break

                node_id = bin_to_hex(st.v['id'])

                # GC may not have run
                if node.node_id != node_id:
                    continue
                datas = {}
                storage = AgentStorage(node.path)
                datas['path'] = node.path
                if storage.getmeta('content_type') != self.CONTENT_TYPE:
                    logger.debug("%s : not identified as a %s", node.path, self.CONTENT_TYPE)
                    with lima_db.transaction():
                        node.status_index = LimaNode.STATUS_INDEX_IGNORE
                        node.save()
                    continue
                datas['album_artist'] = storage.getmeta("audio.album.artist")
                datas['track_artist'] = storage.getmeta("audio.track.artist")
                datas['album'] = storage.getmeta("audio.album.title")
                datas['title'] = storage.getmeta("audio.track.title")
                datas['number'] = storage.getmeta("audio.track.number")
                datas['duration'] = storage.getmeta("audio.track.duration")
                datas['album_year'] = storage.getmeta("audio.album.year")
                datas['mimetype'] = storage.getmeta("mimetype")
                logger.info(u"%(path)s - %(album_artist)s %(track_artist)s on %(album)s : %(title)s" % datas)

                datas['pvw'] = None
                if storage.exists("pvw"):
                    datas['pvw'] = storage.get_storage_fname(u"pvw")

                datas['album_pvw'] = None
                if datas['album_artist'] is not None and datas['album'] is not None:
                    album_storage = AlbumStorage.get_from_pool(datas['album_artist'], datas['album'])
                    if album_storage.exists("pvw"):
                        datas['album_pvw'] = album_storage.get_storage_fname(u"pvw")

                datas['album_artist_pvw'] = None
                if datas['album_artist'] is not None:
                    album_artist_storage = ArtistStorage.get_from_pool(datas['album_artist'])
                    if album_artist_storage.exists("pvw"):
                        datas['album_artist_pvw'] = album_artist_storage.get_storage_fname(u"pvw")

                datas['track_artist_pvw'] = None
                if datas['track_artist'] is not None:
                    track_artist_storage = ArtistStorage.get_from_pool(datas['track_artist'])
                    if track_artist_storage.exists("pvw"):
                        datas['track_artist_pvw'] = track_artist_storage.get_storage_fname(u"pvw")

                self.insert_music(cursor, **datas)
                # self.insert_raw(cursor, **datas)
                indexed += 1
                try:
                    with lima_db.transaction():
                        node.status_index = LimaNode.STATUS_INDEX_INDEXED
                        node.save()
                except Exception as e:
                    logger.exception(u"Could not flag as indexed %s : %s",
                                     datas['path'], getattr(e, 'message', 'Unknown error'))
            except Exception as e:
                logger.exception(u"Could not index %s", node.path)
                # TODO : segment
            finally:
                if storage is not None:
                    storage.unload(persist=False)
                if album_artist_storage is not None:
                    album_artist_storage.unload(persist=False)
                if track_artist_storage is not None:
                    track_artist_storage.unload(persist=False)
                if album_storage is not None:
                    album_storage.unload(persist=False)
        else:
            logger.info(u"Audio indexation : done (%s new(s))", indexed)
        return indexed

    def _fix_missing_albums_pvw(self, idx_db):
        """Fixes albums without preview by taking the last non null of its songs"""
        with idx_db:
            cursor = idx_db.cursor()
            query = """create temp table fix_album_pvw(artist_id, pvw)"""
            self.execute(query, cursor=cursor)
            query = """insert into fix_album_pvw
                select alb.id, album_music.pvw as pvw
                from album alb
                left join artist on alb.artist_id = artist.id
                left join music album_music on album_music.album_id = alb.id and album_music.pvw is not NULL
                where alb.pvw is null
                and album_music.pvw is not NULL
            """
            self.execute(query, cursor=cursor)
            query = """
                    update album
                    set pvw = (select pvw from fix_album_pvw where fix_album_pvw.artist_id = album.id and pvw is not NULL limit 1)
                    where album.id = (select id from fix_album_pvw where fix_album_pvw.artist_id = album.id and pvw is not NULL limit 1)
            """
            self.execute(query, cursor=cursor)
            if cursor.rowcount:
                logger.info(u"Fixed %s albums with no preview", cursor.rowcount)

    def _fix_missing_artists_pvw(self, idx_db):
        """Fixes artist without preview by taking the last non null of its album or songs"""
        with idx_db:
            cursor = idx_db.cursor()
            query = """ SELECT artist.id, artist.name from artist where pvw is null"""
            artists = list(self.execute(query, cursor=cursor))
        for artist_id, artist_name in artists:
            artist_storage = ArtistStorage(artist_name)
            try:
                if artist_storage.exists(u"pvw"):
                    pvw = artist_storage.get_storage_fname(u"pvw")
                    with idx_db:
                        cursor = idx_db.cursor()
                        query = """ update artist set pvw = ? where id =? """
                        values = (pvw, artist_id)
                        self.execute(query, values, cursor=cursor)
                        logger.info(u"Fixed preview for artist %s", artist_name)
            except Exception:
                self.logger.error(u"Could not fix %s preview", artist_name)
            finally:
                artist_storage.unload()
        '''
        with idx_db:
            cursor = idx_db.cursor()
            query = """create temp table fix_artist_pvw(artist_id, album_pvw, music_pvw)"""
            self.execute(query, cursor=cursor)
            query = """insert into fix_artist_pvw
                select artist.id, alb.pvw, coalesce(album_music.pvw, artist_music.pvw) as music_pvw
                from artist
                left join album alb on alb.artist_id = artist.id and alb.pvw is not NULL
                left join music artist_music on artist_music.track_artist_id = artist.id and artist_music.pvw is not NULL
                left join music album_music on album_music.album_id = alb.id and album_music.pvw is not NULL
                where artist.pvw is null
                and lower(artist.name) != 'various artist'
                group by artist.id
            """
            self.execute(query, cursor=cursor)
            query = """
                    update artist
                    set pvw = (
                                select coalesce(album_pvw, music_pvw)
                                from fix_artist_pvw
                                where fix_artist_pvw.artist_id = artist.id
                                    and coalesce(album_pvw, music_pvw) is not NULL
                                    limit 1
                                    )
                    where artist.id = (
                        select id
                        from fix_artist_pvw
                        where fix_artist_pvw.artist_id = artist.id
                        and coalesce(album_pvw, music_pvw) is not NULL
                        limit 1
                        )
            """
            self.execute(query, cursor=cursor)
            if cursor.rowcount:
                logger.info(u"Fixed %s artists with no preview", cursor.rowcount)
            '''

    def _fix_missing_pvw(self, idx_db):
        try:
            self._fix_missing_albums_pvw(idx_db)
        except Exception:
            logger.exception(u"Could not fix missing albums pvw")
        try:
            self._fix_missing_artists_pvw(idx_db)
        except Exception:
            logger.exception(u"Could not fix missing artist pvw")

    def _find_deleteable_items(self, idx_db):
        with self.attach(idx_db, self.filesdb, self.filesdb_name):
            cursor = idx_db.cursor()
            with idx_db:
                create_temptable = "CREATE TEMP TABLE music_to_delete(id int);"
                self.execute(create_temptable, cursor=cursor)
                insert_music_delete = """insert into music_to_delete
                select m.id
                from music m
                left join %s.limanode node on node.path = m.path
                where node.node_id is null
                """ % self.filesdb_name  # assume this attribute is safe
                self.execute(insert_music_delete, cursor=cursor)
        logger.debug(u"Will remove %s music(s) from %s index", cursor.rowcount, self.CONTENT_TYPE)

    def do_cleanup(self, idx_db):
        total_cleaned = 0
        # collect item to be deleted in a temporary table
        self._find_deleteable_items(idx_db)

        cursor = idx_db.cursor()
        delete_query = """
        delete from music
        where music.id in (
            select m.id
            from music_to_delete m
        )
        """
        self.execute(delete_query, cursor=cursor)
        logger.info(u"%s music(s) removed from %s index", cursor.rowcount, self.CONTENT_TYPE)
        total_cleaned += cursor.rowcount

        create_temptable = "CREATE TEMP TABLE album_to_delete(id int, name TEXT, artist_name TEXT);"
        self.execute(create_temptable, cursor=cursor)

        insert_delete_query = """
        insert into album_to_delete (id, name, artist_name)

        select alb.id, alb.name, art.name
        from album alb
        inner join artist art on art.id = alb.artist_id
        left join music mus on mus.album_id = alb.id
        where mus.id is null
        """

        self.execute(insert_delete_query, cursor=cursor)
        logger.debug(u"%s album(s) will be removed from %s index", cursor.rowcount, self.CONTENT_TYPE)

        delete_query = """
        delete from album
        where album.id in (
            select alb.id
            from album_to_delete alb
        )
        """

        self.execute(delete_query, cursor=cursor)
        logger.info(u"%s album(s) removed from %s index", cursor.rowcount, self.CONTENT_TYPE)
        total_cleaned += cursor.rowcount
        # clean up album metas datas
        select_album_to_clean = "select artist_name, name from album_to_delete"
        for line in cursor.execute(select_album_to_clean):
            artist, album = line[0], line[1]
            try:
                cleanup_utils.clean_up_album(artist, album)
            except Exception:
                logger.error("Could not remove metas for %s - %s", artist, album)

        create_temptable = "CREATE TEMP TABLE artist_to_delete(id int, name TEXT);"
        self.execute(create_temptable, cursor=cursor)

        insert_delete_artist = """insert into artist_to_delete
            select art.id, art.name
            from artist art
            left join album alb on alb.artist_id = art.id
            left join music mus on mus.track_artist_id = art.id
            group by art.id having count(alb.id) = 0 and count(mus.id)=0
        """

        self.execute(insert_delete_artist, cursor=cursor)
        # clean up album metas datas
        select_artist_to_clean = "select name from artist_to_delete"

        for line in cursor.execute(select_artist_to_clean):
            artist = line[0]
            try:
                cleanup_utils.clean_up_artist(artist)
            except Exception:
                logger.error("Could not remove metas for %s", artist)

        delete_query = """
        delete from artist
        where artist.id in (
            select art.id
            from artist_to_delete art
        )
        """

        self.execute(delete_query, cursor=cursor)
        total_cleaned += cursor.rowcount
        logger.info(u"%s artist(s) removed from %s index", cursor.rowcount, self.CONTENT_TYPE)

        return total_cleaned

    def indexables(self):
        with db.transaction():
            return LimaNode.reindexables(u"audio")



    def _mark_missing_from_existing(self, lima_db):
        if fs.exists(self.DB_ZIP()):
            db_index = self._load_db_gzip()
            if not db_index:
                return # Mark all ?
            try:
                with self.attach(db_index, self.filesdb, self.filesdb_name):
                        mark_missings = """
                            update %(files_db)s.limanode
                            set status_index = %(new_status_index)s
                            where %(files_db)s.limanode.node_id in (
                                    select node.node_id
                                    from %(files_db)s.limanode as node
                                    left join music m on node.path = m.path
                                    where node.content_type = '%(content_type)s'
                                        and ( node.status_index != %(old_status_index)s or node.status_index is null)
                                        and node.status = %(status)s
                                        and m.id is null
                            )
                        """ % {
                                'files_db': self.filesdb_name,
                                'content_type': self.CONTENT_TYPE,
                                'status': LimaNode.STATUS_PROCESSED,
                                'new_status_index': LimaNode.STATUS_INDEX_NEED_REINDEX,
                                'old_status_index': LimaNode.STATUS_INDEX_IGNORE,
                                }
                        with lima_db.transaction():
                            cursor = db_index.cursor()
                            self.execute(mark_missings, cursor=cursor)
                        if cursor.rowcount:
                            logger.warn(u"Marked %s %s missing from index as to be reindexed", cursor.rowcount, self.CONTENT_TYPE)
            finally:
                db_index.close()

    def _do_index(self, idx_db, lima_db):
        self.create_or_load(idx_db, fs.get_system_path(self.DB_ZIP()))
        try:
            with idx_db:
                indexed = self.do_index_audio(idx_db, lima_db)
            cleaned = 0
            with idx_db:
                cleaned = self.do_cleanup(idx_db)
            self._fix_missing_pvw(idx_db)
            # return True if items as been added or if anythong as been removed
            return ((indexed or 0) != 0) or ((cleaned or 0) != 0)
        except Exception:
            logger.error(u"Indexation error")
            raise
        finally:
            AlbumStorage.unload_all(persist=False)
            ArtistStorage.unload_all(persist=False)
