# coding: utf-8

import peewee
from lima.fs import FileSystem as fs
from lima.fs.fcfs import (
    LimaFsError,
    LimaFsDoesnotExistError,
    LimaFsNotMountedError,
    LimaFsNotReadyError,
    LimaFsConnectionRefusedError,
)
from lima.settings import CONFIG
from lima.tools import segment
from lima.meta.storage import MetaKeyValueStorage, ArtistStorage, SerieStorage

from lima import db
from lima.models import LimaNode


import sqlite3
from lima.meta.indexers.musics import MusicIndexer
from lima.meta.indexers.movies import MovieIndexer
from lima.meta.indexers.episodes import TvIndexer
from lima.meta.scheduler.indexer import FileIndexer

import logging
logger = logging.getLogger(__name__)

IGNORE_PATTERNS = u"((.*(\.fseventsd|\.DS_Store|Thumbs\.db|\.Trashes).*)|(^\..+)$)|(^.*/\..*$)"

META_VERSION = 1


class MigrationMeta2(object):
    order = 0

    def __init__(self):
        self.old_root = u"/.lima/meta"
        self.new_root = CONFIG[u"meta_root"]
        self.music_index = None
        self.movies_index = None
        self.episodes_index = None
        #
        self.indexes_modified = {}

    def _make_pass(self, old_root, new_root):
        if not fs.exists(old_root):
            return
        paths = [(path, st) for path, st in fs.walk_stat(old_root, recursive=True)]
        for path, st in paths:
            if not fs.exists(path):
                # file may already bee moved
                continue
            if st.isdir():
                # remove empty directories
                if fs.is_empty_dir(path):
                    fs.rmdir(path)
                # do not migrate directories to avoid empty ones
                continue

            # path is a link, we should resolve it, copy the real file behind it
            # and recreate the symlink
            if st.islink():
                link_real_path = fs.resolve_symlink(path, fallback_last_resolved=True)
                logger.debug("%s resolved as %s", path, link_real_path)

                # first try at current root level
                # convert old path to new path
                new_path = path.replace(old_root, new_root)
                new_link_real_path = link_real_path.replace(old_root, new_root)
                if new_link_real_path == link_real_path:
                    new_link_real_path = link_real_path.replace(self.old_root, self.new_root)

                # the file does not exists, copy
                if not fs.exists(new_link_real_path) and fs.exists(link_real_path):
                    fs.ensure_directory(fs.dirname(new_link_real_path))
                    fs.ensure_directory(fs.dirname(new_path))
                    fs.rename(link_real_path, new_link_real_path)
                    logger.debug("%s had been migrated to %s", link_real_path, new_link_real_path)
                else:
                    # the file already exists
                    try:
                        fs.unlink(link_real_path)
                    except LimaFsDoesnotExistError:
                        pass  # don't care
                if not fs.exists(new_path):
                    fs.ensure_directory(fs.dirname(new_path))
                    try:
                        fs.symlink(new_link_real_path, new_path)
                        fs.unlink(path)
                        pass
                    except LimaFsDoesnotExistError:
                        pass  # don't care
                    logger.debug("%s had been symlinked to %s", new_link_real_path, new_path)

            if st.isreg():
                new_path = path.replace(old_root, new_root)
                if not fs.exists(new_path):
                    fs.ensure_directory(fs.dirname(new_path))
                    fs.rename(path, new_path)
                    logger.debug("%s had been migrated to %s", path, new_path)
                else:
                    try:
                        fs.unlink(path)  # all job's done
                    except LimaFsDoesnotExistError:
                        pass  # don't care

    def migrate_music(self):
        old_root = fs.join(self.old_root, u'musics')
        new_root = ArtistStorage.ROOT()

        # make two pass
        self._make_pass(old_root, new_root)
        self._make_pass(old_root, new_root)

        self.fixes_pvw_paths(old_root, new_root)
        self.fixes_pvw_root_music(old_root, new_root)

    def migrate_series(self):
        old_root = fs.join(self.old_root, u'series')
        new_root = SerieStorage.ROOT()

        # make two pass
        self._make_pass(old_root, new_root)
        self._make_pass(old_root, new_root)

        self.fixes_pvw_paths(old_root, new_root)
        self.fixes_pvw_root_episodes(old_root, new_root)

    def _make_pass_meta_files(self, old_root, new_root):
        if not fs.exists(old_root):
            return
        paths = [(path, st) for path, st in fs.walk_stat(old_root, recursive=True)]
        for path, st in paths:
            if st.isdir():
                if fs.is_empty_dir(path):
                    logger.info("Removing empty directory %s", path)
                    fs.rmdir(path)
                    continue
                node_cvid = fs.basename(path)
                if len(node_cvid) == 2:
                    # ignore first level folders
                    continue
                if len(node_cvid) != 40:
                    logger.debug("not a cvid : '%s'", node_cvid)
                    continue
                node = None
                with db.transaction():
                    try:
                        node = LimaNode.get(LimaNode.node_cvid == node_cvid)
                    except peewee.DoesNotExist:
                        logger.info("Ignoring not matched id %s", node_cvid)
                        continue
                if node and node.default_forkid:
                    node_old_root = fs.join(old_root, node_cvid[0:2], node_cvid)
                    node_new_root = fs.join(new_root, node.default_forkid[0:2], node.default_forkid)
                    logger.info(u"Migrate metas for file %s", node.path)
                    logger.debug(u"Will migrate %s to %s", node_old_root, node_new_root)

                    self._make_pass(node_old_root, node_new_root)
                    self._make_pass(node_old_root, node_new_root)
                    with db.transaction():
                        # force rescan
                        node.status = LimaNode.STATUS_CREATED
                        node.meta_version = CONFIG['min_meta_version']
                        node.save()

    def migrate_metas_files(self):
        old_root = u"{meta_root}/files".format(meta_root=self.old_root)
        new_root = MetaKeyValueStorage.ROOT()

        self._make_pass_meta_files(old_root, new_root)
        self._make_pass_meta_files(old_root, new_root)

        self.fixes_pvw_paths(old_root, new_root)
        self.fixes_pvw_root_music(old_root, new_root)
        self.fixes_pvw_root_movies(old_root, new_root)
        self.fixes_pvw_root_episodes(old_root, new_root)

    def load_indexes(self):
        try:
            old_music_index = fs.join(self.old_root, 'music.db.gz')
            self.music_index = self._load_index(old_music_index, MusicIndexer(None))
        except Exception:
            logger.error("Could not load old music indexer")

        try:
            old_movies_index = fs.join(self.old_root, 'movies.db.gz')
            self.movies_index = self._load_index(old_movies_index, MovieIndexer(None))
        except Exception:
            logger.error("Could not load old movies indexer")

        try:
            old_episodes_index = fs.join(self.old_root, 'series.db.gz')
            self.episodes_index = self._load_index(old_episodes_index, TvIndexer(None))
        except Exception:
            logger.error("Could not load old episodes indexer")

    def dump_indexes(self):
        file_indexer = FileIndexer(indexer_delay=0, indexer_files_threshold=0, filesdb=None, is_debug=False)
        try:
            if self.music_index and self.indexes_modified.get(self.music_index, False):
                indexer = MusicIndexer(self.new_root)
                file_indexer.persist_index(indexer, self.music_index)
                file_indexer.remove_conflicts(indexer)
                self.indexes_modified[self.music_index] = False
        except Exception:
            logger.exception("Could not persist music index")
            # segment.track_error(event=u"Mucic index persists error")
        try:
            if self.movies_index and self.indexes_modified.get(self.movies_index, False):
                indexer = MovieIndexer(self.new_root)
                file_indexer.persist_index(indexer, self.movies_index)
                file_indexer.remove_conflicts(indexer)
                self.indexes_modified[self.movies_index] = False
        except Exception:
            logger.exception("Could not persist movies index")
            # segment.track_error(event=u"Movie index persists error")
        try:
            if self.episodes_index and self.indexes_modified.get(self.episodes_index, False):
                indexer = TvIndexer(self.new_root)
                file_indexer.persist_index(indexer, self.episodes_index)
                file_indexer.remove_conflicts(indexer)
                self.indexes_modified[self.episodes_index] = False
        except Exception:
            logger.exception("Could not persist episode index")
            # segment.track_error(event=u"Episode index persists error")

    def _load_index(self, old_index_gz, indexer):
        if not fs.exists(old_index_gz):
            return

        new_index_gz = indexer.DB_ZIP()
        self._copy_old_index(old_index_gz, new_index_gz)

        if fs.exists(new_index_gz):
            new_index = sqlite3.connect(":memory:")
            indexer._load_from_gzip(new_index, fs.get_system_path(new_index_gz))
            return new_index

    def _copy_old_index(self, old_index_gz, new_index_gz):
        """ Copy old index to new location"""
        if fs.exists(new_index_gz):
            logger.debug("New index already exists")
            return
        if fs.exists(old_index_gz):
            fs.ensure_directory(fs.dirname(new_index_gz))
            fs.rename(old_index_gz, new_index_gz)
            logger.info("Moved %s to %s", old_index_gz, new_index_gz)

    def fixes_pvw_paths(self, old_root, new_root):
        if self.music_index:
            try:
                self._fixes_cvid_fid_paths(old_root, new_root, self.music_index, "music", "pvw")
            except Exception:
                logger.exception("Could not update music pvw cvid>fid")
        if self.movies_index:
            try:
                self._fixes_cvid_fid_paths(old_root, new_root, self.movies_index, "movie", "pvw")
            except Exception:
                logger.exception("Could not update movie pvw cvid>fid")
            try:
                self._fixes_cvid_fid_paths(old_root, new_root, self.movies_index, "movie", "cover")
            except Exception:
                logger.exception("Could not update movie cover cvid>fid")
        if self.episodes_index:
            try:
                self._fixes_cvid_fid_paths(old_root, new_root, self.episodes_index, "episode", "pvw")
            except Exception:
                logger.exception("Could not update episode pvw cvid>fid")

    def fixes_pvw_root_music(self, old_root, new_root):
        if self.music_index:
            try:
                self._fixes_roots_paths(old_root, new_root, self.music_index, "music", "pvw")
            except Exception:
                logger.exception("Could not update music pvw")

            try:
                self._fixes_roots_paths(old_root, new_root, self.music_index, "album", "pvw")
            except Exception:
                logger.exception("Could not update album pvw")

            try:
                self._fixes_roots_paths(old_root, new_root, self.music_index, "artist", "pvw")
            except Exception:
                logger.exception("Could not update artist pvw")

    def fixes_pvw_root_movies(self, old_root, new_root):
        if self.movies_index:
            try:
                self._fixes_roots_paths(old_root, new_root, self.movies_index, "movie", "pvw")
            except Exception:
                logger.exception("Could not update movie pvw")
            try:
                self._fixes_roots_paths(old_root, new_root, self.movies_index, "movie", "cover")
            except Exception:
                logger.exception("Could not update movie cover")

    def fixes_pvw_root_episodes(self, old_root, new_root):
        if self.episodes_index:
            try:
                self._fixes_roots_paths(old_root, new_root, self.episodes_index, "episode", "pvw")
            except Exception:
                logger.exception("Could not update episode pvw")
            try:
                self._fixes_roots_paths(old_root, new_root, self.episodes_index, "episode", "cover")
            except Exception:
                logger.exception("Could not update episode cover")
            try:
                self._fixes_roots_paths(old_root, new_root, self.episodes_index, "season", "cover")
            except Exception:
                logger.exception("Could not update season cover")
            try:
                self._fixes_roots_paths(old_root, new_root, self.episodes_index, "serie", "cover")
            except Exception:
                logger.exception("Could not update serie cover")

    def _fixes_cvid_fid_paths(self, old_root, new_root, index_db, table_name, column_name):
        cursor = index_db.cursor()
        query = "select path, {column_name} from {table_name} where {column_name} is not null and {column_name} like ?".format(column_name=column_name, table_name=table_name)
        paths = [(path, pvw) for path, pvw in cursor.execute(query, (old_root+"%",))]

        for path, pvw in paths:
            try:
                with db.transaction():
                    node = LimaNode.get(LimaNode.path == path)
            except peewee.DoesNotExist:
                continue
            if not all([node.node_cvid, node.default_forkid, pvw]):
                continue
            node_old_root = fs.join(old_root, node.node_cvid[0:2], node.node_cvid)
            node_new_root = fs.join(new_root, node.default_forkid[0:2], node.default_forkid)

            new_pvw = pvw.replace(node_old_root, node_new_root)
            if new_pvw != pvw:
                cursor.execute("update {table_name} set {column_name}=? where path=?".format(column_name=column_name, table_name=table_name), (new_pvw, path))
                logger.info("%s : updated %s to %s", table_name, pvw, new_pvw)
                self.indexes_modified[index_db] = True

    def _fixes_roots_paths(self, old_root, new_root, index_db, table_name, column_name):
        cursor = index_db.cursor()
        query = "select id, {column_name} from {table_name} where {column_name} is not null and {column_name} like ?".format(column_name=column_name, table_name=table_name)
        paths = [(path, pvw) for path, pvw in cursor.execute(query, (old_root+"%",))]
        for _id, pvw in paths:
            if not pvw:
                continue
            new_pvw = pvw.replace(old_root, new_root)
            if new_pvw != pvw:
                cursor.execute("update {table_name} set {column_name}=? where id=?".format(column_name=column_name, table_name=table_name), (new_pvw, _id))
                logger.info("%s : updated %s to %s", table_name, pvw, new_pvw)
                self.indexes_modified[index_db] = True

    def migrate(self):
        try:
            if not fs.exists(self.old_root):
                return
            logger.info(u"migrating from %s to %s", self.old_root, self.new_root)
        except (LimaFsError, LimaFsDoesnotExistError, LimaFsNotMountedError, LimaFsNotReadyError, LimaFsConnectionRefusedError):
            logger.error(u"Core not ready")
            return
        except Exception:
            logger.exception(u"Could not check old meta directory existence.")
            # segment.track_error(event="Migration error")
            return


        try:
            self.load_indexes()
        except Exception:
            logger.exception(u"Could not load indexes")
            # segment.track_error(event=u"Index load error")

        try:
            fs.ensure_directory(self.new_root)
        except (LimaFsError, LimaFsDoesnotExistError, LimaFsNotMountedError, LimaFsNotReadyError, LimaFsConnectionRefusedError):
            logger.error(u"Core not ready")
            return
        except Exception:
            logger.exception(u"Could not create new meta directory. Aborting")
            # segment.track_error(event="Migration error")
            return

        try:
            self.migrate_music()
        except (LimaFsError, LimaFsDoesnotExistError, LimaFsNotMountedError, LimaFsNotReadyError, LimaFsConnectionRefusedError):
            logger.error(u"Core not ready")
            return
        except Exception:
            logger.exception(u"Could not migrate musics metas files")
            # segment.track_error(event=u"Music migration error")
        try:
            self.migrate_series()
        except (LimaFsError, LimaFsDoesnotExistError, LimaFsNotMountedError, LimaFsNotReadyError, LimaFsConnectionRefusedError):
            logger.error(u"Core not ready")
            return
        except Exception:
            logger.exception(u"Could not migrate series metas files")
            # segment.track_error(event=u"Series migration error")

        try:
            self.migrate_metas_files()
        except (LimaFsError, LimaFsDoesnotExistError, LimaFsNotMountedError, LimaFsNotReadyError, LimaFsConnectionRefusedError):
            logger.error(u"Core not ready")
            return
        except Exception:
            logger.exception(u"Could not migrate metas files")
            # segment.track_error(event=u"Meta migration error")

        try:
            self.dump_indexes()
        except (LimaFsError, LimaFsDoesnotExistError, LimaFsNotMountedError, LimaFsNotReadyError, LimaFsConnectionRefusedError):
            logger.error(u"Core not ready")
            return
        except Exception:
            logger.exception(u"Could not dump indeces")
            # segment.track_error(event=u"Index migration error")
