# coding: utf-8

from lima.tools import bin_to_hex
from lima.mimes import mimetypes
from lima.mimes import is_supported_mimetype

from lima.fs import FileSystem as fs

from lima.fs.limafs import (LimaFsTimeoutError, LimaFsConnectionRefusedError,
                            LimaFsNotReadyError, LimaFsDoesnotExistError,
                            LimaFsNoSpaceError, LimaFsMemoryError, LimaFsNotMountedError,
                            LimaFsNotDirectoryError, LimaFsDeviceNotconfigured
                            )

from lima import db
from lima.db import DbNoSpaceError
from lima.models import LimaNode
import peewee

from lima.tools.threads import threading
from lima.meta.storage import MetaKeyValueStorage
from lima.meta import cleanup_utils

from lima.meta.scheduler import indexer
from lima.meta import locks

import logging
logger = logging.getLogger(__name__)

DEFAULT_WAIT = 10
MAX_WAIT = 60
WAIT_TIMEOUT = 10
WAIT_CONN_REFUSED = 30
WAIT_NOT_READY = 30


# Files to ignore
# Ignore pattern :
# kowns files
# or filename starting with a dot (/.secret)
# or filename in a path where a folder starts with a dot (i.e #
# /home/user/.hidden/folder) (unix only)
# and other boring stuffs
IGNORE_PATTERNS = u"((.*(\.fseventsd|\.DS_Store|Thumbs\.db|\.Trashes|__MACOSX|RECYCLE\.BIN).*)|(^\..+)$)|(^.*/\..*$)"


class FsWatcher(threading.Thread):
    """Class populating the local database whene the filesystem change.
    It does not save all fs changes but only the one we are interested in.
    """

    WAIT_CHANGED = 10

    # DB purge delay in seconds. 3 hours

    def __init__(self, config, force_fisrt_import=False):
        super(FsWatcher, self).__init__()
        self.logger = logging.getLogger("lima.meta.fs.watcher")
        self.logger.name = u"lima.meta.fs.watcher"

        self.config = config
        self.root = u"/"

        self.force_fisrt_import = force_fisrt_import

        self.wait_delay = DEFAULT_WAIT

    def __repr__(self):
        return "fs watcher"

    @property
    def allow_dl(self):
        return self.config.get('download.enabled', False)

    def execute(self):
        self.logger.info(u"Starting fs watcher")
        while not self._kill_event.isSet():
            if not self._resume_event.wait(0.1):
                continue
            if not self._resume_event.isSet():
                continue
            try:
                with locks.FSWATCHER_LOCK:
                    self.scan(self.root)
                self.wait(self.wait_delay)
            except threading.ThreadKilledError:
                self.logger.info("[walk] fs watcher killed")
                break
            except threading.ThreadStoppedError:
                self.logger.info("[walk] fs watcher stopped")
            except (LimaFsTimeoutError, ):
                self.logger.error(u"[walk] FS timeout. Will wait for %s sec", WAIT_TIMEOUT)
                self.wait(WAIT_TIMEOUT)
                self.logger.error(u"[walk] Wait ends")
            except (LimaFsConnectionRefusedError, ):
                self.logger.error(u"[walk] FS connection refused. Will wait for %s sec",
                    WAIT_CONN_REFUSED)
                self.wait(WAIT_CONN_REFUSED)
                self.logger.error(u"[walk] Wait ends")
            except (LimaFsNotReadyError, ):
                self.logger.error(u"[walk] FS not ready yet. Will wait for %s sec",
                    WAIT_NOT_READY)
                self.wait(WAIT_NOT_READY)
                self.logger.error(u"[walk] Wait ends")
            except (LimaFsDoesnotExistError, ):
                # if everything is well catched,  should not happen
                self.logger.error(u"[walk] File has been removed")
            except (LimaFsNotMountedError):
                self.logger.error(u"[walk] Disk unmounted")
                self.kill(fatal=True)
            except (LimaFsDeviceNotconfigured):
                self.logger.error(u"[walk] Device not configured")
                self.kill(fatal=True)
            except (LimaFsNoSpaceError, DbNoSpaceError):
                self.logger.error(u"[walk] No left space")
                self.kill(fatal=True)
            except (LimaFsMemoryError, ):
                self.logger.error(u"[walk] No memory")
                self.kill(fatal=True)

        self.logger.info(u"Bye.")

    def path_filter(self, p, st):
        if st.isreg():
            mime, _ = mimetypes.guess_type(p)
            return is_supported_mimetype(mime)
        else:
            # we need directories to detect removed files
            return st.isdir()
        return False

    def scan(self, root, last_mrtime=None, recursive=False):
        logger.debug(u"Browsing %s", root)
        self.check_continue()

        forced_import = self.force_fisrt_import
        if forced_import:
            logger.info(u"Forcing first reimport")
            last_mrtime = -1
            recursive = True
            self.force_fisrt_import = False

        root_st = fs.stat(root)

        if last_mrtime is None:
            with db.transaction():
                last_mrtime = LimaNode.max_mrtime()

        paths = fs.walk_stat_filter(root, path_filter=self.path_filter, ignore_pattern=IGNORE_PATTERNS, since_mrtime=last_mrtime, recursive=recursive)
        paths = fs.isort_walk(paths)

        have_seen_change = False

        # check if we should continue before iteration
        for p, st in paths:
            self.check_continue()
            if p == root:
                continue
            try:
                self._process_path(p, st)
            except (LimaFsDoesnotExistError, LimaFsNotDirectoryError):
                logger.warn("%s removed when scanning", p)
                continue
            have_seen_change = True
            self.wait(0.05)
        try:
            self._process_path(root, root_st, allow_recursion=False)
        except (LimaFsDoesnotExistError, LimaFsNotDirectoryError):
            logger.warn("%s removed when scanning", root)

        if forced_import:
            logger.info(u"Done")

        # if no change seen, wait more
        if not have_seen_change:
            new_delay = min(self.wait_delay + 5, MAX_WAIT)
            if self.wait_delay != new_delay:
                logger.debug(u"No change, increasing wait delay to %s sec", new_delay)
                self.wait_delay = new_delay
        else:
            new_delay = DEFAULT_WAIT
            if self.wait_delay != new_delay:
                logger.debug(u"Got change - reseting to %s sec", new_delay)
                self.wait_delay = new_delay

    def _process_path(self, p, st, allow_recursion=True):
        node_id = bin_to_hex(st.v['id'])
        node_cvid = getattr(st, 'st_cvid', None)
        node_cvid = bin_to_hex(node_cvid) if node_cvid else None

        default_forkid = getattr(st, 'st_default_forkid', None)
        default_forkid = bin_to_hex(default_forkid) if default_forkid else None

        mimetype, _ = mimetypes.guess_type(p)

        node = None
        created = False
        with db.transaction():
            try:
                node = LimaNode.get(node_id=node_id)
            except peewee.DoesNotExist:
                LimaNode.delete().where(
                    (LimaNode.path == p) &
                    (LimaNode.node_id != node_id)
                ).execute()
                node = LimaNode.create(
                    node_id=node_id,
                    path=p,
                    status=LimaNode.STATUS_CREATED,
                )
                created = True
        # }}
        # Do not continue if the file have not changed
        if not created and (node.mrtime == st.st_mrtime and node.path == p):
            return
        # new file
        if created and st.isreg():
            self._process_file_created(p, st, node)
        # updated file
        elif st.isreg():
            self._process_file_modified(p, st, node)
        # new directory
        elif created and st.isdir():
            self._process_dir_created(p, st, node, allow_recursion)
        # updated directory
        elif st.isdir():
            self._process_dir_modified(p, st, node, allow_recursion)

    def _update_node_infos(self, node, st, p, is_new=False):

        node_cvid = getattr(st, 'st_cvid', None)
        node_cvid = bin_to_hex(node_cvid) if node_cvid else None

        node.is_path_modified = False
        # path changed
        if not is_new and (node.path and node.path != p):
            logger.info(u"%s -> %s", node.path, p)
            node.status = LimaNode.STATUS_NEED_UPDATE
            # flag the path modification
            node.is_path_modified = True

            if fs.basename(node.path) != fs.basename(p):
                # flag the filename modification
                node.is_basename_modified = True
            else:
                node.is_basename_modified = False
        node.path = p

        # file was opreviously marked as deleted but was still present in db
        if node.status == LimaNode.STATUS_DELETED:
            node.status = LimaNode.STATUS_NEED_UPDATE

        # content changed
        default_forkid = getattr(st, 'st_default_forkid', None)
        default_forkid = bin_to_hex(default_forkid) if default_forkid else None

        node.is_content_modified = False
        if not is_new and (node.default_forkid != default_forkid):
            logger.info(u"Content modified %s", p)
            # flag the content modification
            node.is_content_modified = True
            node.status = LimaNode.STATUS_NEED_UPDATE
            try:
                # Try to switch old metas for this file
                with MetaKeyValueStorage.get_from_pool(node.path, 'meta') as storage:
                    storage.switch_from(node.default_forkid, default_forkid)
                logger.info("Moved old metas")
            except Exception:
                logger.exception("Could not move old metas")
        node.default_forkid = default_forkid

        # Detect mimetype
        mimetype, _ = mimetypes.guess_type(node.path)
        if not is_supported_mimetype(mimetype):
            node.status = LimaNode.STATUS_MIME_NOT_SUPPORTED

        # get content type if possible
        if st.isreg() and is_supported_mimetype(mimetype) and not node.content_type:
            try:
                with MetaKeyValueStorage.get_from_pool(node.path, 'meta') as storage:
                    content_type = storage.getmeta('content_type', None)
                    if content_type:
                        node.content_type = content_type
                        logger.info(u"%s has content type %s", p, content_type)
            except Exception:
                logger.exception(u"Could not read content type")

        # file is not fully downloaded
        if not fs.is_content_local(st) and not self.allow_dl:
            node.status = LimaNode.STATUS_NOT_LOCAL
        else:
            # not a regular file
            if not st.isreg():
                node.status = LimaNode.STATUS_NOT_A_FILE
            elif node.status != LimaNode.STATUS_NEED_UPDATE:
                # only flag as new if the file was not already flagged as to update
                node.status = LimaNode.STATUS_CREATED

        node.node_cvid = node_cvid
        node.mimetype = mimetype
        node.mrtime = st.st_mrtime
        node.size = st.st_size
        node.default_forkid = default_forkid
        node.total_file_count = getattr(st, "st_total_file_count", None)

        node.is_local_source = getattr(st, 'st_file_local_origin', False)

    def _save_node(self, node):
        with db.transaction():
            # path have unique constraint.
            # in some case, can have remaining node with different id / same path.
            # We delete it before saving the node.
            LimaNode.delete().where(
                (LimaNode.path == node.path) &
                (LimaNode.node_id != node.node_id)
            ).execute()
            node.save()
            logger.debug(u"%s saved", node.path)

    def _process_file_created(self, p, st, node):
        logger.info(u"New file %s", p)
        # nothing special
        self._update_node_infos(node, st, p, is_new=True)
        self._save_node(node)

    def _process_file_modified(self, p, st, node):
        # logger.info(u"File modified %s", p)
        self._update_node_infos(node, st, p)
        self._save_node(node)

    def _process_dir_created(self, p, st, node, allow_recursion=True):
        logger.info(u"Dir created %s", p)
        self._update_node_infos(node, st, p, is_new=True)
        if allow_recursion:
            self.scan(p, -1, recursive=False)
        self._save_node(node)

    def _process_dir_modified(self, p, st, node, allow_recursion=True):
        logger.info(u"Dir modified, need rescan %s", p)
        self._update_node_infos(node, st, p)
        # will check first level files and directories for existence
        if allow_recursion:
            self.scan(p, -1, recursive=False)
        self._process_deleted_files(p, recursive=allow_recursion)
        self._save_node(node)

    def _process_file_deleted(self, node):
        _type = node.content_type if node.content_type else 'node'
        with db.transaction():
            logger.info(u"Removing non existing %s %s", _type, node.path)
            try:
                if node.default_forkid:
                    cleanup_utils._remove_metas(node)
            except Exception:
                logger.exception("Could not remove metas for %s", node.path)
            node.delete_instance()

    def _process_deleted_files(self, root, recursive=False):
        content_types = []
        try:
            for node in self._iter_deletables(root, recursive):
                # keep a trace of content type we delete to force reindexation on it
                if node.content_type and node.content_type not in content_types:
                    content_types.append(node.content_type)
                self._process_file_deleted(node)
        except Exception:
            logger.exception("Error while browsing %s for deleted items", root)

        if content_types:
            indexer.force_reindexation(content_types=set(content_types))

    def _iter_deletables_all(self, root):
        with db.transaction():
            files = LimaNode.list(root, True)[:]
        for node in files:
            yield node

    def _iter_deletables(self, root, recursive=False):
        with db.transaction():
            files = LimaNode.list(root, recursive)[:]

        step = 1000
        for r in xrange(0, len(files), step):
            nodes = files[r:r + step]
            stats = list(fs.bulkstat([x.path for x in nodes]))
            for x in xrange(len(stats)):
                node = nodes[x]
                exc, st = stats[x]
                if isinstance(exc, LimaFsDoesnotExistError):
                    yield node
                    # if the node had been deleted, find all subnode and process them
                    for subnode in self._iter_deletables_all(node.path):
                        yield subnode
