# coding: utf-8

import os
import sqlite3
import gzip
from lima.models import LimaNode
from StringIO import StringIO
import datetime
from lima.fs import FileSystem as fs
from lima.fs.limafs import (LimaFsError, LimaFsTimeoutError,
                            LimaFsDoesnotExistError,
                            LimaFsConnectionRefusedError,
                            LimaFsNotReadyError,
                            LimaFsNoSpaceError,
                            LimaFsNotMountedError,
                            LimaFsMemoryError
                            )
from lima.tools.threads import threading

from lima.meta.indexers import (
    BaseIndexer,
    MusicIndexer,
    MovieIndexer,
    TvIndexer
)

from lima import db
from lima.db import DbNoSpaceError
from lima.tools import segment
from lima.meta import locks

import logging
logger = logging.getLogger(__name__)

ENABLED = threading.Event()

CONTENT_TYPES = (("movie", "video"), ("episode", "video"),
                 ("music", "audio"), )
FORCE_INDEX = {}

for content_type, mimetype in CONTENT_TYPES:
    FORCE_INDEX[content_type] = {
        'mimetype': mimetype,
        'event': threading.Event()
    }


def configure():
    pin()


def pin():
    """ Pin meta datas root folder """
    try:
        stat = fs.stat(BaseIndexer.INDEXES_FOLDER)
        if stat.st_total_file_count != stat.st_total_pin_file_count:
            fs.pin(True, BaseIndexer.INDEXES_FOLDER)
            logger.info(u"%s pined", BaseIndexer.INDEXES_FOLDER)
    except LimaFsDoesnotExistError:
        logger.debug(u"[indexes pin] %s does not exists", BaseIndexer.INDEXES_FOLDER)
    except LimaFsTimeoutError:
        logger.error("[indexes pin] Core timeout")
    except LimaFsConnectionRefusedError:
        logger.error("[indexes pin] Connection refused")
        # no need to continue
        return
    except LimaFsNotMountedError:
        logger.error("[indexes pin] not mounted")
        # no need to continue
        return
    except LimaFsNotReadyError:
        logger.error("[indexes pin] Core not ready")
    except LimaFsError:
        logger.error(u"[indexes pin] Could not pin indexes")
    for indexer in FileIndexer.INDEXERS:
        try:
            indexer.pin()
        except LimaFsTimeoutError:
            logger.error("[indexer pin] Core timeout")
        except LimaFsConnectionRefusedError:
            logger.error("[indexer pin] Connection refused")
            # no need to continue
            break
        except LimaFsNotMountedError:
            logger.error("[indexer pin] Core not mounted")
            # no need to continue
            break
        except LimaFsNotReadyError:
            logger.error("[indexer pin] Core not ready")
        except LimaFsError:
            logger.error(u"[indexer pin] Could not pin indexes")


def force_reindexation(content_type=None, content_types=None):
    if content_type:
        _force_reindexation(content_type)
    if content_types:
        for c in content_types:
            _force_reindexation(c)


def _force_reindexation(content_type):
    event = FORCE_INDEX.get(content_type, None)
    if event:
        event['event'].set()
    ENABLED.set()


class FileIndexer(threading.Thread):

    INDEXERS = (MusicIndexer, MovieIndexer, TvIndexer)

    def __init__(self, indexer_delay, indexer_files_threshold, filesdb,
                 is_debug, *args, **kwargs):
        super(FileIndexer, self).__init__(*args, **kwargs)
        self.logger = logger

        self.indexer_delay = indexer_delay
        # we may increase indexer_delay on error, so we keep the original
        # indexer delay to reset it at its original value
        self.original_indexer_delay = indexer_delay
        # max indexer delay : indexer_delay won't be > to this value
        # if ever we have to increase its value on errors
        self.max_indexer_delay = indexer_delay * 3

        self.indexer_threshold = indexer_files_threshold

        # how often will we check for missing entries ?
        # this operation is as heavy as the indexation
        # so we don't want to do it too often
        self.check_missing_delay = indexer_delay * 10
        self.check_missing_last_check = None

        self.filesdb = filesdb
        self.is_debug = is_debug

        self.indexers = []

    def __repr__(self):
        return u"lima.meta.files.indexer"

    def initialize_indexers(self):
        for indexer_cls in self.INDEXERS:
            indexer = indexer_cls(filesdb=self.filesdb)
            self.indexers.append(indexer)
            self.logger.debug(u"Initialized %s", indexer)

    def mark_missings(self, lima_db, indexer):
        """ Mark missings files from the indexer index as to be reindexed"""
        check_missing_last_check = self.check_missing_last_check or datetime.datetime.now()
        delta = datetime.datetime.now() - check_missing_last_check

        if delta.seconds > self.check_missing_delay or delta.seconds == 0:
            try:
                self.check_missing_last_check = datetime.datetime.now()
                indexer.mark_missings(lima_db)
            except (LimaFsTimeoutError):
                logger.error("Could not mark missing files (timeout)")
                raise
            except (LimaFsNotMountedError, LimaFsNotReadyError, LimaFsConnectionRefusedError):
                logger.error("Could not mark missing files (core)")
                raise
            except (LimaFsNoSpaceError, DbNoSpaceError):
                logger.error("Could not mark missing files (no space)")
                raise
            except (LimaFsMemoryError,):
                logger.error("Could not mark missing files (no memory)")
                raise
            except Exception:
                logger.exception("Could not mark missing files")
                # segment.track_error(event="Marking missing files error")

    def do_index(self, lima_db):
        for indexer in self.indexers:
            if not self.should_continue:
                break
            try:
                with indexer:
                    try:
                        db_index = None

                        self.mark_missings(lima_db, indexer)

                        if not self.should_continue:
                            break
                        with lima_db.transaction():
                            indexables = indexer.indexables().count()
                        if indexables > 0 or\
                                FORCE_INDEX[indexer.CONTENT_TYPE]["event"].isSet():
                                db_index = sqlite3.connect(":memory:")
                                try:
                                    if indexer.do_index(db_index, lima_db) != 0:
                                        self.persist_index(indexer, db_index)
                                finally:
                                    db_index.close()
                                self.remove_conflicts(indexer)
                        else:
                            logger.debug(u"%s : Nothing to index", repr(indexer))
                    finally:
                        if db_index is not None:
                            db_index.close()
                        FORCE_INDEX[indexer.CONTENT_TYPE]["event"].clear()
                    self.indexer_delay = self.original_indexer_delay
            except LimaFsTimeoutError:
                logger.error("[indexation] Core timeout")
                self.indexer_delay = min(self.indexer_delay + 5, self.max_indexer_delay)
            except LimaFsConnectionRefusedError:
                logger.error("[indexation] Connection refused")
                self.indexer_delay = min(self.indexer_delay + 5, self.max_indexer_delay)
            except LimaFsNotReadyError:
                logger.error("[indexation] Core not ready")
                self.indexer_delay = min(self.indexer_delay + 5, self.max_indexer_delay)
            except LimaFsNotMountedError:
                logger.error("[indexation] Lima is not mounted.")
                self.kill(fatal=True)
            except (LimaFsNoSpaceError, DbNoSpaceError):
                logger.error("[indexation] No more space.")
                self.kill(fatal=True)
            except (LimaFsMemoryError):
                logger.error("[indexation] No memory.")
                self.kill(fatal=True)
            except (LimaFsNotMountedError):
                logger.error("[indexation] Not mounted.")
                self.kill(fatal=True)
            except LimaFsError:
                logger.exception(u"[indexation] Error when creating index")
                self.indexer_delay = min(self.indexer_delay + 5, self.max_indexer_delay)
                # segment.track_error(event="Indexation error")

    def persist_index(self, indexer, db_index):
        # first, extract memory db ot a file like object containing SQL
        sql_file = self.dump_db_as_sql_file(db_index)
        try:
            # Write SQL to a tmp DB file
            try:
                os.unlink(fs.get_system_path(indexer.DB_TEMP()))
            except (OSError,) as e:
                # errno 2 : No such file or directory
                if getattr(e, 'errno', -1) != 2:
                    raise
            except Exception:
                logger.exception('Could not unlink %s', fs.get_system_path(indexer.DB_TEMP()))
            fs.ensure_directory(fs.dirname(indexer.DB_TEMP()))
            db_tmp = sqlite3.connect(fs.get_system_path(indexer.DB_TEMP()))
            db_tmp.cursor().executescript(sql_file.read())
            db_tmp.commit()
            db_tmp.close()
        except Exception:
            logger.exception(u'Could not dump DB')
            try:
                os.unlink(fs.get_system_path(indexer.DB_TEMP()))
                logger.warn(u'Removed %s', indexer.DB_TEMP())
            except (OSError,) as e:
                # errno 2 : No such file or directory
                if getattr(e, 'errno', -1) != 2:
                    raise
            except Exception:
                logger.exception('Could not unlink %s', fs.get_system_path(indexer.DB_TEMP()))
            raise

        try:
            try:
                # remove any previous file if present
                os.unlink(fs.get_system_path(indexer.DB_ZIP_TEMP()))
            except (OSError,) as e:
                # errno 2 : No such file or directory
                if getattr(e, 'errno', -1) != 2:
                    raise
            except Exception:
                logger.exception('Could not unlink %s', fs.get_system_path(indexer.DB_ZIP_TEMP()))

            # Compress DB to TMP GZ
            with open(fs.get_system_path(indexer.DB_TEMP()), 'rb') as db_tmp:
                with gzip.open(fs.get_system_path(indexer.DB_ZIP_TEMP()), 'wb') as dbzip_tmp:
                    dbzip_tmp.write(db_tmp.read())
                    dbzip_tmp.flush()
        except Exception:
            logger.exception(u'Could not compress DB')
            if os.path.exists(fs.get_system_path(indexer.DB_ZIP_TEMP())):
                logger.warn(u'Removing %s', indexer.DB_ZIP_TEMP())
                try:
                    os.unlink(fs.get_system_path(indexer.DB_ZIP_TEMP()))
                except (OSError,) as e:
                    # errno 2 : No such file or directory
                    if getattr(e, 'errno', -1) != 2:
                        raise
                except Exception:
                    logger.exception('Could not unlink %s', indexer.DB_ZIP_TEMP())
            raise

        # if everything is ok, persist
        # Replace old DB
        try:
            if os.path.exists(fs.get_system_path(indexer.DB())):
                try:
                    os.unlink(fs.get_system_path(indexer.DB()))
                except (OSError,) as e:
                    # errno 2 : No such file or directory
                    if getattr(e, 'errno', -1) != 2:
                        raise
                except Exception:
                    logger.exception('Could not unlink %s', indexer.DB())
            os.rename(fs.get_system_path(indexer.DB_TEMP()), fs.get_system_path(indexer.DB()))
            logger.info("%s writed", indexer.DB())
        except Exception:
            logger.exception(u"Could not replace %s", indexer.DB())
        # replace OLD gz with new GZ

        try:
            os.unlink(fs.get_system_path(indexer.DB_ZIP()))
        except (OSError,) as e:
            # errno 2 : No such file or directory
            if getattr(e, 'errno', -1) != 2:
                raise
        except Exception:
            logger.exception('Could not unlink %s', indexer.DB_ZIP())
        try:
            os.rename(fs.get_system_path(indexer.DB_ZIP_TEMP()), fs.get_system_path(indexer.DB_ZIP()))
            logger.info("%s writed", indexer.DB_ZIP())
        except Exception:
            logger.exception(u"Could not replace %s", indexer.DB_ZIP())

    def dump_db_as_sql_file(self, db_from):
        tempfile = StringIO()

        for line in db_from.iterdump():
            tempfile.write('%s\n' % line)
        tempfile.seek(0)
        return tempfile

    def compress_db_tmp(self, indexer):
        self.logger.debug(u"Compressing %s", indexer.DB())
        try:
            with open(fs.get_system_path(indexer.DB()), 'rb') as dbin:
                with gzip.open(fs.get_system_path(indexer.DB_ZIP_TEMP()), 'wb') as dbzip:
                    dbzip.write(dbin.read())
                    dbzip.flush()
            self.logger.debug(u"Compressed to %s", indexer.DB_ZIP_TEMP())
        except Exception:
            logger.exception(u"Could not zip file")

    def execute(self):
        self.logger.debug(
            u"Indexer configuration : delay : %s, threshold : %s",
            self.indexer_delay, self.indexer_threshold)
        self.initialize_indexers()
        last_run = datetime.datetime.now()

        while not self._kill_event.isSet():
            self._kill_event.wait(self.indexer_delay)
            if not self._resume_event.isSet():
                continue

            with locks.FSWATCHER_LOCK:
                with locks.BLACKBOARD_LOCK:
                    with db.transaction():
                        total_indexables = LimaNode.reindexables().count()
                    if total_indexables >= self.indexer_threshold:
                        ENABLED.set()

                    for indexer in self.indexers:
                        if FORCE_INDEX[indexer.CONTENT_TYPE]["event"].isSet():
                            ENABLED.set()
                            break

                    if not ENABLED.isSet():
                        continue

                    if total_indexables >= self.indexer_threshold:
                        logger.debug(u"Indexables files threshold reached")

                    now = datetime.datetime.now()
                    delta = now - last_run

                    if delta.seconds < self.indexer_delay:
                        continue

                    logger.debug(u"Indexer delay reached")
                    last_run = datetime.datetime.now()

                    pin()
                    try:
                        self.logger.debug(u"Starting indexer")
                        lima_db = db.get_database()
                        self.do_index(lima_db)
                    finally:
                        try:
                            pin()
                        finally:
                            # always clean lock
                            ENABLED.clear()
        self.logger.info(u"Bye.")

    def remove_conflicts(self, indexer):
        """ Remove genereated indexes conflicts if any."""
        directory = fs.dirname(indexer.DB_ZIP())
        if not fs.exists(directory):
            return
        try:
            for fname, stat in fs.listdirstat(directory):
                if not stat:
                    continue
                if not stat.isreg():
                    continue
                try:
                    if fs.is_conflict(fname):
                        try:
                            fs.unlink(fs.join(directory, fname))
                            self.logger.debug(u"Removed conflict %s", fname)
                        except LimaFsDoesnotExistError:
                            pass
                        except Exception:
                            self.logger.debug(u"Could not remove conflict %s", fname)
                except LimaFsDoesnotExistError:
                    self.logger.warn(u"%s removed when checking for conflict", fname)
        except LimaFsDoesnotExistError:
            self.logger.debug(u"%s removed during listdir", directory)
