# coding: utf-8
"""Killswitch module."""

import msgpack
import datetime

from lima.fs import FileSystem as fs
from lima.fs.fcfs import (LimaFsError,
        LimaFsDoesnotExistError,
        LimaFsTimeoutError,
        LimaFsConnectionRefusedError,
        LimaFsNotReadyError
)
from lima.settings import CONFIG

import logging
logger = logging.getLogger(__name__)


class KillSwitchEnableError(Exception):
    pass


class MinimumVersionRequiredError(KillSwitchEnableError):
    pass


def configure():
    pin()


def pin():
    """ Pin meta datas root folder """
    config_root = CONFIG[u"config_root"]
    system_dir = fs.join(config_root, 'config')
    try:
        stat = fs.stat(system_dir)
        if stat.st_total_file_count != stat.st_total_pin_file_count:
            fs.pin(True, system_dir)
            logger.info(u"%s pined", system_dir)
    except LimaFsDoesnotExistError:
        logger.debug(u"[config pin] %s does not exists", system_dir)
    except LimaFsTimeoutError:
        logger.error("[config pin] Core timeout")
    except LimaFsConnectionRefusedError:
        logger.error("[config pin] Connection refused")
    except LimaFsNotReadyError:
        logger.error("[config pin] Core not ready")
    except LimaFsError:
        logger.error(u"[config pin] Could not pin indexes")


def _check_killswitch():
    config_root = CONFIG[u"config_root"]
    my_min_version = CONFIG[u"min_meta_version"]

    required_version = get_min_version(config_root)

    if my_min_version < required_version:
        logger.error(u"Required meta version %s (current : %s). Please upgrade.", required_version, my_min_version)

        raise MinimumVersionRequiredError(
            u"Minimum version requirement not met.")
    if my_min_version > required_version:
        set_min_version(config_root, my_min_version)


def _read_config(config_file):
    """ Read the file containing the config"""
    with fs.open(config_file, 'r') as config:
        data = config.read()
        if data:
            return msgpack.unpackb(data)
        else:
            return {}


def _write_config(config_file, meta_config):
    """ Write the file containing the config"""
    # create the directory for the file if it does not exists
    fs.ensure_directory(u'/'.join(config_file.split(u'/')[:-1]))
    with fs.open(config_file, 'w') as config:
        config.write(msgpack.packb(meta_config))


def set_min_version(config_root, min_meta_version):
    """ Set the minimum version"""
    # write in /.lima/meta/system/config/killswitch
    system_dir = fs.join(config_root, 'config')
    config_file = fs.join(system_dir, 'killswitch')

    meta_config = {}
    meta_config['min_meta_version'] = min_meta_version
    logger.warn("Set minimum meta version to %s", min_meta_version)
    _write_config(config_file, meta_config)


def get_min_version(config_root):
    """ Get the minimum version"""
    # read in /.lima/meta/system/config/killswitch
    killswitch_fname = 'killswitch'
    system_dir = fs.join(config_root, 'config')
    config_file = fs.join(system_dir, killswitch_fname)

    min_version = -1
    try:
        current_min_version = _read_config(config_file).get("min_meta_version", -1)
    except LimaFsDoesnotExistError:
        current_min_version = -1

    min_version = max(min_version, current_min_version)
    try:
        min_version = max(min_version, _get_conflicts_min_version(config_root, min_version))
    except Exception:
        pass

    return min_version


def _get_conflicts_min_version(config_root, current_min_version):
    killswitch_fname = 'killswitch'
    system_dir = fs.join(config_root, 'config')
    config_file = fs.join(system_dir, killswitch_fname)

    min_version = -1
    # {{ Conflicts resolution
    for path in fs.walk(system_dir, recursive=False):
        try:
            basename = fs.basename(path)
            # conflicts will be named 'killswitch (other host)'
            if basename.startswith(killswitch_fname) and not basename == killswitch_fname:
                file_min_version = _read_config(path).get("min_meta_version", -1)
                min_version = max(min_version, file_min_version)
                # The file is a conflict and contains a minimum version <= to current : we remove it
                if fs.is_conflict(path) and file_min_version <= current_min_version:
                    fs.unlink(path)
                # The file is a conflict and contains a minimum version > to current : we
                # replace the old file with this one
                elif fs.is_conflict(path) and file_min_version > current_min_version:
                    logger.warning(u"Detected minimum %s, overrinding.", file_min_version)
                    try:
                        fs.unlink(config_file)
                    except LimaFsError:
                        pass
                    fs.rename(path, config_file)
        except LimaFsError:
            pass
    # }} Conflicts resolution
    return min_version


class KillSwitch(object):
    def __init__(self, delay=5 * 60):
        self.last_check = None
        self.delay = delay
        self.is_ks_enabled = False

    def reset(self):
        """ Reset method"""
        self.last_check = None
        self.is_ks_enabled = False

    def check(self, force=False):
        """ Check if a killswitch is enabled
        for now, only check if the minimum meta_version is inferior to the current meta_version.
        raise MinimumVersionRequiredError
        """
        if self.last_check is not None and not force:
            # we use a delay to avoid to many reads
            tdelta = datetime.datetime.now() - self.last_check
            if tdelta.seconds < self.delay:
                return self.is_ks_enabled
        try:
            logger.debug(u"Checking killswitch")
            self.last_check = datetime.datetime.now()
            _check_killswitch()
            self.is_ks_enabled = False
        except MinimumVersionRequiredError:
            logger.warn(u"Minimum version is too low.")
            self.is_ks_enabled = True
        except KillSwitchEnableError:
            logger.warn(u"Detected kill switch.")
            self.is_ks_enabled = True
        except LimaFsError:
            # oter fs error. Assume False
            self.is_ks_enabled = False
        return self.is_ks_enabled


switch = KillSwitch()
