# coding: utf-8
import datetime
import peewee

from lima.tools.threads import threading
from lima.meta import cleanup_utils
from lima.models import LimaState, LimaNode
from lima import db
from lima.db import DbNoSpaceError

from lima.fs import FileSystem as fs
from lima.fs.fcfs import (
    LimaFsError,
    LimaFsTimeoutError,
    LimaFsConnectionRefusedError,
    LimaFsNotMountedError,
    LimaFsNoSpaceError,
    LimaFsMemoryError,
    LimaFsNotReadyError,
    LimaFsDeviceNotconfigured,
)

import logging
logger = logging.getLogger(__name__)

from lima.meta.scheduler import indexer
from lima.meta.scheduler import killswitch

from lima.meta import locks


class MetaGarbageCollector(threading.Thread):

    """ Garbage collector for metadata"""
    # default purge delay : 3h
    PURGE_DELAY = 14400

    def __init__(self, gcdelay, gc_fake, *args, **kwargs):
        super(MetaGarbageCollector, self).__init__(*args, **kwargs)
        self.logger = logger

        self.gcdelay = gcdelay if gcdelay is not None else MetaGarbageCollector.PURGE_DELAY
        self.gc_fake = gc_fake
        self.min_fcfs_alive_time = 60 * 15  # 15 minutes

        self.last_purg_date = None

    def __repr__(self):
        return u"meta gc"

    def execute(self):
        self.logger.info(u"Starting meta gc with %s sec delay", self.gcdelay)
        while not self._kill_event.isSet():
            if not self._resume_event.wait(0.1):
                continue
            if not self._resume_event.isSet():
                continue
            # Check if the core is up since enough time
            try:
                fcfs_uptime = fs.status.get('fcfs_uptime', -1)
                if fcfs_uptime >= self.min_fcfs_alive_time:
                    with locks.FSWATCHER_LOCK:
                        with locks.BLACKBOARD_LOCK:
                            indexer.ENABLED.clear()
                            # Only run GC if local index is synced with fs
                            # we do this by comparing the root mrtime (/) on the
                            # filesystem and the one in local index
                            with db.transaction():
                                root_known_mrtime = LimaNode.root_mrtime()
                            current_mrrtime = fs.stat(u"/").st_mrtime
                            if root_known_mrtime == current_mrrtime:
                                self.logger.info(u"Local index is up to date with fs. meta-gc enabled")
                                if killswitch.switch.check(force=True):
                                    self.logger.error(u"Kill switch detected. not pugring.")
                                    break
                                self.purge()
                            else:
                                self.logger.info(u"Local index not up to date with fs. meta-gc won't run.")
                else:
                    self.logger.info(u"fcfs started too soon (%s min ago).", fcfs_uptime / 60.0)
            except(LimaFsTimeoutError):
                # RPC timeout or filesystem not ready, we just wait
                self.logger.error(u"Timeout")
            except(LimaFsConnectionRefusedError):
                # RPC timeout or filesystem not ready, we just wait
                self.logger.error(u"Connection refused")
            except(LimaFsNotReadyError):
                # RPC timeout or filesystem not ready, we just wait
                self.logger.error(u"Connection refused")
            except (LimaFsNotMountedError):
                self.logger.error(u"Disk unmounted during purge")
                self.kill(fatal=True)
            except (LimaFsDeviceNotconfigured):
                self.logger.error(u"Device not configured during purge")
                self.kill(fatal=True)
            except (LimaFsNoSpaceError, DbNoSpaceError):
                self.logger.error(u"No space during purge")
                self.kill(fatal=True)
            except (LimaFsMemoryError, ):
                self.logger.error(u"No memory purge")
                self.kill(fatal=True)
            except (LimaFsError):
                self.logger.exception(u"Could not purge meta datas")

            self.wait(self.gcdelay / 2)
        self.logger.info(u"Bye.")

    def get_purge_state(self):
        try:
            with db.transaction():
                return LimaState.get(key=self.CONFIG_PURGE_KEY)
        except peewee.DoesNotExist:
            return None

    def set_last_purge_date(self):
        self.logger.info(u"Saving purge time")
        self.last_purg_date = datetime.datetime.now()

    def purge(self):
        """ Purge non exisintgs files from database"""
        now = datetime.datetime.now()

        if self.last_purg_date is None:
            self.logger.info(u"First launch. Waiting for next round.")
            self.set_last_purge_date()
            return 0

        purged_files = {}
        purged_files['orphans'] = 0

        if (now - self.last_purg_date).seconds > self.gcdelay:
            self.logger.info(u"Enabling db purge")
            with db.database() as database:
                if not self.should_continue:
                    return purged_files
                try:
                    cleanup_utils.remove_status_deleted(database)
                except Exception:
                    self.logger.exception("Could not remove files flagged as deleted")
                if not self.should_continue:
                    return purged_files
                try:
                    cleanup_utils.purge_deleted_files(database, recursive=True)
                except Exception:
                    self.logger.exception("Could not remove non existing files")
                if not self.should_continue:
                    return purged_files
                try:
                    purged_files['orphans'] = cleanup_utils.remove_orphan_metas(database, fake=self.gc_fake)
                except Exception:
                    self.logger.exception("Could not remove orphans metas")
                if not self.should_continue:
                    return purged_files
                try:
                    cleanup_utils.remove_storages_conflicts()
                except Exception:
                    self.logger.exception("Could not remove meta conflicts")

                self.set_last_purge_date()
        return purged_files

# This is a throwaway variable to deal with a python bug
# see http://bugs.python.org/issue7980
_ = datetime.datetime.strptime('20110101', '%Y%m%d')
