# coding: utf-8

import sys
import traceback

from concurrent import futures

import logging
logger = logging.getLogger(__name__)

from lima.config import SharedConfig
from lima.fs import FileSystem as fs

from lima import db
from lima.models import LimaNode
from lima.meta.blackboard.blackboard import BlackboardController

from lima.meta.scheduler.fswatcher import (
    FsWatcher,
)
from lima.meta.scheduler import killswitch
from lima.meta.scheduler.metagc import MetaGarbageCollector
from lima.meta.scheduler.indexer import FileIndexer
from lima.tools.threads import threading
from lima.meta.migrations.migrate import MetaMigrator
from lima.meta import locks

import time
from lima.meta.scheduler.tasks import (
    TaskRunner,
)
from lima import version


class LimaMetaApp(threading.Thread):

    """ Main application thread """

    def __init__(self,
                 config_obj, pool_size=1,
                 disable_fs_watcher=False,
                 disable_tasks_worker=False,
                 disable_file_processor=False,
                 disable_gc=False,
                 gc_delay=0,
                 gc_fake=True,
                 disable_indexer=False,
                 indexer_delay=0,
                 indexer_files_threshold=100,
                 is_debug=False
                 ):
        super(LimaMetaApp, self).__init__()
        self.have_killswitch = False
        self.killswitch_enabled_wait = 180
        # method executed on non handled exceptions
        sys.excepthook = self.log_uncaught_exceptions

        self.thread_pools = []
        # start as stopped by default
        self._stop = True
        self._kill = False

        self.config_obj = config_obj
        self.config = SharedConfig(restrict_keys_flag_dirty=('gracenote.web_user_token', ), configuration_dirty_handler=self.dump_config)

        self.gc_fake = gc_fake
        # update config from config object

        # flag indicating if threads have been started
        self._initial_start = False

        self.is_debug = is_debug

        self.migrator = MetaMigrator()
        self.thread_pools.append(self.migrator)

        if not disable_gc:
            self.gc = MetaGarbageCollector(gc_delay, gc_fake)
            self.thread_pools.append(self.gc)

        # Run the filesystem watcher
        if not disable_fs_watcher:
            self.init_feeder = FsWatcher(self.config, force_fisrt_import=False)
            self.thread_pools.append(self.init_feeder)

        # Run the task scheduler
        if not disable_tasks_worker:
            self.task_runner = TaskRunner(pool_size=pool_size)
            self.thread_pools.append(self.task_runner)

        # Run the blackboards controller
        if not disable_file_processor:
            self.task_blackboard = BlackboardController(
                self.config, self.task_runner, pool_size=pool_size,
                is_debug=self.is_debug
            )
            self.thread_pools.append(self.task_blackboard)

        if not disable_indexer:
            self.indexer = FileIndexer(indexer_delay, indexer_files_threshold, db.get_db_path(), self.is_debug)
            self.thread_pools.append(self.indexer)

    def execute(self):
        while not self._kill:
            for t in self.thread_pools:
                if t.exc_info()[0] is not None:
                    logger.error(u"%s died - mission aborted. killing myself", t)
                    self.kill()
            try:
                if killswitch.switch.check():
                    self.have_killswitch = True
                    # kill switch detected, just wait and do nothing
                    # we do not quit because we will be relaunch by the service.
                    self.stop()
                    time.sleep(self.killswitch_enabled_wait)
                else:
                    time.sleep(1)
            except IOError:
                # likely a [Errno 4] Interrupted function call
                logger.info(u"Main loop interrupted")
                break

    def stop(self, signum=None, frame=None):
        logger.info(u"Stoping scheduler")
        start = time.time()
        for t in self.thread_pools[::-1]:
            try:
                logger.info(u"Stoping %s", t)
                t.stop()
                logger.info(u"Stoped %s", t)
            except(Exception, ):
                logger.exception("Error while stopping thread")
        self._stop = True
        end = time.time()
        logger.info('Main thread stopped in %s sec', end - start)

    def _do_start_up(self, reset_non_local=False):
        """ Start all thread child"""
        self._initial_start = True
        with db.database("DB CLEAN") as _database:
            try:
                db.maintenance(_database)
            except Exception:
                logger.exception(u"Could not execute DB maintenance")
            try:
                LimaNode.reset_status(_database, reset_non_local)
            except Exception:
                logger.exception(u"Could not reset inermediate status")

        if killswitch.switch.check():
            self.have_killswitch = True
            # kill switch detected, remove all threads and do nothing
            self.thread_pools = []
            return

        for t in self.thread_pools:
            try:
                logger.info(u"Starting %s", repr(t))
                t.start()
            except(Exception, ):
                logger.exception("Error while starting thread")
                self.stop()
        self._stop = False

    def resume(self, config=None):
        logger.info(u"Resuming scheduler")

        if killswitch.switch.check():
            # kill switch detected, remove all threads and do nothing
            self.have_killswitch = True
            logger.warn(u"Kill switch enabled. Won't run")
            self.stop()
            return

        if config is not None:
            self.update_config(config)

        if not self._initial_start:
            self._do_start_up(reset_non_local=self.config['download.enabled'])
            return
        start = time.time()
        self._stop = False
        for t in self.thread_pools[::-1]:
            try:
                logger.info(u"Resuming %s", t)
                t.resume()
                logger.info(u"Resumed %s", t)
            except(Exception, ):
                logger.exception("Error while resuming thread")
        end = time.time()
        logger.info('Main thread resumed in %s sec', end - start)

    def _kill_thread(self, thread):
        start = time.time()
        logger.info("Killing %s", thread)
        try:
            thread.kill()
        except(Exception, ):
            logger.exception("Error while killing thread")
        try:
            if thread.is_alive():
                thread.join()
            else:
                logger.info(u"%s not started - not waiting for it", thread)
        except(Exception, ):
            logger.exception("Error while killing thread")

        end = time.time()
        return end - start

    def kill(self, signum=None, frame=None):
        logger.info(u"Killing scheduler")

        self._kill = True
        start = time.time()
        fs.kill()
        # We do not need thread to be kill synchronously
        _futures = {}
        with futures.ThreadPoolExecutor(max_workers=len(self.thread_pools)) as executor:
            for t in self.thread_pools[::-1]:
                _future = executor.submit(self._kill_thread, t)
                _futures[_future] = t

        for _future in futures.as_completed(_futures):
            t = _futures[_future]
            logger.info("Killed %s in %s sec", t, _future.result())

        end = time.time()
        self._stop = True
        locks.release()
        self.webapp.server.stop()
        logger.info('Main thread killed in %s sec', end - start)

    def is_running(self):
        return not self._stop

    def is_processing(self):
        return self.task_blackboard.is_processing()

    def processed_files_count(self):
        return self.task_blackboard.processed_files_count()

    def all_files_count(self):
        return self.task_blackboard.all_files_count()

    def update_config(self, config=None):
        with self.config.lock:
            # reload configureation from file
            self.config_obj.reload_config()

            if config is None:
                config = {}

            externals_enabled = bool(int(config.get("externals", 0)))
            osdb_enabled = externals_enabled is True or bool(int(config.get("osdb.enabled", 0)))
            self.config['osdb.enabled'] = osdb_enabled
            if osdb_enabled:
                logger.info(u"OSDB is enabled")

            tmdb_enabled = externals_enabled is True or bool(int(config.get("tmdb.enabled", 0)))
            self.config['tmdb.enabled'] = tmdb_enabled

            if tmdb_enabled:
                logger.info(u"TMDB is enabled")
                self.config['tmdb.api_key'] = self.config_obj.get("tmdb", "api_key", True)

            gn_enabled = externals_enabled is True or bool(int(config.get("gracenote.enabled", 0)))
            self.config['gracenote.enabled'] = gn_enabled

            dl_enabled = bool(int(config.get("download.enabled", 0)))
            self.config['download.enabled'] = dl_enabled

            if gn_enabled:
                logger.info(u"Gracenote is enabled")
                self.config['gracenote.client_id'] =\
                    self.config_obj.get( "gracenote", "client_id", True)
                self.config['gracenote.client_id_tag'] =\
                    self.config_obj.get("gracenote", "client_id_tag", True)
                self.config['gracenote.licence'] =\
                    self.config_obj.get("gracenote", "licence", True)
                self.config['gracenote.app_version'] = version.full_version

                self.config['gracenote.user_token'] =\
                    self.config_obj.get("gracenote", "user_token", True)

                web_user_token = self.config_obj.get_user("gracenote", "web_user_token")

                # first, try to get config at user config level
                if web_user_token is not None and\
                        len(web_user_token.strip()) == 0:
                    web_user_token = None
                # If no result, try to get token at system level
                if web_user_token is None:
                    web_user_token = self.config_obj.get("gracenote", "web_user_token", True)
                    if web_user_token is not None and \
                            len(web_user_token.strip()) == 0:
                        web_user_token = None
                self.config['gracenote.web_user_token'] = web_user_token

            self.config.is_dirty = False

    def dump_config(self, config):
        """ Dump the config in a user config file"""
        if config.is_dirty:
            web_user_token = config.get('gracenote.web_user_token', None)
            # only save if not None and not empty
            if web_user_token and len(web_user_token.strip()) != 0:
                self.config_obj.set("gracenote", "web_user_token", web_user_token)
                self.config_obj.save()
            config.is_dirty = False

    def log_uncaught_exceptions(self, ex_cls, ex, tb):
        logger.critical(''.join(traceback.format_tb(tb)))
        logger.critical('{0}: {1}'.format(ex_cls, ex))
