# coding: utf-8


import sys
import time

from lima.meta.utils.timedpriorityqueue import (
    TimedPriorityQueue,
    TimedPriorityQueueWaitException
)
from lima.meta.base.exceptions import LimaMetaError
from lima.tools.threads import threading
from lima.fs.fcfs import (
    LimaFsError,
    LimaFsNoSpaceError,
    LimaFsMemoryError,
    LimaFsDeviceNotconfigured,
)

import logging
logger = logging.getLogger(__name__)


from concurrent import futures


class TaskError(LimaMetaError):
    pass


class TaskAbortedError(TaskError):

    def __init__(self, cause=None):
        super(TaskAbortedError, self).__init__(self)
        self.cause = cause


class TaskKilledError(TaskError):

    def __init__(self, cause=None):
        super(TaskKilledError, self).__init__(self)
        self.cause = cause


class TaskStopedError(TaskError):

    def __init__(self, cause=None):
        super(TaskStopedError, self).__init__(self)
        self.cause = cause


class TaskRetryError(TaskError):

    def __init__(self, time_to_wait=None):
        self.time_to_wait = time_to_wait


class TaskDisabledError(TaskError):

    def __init__(self, tryied_times=None):
        self.tryied_times = tryied_times


class TooManyRetryError(TaskError):

    def __init__(self, tryied_times=None):
        self.tryied_times = tryied_times


class TooManyFailureError(TaskError):

    def __init__(self, tryied_times=None):
        self.tryied_times = tryied_times


class ThreadPoolExecutorStackTraced(futures.ThreadPoolExecutor):

    def submit(self, fn, *args, **kwargs):
        """Submits the wrapped function instead of `fn`"""

        return super(ThreadPoolExecutorStackTraced, self).submit(
            self._function_wrapper, fn, *args, **kwargs)

    def _function_wrapper(self, fn, *args, **kwargs):
        """Wraps `fn` in order to preserve the traceback of any kind of
        raised exception

        """
        try:
            return fn(*args, **kwargs)
        except Exception:
            # Creates an exception of the  same type with the
            # traceback
            raise sys.exc_info()[0], \
                sys.exc_info()[1], \
                sys.exc_info()[2]


class TaskRunner(threading.Thread):

    def __init__(self, pool_size=1):
        threading.Thread.__init__(self)
        self.daemon = True
        self.logger = logger

        self.task_queue = TimedPriorityQueue()

        self.pool_size = pool_size

        self.running_tasks = {}
        self.executor = ThreadPoolExecutorStackTraced(max_workers=self.pool_size)

    def __repr__(self):
        return u"lima.meta.tasks.runner"

    def stop(self):
        super(TaskRunner, self).stop()
        self.task_queue.stop()
        for f, t in self.running_tasks.items():
            try:
                t.stop()
            except:
                self.logger.exception(
                    u"Exception on task stop %s", str(t).decode('utf-8'))

    def resume(self):
        super(TaskRunner, self).resume()
        [t.resume() for f, t in self.running_tasks.items()]
        self.task_queue.resume()

    def kill(self, fatal=False):
        self.task_queue.kill()
        for f, t in self.running_tasks.items():
            try:
                t.kill()
            except:
                self.logger.exception(
                    u"Exception on task kill %s", str(t).decode('utf-8')
                )
        super(TaskRunner, self).kill(fatal)

    def push(self, *args):
        self.task_queue.push(*args)

    def execute(self):
        self.logger.info("Starting Task runner")

        while not self._kill_event.isSet():
            if not self._resume_event.wait(0.1):
                continue
            if not self._resume_event.isSet():
                continue
            try:
                self.run_tasks()
            except(TimedPriorityQueueWaitException,) as e:
                reltime = e.abstime - time.time()
                self.logger.debug(
                    u"Waiting for next task in max %s sec", reltime)
                self.task_queue.wait(reltime)
                self.logger.debug(u"Stopped waiting")
            self._kill_event.wait(0.01)

    def run_tasks(self):
        self.running_tasks = {}

        for prio_task in self.task_queue.iterpop():
            prio, task = prio_task
            _future = self.executor.submit(task.run)
            self.running_tasks[_future] = task

            if len(self.running_tasks) >= self.pool_size\
                    or self.task_queue.has_only_delayed_tasks\
                    or len(self.task_queue) == 0:
                self._wait_tasks_complete()

    def _wait_tasks_complete(self):
        for future in futures.as_completed(self.running_tasks):
            task = self.running_tasks[future]
            e = future.exception()
            try:
                if e is not None:
                    self._task_ends_exception(task, e)
                    self.logger.debug(u"Still %s tasks in queue",
                                      len(self)
                                      )
                else:
                    self._task_ends(task)
            finally:
                # Remove task from running tasks
                del self.running_tasks[future]
                # explicitly delete references
                del future
                del task

        self.logger.debug(u"%s tasks in queue. Next task in %s sec", len(self), self.task_queue.next_task_in)

    def _task_ends(self, task):
        self.logger.info(u"[%s] Task ends (%s remaining)", str(task).decode('utf-8'), len(self))

    def _task_ends_exception(self, task, exc):
        str_task = str(task).decode('utf-8')

        if issubclass(type(exc), TaskRetryError):
            # We have a retry error, we repush the task in the
            # timed priority queue
            if not self.should_continue:
                return
            now = time.time()
            delay = exc.time_to_wait
            self.logger.warning(
                u"%s Delayed for %s sec (%s - %s)",
                exc,
                delay,
                len(self), self.task_queue.next_task_in
            )
            # repush
            self.task_queue.push(task.priority, task, now + delay)

        elif issubclass(type(exc), TaskStopedError):
            if not self.should_continue:
                return
            # Task has been stopped
            self.logger.info(
                u"Task %s Stopped (%s - %s)",
                str_task,
                len(self), self.task_queue.next_task_in
            )
            # repush in case of resume
            self.task_queue.push(task.priority, task)
        elif issubclass(type(exc), TaskKilledError):
            # Task has been killed
            self.logger.info(
                u"Task %s killed (%s - %s)",
                str_task,
                len(self), self.task_queue.next_task_in
            )
        elif issubclass(type(exc), TaskAbortedError):
            # Task has been aborted
            self.logger.error(
                u"%s aborted : %s (%s - %s)",
                str_task,
                exc.cause,
                len(self), self.task_queue.next_task_in
            )
        elif issubclass(type(exc), TaskDisabledError):
            # Task has been disabled
            self.logger.error(
                u"Class for task %s was disabled - %s (%s - %s)",
                str_task,
                exc.tryied_times,
                len(self), self.task_queue.next_task_in
            )
        elif issubclass(type(exc), TooManyRetryError):
            # Task has been retryied too many
            # times
            self.logger.error(
                u"%s retried too many times - %s (%s - %s)",
                str_task,
                exc.tryied_times,
                len(self), self.task_queue.next_task_in
            )
        elif issubclass(type(exc), TooManyFailureError):
            # Task has failed too many times
            self.logger.error(
                u"%s failed too many times - %s (%s - %s)",
                str_task,
                exc.tryied_times,
                len(self), self.task_queue.next_task_in
            )
        elif issubclass(type(exc), LimaFsNoSpaceError):
            self.logger.error(u"No space left on device. Killing myself")
            try:
                self.kill(fatal=True)
            except:
                self.logger.error(u"Error when killing task scheduler")
                raise
        elif issubclass(type(exc), LimaFsMemoryError):
            self.logger.error(u"No memory. Killing myself")
            try:
                self.kill(fatal=True)
            except:
                self.logger.error(u"Error when killing task scheduler")
                raise
        elif issubclass(type(exc), LimaFsDeviceNotconfigured):
            self.logger.error(u"Device not configured. Killing myself")
            try:
                self.kill(fatal=True)
            except:
                self.logger.error(u"Error when killing task scheduler")
                raise
        elif issubclass(type(exc), LimaFsError):
            self.logger.error(u"FS error.")
        else:
            # Unknown exception
            self.logger.error(u"Task failed hard : %s (%s - %s)",
                              exc,
                              len(self), self.task_queue.next_task_in
                              )

    def __len__(self):
        return len(self.task_queue)
