# coding: utf-8
import abc

import msgpack

from lima.fs import FileSystem as fs
from lima.fs.limafs import (LimaFsError, LimaFsTimeoutError,
                            LimaFsConnectionRefusedError,
                            LimaFsDoesnotExistError, LimaFsNotReadyError, )
from lima.fs.utils import DeleteForkOnError
from lima.tools import bin_to_hex
from lima.tools import segment
from lima.tools import strings
from lima.settings import CONFIG

from lima.tools import logs
logger = logs.getLogger(__name__)

WINDOWS_FORBIDDEN_CHARS = r'*|\:"<>?/ '
NIX_FORBIDDEN_CHARS = '\0'

FORBIDDEN_CHARS = list(set(WINDOWS_FORBIDDEN_CHARS)) + list(
    set(NIX_FORBIDDEN_CHARS))


def cleanup_for_fs(fname):
    fname = strings.normalize(fname, lowercase=False)
    for c in FORBIDDEN_CHARS:
        fname = fname.replace(c, '_')
    return fname


def configure():
    """ Configure storage"""
    try:
        pin()
    except LimaFsError:
        logger.error("Could not pin meta folder")


def pin():
    """ Pin meta datas root folder """
    try:
        classes = [SerieStorage, SeasonStorage, ArtistStorage, AlbumStorage, MetaKeyValueStorage]
        folders = list(set(map(lambda x: x.ROOT(), classes)))

        for f in folders:
            try:
                stat = fs.stat(f)
                if stat.st_total_file_count != stat.st_total_pin_file_count:
                    fs.pin(True, f)
                    logger.info(u"%s pinned", f)
            except LimaFsDoesnotExistError:
                logger.debug(u"Could not pin %s :  does not exists", f)

    except LimaFsTimeoutError:
        logger.error("Core timeout")
        raise
    except LimaFsConnectionRefusedError:
        logger.error("Connection refused")
        raise
    except LimaFsNotReadyError:
        logger.error("Core not ready")
        raise
    except LimaFsError:
        logger.exception(u"Could not ensure meta folder existence")
        raise


class Storage(object):
    """ Base class for metadata storage abstraction layer """

    __metaclass__ = abc.ABCMeta
    META_FOLDER = CONFIG['meta_root']
    KEY_PREFIX = u'limameta'

    @classmethod
    def ROOT(cls):
        return fs.join(cls.META_FOLDER, cls.DIRECTORY)

    @abc.abstractproperty
    def DIRECTORY(self):
        """ Define storage kind. Used by default for the root folder."""
        return None  # pragma: no cover

    @abc.abstractproperty
    def get_storage_folder(self):
        """ Storage folder for an item"""
        pass

    def get_storage_fname(self, name):
        path = self.get_storage_folder()
        return fs.join(path, u"%s.%s" % (self.KEY_PREFIX, cleanup_for_fs(name)))

    def exists(self, *args, **kwargs):
        path = self.get_storage_fname(*args, **kwargs)
        return fs.exists(path)

    def exists_non_empty(self, *args, **kwargs):
        path = self.get_storage_fname(*args, **kwargs)
        try:
            stat = fs.stat(path)
            return stat.st_size > 0
        except LimaFsDoesnotExistError:
            return False

    def get_storage(self, *args, **kwargs):
        mode = kwargs.pop('mode', u'r')
        path = self.get_storage_fname(*args, **kwargs)
        if u'w' in mode:
            try:
                storage = DeleteForkOnError(fs, path, mode=mode)
            except LimaFsDoesnotExistError:
                # Tell, don't ask
                # this may happen if the folder of path_to does not exists yet
                fs.ensure_directory(fs.dirname(path))
                storage = DeleteForkOnError(fs, path, mode=mode)
        else:
            try:
                storage = fs.open(path, mode=mode)
            except LimaFsDoesnotExistError:
                # Tell, don't ask
                # this may happen if the folder of path_to does not exists yet
                fs.ensure_directory(fs.dirname(path))
                storage = DeleteForkOnError(fs, path, mode=mode)
        return storage

    def remove(self):
        path = self.get_storage_folder()
        fs.rmdirs(path)

    def link(self, other_storage, key_from, key_to):
        path_from = other_storage.get_storage_fname(key_from)
        path_to = self.get_storage_fname(key_to)
        self.logger.debug(u"Linking %s to %s", path_from, path_to)

        if fs.exists(path_to):
            fs.unlink(path_to)

        try:
            fs.symlink(path_from, path_to)
        except LimaFsDoesnotExistError:
            # Tell, don't ask
            # this may happen if the folder of path_to does not exists yet
            # if this happens because of path_from which does not exists, this
            # is an error so we let it be raise
            fs.ensure_directory(fs.dirname(path_to))
            fs.symlink(path_from, path_to)

    @classmethod
    def remove_key(cls, key, delete_empty_parent=False):
        path = fs.join(cls.META_FOLDER, cls.DIRECTORY, cleanup_for_fs(key[0:2]),
                       cleanup_for_fs(key), )
        fs.rmdirs(path)

        if delete_empty_parent:
            parent = fs.dirname(path)
            if fs.is_empty_dir(parent):
                fs.rmdirs(parent)

    @classmethod
    def remove_empty(cls):
        """ Removes empty directories from this storage root """
        root = cls.ROOT()
        for path, st in fs.walk_stat(root):
            try:
                if st.isdir() and fs.is_empty_dir(path):
                    fs.rmdir(path)
            except LimaFsDoesnotExistError:
                pass

    @classmethod
    def iter_files(cls, files_only=False):
        for f, _ in cls.iter_files_stat(files_only):
            yield f

    @classmethod
    def iter_files_stat(cls, files_only=False):
        root = fs.join(cls.META_FOLDER, cls.DIRECTORY)
        for f, stat in fs.walk_stat(root):
            if files_only:
                if stat.isreg():
                    yield f, stat
            else:
                yield f, stat

    def pin(self):
        """ Pin storage folder"""
        try:
            directory = self.get_storage_folder()
            stat = fs.stat(directory)
            # is_dir = stat.isdir()
            if getattr(stat, 'st_total_file_count', -1) < getattr(stat, 'st_total_pin_file_count', -1):
                fs.pin(True, directory)
                self.logger.debug(u"%s pinned", self.get_storage_folder())
        except LimaFsDoesnotExistError:
            pass
        except LimaFsTimeoutError:
            self.logger.error("Core timeout")
        except LimaFsConnectionRefusedError:
            self.logger.error("Connection refused")
        except LimaFsNotReadyError:
            self.logger.error("Core not ready")
        except LimaFsError:
            self.logger.exception(u"Could not pin %s", directory)
        except AttributeError:
            # this appeared sometimes on stat.st_total_file_count
            # logger.exception(u"Could not pin %s", directory)
            # segment.track_error(event="Could not pin meta folder (isdir=%s)" % is_dir)
            raise


class MetaForkStorage(Storage):
    DIRECTORY = fs.join('data', 'files', 'fid')

    def __init__(self, path):
        super(MetaForkStorage, self).__init__()
        self.path = path
        self.logger = logs.getLogger(__name__, fs.basename(path))

    def get_storage_folder(self, key=None):
        if key is None:
            key = self._get_storage_key()
        return fs.join(self.ROOT(), cleanup_for_fs(key[0:2]), cleanup_for_fs(key))

    def _get_storage_key(self):
        # lazy load storage key to avoid too many stat call on the core
        if not hasattr(self, '_storage_key'):
            stat = fs.stat(self.path)
            self._storage_key = bin_to_hex(stat.st_default_forkid)
        return self._storage_key

    def switch_from(self, old_key, new_key):
        """ Moe meta datas files from old key to new key"""

        old_root = self.get_storage_folder(old_key)
        new_root = self.get_storage_folder(new_key)

        if not fs.exists(old_root):
            return
        if not fs.exists(new_root):
            fs.ensure_directory(new_root)

        for path in fs.walk(old_root):
            fname = fs.basename(path)
            new_path = fs.join(new_root, fname)
            if not fs.exists(new_path):
                fs.rename(path, new_path)

        fs.rmdirs(old_root)


class PoolableStorageMixin(object):
    # will store metastorage, singleton style
    # key is the path
    METAS_STORAGES = {}

    def unload(self, persist=False):
        self.__class__.forget(self.path, self.fork_meta, persist=persist)

    @classmethod
    def get_from_pool(cls, *args, **kwargs):
        path = None
        if len(args) > 0:
            path = args[0]
        if path is None:
            path = kwargs.get('path')

        fork_meta = None
        if len(args) > 1:
            fork_meta = args[1]
        if fork_meta is None:
            fork_meta = kwargs.get('fork_meta')
        try:
            return cls.METAS_STORAGES[(cls, path, fork_meta)]
        except KeyError:
            obj = cls(*args, **kwargs)
            cls.METAS_STORAGES[(cls, path, obj.fork_meta)] = obj
            return obj

    @classmethod
    def persist_metas(cls, path, fork_meta, forget=False):
        """ Flush file metadatas.
        Meta datas will be effectively writed only if they are differents from
        existing metas
        """
        storage = cls.METAS_STORAGES.get((cls, path, fork_meta), None)

        if storage is not None:
            storage.persist()
            logger.debug(u"Persisted %s for %s", fork_meta, path)

            if forget is True:
                cls.forget(storage.path, fork_meta)
        else:
            logger.debug(u"Metas was null : %s for %s", fork_meta, path)

    @classmethod
    def unload_all(cls, persist=False):
        for key, _ in cls.METAS_STORAGES.items():
            if key[0] == cls:
                try:
                    if persist:
                        cls.METAS_STORAGES[key].persist()
                    del cls.METAS_STORAGES[key]
                except KeyError:
                    logger.debug(u"key %s already removed")
                except LimaFsTimeoutError:
                    logger.error("Core timeout")
                except LimaFsConnectionRefusedError:
                    logger.error("Connection refused")
                except LimaFsNotReadyError:
                    logger.error("Core not ready")
                except LimaFsDoesnotExistError:
                    logger.warn(u"Does not exists")
                except LimaFsError:
                    logger.error(u"Could not persist metas")

    @classmethod
    def forget(cls, path, fork_meta, persist=True):
        try:
            key = (cls, path, fork_meta)
            if persist:
                cls.METAS_STORAGES[key].persist()
            del cls.METAS_STORAGES[key]
        except KeyError:  # persist called multiples times
            pass

    def __enter__(self, persist=False):
        """ Context manager methods"""
        # set an internal attribute to know if we shoul persists
        # the object on __exit__
        setattr(self, '__ctx_persists', persist)
        return self

    def __exit__(self, *args, **kwargs):
        self.unload(persist=getattr(self, '__ctx_persists', False))


class MetaKeyValueStorage(PoolableStorageMixin, MetaForkStorage):
    KEY_CONTEXT = u"context"

    def __init__(self, path, fork_meta, *args, **kwargs):
        super(MetaKeyValueStorage, self).__init__(path, *args, **kwargs)
        self.fork_meta = fork_meta
        self.metas = {}
        self.load()

    def _load_metas(self):
        """ Load datas from the fork meta.
        """
        try:
            if self.exists(self.fork_meta):
                with self.get_storage(self.fork_meta, mode='r') as fork_meta:
                    try:
                        msgpack_data = fork_meta.read()
                    except (LimaFsNotReadyError, LimaFsConnectionRefusedError):
                        # fatals, re-raise
                        raise
                    except Exception:
                        self.logger.warning(u"Could not reads metas for %s", self.path)
                        return None
                    try:
                        return msgpack.unpackb(msgpack_data, encoding="utf-8")
                    except (msgpack.exceptions.UnpackException):
                        self.logger.error(u"Could not unpack metas for %s", self.path)
                    except (msgpack.exceptions.ExtraData):
                        self.logger.error(u"Meta corrupted for %s", self.path)
                    except Exception:
                        self.logger.error(u"could not deserialize metas for %s", self.path)
        except LimaFsDoesnotExistError:
            # may occur if the file is removed between exists check and read
            pass
        return None

    def _get_or_load_metas(self):
        """Return the dictionary containing metas for the given path.

        If metas are already loaded, returns the loaded metas
        If metas are not already loaded and if the fork meta exists, read its
        content and return it.
        If metas are not already loaded and if the fork meta does not exists,
        return a new dic instance

        """
        if len(self.metas) == 0:
            existings_metas = self._load_metas()
            if existings_metas is None:
                self.metas = {}
            else:
                self.metas = existings_metas
        return self.metas

    def load(self):
        """Load metadatas from the core"""
        self.metas = self._get_or_load_metas()

    def getmeta(self, key, default=None):
        """ Return data for the given key"""
        metas = self._get_or_load_metas()
        return metas.get(key, default)

    def setmeta(self, key, value):
        """ Set a meta for the given path.
        Metadatas are stored in the fork meta.

        To avoid too many read/write, the fork content will be loaded in memory
        and wont be persisted until the method persist_metas is called.
        """
        self._get_or_load_metas()
        self.metas[key] = value

    def update(self, values):
        """ Update the meta dictionary once in a row.
        """
        self._get_or_load_metas()
        self.metas.update(values)

    def hasmeta(self, key):
        """Return True if the key is present """
        self._get_or_load_metas()
        return key in self.metas

    def persist(self):
        """Persists metasto the core"""
        # Current metas : current metas in memory
        current_metas = self.metas
        # If current meta is None, it as never been loaded
        #
        if current_metas is None:
            return
        # Old metas for comparison
        old_metas = self._load_metas()

        # only save if changed
        if cmp(current_metas, old_metas) != 0:
            # special case : if the context is empty, we do not store it
            if self.KEY_CONTEXT in self.metas:
                if len(self.metas[self.KEY_CONTEXT]) == 0:
                    try:
                        del self.metas[self.KEY_CONTEXT]
                    except KeyError:
                        # should not happen
                        pass
            if len(current_metas) > 0:
                self.logger.info("Metas changed, persists datas")
                with self.get_storage(self.fork_meta, mode='wb') as fork_meta:
                    self.logger.debug(u"Writing to %s", fork_meta.path)
                    fork_meta.write(msgpack.packb(current_metas,
                                                  use_bin_type=True))

    def remove(self, *args, **kwargs):
        super(MetaKeyValueStorage, self).remove(*args, **kwargs)
        MetaKeyValueStorage.forget(self.path, self.fork_meta)

    def get_context(self):
        context = self.getmeta(self.KEY_CONTEXT)
        if context is None:
            context = {}
            self.setmeta(self.KEY_CONTEXT, context)
        return context

    def save_context(self, context):
        self.setmeta(self.KEY_CONTEXT, context)

    def is_same_link(self, other_storage, key_from, key_to):
        path_from = self.get_storage_fname(key_from)
        path_to = other_storage.get_storage_fname(key_to)
        try:
            st = fs.stat(path_from)
        except LimaFsDoesnotExistError:
            return False

        if not st.islink():
            return False
        try:
            real_path = fs.resolve_symlink(path_from)
            return real_path == path_to
        except LimaFsDoesnotExistError:
            return False

class ArtistStorage(MetaKeyValueStorage):
    DIRECTORY = fs.join('data', 'musics',)

    def __init__(self, artist):
        if artist is None:
            raise ValueError(u"Artist can't be None")
        # use artist as path
        self.artist = artist
        super(ArtistStorage, self).__init__(path=artist, fork_meta='meta')

    def get_storage_folder(self):
        return fs.join(self.ROOT(), cleanup_for_fs(self.artist.lower()), )


class AlbumStorage(MetaKeyValueStorage):
    """Storage for albums"""
    DIRECTORY = fs.join('data', 'musics',)

    def __init__(self, artist, album):
        if artist is None:
            raise ValueError(u"Artist can't be None")
        if album is None:
            raise ValueError(u"Album can't be None")
        # use artist/album as reference path
        path = u"%s/%s" % (artist, album)
        self.artist = artist
        self.album = album
        super(AlbumStorage, self).__init__(path, 'meta')

    def get_storage_folder(self):
        return fs.join(self.ROOT(), cleanup_for_fs(self.artist.lower()),
                       cleanup_for_fs(self.album.lower()))


class SerieStorage(MetaKeyValueStorage):
    """Storage for albums"""
    DIRECTORY = fs.join('data', 'series',)

    def __init__(self, serie):
        if serie is None:
            raise ValueError(u"Serie name can't be None")
        # use artist/album as reference path
        path = serie
        self.serie = serie
        super(SerieStorage, self).__init__(path, 'meta')

    def get_storage_folder(self):
        return fs.join(self.ROOT(), cleanup_for_fs(self.serie.lower()))


class SeasonStorage(MetaKeyValueStorage):
    """Storage for albums"""
    DIRECTORY = fs.join('data', 'series',)

    def __init__(self, serie, season):
        if serie is None:
            raise ValueError(u"Serie name can't be None")
        # use artist/album as reference path
        path = u"%s/%s" % (serie, season)
        self.serie = serie
        self.season = season
        super(SeasonStorage, self).__init__(path, 'meta')

    def get_storage_folder(self):
        return fs.join(self.ROOT(), cleanup_for_fs(self.serie.lower()), cleanup_for_fs(str(self.season).lower()), )
