import heapq
import threading
import time


def heapiter(heap):
    heap = heap[:]

    while True:
        try:
            yield heapq.heappop(heap)
        except IndexError:
            break


class TimedPriorityQueueWaitException(Exception):

    def __init__(self, abstime):
        super(TimedPriorityQueueWaitException, self).__init__()

        self.abstime = abstime


class TimedPriorityQueue(object):

    def __init__(self):
        super(TimedPriorityQueue, self).__init__()

        self.q = []
        self.timed_q = []
        self.lock = threading.RLock()
        self.cond = threading.Condition(self.lock)

        self._kill_event = threading.Event()
        self._resume_event = threading.Event()
        self._resume_event.set()

    def __len__(self):
        return len(self.q) + len(self.timed_q)

    def push(self, priority, v, abstime=None):
        with self.lock:
            if abstime:
                heapq.heappush(self.timed_q, (abstime, priority, v))
            else:
                heapq.heappush(self.q, (priority, v))
            self.cond.notify()

    def pop_nonblocking(self):
        with self.lock:
            if not self._resume_event.wait(0.1):
                return None
            if not self._resume_event.isSet():
                return None
            # 1st step: merge self.timed_q to self.q
            now = time.time()
            new_timed_q = []
            for abstime, priority, v in heapiter(self.timed_q):
                if now > abstime:
                    heapq.heappush(self.q, (priority, v))
                else:
                    heapq.heappush(new_timed_q, (abstime, priority, v))
            self.timed_q = new_timed_q

            # 2nd step: pop from self.q if element
            try:
                priority, v = heapq.heappop(self.q)
            except IndexError:
                if self.timed_q:
                    raise TimedPriorityQueueWaitException(self.timed_q[0][0])
                else:
                    return None
            else:
                return priority, v

    def pop_blocking(self, wait_timeout=0.01):
        while not self._kill_event.isSet():
            try:
                priority_v = self.pop_nonblocking()
            except TimedPriorityQueueWaitException, e:
                reltime = e.abstime - time.time()
                self.wait(reltime)
            else:
                if priority_v is None:
                    self.wait(wait_timeout)
                else:
                    return priority_v
        raise StopIteration

    def iterpop_non_blocking(self):
        while not self._kill_event.isSet():
            yield self.pop_nonblocking()
        raise StopIteration

    def iterpop(self, wait_timeout=0.01):
        while not self._kill_event.isSet():
            yield self.pop_blocking(wait_timeout)
        raise StopIteration

    def resume(self):
        self._resume_event.set()

    def stop(self):
        self._resume_event.clear()

    def kill(self):
        self._kill_event.set()
        try:
            self.cond.notify_all()
        except RuntimeError:
            # lock was not aquired
            pass

    def wait(self, wait_timeout):
        with self.lock:
            self.cond.wait(wait_timeout)

    @property
    def has_delayed_tasks(self):
        return len(self.timed_q) > 0

    @property
    def has_non_delayed_tasks(self):
        return len(self.q) > 0

    @property
    def has_only_delayed_tasks(self):
        return self.has_delayed_tasks and not self.has_non_delayed_tasks

    @property
    def next_task_time(self):
        if len(self.timed_q) > 0:
            return self.timed_q[0][0]
        return None

    @property
    def next_task_in(self):
        if len(self.q) > 0:
            return 0
        else:
            now = time.time()
            nex_time = self.next_task_time
            if nex_time:
                return nex_time - now
            return 0
