# coding: utf-8

from peewee import *
import datetime

from lima import db
from lima.settings import CONFIG


database = db.get_database()

import logging
logger = logging.getLogger("lima.meta.models")


class LimaModel(Model):

    class Meta:
        database = database


class LimaNode(LimaModel):
    # meta won't be processed

    STATUS_NOT_LOCAL = -5
    STATUS_MIME_NOT_SUPPORTED = -4
    STATUS_NOT_A_FILE = -3
    STATUS_DELETED = -2
    STATUS_IGNORE = -1
    # meta will be processed in mode update = False
    STATUS_CREATED = 0
    # meta will be processed in mode update = True
    STATUS_NEED_UPDATE = 1

    # intermediate status
    # file is waiting to be processed
    STATUS_WAIT_PROCESSING = 10
    STATUS_WAIT_PROCESSING_UPDATE = 11
    # file is being processed
    STATUS_PROCESSING = 12
    STATUS_PROCESSING_UPDATE = 13

    # file process ends with error(s)
    STATUS_ERROR = 20
    # file process ends with no error
    STATUS_NO_ERROR = 21

    STATUS_NEED_RETRY = 30

    # file has been processed
    STATUS_PROCESSED = 40

    # {{ Indexation status
    STATUS_INDEX_IGNORE = -1
    STATUS_INDEX_NONE = 0
    STATUS_INDEX_NEED_REINDEX = 1
    STATUS_INDEX_INDEXED = 2
    # }} Indexation status


    # ## Content type
    CONTENT_TYPE_MOVIE = u'movie'
    CONTENT_TYPE_EPISODE = u'episode'
    CONTENT_TYPE_MUSIC = u'music'

    node_id = CharField(primary_key=True)
    path = CharField(unique=True)
    node_cvid = CharField(null=True)

    mrtime = IntegerField(default=0, index=True)
    size = IntegerField(default=0, index=True)
    default_forkid = CharField(null=True)

    # Status. See constants above
    status = IntegerField(index=True)
    last_updated = DateTimeField(default=datetime.datetime.now)

    # date of the last time we tryied to process the item
    last_try = DateTimeField(default=datetime.datetime.now)
    try_count = IntegerField(default=0, index=True)
    next_try = DateTimeField(null=True, index=True)

    mimetype = CharField(null=True, index=True)
    # content type : specialization on the mimetype, eg 'movie', 'episode'
    content_type = CharField(null=True, index=True)

    # Indexation status. See constants above
    status_index = IntegerField(null=False, index=True, default=STATUS_INDEX_NONE)

    # Total file count : for directory only
    total_file_count = IntegerField(null=True)
    meta_version = IntegerField(null=False, default=CONFIG[u'min_meta_version'])

    # handle levels of file modification
    is_content_modified = BooleanField(default=False)
    is_path_modified = BooleanField(default=False)
    is_basename_modified = BooleanField(default=False)

    # we need to know if the file had been originally created on this device
    is_local_source = BooleanField(default=False)


    def save(self, *args, **kwargs):
        """Override save method"""
        # update las updated timestamp
        self.last_updated = datetime.datetime.now()
        super(LimaNode, self).save(*args, **kwargs)

    @staticmethod
    def max_mrtime():
        return LimaNode.select(fn.Max(LimaNode.mrtime)).scalar() or -1

    @staticmethod
    def root_mrtime():
        try:
            return LimaNode.get(LimaNode.path == u'/').mrtime
        except DoesNotExist:
            return -1

    @staticmethod
    def reset_status(database, reset_non_local=False):
        """Reset intermediate status"""
        logger.info(u"Reset old intermediate status")
        with database.transaction():
            LimaNode.update(
                status=LimaNode.STATUS_NEED_UPDATE,
            ).where(
                LimaNode.status << [
                    LimaNode.STATUS_PROCESSING_UPDATE,
                    LimaNode.STATUS_WAIT_PROCESSING_UPDATE,
                ]
            ).execute()

            LimaNode.update(
                status=LimaNode.STATUS_NEED_RETRY,
                next_try=datetime.datetime.now(),
            ).where(
                LimaNode.status << [
                    LimaNode.STATUS_WAIT_PROCESSING,
                    LimaNode.STATUS_PROCESSING,
                    LimaNode.STATUS_ERROR,
                    LimaNode.STATUS_NO_ERROR,
                ]
            ).execute()

            if reset_non_local:
                LimaNode.update(
                    status=LimaNode.STATUS_NEED_UPDATE,
                ).where(
                    LimaNode.status << [
                        LimaNode.STATUS_NOT_LOCAL,
                    ]
                ).execute()

            logger.info(u"Reset old intermediate status : DONE")

    @staticmethod
    def reindexables(mimetype=None, content_type=None):
        condition = (LimaNode.status_index == LimaNode.STATUS_INDEX_NEED_REINDEX) &\
                (LimaNode.status == LimaNode.STATUS_PROCESSED)
        if mimetype is not None:
            mimetype+=u'%'
            condition = condition & (LimaNode.mimetype ** mimetype)
        if content_type is not None:
            condition = condition & (LimaNode.content_type == content_type)
        return LimaNode.select().where(condition).order_by(LimaNode.path)

    @staticmethod
    def force_reindexation(content_type):
        """ Force reindeaxtion for the given content type"""
        query = LimaNode.update(status_index=LimaNode.STATUS_INDEX_NEED_REINDEX)\
                .where(
                        (LimaNode.content_type == content_type) & \
                                (LimaNode.status_index!=LimaNode.STATUS_INDEX_NEED_REINDEX) &\
                                (LimaNode.status_index!=LimaNode.STATUS_INDEX_IGNORE) &\
                                (LimaNode.status == LimaNode.STATUS_PROCESSED)
                        )
        return query.execute()

    @staticmethod
    def processed_music():
        condition = (LimaNode.mimetype ** u'audio/%') & \
            (LimaNode.status == LimaNode.STATUS_PROCESSED) &\
            (LimaNode.content_type == LimaNode.CONTENT_TYPE_MUSIC)
        return LimaNode.select().where(condition)

    @staticmethod
    def processed_audio(content_type='music'):
        condition = (LimaNode.mimetype ** u'audio/%') & \
            (LimaNode.status == LimaNode.STATUS_PROCESSED) &\
            (LimaNode.content_type == content_type)
        return LimaNode.select().where(condition)

    @staticmethod
    def processed_video(content_type='movie'):
        condition = (LimaNode.mimetype ** u'video/%') & \
            (LimaNode.status == LimaNode.STATUS_PROCESSED) &\
            (LimaNode.content_type == content_type)
        return LimaNode.select().where(condition)


    @staticmethod
    def list(root, recursive=False):
        if not root.endswith('/'):
            root += '/'

        if recursive is True:
            cond = (LimaNode.path ** (root+'%') )
        else:
            cond = (
                    (LimaNode.path ** (root+'%') &
                    ~ (LimaNode.path ** (root+'%/%'))
                    )
                )
        return LimaNode.select().where(cond)

    @classmethod
    def processed_nodes(cls):
        return cls.select().where(LimaNode.status << [LimaNode.STATUS_PROCESSED, LimaNode.STATUS_IGNORE])

#
class LimaState(LimaModel):

    """ General purpose key/value store """
    key = CharField(primary_key=True)
    value = CharField()
