# coding: utf-8

import os
import subprocess as sp
import sys
import time

import threading

import abc


from decimal import Decimal as _D
try:
    import queue
except ImportError:
    import Queue as queue

import logging
logger = logging.getLogger(__name__)

__all__ = [
    "VideoFile",
    "VideoFilter",
    "VideoThumbailer",
    "FFmpegError",
    "NotInstalledError",
    "NoOutputError",
    "ThumbnailError",
    "ReadTimeoutError",
]

class FFmpegError(Exception):

    """Base class for ffmpeg errors"""
    pass


class NotInstalledError(FFmpegError):

    """Could not find the ffmpeg binary."""


class NoOutputError(FFmpegError):

    """ Could not get ffmpeg output """


class ThumbnailError(FFmpegError):
    pass


class ReadTimeoutError(FFmpegError):

    """Reading from the ffmpeg command-line tool timed out."""
    pass

FFMPEG_BIN = "ffmpeg"
"""ffmpeg"s path"""


FFPROBE_BIN = u"ffprobe"
"""ffprobe"s path"""


class QueueReaderThread(threading.Thread):

    """A thread that consumes data from a filehandle and sends the data
    over a Queue.
    """

    def __init__(self, fh, blocksize=1024, discard=False):
        super(QueueReaderThread, self).__init__()
        self.fh = fh
        self.blocksize = blocksize
        self.daemon = True
        self.discard = discard
        self.queue = None if discard else queue.Queue()

    def run(self):
        while True:
            data = self.fh.read()
            if not self.discard:
                self.queue.put(data)
                if not data:
                    # Stream closed (EOF).
                    break


class FFInput(object):

    def __init__(self, ff_file):
        # fp is a path
        if type(ff_file.fp) in (str, unicode):
            self.stdin = sp.PIPE
            self.input_name = ff_file.fp
            if type(ff_file.fp) is unicode:
                self.input_name = self.input_name.encode(
                    sys.getfilesystemencoding() or 'utf-8')
        else:
            self.stdin = ff_file.fp
            self.input_name = u"pipe:0"

    def seek(self, seek_to):
        self.seek_to = seek_to
        return self

    def as_args(self):
        ret = []
        if hasattr(self, u"seek_to"):
            # It"s better to tell ffmpeg where to seek before givin it the input
            # name
            ret += [u"-ss", unicode(self.seek_to)]
        ret += [u"-i", self.input_name]

        return ret


class FFProcess(object):

    def __init__(self, processor, task):
        self.logger = logger

        self.processor = processor
        self.task = task
        self.processor.ff_processes.append(task)

    def __enter__(self):
        return self.task

    def __exit__(self, *exc):
        ''' Task have finish  we remove it from tasklist'''
        if self.task in self.processor.ff_processes :
            self.processor.ff_processes.remove(self.task)
        del self.task
        del self.processor

        # preserve original exception
        exc_type, exc_inst, traceback = exc
        if exc_type is not None:
            # raise original exception
            raise exc_type, exc_inst, traceback


class FFMpegFile(object):

    def __init__(self, fp):
        self.logger = logger

        self.fp = fp
        self.ff_processes = []

    """ Base class for ffmpeg files"""

    def seek(self, seek_to):
        if not type(self.fp) in (str, unicode):
            self.fp.seek(seek_to)

    @property
    def infos(self):
        # lazy loading
        if not hasattr(self, u"_infos"):
            with FFFileInfos(FFInput(self)) as ffprobe:
                # get_info returns json
                self._infos = ffprobe.get_infos()
        return self._infos

    @property
    def duration(self):
        # lazy loading
        if not hasattr(self, u"_duration"):
            self._duration = _D(
                self.infos.get(u"format", {}).get(u"duration", u"0")
            )
        return self._duration

    def stop(self):
        """ Stop all ffmpeg spawned processes  """
        self.logger.info(u"Stoping ff processes")
        for t in self.ff_processes:
            try:
                t.stop()
            except:
                self.logger.error("Could not stop process")
        return True

    def kill(self):
        """ Kill all ffmpeg spawned processes  """
        self.logger.info(u"Killing ff processes")
        for t in self.ff_processes[:]:
            try:
                t.kill()
            except:
                self.logger.error("Could not stop process")
            finally:
                self.ff_processes.remove(t)
                del t

        return True

OUTPUT_CODECS = {
    u"jpeg": u"mjpeg",
    u"jpg": u"mjpeg",
    u"png": u"png",
    u"ppm": u"ppm",
}


class VideoFile(FFMpegFile):

    """ Base class for Video File """

    def thumbnail(self, size=None, fp_out=None, offset=u"0",
                  vframes=u"1", format=u"jpeg", scene_detection_factor=u"0.4",
                  timeout=10):
        # Seek on the file descriptor.
        self.seek(0)
        # vf = VideoFilter().select_scene(scene_detection_factor).scale(*size)
        vf = VideoFilter().select_keyframe().scale(*size)
        # Which video frame to use
        vf.use_video_frame(1)
        # How many frames to output
        vf.output_frames(1)
        # output format
        of = OutputFormat().format(u"image2").codec(
            OUTPUT_CODECS.get(format, u"jpg"))
        # Here we seek, but in an ffpmeg argument context : at what duration
        # should we take the thumbnail

        # FFProcess will keep a trace on ffmpeg process executed
        # VideoThumbailer is an object extending FFMpeg
        with FFProcess(self, VideoThumbailer(FFInput(self).seek(offset))) as ffmpeg:
            ffmpeg.thumbnail(vf, of, fp_out)
            if not ffmpeg.is_success:
                raise ThumbnailError(ffmpeg.errors)
            return True
        return False


class FFRunner(object):

    """ Base class for spawning ffmpeg and ffprobe"""

    __metaclass__ = abc.ABCMeta

    @abc.abstractproperty
    def executable(self):
        pass

    def __init__(self, ff_input):
        self.input = ff_input

    def __enter__(self):
        return self

    def __exit__(self, *exc):
        # stop nicely
        self.stop()

    def _run_(self, opts, pre_opts=None):
        """ Run a ffmpeg command """

        cmd = [
            self.executable,
        ]
        cmd += self.input.as_args()
        cmd += opts

        # sanitize encoding for output
        cmd = map(lambda x: x.encode(sys.getfilesystemencoding() or 'utf-8')
                  if type(x) == unicode else x, cmd)
        # now that cmd if full of str, we use unicode for logging
        log_cmd = " ".join(cmd).decode(sys.getfilesystemencoding() or 'utf-8')
        logger.debug(u"Running command : %s", log_cmd)

        is_windows = sys.platform.startswith("win")
        try:
            if is_windows:
                    # by default, windows will display a dialog if the subprocess dies.
                    # passing SEM_NOGPFAULTERRORBOX to SetErrorMode disable this behavior
                    SEM_NOGPFAULTERRORBOX=0x0002
                    import ctypes
                    # we call SetErrorMode in 2 steps to avoid overriding existing error mode
                    previous_error_mode = ctypes.windll.kernel32.SetErrorMode(SEM_NOGPFAULTERRORBOX)
                    ctypes.windll.kernel32.SetErrorMode(previous_error_mode|SEM_NOGPFAULTERRORBOX)

            self.proc = sp.Popen(
                cmd,
                stdin=self.input.stdin,
                stdout=sp.PIPE,
                stderr=sp.PIPE,
            )
        finally:
            if is_windows:
                import ctypes
                # reset previous error mode
                ctypes.windll.kernel32.SetErrorMode(previous_error_mode)

        self.stdout_reader = QueueReaderThread(self.proc.stdout)
        self.stdout_reader.start()

        self.stderr_reader = QueueReaderThread(self.proc.stderr)
        self.stderr_reader.start()

    @property
    def returncode(self):
        if hasattr(self, u"proc"):
            return self.proc.poll()
        return None

    @property
    def is_success(self):
        return self.returncode == 0

    def stop(self):
        """ Stop nicely the underlying process"""
        if hasattr(self, u"proc"):
            # if None, processus is still running
            if self.proc.poll() is None:
                self.proc.terminate()

    def kill(self):
        """ Kill the underlying process """
        if hasattr(self, u"proc"):
            # if None, processus is still running
            if self.proc.poll() is None:
                self.proc.kill()

    def read_data(self, timeout=10.0):
        """Generator that read blocks datas from ffmpeg stdout"""
        # Read from stdout in a separate thread and consume data from
        # the queue.
        start_time = time.time()
        while True:
            # Wait for data to be available or a timeout.
            data = None
            try:
                data = self.stdout_reader.queue.get(timeout=timeout)
                if data:
                    yield data
                else:
                    # End of file.
                    break
            except queue.Empty:
                # Queue read timed out.
                end_time = time.time()
                if not data:
                    if end_time - start_time >= timeout:
                        # Nothing interesting has happened for a while --
                        # FFmpeg is probably hanging.
                        raise ReadTimeoutError(
                            u"ffmpeg output: {}".format(self.errors))
                    else:
                        start_time = end_time
                        # Keep waiting.
                        continue

    @property
    def errors(self):
        return "".join(self.stderr_reader.queue.queue).decode('utf-8')


class FFMpeg(FFRunner):

    """ Base class for ffmpeg """

    executable = FFMPEG_BIN


class FFArgument(object):

    def __init__(self):
        self.params = {}

    def to_list(self, key, **kwargs):
        return [
            key,
            ",".join(u"%s=%s" % (key, value) for key, value in kwargs.items())
        ]

    def _get_parameters(self, *args):
        ret = []
        for param in args:
            ret += getattr(self, param, [])
        return ret


class VideoFilter(FFArgument):

    def scale(self, width, height):
        scale = u"%s:%s" % (width, height)
        self.params[u"scale"] = scale
        return self

    def select_scene(self, factor):
        return self.select(self.gt(u"scene", factor))

    def select_keyframe(self):
        return self.select(self.eq(u"pict_type", u"I"))

    def gt(self, what, threshold):
        return u"gt(%s\,%s)" % (what, _D(threshold))

    def eq(self, what, value):
        return u"eq(%s\,%s)" % (what, value)

    def select(self, what):
        self.params[u"select"] = what
        return self

    def output_frames(self, frame):
        self._vframes = [u"-vframes", unicode(int(frame))]
        return self

    def use_video_frame(self, frame):
        self._framesv = [u"-frames:v", unicode(int(frame))]
        return self

    def as_args(self):
        ret = self.to_list(u"-vf", **self.params)
        ret += self._get_parameters(u"_framesv", u"_vframes")
        return ret


class OutputFormat(FFArgument):

    def format(self, format):
        self._format = [u"-f", format]
        return self

    def codec(self, codec):
        self._codec = [u"-c", codec]
        return self

    def as_args(self):
        ret = []
        ret += self._get_parameters(u"_format", u"_codec")
        return ret


class VideoThumbailer(FFMpeg):

    def thumbnail(self, video_filter, output_format, fp_out, timeout=10):
        opts = []
        opts += video_filter.as_args()
        opts += output_format.as_args()

        opts += [
            u"-"  # it means output to pipe
        ]

        self._run_(opts)

        had_data = False
        try:
            for data in self.read_data(timeout):
                if data is not None and len(data) >= 0:
                    had_data = True
                    fp_out.write(data)
            if not had_data:
                raise ThumbnailError(
                    u"Could not extract thumbnail : %s" % self.errors)
        except (ReadTimeoutError, ):
            raise ThumbnailError(
                u"Timeout while reading thumbnail data : %s" % self.errors
            )


class FFProbe(FFRunner):

    """ Base class for ffmpeg """

    executable = FFPROBE_BIN


import json


class FFFileInfos(FFProbe):

    def get_infos(self):
        output = None
        opts = [
            u"-print_format", u"json",
            u"-show_format",
            u"-show_streams"
        ]
        self._run_(opts)
        had_data = False
        for data in self.read_data():
            had_data = True
            if output is None:
                output = data
            else:
                output += data

        if not had_data:
            raise NoOutputError(u"Could not get file info.")

        return json.loads(output)
