# coding: utf-8
import os
import sys


def is_filename_encoding_consistent(fname):
    """ Return True if the filename encoding is consistent.
    In somes case, we need to manipulate a filename as a bytestring (like when
    calling subprocess.Popen.

    If the filename contains invalid data, we may be able to access it by
    using its filename in unicode but not in bytestring.

    This function returns True if we can access it in both way.
    """
    fname_enc = None
    fs_enc = sys.getfilesystemencoding()
    if type(fname) == unicode:
        fname_enc = fname.encode(fs_enc)
    elif type(fname) == str:
        fname_enc = fname.decode(fs_enc)
    return os.path.exists(fname) and os.path.exists(fname_enc)


# {{ strings normalization functions
import unicodedata
def remove_accents(s):
    """ Remove accents from the given string"""
    return u''.join((c for c in unicodedata.normalize('NFD', s) if unicodedata.category(c) != 'Mn'))


import string
def remove_puctuation(s):
    return s.encode('utf-8').translate(string.maketrans("",""), string.punctuation).decode('utf-8')

def normalize(s, lowercase=True):
    if isinstance(s, str):
        s=s.decode('utf-8')
    ret = remove_puctuation(remove_accents(s)).strip().replace(' ', '_')
    if lowercase:
        ret = ret.lower()
    return ret
# }} strings normalization functions

# {{ strings matchings functions
from fuzzywuzzy import fuzz
def ratio(s1, s2):
    s1 = s1 or u''
    s2 = s2 or u''
    return fuzz.ratio(normalize(s1), normalize(s2))


def token_ratio(s1, s2):
    s1 = s1 or u''
    s2 = s2 or u''
    s1 = u' '.join([normalize(w) for w in s1.split(' ')])
    s2 = u' '.join([normalize(w) for w in s2.split(' ')])
    return fuzz.token_sort_ratio(s1, s2)
# }} strings matchings functions



# {{ strings cosine similarity
import re, math
from collections import Counter

def get_vec_cosine(vec1, vec2):
    """ Returns cosine similarity for two vectors"""
    intersection = set(vec1.keys()) & set(vec2.keys())
    numerator = sum([vec1[x] * vec2[x] for x in intersection])
    sum1 = sum([vec1[x]**2 for x in vec1.keys()])
    sum2 = sum([vec2[x]**2 for x in vec2.keys()])
    denominator = math.sqrt(sum1) * math.sqrt(sum2)
    if not denominator:
        return 0.0
    else:
        return float(numerator) / denominator

def text_to_vector(text):
    """ Transform a string to a vector"""
    text = text or u''
    words = text.split(' ')
    words = [normalize(w) for w in words]
    words = [w for w in words if w]
    return Counter(words)

def cosine_similarity(s1, s2):
    """Returns the cosine similarity for 3 strings"""
    vector1 = text_to_vector(s1)
    vector2 = text_to_vector(s2)
    return int(get_vec_cosine(vector1, vector2) * 100)
# }} strings cosine similarity


def score(s1, s2):
    return max([ratio(s1, s2), token_ratio(s1, s2), cosine_similarity(s1, s2)])


def partial_score(s1, s2):
    s1 = s1 or u''
    s2 = s2 or u''
    s1 = u' '.join([normalize(w) for w in s1.split(' ')])
    s2 = u' '.join([normalize(w) for w in s2.split(' ')])
    return [fuzz.partial_ratio(s1, s2), fuzz.partial_token_sort_ratio(s1, s2)]
