import os
# original project https://github.com/cam-stitt/arnold
# code copied here mainly because of this issue : https://github.com/cam-stitt/arnold/issues/16
# also having the code locally avoid monkey patching

from .exceptions import (ArgumentException, DBAttrNotFound,
                         DirectionNotFoundException, InvalidConfiguration,
                         MigrationNotFoundException, )
from .models import Migration
from importlib import import_module

import logging
logger = logging.getLogger(__name__)

IGNORED_FILES = ["__init__"]


def _setup_table(model):
    if model.table_exists():
        return
    else:
        model.create_table()
    return True


def _retreive_filenames(files):
    # Redifine the original _retreive_filenames to take in account
    # pyc files
    filenames = list()
    for f in files:
        splits = f.rsplit(".", 1)
        if len(splits) <= 1 or splits[1] not in ("py", "pyc") or \
           splits[0] in IGNORED_FILES:
            continue
        filenames.append(splits[0])
    return sorted(set(filenames), key=lambda fname: int(fname.split("_")[0]))


def _perform_single_migration(direction, model, **kwargs):
    """Runs a single migration method (up or down)"""
    fake = kwargs.get("fake")

    migration = kwargs.get("migration")
    if not migration:
        raise MigrationNotFoundException()

    migration_exists = model.select().where(
        model.migration == migration).limit(1).exists()

    if migration_exists and direction == "up":
        logger.debug("Migration {0} already exists, {1}".format(migration, "skipping"))
        return False
    if not migration_exists and direction == "down":
        logger.warn("Migration {0} does not exist, {1}".format(migration, "skipping"))
        return False

    logger.debug("Migration {0} going {1}".format(migration, direction))

    if fake:
        _update_migration_table(direction, model, migration)
        logger.warn("Faking {0}".format(migration))
    else:
        module_name = "{0}.{1}".format(kwargs.get("migration_module"),
                                       migration)
        logger.debug("Importing {0}".format(module_name))
        migration_module = import_module(module_name)

        if hasattr(migration_module, direction):
            getattr(migration_module, direction)()
            _update_migration_table(direction, model, migration)
        else:
            raise DirectionNotFoundException

    logger.debug("Migration {0} went {1}".format(migration, direction))
    return True


def _update_migration_table(direction, model, migration):
    if direction == "up":
        model.insert(migration=migration).execute()
    else:
        model.delete().where(model.migration == migration).execute()


def _perform_migrations(direction, model, **kwargs):
    """
    Find the migration if it is passed in and call the up or down method as
    required. If no migration is passed, loop through list and find
    the migrations that need to be run.
    """
    migration = kwargs.get("migration")
    files = os.listdir(kwargs.get("directory"))
    filenames = _retreive_filenames(files)

    if direction == "down":
        filenames.reverse()

    if len(filenames) <= 0:
        return True

    if migration:
        if migration not in filenames:
            raise MigrationNotFoundException()
        _perform_single_migration(direction, model, **kwargs)
        return True

    for f in filenames:
        _perform_single_migration(direction, model, migration=f, **kwargs)
    return True


def migrate(direction="up", **kwargs):
    """The main method, handle exceptions and start migrations"""
    # Pop ignored and db, they aren't required later
    ignored = kwargs.pop("ignored", None)
    try:
        db = kwargs.pop("database")
    except KeyError:
        raise DBAttrNotFound()
    # Get directory and module, we need them later
    directory = kwargs.get("directory")
    migration_module = kwargs.get("migration_module")

    if direction not in ["up", "down"]:
        raise ArgumentException()

    if ignored:
        IGNORED_FILES.extend(ignored)

    if db:
        model = Migration
        model._meta.database = db
    else:
        raise DBAttrNotFound()

    _setup_table(model)

    if not directory or not migration_module:
        raise InvalidConfiguration()

    _perform_migrations(direction, model, **kwargs)

    return True
